Package: nonprobsampling
Type: Package
Title: Inference for Nonprobability Samples Using Multiple Reference
        Surveys
Version: 0.1.0
Authors@R: 
    c(
      person(given = "Jiakun",
             family = "Lin",
             email = "jiak.lin@alumni.utoronto.ca",
             role = c("aut", "cre")),
      person(given = "Victoria",
             family = "Landsman",
             role = "aut"),
      person(given = "Aya A.",
             family = "Mitani",
             role = "aut")
    )
Description: Provides pseudo-weighted estimates of means and
    prevalences for finite population inference from nonprobability
    samples using auxiliary information from one or multiple probability
    reference surveys. The package supports estimation with multiple
    reference surveys, allowing auxiliary information to be combined
    when no single survey contains all variables relevant to
    participation. Optional cumulative precalibration can be applied to
    align weighted totals of shared variables across surveys. Methods
    are based on the generalized estimating equations framework of
    Landsman et al. (2026) <doi:10.1002/sim.70403> for correcting
    participation bias. For a single reference survey, the package
    implements the raking ratio calibration method and includes the
    adjusted logistic propensity (ALP) method of Wang, Valliant, and Li
    (2021) <doi:10.1002/sim.9122>, as well as the Chen-Li-Wu (CLW)
    method of Chen, Li, and Wu (2020) <doi:10.1080/01621459.2019.1677241>.
    Analytic variance estimation uses Taylor linearization and accounts
    for complex sampling designs in the reference surveys via integration
    with the 'survey' package.
Depends: R (>= 4.0.0)
Imports: stats, survey, nleqslv, utils
Suggests: testthat (>= 3.0.0), knitr, rmarkdown
VignetteBuilder: knitr
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.2
Language: en-US
URL: https://github.com/Jiakun0611/nonprobsampling,
        https://jiakun0611.github.io/nonprobsampling/
BugReports: https://github.com/Jiakun0611/nonprobsampling/issues
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-07-02 03:20:06 UTC; Kevin
Author: Jiakun Lin [aut, cre],
  Victoria Landsman [aut],
  Aya A. Mitani [aut]
Maintainer: Jiakun Lin <jiak.lin@alumni.utoronto.ca>
Repository: CRAN
Date/Publication: 2026-07-07 10:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(na.action,pw_fit)
S3method(na.action,pwmean)
S3method(print,pw_fit)
S3method(print,pw_na_summary)
S3method(print,pwmean)
S3method(print,pwmean_factor)
S3method(summary,pw_fit)
S3method(summary,pwmean)
S3method(summary,pwmean_factor)
export(est_pw)
export(pw_solver_control)
export(pwmean)
importFrom(stats,as.formula)
importFrom(stats,coef)
importFrom(stats,model.matrix)
importFrom(stats,na.action)
importFrom(stats,setNames)
importFrom(stats,weights)
importFrom(survey,calibrate)
importFrom(survey,svydesign)
importFrom(survey,svytotal)

# nonprobsampling

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/nonprobsampling)](https://CRAN.R-project.org/package=nonprobsampling)
[![pkgdown site](https://img.shields.io/badge/docs-pkgdown-blue.svg)](https://jiakun0611.github.io/nonprobsampling/)
[![License: GPL v3](https://img.shields.io/badge/License-GPL%20v3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)
<!-- badges: end -->

**Inference for Nonprobability Samples Using Multiple Reference Surveys**

`nonprobsampling` 
implements pseudo-weighting methods for finite population inference from nonprobability samples, such as convenience samples, volunteer cohorts, and opt-in panels.
Because the participation mechanism in a
nonprobability sample is unknown, unadjusted estimates of population means and
prevalences may be biased. The package addresses this issue by leveraging auxiliary information from one or multiple probability reference surveys to estimate participation probabilities and using their inverses as pseudo-weights to obtain bias-corrected estimates of finite population means and prevalences. 

The implemented methods are based on the generalized estimating equations framework of Landsman et al. (2026). This includes a multi-reference extension that enables the integration of auxiliary information across multiple reference surveys.

## Installation

Install the released version from CRAN (once available):

```r
install.packages("nonprobsampling")
```

Install the development version from GitHub:

```r
# install.packages("remotes")
remotes::install_github("Jiakun0611/nonprobsampling")
```

## Methods

With **one** reference survey:

- **Calibration** — raking ratio calibration (Landsman et al., 2026)
- **ALP** — adjusted logistic propensity weighting (Wang, Valliant, and Li, 2021)
- **CLW** — Chen, Li, and Wu (2020)


With **multiple** reference surveys:

- **Multi-reference calibration** — enables the integration of auxiliary information across multiple surveys when no single reference survey contains all variables relevant to participation (Landsman et al., 2026), with an optional cumulative precalibration step to preserve information on the relationships between overlapping and unique auxiliary variables across reference surveys.

Variance estimation is based on Taylor linearization, with complex sampling designs in the reference surveys handled through integration with the
[`survey`](https://CRAN.R-project.org/package=survey) package; when bootstrap replicate weights are provided, bootstrap-based variance estimation is also supported.

## Usage

Estimation proceeds in two steps: `est_pw()` estimates pseudo-weights, then
`pwmean()` estimates a pseudo-weighted mean or prevalence for an outcome.

```r
library(nonprobsampling)

data(sc)    # nonprobability sample (outcome: psa_level)
data(sp1)   # probability reference survey

# Reference survey design
ref1_design <- survey::svydesign(
  ids     = ~psu_sp1,
  strata  = ~strata_sp1,
  weights = ~wts_sp1,
  data    = sp1,
  nest    = TRUE
)

# Step 1: estimate pseudo-weights (one-reference calibration)
fit <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "calibration",
  control   = pw_solver_control(ftol = 1e-6)
)
print(fit)
summary(fit)

# Step 2: pseudo-weighted mean of the outcome, by BMI categories
out <- pwmean(fit, y = "psa_level", zcol = "BMI")
print(out)
summary(out)
```

With multiple reference surveys, users provide one survey design object for each reference survey and a corresponding list of participation model formulas, with one formula specified for each survey.

```r
data(sp2)   # second probability reference survey

ref2_design <- survey::svydesign(
  ids     = ~psu_sp2,
  strata  = ~strata_sp2,
  weights = ~wts_sp2,
  data    = sp2,
  nest    = TRUE
)

fit2 <- est_pw(
  data = list(sc, ref1_design, ref2_design),
  p_formula = list(
    ~ agecat + race + education + psa_level + pros_enlarged + comorbidity,
    ~ agecat + race + BMI + diabetes + comorbidity
  ),
  sp_order = "size",
  precali = TRUE,
  control = pw_solver_control(ftol = 1e-6)
)
print(fit2)
summary(fit2)
```

See `vignette("nonprobsampling")` for more details.

## Datasets

The package includes example datasets used throughout the documentation:

| Dataset         | Description                                              |
|-----------------|----------------------------------------------------------|
| `sc`            | Nonprobability sample (synthetic, NHANES-based)          |
| `sp1`           | First probability reference survey (NHANES 1999–2010)    |
| `sp2`           | Second probability reference survey (NHIS 1997–2008)     |
| `sp1_bootstrap` | `sp1` with bootstrap replicate weights                   |

## Getting help

Run any of the following commands in R to open the corresponding help page
(clickable links are available on the
[package website](https://jiakun0611.github.io/nonprobsampling/)):

<div style="font-family: monospace; line-height: 1.6; padding: 1em; border-radius: 6px; overflow-x: auto;">
<span style="color:#6a737d;">## Package overview</span><br>
?nonprobsampling<br>
<br>
<span style="color:#6a737d;">## Main functions</span><br>
?est_pw<br>
?pwmean<br>
?pw_solver_control<br>
<br>
<span style="color:#6a737d;">## Datasets</span><br>
?sc<br>
?sp1<br>
?sp2<br>
?sp1_bootstrap<br>
<br>
<span style="color:#6a737d;">## Vignette</span><br>
vignette("nonprobsampling")
</div>

## Citation

To cite the package in publications, run:

```r
citation("nonprobsampling")
```

## References

- Chen, Y., Li, P., and Wu, C. (2020). Doubly robust inference with
  nonprobability survey samples. *Journal of the American Statistical
  Association*, 115(532), 2011–2021.
  [doi:10.1080/01621459.2019.1677241](https://doi.org/10.1080/01621459.2019.1677241)
- Landsman, V., Wang, L., Carrillo-Garcia, I., Mitani, A. A., Smith, P. M.,
  Graubard, B. I., Bui, T., and Carnide, N. (2026). Correction for
  participation bias in nonprobability samples using multiple reference
  surveys. *Statistics in Medicine*, 45(3–5).
  [doi:10.1002/sim.70403](https://doi.org/10.1002/sim.70403)
- Wang, L., Valliant, R., and Li, Y. (2021). Adjusted logistic propensity
  weighting methods for population inference using nonprobability
  volunteer-based epidemiologic cohorts. *Statistics in Medicine*, 40(24),
  5237–5250.
  [doi:10.1002/sim.9122](https://doi.org/10.1002/sim.9122)

## License

GPL-3 © Jiakun Lin, Victoria Landsman, Aya A. Mitani

#' Handle missing values at the build stage (first layer)
#'
#' Process NA for `est_pw()` before estimation step.
#' Resolves the `na.action` argument to a standard mode string, applies NA
#' filtering to `sc` and `sp`, subsets the survey design object(s) in
#' `sp_des` to match the filtered `sp`, and produces a summary of
#' exclusions.
#'
#' @param sc A data frame. The nonprobability sample (before NA removal).
#' @param sp A data frame (one-reference case) or a named list of data frames
#'   (multi-reference case). Analysis data extracted from the reference
#'   survey design(s).
#' @param sp_des A single survey design object (one-reference case) or a
#'   named list of survey design objects (multi-reference case). Subsetted
#'   to align with the filtered `sp`.
#' @param p_formula A one-sided formula (one-reference case) or a list of
#'   one-sided formulas (multi-reference case) specifying the participation
#'   model variables used to identify rows with missing values.
#' @param na.action A function (`stats::na.omit`, `stats::na.exclude`,
#'   `stats::na.fail`, or `stats::na.pass`), an equivalent character string,
#'   or `NULL` (which inherits from `getOption("na.action")`).
#' @param n_ref Integer. Number of reference surveys. Controls whether
#'   one-reference or multi-reference logic is applied when subsetting
#'   `sp_des` and computing `n_sp_orig`.
#' @param verbose Logical. If `TRUE`, prints per-dataset row counts and
#'   exclusion totals via `message()`.
#'
#' @return A list with the following elements:
#' \describe{
#'   \item{`na_mode`}{Character string: one of `"omit"`, `"exclude"`,
#'     `"fail"`, or `"pass"`.}
#'   \item{`sc`}{The cleaned nonprobability sample data frame.}
#'   \item{`sp`}{The cleaned reference survey data frame(s).}
#'   \item{`sp_des`}{The subsetted survey design object(s).}
#'   \item{`keep_sc`}{Logical vector indicating which rows of the original
#'     `sc` are retained.}
#'   \item{`keep_sp`}{Logical vector (one-reference) or list of logical
#'     vectors (multi-reference) indicating which rows of each `sp` are
#'     retained.}
#'   \item{`n_sp_orig`}{Original row count(s) of `sp_des` before
#'     subsetting.}
#'   \item{`na_action_obj`}{An lm-style `na.action` attribute for `sc`,
#'     or `NULL` if no rows were removed.}
#'   \item{`log_messages`}{Character vector of per-variable NA detail
#'     messages from `handle_na_for_ipwm()`, suitable for appending to
#'     the running `log_messages` in `est_pw()`.}
#'   \item{`na_summary`}{A `pw_na_summary` object with row counts before
#'     and after NA removal, or `NULL` if no rows were excluded.}
#' }
#'
#' @keywords internal
process_na_build <- function(sc, sp, sp_des, p_formula, na.action, n_ref, verbose = FALSE) {

  na_mode <- resolve_na_action(na.action)

  na_res <- handle_na_for_ipwm(
    sc        = sc,
    sp        = sp,
    p_formula = p_formula,
    na_mode   = na_mode
  )

  # Capture original sp size(s) before subsetting sp_des
  if (n_ref == 1) {
    n_sp_orig <- nrow(sp_des$variables)
    keep_sp   <- na_res$keep_sp
    if (!all(keep_sp)) sp_des <- sp_des[keep_sp, ]
  } else {
    n_sp_orig <- vapply(sp_des, function(d) nrow(d$variables), integer(1))
    keep_sp   <- na_res$keep_sp
    for (.i in seq_along(sp_des)) {
      if (!all(keep_sp[[.i]])) sp_des[[.i]] <- sp_des[[.i]][keep_sp[[.i]], ]
    }
  }

  result <- list(
    na_mode       = na_mode,
    sc            = na_res$sc,
    sp            = na_res$sp,
    sp_des        = sp_des,
    keep_sc       = na_res$keep_sc,
    keep_sp       = keep_sp,
    n_sp_orig     = n_sp_orig,
    na_action_obj = na_res$na_action,
    log_messages  = na_res$log
  )

  result$na_summary <- .report_na_exclusions(result, nrow(sc), n_ref, verbose = verbose)

  result
}

#' Construct block design matrices for multi-reference calibration
#'
#' @param vars_XC Character vector of all predictor variables (union across
#'   reference samples), used to build the `sc` design matrix `Xc`.
#' @param sc Data frame. The nonprobability sample.
#' @param sp_list List of reference sample data frames.
#' @param xcol List of character vectors: the variables contributed by each
#'   reference sample (as returned by `check_input_multi`).
#' @param wts_cols Character vector of survey-weight column names,
#'   one per element of `sp_list`.
#'
#' @return A list with components:
#'   \item{Xc}{Numeric design matrix for `sc`, with intercept prepended,
#'     of dimension \eqn{n_c \times p}.}
#'   \item{Xp_list}{List of numeric design matrices for each reference sample.
#'     The first block includes an intercept column; subsequent blocks do not.}
#'   \item{wts_list}{List of numeric weight vectors, one per reference sample.}
#'
#' @keywords internal
multi_matrix_construction <- function(vars_XC, sc, sp_list, xcol, wts_cols) {
  # -----------------------------
  # basic checks
  # -----------------------------
  if (!is.data.frame(sc)) {
    stop("`sc` must be a data.frame.", call. = FALSE)
  }

  if (!is.list(sp_list) || length(sp_list) == 0L) {
    stop("`sp_list` must be a non-empty list of reference samples.", call. = FALSE)
  }

  n_ref <- length(sp_list)

  if (!is.list(xcol) || length(xcol) != n_ref) {
    stop("`xcol` must be a list with the same length as `sp_list`.", call. = FALSE)
  }

  if (!is.character(wts_cols) || length(wts_cols) != n_ref) {
    stop("`wts_cols` must be a character vector with the same length as `sp_list`.", call. = FALSE)
  }

  # -----------------------------
  # build Xc
  # -----------------------------
  Xc <- add_intercept(vars_XC, data = sc, intercept = TRUE)
  check_design_identifiability(Xc, label = "Multi_Calibration")

  # -----------------------------
  # build Xp_list and wts_list
  # -----------------------------
  Xp_list  <- vector("list", n_ref)
  wts_list <- vector("list", n_ref)

  for (i in seq_len(n_ref)) {
    sp_i <- sp_list[[i]]

    if (!is.data.frame(sp_i)) {
      stop(sprintf("`sp_list[[%d]]` must be a data.frame.", i), call. = FALSE)
    }

    if (!(wts_cols[i] %in% names(sp_i))) {
      stop(
        sprintf("Weight column '%s' not found in `sp_list[[%d]]`.", wts_cols[i], i),
        call. = FALSE
      )
    }

    # intercept only in the first reference sample
    Xp_i <- add_intercept(
      vars      = xcol[[i]],
      data      = sp_i,
      intercept = (i == 1)
    )


    check_design_identifiability(Xp_i, label = "Multi_Calibration")


    w_i <- sp_i[[wts_cols[i]]]

    if (!is.numeric(w_i)) {
      stop(
        sprintf("Weight column '%s' in `sp_list[[%d]]` must be numeric.", wts_cols[i], i),
        call. = FALSE
      )
    }

    if (anyNA(w_i)) {
      stop(
        sprintf("Weight column '%s' in `sp_list[[%d]]` contains NA.", wts_cols[i], i),
        call. = FALSE
      )
    }

    if (any(!is.finite(w_i))) {
      stop(
        sprintf("Weight column '%s' in `sp_list[[%d]]` contains non-finite values.", wts_cols[i], i),
        call. = FALSE
      )
    }

    if (any(w_i <= 0)) {
      stop(
        sprintf("Weight column '%s' in `sp_list[[%d]]` must be strictly positive.", wts_cols[i], i),
        call. = FALSE
      )
    }

    Xp_list[[i]]  <- Xp_i
    wts_list[[i]] <- w_i
  }

  # -----------------------------
  # dimension consistency check
  # In multi-raking, columns of Xc should match the total number of columns
  # across all Xp blocks (first block contains intercept).
  # -----------------------------
  p_xc <- ncol(Xc)
  p_xp <- sum(vapply(Xp_list, ncol, integer(1)))

  if (p_xc != p_xp) {
    stop(
      paste0(
        "Column mismatch between `Xc` and combined `Xp_list`: ",
        "ncol(Xc) = ", p_xc, ", but sum(ncol(Xp_list[[i]])) = ", p_xp, ". ",
        "Please check `vars_XC` and `xcol` construction."
      ),
      call. = FALSE
    )
  }

  return(list(
    Xc       = Xc,
    Xp_list  = Xp_list,
    wts_list = wts_list
  ))
}

#' Print method for pwmean objects
#'
#' Displays the pseudo-weighted mean estimate and its uncertainty.
#' For factor-like domain variables, prints one row per domain level.
#'
#' @param x An object of class \code{"pwmean"}, returned by \code{\link{pwmean}}.
#' @param ... Additional arguments, currently unused.
#'
#' @return Invisibly returns \code{x}.
#'
#' @method print pwmean
#' @export
print.pwmean <- function(x, ...) {
  print_pwmean_impl(x, group_label = "Domain", estimate_label = "Mean")
}

#' Print method for pwmean objects with categorical outcomes
#'
#' Displays pseudo-weighted prevalence estimates and their uncertainty.
#'
#' @param x An object of class \code{"pwmean_factor"}, returned by
#'   \code{\link{pwmean}} when \code{y} is a factor.
#' @param ... Additional arguments, currently unused.
#'
#' @return Invisibly returns \code{x}.
#'
#' @method print pwmean_factor
#' @export
print.pwmean_factor <- function(x, ...) {
  print_pwmean_impl(x, group_label = "Category", estimate_label = "Prevalence")
}

print_pwmean_impl <- function(x, group_label, estimate_label) {

  m_raw <- trimws(tolower(as.character(x$method %||% "")))
  m <- if (m_raw %in% c("alp", "clw")) toupper(m_raw) else m_raw

  cat("\nPseudo-weighted (", m, ") Estimators:\n", sep = "")

  for (i in seq_len(nrow(x$estimates))) {

    d <- x$estimates[i, ]

    if ("category" %in% names(d)) {
      cat(sprintf("  Category: %s\n", d$category))
      cat(sprintf("  Domain: %s\n", d$domain))
    } else {
      cat(sprintf("  %s: %s\n", group_label, d$domain))
    }
    cat(sprintf("  %-15s %10.6f\n", paste0(estimate_label, ":"), d$adjusted_mean))
    cat(sprintf("  %-15s %10.6f\n", "Std. Error:",               d$adjusted_se))
    cat(sprintf("  %-15s [%0.6f, %0.6f]\n",
                "95% CI:", d$adjusted_lower, d$adjusted_upper))

    if (i < nrow(x$estimates)) cat("\n")
  }

  invisible(x)
}

#' Map each reference-sample block to its column indices in Xc
#'
#' @param Xc Numeric design matrix for the nonprobability sample, with named
#'   columns.
#' @param Xp_list List of reference-sample design matrices, each with named
#'   columns that are a subset of `colnames(Xc)`.
#' @param label Character scalar used as a prefix in error messages.
#'
#' @return A list of integer vectors, one per element of `Xp_list`, giving the
#'   column positions in `Xc` that correspond to the columns of each
#'   `Xp_list[[j]]`.
#'
#' @keywords internal
multi_raking_block_cols <- function(Xc, Xp_list, label = "Multi_Calibration") {
  if (is.null(colnames(Xc))) {
    stop(
      sprintf("[%s] Xc must have column names for multi-reference calibration.", label),
      call. = FALSE
    )
  }

  lapply(seq_along(Xp_list), function(j) {
    Xp <- Xp_list[[j]]

    if (is.null(colnames(Xp))) {
      stop(
        sprintf("[%s] Xp_list[[%d]] must have column names.", label, j),
        call. = FALSE
      )
    }

    idx <- match(colnames(Xp), colnames(Xc))

    if (any(is.na(idx))) {
      missing_cols <- colnames(Xp)[is.na(idx)]
      stop(
        sprintf(
          paste0(
            "[%s] Some columns in Xp_list[[%d]] are not found in Xc: %s.\n",
            "This usually means the multi-reference design matrices are not aligned."
          ),
          label, j, paste(missing_cols, collapse = ", ")
        ),
        call. = FALSE
      )
    }

    idx
  })
}

#' Compute reference weighted totals for multi-reference calibration
#'
#' @param Xp_list List of reference-sample design matrices.
#' @param wts_list List of numeric weight vectors, one per element of
#'   `Xp_list`.
#' @param label Character scalar used as a prefix in error messages.
#'
#' @return A named numeric vector of length equal to the total number of
#'   columns across all elements of `Xp_list`, containing the survey-weighted
#'   column totals \eqn{\sum_j w_{ij} X_{ij}} stacked in block order.
#'
#' @keywords internal
multi_raking_fp <- function(Xp_list, wts_list, label = "Multi_Calibration") {
  if (length(Xp_list) != length(wts_list)) {
    stop(
      sprintf(
        "[%s] Xp_list and wts_list must have the same length.",
        label
      ),
      call. = FALSE
    )
  }

  unlist(
    lapply(seq_along(Xp_list), function(j) {
      Xp <- Xp_list[[j]]
      wts <- wts_list[[j]]

      if (nrow(Xp) != length(wts)) {
        stop(
          sprintf(
            "[%s] nrow(Xp_list[[%d]]) does not match length(wts_list[[%d]]).",
            label, j, j
          ),
          call. = FALSE
        )
      }

      out <- colSums(wts * Xp)

      if (any(!is.finite(out))) {
        stop(
          sprintf(
            "[%s] Reference weighted totals contain non-finite values for Xp_list[[%d]].",
            label, j
          ),
          call. = FALSE
        )
      }

      out
    }),
    use.names = TRUE
  )
}

#' Compute starting values for multi-reference calibration
#'
#' @param Xc Numeric design matrix for the nonprobability sample.
#' @param Xp_list List of reference-sample design matrices.
#' @param block_cols List of integer vectors mapping each reference-sample
#'   block to its column indices in `Xc`, as returned by
#'   `multi_raking_block_cols`.
#' @param f_p Named numeric vector of reference weighted totals, as returned
#'   by `multi_raking_fp`.
#' @param label Character scalar used as a prefix in error messages.
#'
#' @return A numeric vector of length `ncol(Xc)` containing the starting
#'   values for the coefficient vector \eqn{\beta}, with the intercept
#'   initialized to match the ratio of reference to sample weighted totals
#'   and all other coefficients set to zero.
#'
#' @keywords internal
multi_raking_start <- function(Xc, Xp_list,
                               block_cols,
                               f_p,
                               label = "Multi_Calibration") {
  p_dim <- ncol(Xc)
  beta_start <- rep(0, p_dim)

  eta <- as.vector(Xc %*% beta_start)

  if (any(!is.finite(eta))) {
    stop(
      sprintf("[%s] Cannot construct initial values because initial linear predictor is non-finite.", label),
      call. = FALSE
    )
  }

  wts_sc <- as.vector(exp(-eta))

  if (any(!is.finite(wts_sc))) {
    stop(
      sprintf("[%s] Cannot construct initial values because initial pseudo-weights are non-finite.", label),
      call. = FALSE
    )
  }

  f_c <- unlist(
    lapply(block_cols, function(cols) {
      colSums(wts_sc * Xc[, cols, drop = FALSE])
    }),
    use.names = TRUE
  )

  if (length(f_p) != length(f_c)) {
    stop(
      sprintf(
        paste0(
          "[%s] Cannot construct initial values because the number of reference totals ",
          "(%d) does not match the number of sample-side totals (%d)."
        ),
        label, length(f_p), length(f_c)
      ),
      call. = FALSE
    )
  }

  if (!is.finite(f_p[1]) || !is.finite(f_c[1]) || f_p[1] <= 0 || f_c[1] <= 0) {
    stop(
      sprintf(
        paste0(
          "[%s] Cannot construct initial intercept because intercept totals are invalid. ",
          "Check survey weights and intercept totals."
        ),
        label
      ),
      call. = FALSE
    )
  }

  beta_start[1] <- -log(f_p[1] / f_c[1])

  if (any(!is.finite(beta_start))) {
    stop(
      sprintf("[%s] Initial coefficient vector contains non-finite values.", label),
      call. = FALSE
    )
  }

  beta_start
}


multi_raking_fn <- function(beta, Xc,
                            f_p,
                            block_cols,
                            label = "Multi_Calibration",
                            ...) {
  eta <- as.vector(Xc %*% beta)

  if (any(!is.finite(eta))) {
    stop(
      sprintf("[%s] Multi-reference calibration failed: linear predictor became non-finite.", label),
      call. = FALSE
    )
  }

  if (any(eta < -log(.Machine$double.xmax))) {
    stop(
      sprintf(
        paste(
          "[%s] Multi-reference calibration failed: exp(-eta) overflow.",
          "This usually indicates unstable coefficient values."
        ),
        label
      ),
      call. = FALSE
    )
  }

  wts_sc <- as.vector(exp(-eta))

  if (any(!is.finite(wts_sc))) {
    stop(
      sprintf("[%s] Multi-reference calibration failed: pseudo-weights became non-finite.", label),
      call. = FALSE
    )
  }

  f_c <- unlist(
    lapply(block_cols, function(cols) {
      colSums(wts_sc * Xc[, cols, drop = FALSE])
    }),
    use.names = TRUE
  )

  if (length(f_c) != length(f_p)) {
    stop(
      sprintf(
        "[%s] Estimating-equation length mismatch: length(f_c) = %d, length(f_p) = %d.",
        label, length(f_c), length(f_p)
      ),
      call. = FALSE
    )
  }

  f_c - f_p
}

multi_raking_jac <- function(beta, Xc,
                             f_p,
                             block_cols,
                             label = "Multi_Calibration",
                             tol_singular = 1e-7,
                             ...) {
  eta <- as.vector(Xc %*% beta)

  if (any(!is.finite(eta))) {
    stop(
      sprintf("[%s] Multi-reference calibration failed: linear predictor became non-finite when computing Jacobian.", label),
      call. = FALSE
    )
  }

  if (any(eta < -log(.Machine$double.xmax))) {
    stop(
      sprintf(
        paste(
          "[%s] Multi-reference calibration failed: exp(-eta) overflow when computing Jacobian.",
          "This usually indicates unstable coefficient values."
        ),
        label
      ),
      call. = FALSE
    )
  }

  wts_sc <- as.vector(exp(-eta))

  if (any(!is.finite(wts_sc))) {
    stop(
      sprintf("[%s] Multi-reference calibration failed: pseudo-weights became non-finite when computing Jacobian.", label),
      call. = FALSE
    )
  }

  f_c <- unlist(
    lapply(block_cols, function(cols) {
      colSums(wts_sc * Xc[, cols, drop = FALSE])
    }),
    use.names = TRUE
  )

  g <- f_c - f_p

  J <- do.call(
    rbind,
    lapply(block_cols, function(cols) {
      Xc_block <- Xc[, cols, drop = FALSE]
      -t(wts_sc * Xc_block) %*% Xc
    })
  )

  check_estimating_system(
    J = J,
    g = g,
    label = label,
    tol_singular = tol_singular
  )

  J
}

# test-build_s5_multi.R
# Tests for multi-reference calibration helper functions and ipwm_multi_build()
#
# Covered functions:
#   multi_raking_block_cols()
#   multi_raking_fp()
#   multi_raking_start()
#   multi_raking_fn()
#   multi_raking_jac()
#   ipwm_multi_build()
#
# Setup:
#   Two reference samples, each contributing different variables, so the
#   number of columns lines up exactly (ncol(Xc) = sum(ncol(Xp_list))):
#     sp1 covers x1, x2  -> Xp1 is 3 columns: intercept, x1, x2
#     sp2 adds x3 only   -> Xp2 is 1 column:  x3 (no intercept)
#     Xc is 4 columns:   intercept, x1, x2, x3
#   So block_cols[[1]] = c(1,2,3), block_cols[[2]] = c(4), and f_p has length 4.

# Setup ----
suppressWarnings(library(survey))

# shared test fixtures ----

set.seed(2026)
n_sc  <- 100L
n_sp1 <- 200L
n_sp2 <- 150L

sc_df  <- data.frame(x1 = rnorm(n_sc),  x2 = rnorm(n_sc),  x3 = rnorm(n_sc))
sp1_df <- data.frame(
  x1      = rnorm(n_sp1),
  x2      = rnorm(n_sp1),
  sp_wts1 = runif(n_sp1, 10, 100)
)
sp2_df <- data.frame(
  x1      = rnorm(n_sp2),
  x2      = rnorm(n_sp2),
  x3      = rnorm(n_sp2),
  sp_wts2 = runif(n_sp2, 10, 100)
)

# xcol[[1]] covers {x1,x2}; xcol[[2]] = {x3} (new variable only)
# Xp1 has intercept (first block); Xp2 does NOT (subsequent block)
vars_XC <- c("x1", "x2", "x3")
Xc  <- add_intercept(vars_XC,        data = sc_df,  intercept = TRUE)   # 150 x 4
Xp1 <- add_intercept(c("x1", "x2"), data = sp1_df, intercept = TRUE)   # 200 x 3
Xp2 <- add_intercept("x3",           data = sp2_df, intercept = FALSE)  # 120 x 1

Xp_list  <- list(Xp1, Xp2)

wts1     <- sp1_df$sp_wts1
wts2     <- sp2_df$sp_wts2
wts_list <- list(wts1, wts2)

sp1_des <- survey::svydesign(ids = ~1, weights = ~sp_wts1, data = sp1_df)
sp2_des <- survey::svydesign(ids = ~1, weights = ~sp_wts2, data = sp2_df)

ctrl <- pw_solver_control(ftol = 1e-4)

# Pre-compute shared derived objects for reuse in later sections
block_cols <- multi_raking_block_cols(Xc, Xp_list)
f_p        <- multi_raking_fp(Xp_list, wts_list)

# Generalized estimating equation: multiple reference samples ----
#
# SIM paper equation (9):
#
#   sum_sc w(x_i, beta) x_i^(pk)
#     - sum_spk d_i^(pk) x_i^(pk) = 0,
#   for k = 1, ..., K.
#
# The factor N^{-1} is omitted because it does not affect whether
# the estimating equation equals zero.
#
# For multi-reference calibration:
#   w(x, beta) = exp(-x beta)
#
# Each reference sample contributes a block of variables. The columns
# in Xc corresponding to each reference block are given by block_cols.

general_ee_multi_ref <- function(beta, Xc, Xp_list, wts_list, block_cols) {
  eta_c <- drop(Xc %*% beta)
  w_c   <- exp(-eta_c)

  ee_list <- vector("list", length(Xp_list))

  for (k in seq_along(Xp_list)) {
    Xc_block <- Xc[, block_cols[[k]], drop = FALSE]
    Xp_block <- Xp_list[[k]]
    wts_k    <- wts_list[[k]]

    ee_list[[k]] <- colSums(w_c * Xc_block) - colSums(wts_k * Xp_block)
  }

  unlist(ee_list, use.names = TRUE)
}

# multi_raking_block_cols ----

test_that("multi_raking_block_cols: returns a list of length equal to Xp_list", {
  expect_type(block_cols, "list")
  expect_length(block_cols, 2L)
})

test_that("multi_raking_block_cols: each element is an integer vector", {
  expect_type(block_cols[[1]], "integer")
  expect_type(block_cols[[2]], "integer")
})

test_that("multi_raking_block_cols: block 1 maps to columns 1-3 of Xc", {
  # Xp1 columns: "(Intercept)", "x1", "x2" — all present in Xc at positions 1,2,3
  expect_equal(block_cols[[1]], c(1L, 2L, 3L))
})

test_that("multi_raking_block_cols: block 2 maps to column 4 of Xc", {
  # Xp2 column: "x3" — present in Xc at position 4
  expect_equal(block_cols[[2]], 4L)
})

test_that("multi_raking_block_cols: column names at mapped positions match Xp colnames", {
  expect_equal(colnames(Xc)[block_cols[[1]]], colnames(Xp1))
  expect_equal(colnames(Xc)[block_cols[[2]]], colnames(Xp2))
})

test_that("multi_raking_block_cols: error when Xc has no colnames", {
  Xc_bad <- Xc
  colnames(Xc_bad) <- NULL
  expect_error(
    multi_raking_block_cols(Xc_bad, Xp_list),
    "Xc must have column names",
    fixed = TRUE
  )
})

test_that("multi_raking_block_cols: error when Xp_list[[j]] has no colnames", {
  Xp_bad    <- Xp_list
  colnames(Xp_bad[[2]]) <- NULL
  expect_error(
    multi_raking_block_cols(Xc, Xp_bad),
    "must have column names",
    fixed = TRUE
  )
})

test_that("multi_raking_block_cols: error when Xp_list[[j]] has a column not in Xc", {
  Xp_bad           <- Xp_list
  colnames(Xp_bad[[2]]) <- "z_unknown"
  expect_error(
    multi_raking_block_cols(Xc, Xp_bad),
    "not found in Xc",
    fixed = TRUE
  )
})


# multi_raking_fp ----

test_that("multi_raking_fp: returns a named numeric vector", {
  expect_type(f_p, "double")
  expect_false(is.null(names(f_p)))
})

test_that("multi_raking_fp: length equals total columns across Xp_list", {
  expected_len <- ncol(Xp1) + ncol(Xp2)   # 3 + 1 = 4
  expect_length(f_p, expected_len)
})

test_that("multi_raking_fp: all values are finite", {
  expect_true(all(is.finite(f_p)))
})

test_that("multi_raking_fp: first block matches manual colSums(wts1 * Xp1)", {
  expected_block1 <- colSums(wts1 * Xp1)
  expect_equal(f_p[seq_len(ncol(Xp1))], expected_block1)
})

test_that("multi_raking_fp: second block matches manual colSums(wts2 * Xp2)", {
  idx <- seq(ncol(Xp1) + 1L, ncol(Xp1) + ncol(Xp2))
  expected_block2 <- colSums(wts2 * Xp2)
  expect_equal(f_p[idx], expected_block2)
})

test_that("multi_raking_fp: error when Xp_list and wts_list have different lengths", {
  expect_error(
    multi_raking_fp(Xp_list, list(wts1)),   # 2 vs 1
    "must have the same length",
    fixed = TRUE
  )
})

test_that("multi_raking_fp: error when nrow(Xp) != length(wts)", {
  wts_bad <- list(wts1, wts2[-1])   # wts2 shortened by 1
  expect_error(
    multi_raking_fp(Xp_list, wts_bad),
    "does not match",
    fixed = TRUE
  )
})


# multi_raking_start ----

beta_start <- multi_raking_start(
  Xc         = Xc,
  Xp_list    = Xp_list,
  block_cols = block_cols,
  f_p        = f_p
)

test_that("multi_raking_start: returns a numeric vector of length ncol(Xc)", {
  expect_type(beta_start, "double")
  expect_length(beta_start, ncol(Xc))
})

test_that("multi_raking_start: all values are finite", {
  expect_true(all(is.finite(beta_start)))
})

test_that("multi_raking_start: non-intercept values are zero", {
  # Only position 1 (the intercept) should be non-zero
  expect_equal(beta_start[-1], rep(0, ncol(Xc) - 1L))
})

test_that("multi_raking_start: intercept initialized to -log(sum(wts1) / n_sc)", {
  # At beta=0: eta=0, wts_sc=1, f_c[1] = sum(Xc[,1]) = n_sc
  # f_p[1] = sum(wts1 * Xp1[,1]) = sum(wts1)
  expected_intercept <- -log(sum(wts1) / n_sc)
  expect_equal(beta_start[1], expected_intercept, tolerance = 1e-10)
})


# multi_raking_fn ----

test_that("multi_raking_fn: returns a numeric vector of length equal to length(f_p)", {
  fn_val <- multi_raking_fn(
    beta       = beta_start,
    Xc         = Xc,
    f_p        = f_p,
    block_cols = block_cols
  )
  expect_type(fn_val, "double")
  expect_length(fn_val, length(f_p))
})

test_that("multi_raking_fn: all values are finite at beta_start", {
  fn_val <- multi_raking_fn(
    beta       = beta_start,
    Xc         = Xc,
    f_p        = f_p,
    block_cols = block_cols
  )
  expect_true(all(is.finite(fn_val)))
})

test_that("multi_raking_fn: not zero at beta_start (system not solved)", {
  fn_val <- multi_raking_fn(
    beta       = beta_start,
    Xc         = Xc,
    f_p        = f_p,
    block_cols = block_cols
  )
  # Non-intercept components should differ from zero
  expect_true(max(abs(fn_val)) > 1e-6)
})

test_that("multi_raking_fn: near zero at converged beta", {
  sol <- solve_participation_model(
    beta_start = beta_start,
    fn         = multi_raking_fn,
    jac        = multi_raking_jac,
    label      = "Test",
    control    = ctrl,
    Xc         = Xc,
    f_p        = f_p,
    block_cols = block_cols
  )
  fn_val <- multi_raking_fn(
    beta       = sol$coefficients,
    Xc         = Xc,
    f_p        = f_p,
    block_cols = block_cols
  )
  expect_true(max(abs(fn_val)) < 1e-4)
})

test_that("multi_raking_fn: error on non-finite beta", {
  beta_bad    <- beta_start
  beta_bad[1] <- Inf
  expect_error(
    multi_raking_fn(
      beta       = beta_bad,
      Xc         = Xc,
      f_p        = f_p,
      block_cols = block_cols
    ),
    "non-finite",
    fixed = TRUE
  )
})


# multi_raking_jac ----

test_that("multi_raking_jac: returns a matrix of dim (length(f_p) x ncol(Xc))", {
  J <- multi_raking_jac(
    beta       = beta_start,
    Xc         = Xc,
    f_p        = f_p,
    block_cols = block_cols
  )
  expect_true(is.matrix(J))
  expect_equal(dim(J), c(length(f_p), ncol(Xc)))
})

test_that("multi_raking_jac: diagonal values are all negative at beta_start", {
  J <- multi_raking_jac(
    beta       = beta_start,
    Xc         = Xc,
    f_p        = f_p,
    block_cols = block_cols
  )
  # The p x p Jacobian should be negative on the diagonal
  diag_values <- diag(J)
  expect_true(all(diag_values < 0))
})

test_that("multi_raking_jac returns the same Jacobian as numerical differentiation", {
  beta0 <- beta_start
  eps   <- 1e-6
  p     <- ncol(Xc)
  m     <- length(f_p)

  J_num <- matrix(0, nrow = m, ncol = p)
  for (k in seq_len(p)) {
    b_up      <- beta0; b_up[k] <- beta0[k] + eps
    b_dn      <- beta0; b_dn[k] <- beta0[k] - eps
    fn_up <- multi_raking_fn(b_up, Xc = Xc, f_p = f_p, block_cols = block_cols)
    fn_dn <- multi_raking_fn(b_dn, Xc = Xc, f_p = f_p, block_cols = block_cols)
    J_num[, k] <- (fn_up - fn_dn) / (2 * eps)
  }

  J_ana <- multi_raking_jac(
    beta       = beta0,
    Xc         = Xc,
    f_p        = f_p,
    block_cols = block_cols
  )

  # unname() removes dimnames inherited by J_ana from matrix multiplication
  expect_equal(unname(J_ana), J_num, tolerance = 1e-4)
})

test_that("multi_raking_jac: error on non-finite beta", {
  beta_bad    <- beta_start
  beta_bad[2] <- NaN
  expect_error(
    multi_raking_jac(
      beta       = beta_bad,
      Xc         = Xc,
      f_p        = f_p,
      block_cols = block_cols
    ),
    "non-finite",
    fixed = TRUE
  )
})


# ipwm_multi_build ----

# For ipwm_multi_build the sp list and vars list must align with the
# design objects.  We pass the FULL variable sets per sample; the
# builder internally derives xcol (new-variables-per-sample) via
# check_input_multi.

sp_list    <- list(sp1_df, sp2_df)
vars_list  <- list(c("x1", "x2"), c("x1", "x2", "x3"))
weight_vec <- c("sp_wts1", "sp_wts2")
des_list   <- list(sp1_des, sp2_des)

build_multi <- function(...) {
  ipwm_multi_build(
    sc       = sc_df,
    sp       = sp_list,
    vars     = vars_list,
    weight   = weight_vec,
    sp_des   = des_list,
    sp_order = "given",
    control  = ctrl,
    ...
  )
}

out_multi <- build_multi()

# return structure ----

test_that("ipwm_multi_build: returns a list with required top-level names", {
  expected <- c("pseudo_weights", "coefficients", "method",
                "solver_diagnostics", "internal")
  expect_true(all(expected %in% names(out_multi)))
})

test_that("ipwm_multi_build: method is 'multi'", {
  expect_equal(out_multi$method, "multi")
})

# pseudo-weights ----

test_that("ipwm_multi_build: pseudo_weights has length n_sc", {
  expect_length(out_multi$pseudo_weights, n_sc)
})

test_that("ipwm_multi_build: all pseudo-weights are positive", {
  expect_true(all(out_multi$pseudo_weights > 0))
})

test_that("ipwm_multi_build: all pseudo-weights are finite", {
  expect_true(all(is.finite(out_multi$pseudo_weights)))
})

# coefficients ----

test_that("ipwm_multi_build: coefficients has length ncol(Xc) = 4", {
  expect_length(out_multi$coefficients, 4L)
})

test_that("ipwm_multi_build: coefficients are named", {
  expect_false(is.null(names(out_multi$coefficients)))
  expect_true("(Intercept)" %in% names(out_multi$coefficients))
})

test_that("ipwm_multi_build: all coefficients are finite", {
  expect_true(all(is.finite(out_multi$coefficients)))
})


# estimation equation (9) in SIM satisfied ---
test_that("ipwm_multi_build satisfies the multi-reference estimating equation", {
  int_obj <- out_multi$internal

  ee <- general_ee_multi_ref(
    beta       = out_multi$coefficients,
    Xc         = int_obj$Xc,
    Xp_list    = int_obj$Xp_list,
    wts_list   = int_obj$wts_list,
    block_cols = int_obj$block_cols
  )

  expect_lt(max(abs(ee)), 1e-4)
})


test_that("multi_raking_fn matches the multi-reference estimating equation helper", {
  ee_manual <- general_ee_multi_ref(
    beta       = beta_start,
    Xc         = Xc,
    Xp_list    = Xp_list,
    wts_list   = wts_list,
    block_cols = block_cols
  )

  ee_function <- multi_raking_fn(
    beta       = beta_start,
    Xc         = Xc,
    f_p        = f_p,
    block_cols = block_cols
  )

  expect_equal(unname(ee_function), unname(ee_manual), tolerance = 1e-10)
})




# solver_diagnostics ----

test_that("ipwm_multi_build: solver_diagnostics has expected names", {
  expected <- c("solver", "termcd", "message", "method", "iter", "fmax")
  expect_true(all(expected %in% names(out_multi$solver_diagnostics)))
})

test_that("ipwm_multi_build: solver converged (termcd 1 or 2)", {
  expect_true(out_multi$solver_diagnostics$termcd %in% c(1L, 2L))
})

test_that("ipwm_multi_build: fmax is small after convergence", {
  expect_true(out_multi$solver_diagnostics$fmax < 1e-4)
})

# internal ----

test_that("ipwm_multi_build: internal contains expected components", {
  expected <- c("Xc", "Xp_list", "wts_list", "f_p",
                "block_cols", "D", "S_beta")
  expect_true(all(expected %in% names(out_multi$internal)))
})

test_that("ipwm_multi_build: internal$Xc has n_sc rows and 4 columns", {
  expect_equal(nrow(out_multi$internal$Xc), n_sc)
  expect_equal(ncol(out_multi$internal$Xc), 4L)
})

test_that("ipwm_multi_build: internal$Xp_list has length 2", {
  expect_length(out_multi$internal$Xp_list, 2L)
})

test_that("ipwm_multi_build: internal$f_p has length 4", {
  expect_length(out_multi$internal$f_p, 4L)
})

test_that("ipwm_multi_build: internal$block_cols has length 2", {
  expect_length(out_multi$internal$block_cols, 2L)
})

test_that("ipwm_multi_build: internal$S_beta is a 4x4 matrix", {
  expect_true(is.matrix(out_multi$internal$S_beta))
  expect_equal(dim(out_multi$internal$S_beta), c(4L, 4L))
})

test_that("ipwm_multi_build: internal$D is a square matrix of dim 4", {
  # D is block-diagonal: 3x3 for block 1 + 1x1 for block 2
  expect_true(is.matrix(out_multi$internal$D))
  expect_equal(dim(out_multi$internal$D), c(4L, 4L))
})

# sp_order options ----

test_that("ipwm_multi_build: sp_order = 'size' runs without error", {
  expect_no_error(
    ipwm_multi_build(
      sc       = sc_df,
      sp       = sp_list,
      vars     = vars_list,
      weight   = weight_vec,
      sp_des   = des_list,
      sp_order = "size",
      control  = ctrl
    )
  )
})

test_that("ipwm_multi_build: sp_order = 'given' and 'size' produce positive pseudo-weights", {
  out_size  <- ipwm_multi_build(
    sc       = sc_df,
    sp       = sp_list,
    vars     = vars_list,
    weight   = weight_vec,
    sp_des   = des_list,
    sp_order = "size",
    control  = ctrl
  )
  expect_true(all(out_size$pseudo_weights > 0))
})

# log_messages forwarded ----

test_that("ipwm_multi_build: log_messages argument is passed to internal", {
  out_log <- ipwm_multi_build(
    sc           = sc_df,
    sp           = sp_list,
    vars         = vars_list,
    weight       = weight_vec,
    sp_des       = des_list,
    sp_order     = "given",
    control      = ctrl,
    log_messages = c("upstream_msg")
  )
  expect_true("upstream_msg" %in% out_log$internal$log_messages)
})

# verbose does not error ----

test_that("ipwm_multi_build: verbose = TRUE runs without error", {
  expect_no_error(
    build_multi(verbose = TRUE)
  )
})

#' Handle missing values in the outcome and domain variables
#'
#' Identifies rows with missing values in `y` (and `zcol` if supplied),
#' applies the chosen `na.action` strategy, and returns the complete-case
#' subsets of the outcome vector, covariate matrix, and pseudo-weight vector.
#' Also standardizes the domain variable via `standardize_zcol()`.
#'
#' @param sc_data A list returned by `prepare_sc_data()`, with components
#'   `sc`, `X`, `w`, and `idx_keep`.
#' @param y Single character string naming the outcome variable in `sc_data$sc`.
#' @param zcol Single character string naming the domain variable in
#'   `sc_data$sc`, or NULL for the overall mean.
#' @param na.action NA-handling function; one of `stats::na.omit` (default),
#'   `stats::na.exclude`, or `stats::na.fail`. `na.pass` is not supported.
#'
#' @return A list with components:
#'
#' - `Y`: numeric outcome vector (complete cases only).
#' - `X`: covariate matrix (complete cases only).
#' - `w`: pseudo-weight vector (complete cases only).
#' - `sc`: data frame (complete cases only).
#' - `y_name`: the value of `y`.
#' - `zcol`: the value of `zcol`.
#' - `domain`: list returned by `standardize_zcol()`.
#' - `na_info`: list with `na_action`, `n_omitted`, `n_used`,
#'   `omitted_raw`, and `kept_raw`.
#'
#' @keywords internal
process_na_yz <- function(sc_data, y, zcol = NULL, na.action = stats::na.omit) {

  #----------------------------------#
  # Step 1: unpack sc_data
  #----------------------------------#
  sc <- sc_data$sc
  X  <- sc_data$X
  w  <- sc_data$w
  idx_keep <- sc_data$idx_keep

  #----------------------------------#
  # Step 2: prepare raw y / z
  #----------------------------------#
  Y0 <- sc[[y]]

  if (is.null(zcol)) {
    ok <- !is.na(Y0)
  } else {
    Z0 <- sc[[zcol]]
    ok <- !is.na(Y0) & !is.na(Z0)
  }

  #----------------------------------#
  # Step 3: NA policy
  #----------------------------------#
  na_mode <- resolve_na_action(na.action)

  if (na_mode == "pass") {
    stop("`na.pass` is not supported at estimate stage.", call. = FALSE)
  }

  if (na_mode == "fail" && any(!ok)) {
    stop("Missing values detected in `y` or `zcol`.", call. = FALSE)
  }

  if (!any(ok))
    stop("No complete cases remain after removing missing values in 'y' and 'zcol'.",
         call. = FALSE)

  # both na.omit and na.exclude use complete cases for fitting
  sc_cc <- sc[ok, , drop = FALSE]
  Y     <- Y0[ok]
  X_cc  <- X[ok, , drop = FALSE]
  w_cc  <- w[ok]

  omitted_raw <- idx_keep[!ok]

  if (length(omitted_raw) > 0L) {
    na_obj <- omitted_raw
    class(na_obj) <- if (na_mode == "exclude") "exclude" else "omit"
  } else {
    na_obj <- NULL
  }

  #----------------------------------#
  # Step 4: standardize zcol after NA filtering
  #----------------------------------#
  domain_info <- standardize_zcol(data = sc_cc, zcol = zcol)

  #----------------------------------#
  # Step 5: return
  #----------------------------------#
  list(
    Y = Y,
    X = X_cc,
    w = w_cc,
    sc = sc_cc,
    y_name = y,
    zcol = zcol,
    domain = domain_info,
    na_info = list(
      na_action = na_obj,
      n_omitted = sum(!ok),
      n_used = length(Y),
      omitted_raw = omitted_raw,
      kept_raw = idx_keep[ok]
    )
  )
}

# Setup ----

data(sc)

# helper: build a minimal sc_data list as prepare_sc_data() would return
make_sc_data <- function(sc_df, idx = NULL) {
  if (is.null(idx)) idx <- seq_len(nrow(sc_df))
  n <- length(idx)
  list(
    sc       = sc_df[idx, , drop = FALSE],
    X        = matrix(1.0, nrow = n, ncol = 1L),
    w        = rep(1.0, n),
    idx_keep = idx
  )
}

sc_data <- make_sc_data(sc)


# clean path ----

test_that("process_na_yz: clean path returns correct list names", {
  res <- process_na_yz(sc_data, y = "psa_level")
  expect_named(res, c("Y", "X", "w", "sc", "y_name", "zcol", "domain", "na_info"))
})

test_that("process_na_yz: Y is a numeric vector with no missing values", {
  res <- process_na_yz(sc_data, y = "psa_level")
  expect_type(res$Y, "double")
  expect_false(anyNA(res$Y))
})

test_that("process_na_yz: na_info contains correct fields", {
  res <- process_na_yz(sc_data, y = "psa_level")
  expect_named(res$na_info, c("na_action", "n_omitted", "n_used", "omitted_raw", "kept_raw"))
})

test_that("process_na_yz: n_omitted + n_used equals nrow of input sc", {
  sc_na              <- sc
  sc_na$psa_level[1:5] <- NA_real_
  sc_data_na         <- make_sc_data(sc_na)
  res                <- process_na_yz(sc_data_na, y = "psa_level")
  expect_equal(res$na_info$n_omitted + res$na_info$n_used, nrow(sc_na))
})

test_that("process_na_yz: no NAs in y gives n_omitted = 0 and n_used = nrow(sc)", {
  res <- process_na_yz(sc_data, y = "psa_level")
  expect_equal(res$na_info$n_omitted, 0L)
  expect_equal(res$na_info$n_used, nrow(sc))
})


# all-rows-missing guard ----

test_that("process_na_yz: errors when all y values are missing", {
  sc_all_na              <- sc
  sc_all_na$psa_level    <- NA_real_
  sc_data_all_na         <- make_sc_data(sc_all_na)
  expect_error(
    process_na_yz(sc_data_all_na, y = "psa_level"),
    "No complete cases remain after removing missing values in 'y' and 'zcol'.",
    fixed = TRUE
  )
})

test_that("process_na_yz: errors when all zcol values are missing", {
  sc_all_na         <- sc
  sc_all_na$grp_na  <- rep(NA_character_, nrow(sc))
  sc_data_all_na    <- make_sc_data(sc_all_na)
  expect_error(
    process_na_yz(sc_data_all_na, y = "psa_level", zcol = "grp_na"),
    "No complete cases remain after removing missing values in 'y' and 'zcol'.",
    fixed = TRUE
  )
})


# na.action checks ----

test_that("process_na_yz: na.fail errors when all y values are missing", {
  sc_na           <- sc
  sc_na$psa_level <- NA_real_
  sc_data_na      <- make_sc_data(sc_na)
  expect_error(
    process_na_yz(sc_data_na, y = "psa_level", na.action = stats::na.fail),
    "Missing values detected in `y` or `zcol`.",
    fixed = TRUE
  )
})

test_that("process_na_yz: na.pass is rejected even when all values are missing", {
  sc_na           <- sc
  sc_na$psa_level <- NA_real_
  sc_data_na      <- make_sc_data(sc_na)
  expect_error(
    process_na_yz(sc_data_na, y = "psa_level", na.action = stats::na.pass),
    "`na.pass` is not supported at estimate stage.",
    fixed = TRUE
  )
})

test_that("process_na_yz: na.fail errors when y has missing values", {
  sc_na              <- sc
  sc_na$psa_level[1] <- NA_real_
  sc_data_na         <- make_sc_data(sc_na)
  expect_error(
    process_na_yz(sc_data_na, y = "psa_level", na.action = stats::na.fail),
    "Missing values detected in `y` or `zcol`.",
    fixed = TRUE
  )
})

test_that("process_na_yz: na.fail passes when no missing values present", {
  res <- process_na_yz(sc_data, y = "psa_level", na.action = stats::na.fail)
  expect_equal(res$na_info$n_omitted, 0L)
})

test_that("process_na_yz: na.pass is rejected", {
  expect_error(
    process_na_yz(sc_data, y = "psa_level", na.action = stats::na.pass),
    "`na.pass` is not supported at estimate stage.",
    fixed = TRUE
  )
})

test_that("process_na_yz: unsupported custom na.action function is rejected", {
  expect_error(
    process_na_yz(sc_data, y = "psa_level", na.action = mean),
    "Unsupported na.action function",
    fixed = TRUE
  )
})


# na_obj class and na_info counts ----

test_that("process_na_yz: na.omit gives na_action class 'omit'", {
  sc_na              <- sc
  sc_na$psa_level[1:3] <- NA_real_
  sc_data_na         <- make_sc_data(sc_na)
  res                <- process_na_yz(sc_data_na, y = "psa_level",
                                      na.action = stats::na.omit)
  expect_equal(class(res$na_info$na_action), "omit")
})

test_that("process_na_yz: na.exclude gives na_action class 'exclude'", {
  sc_na              <- sc
  sc_na$psa_level[1:3] <- NA_real_
  sc_data_na         <- make_sc_data(sc_na)
  res                <- process_na_yz(sc_data_na, y = "psa_level",
                                      na.action = stats::na.exclude)
  expect_equal(class(res$na_info$na_action), "exclude")
})

test_that("process_na_yz: omitted_raw contains original indices of missing rows", {
  sc_na                    <- sc
  sc_na$psa_level[c(2, 4)] <- NA_real_
  sc_data_na               <- make_sc_data(sc_na)
  res                      <- process_na_yz(sc_data_na, y = "psa_level")
  expect_equal(as.integer(res$na_info$omitted_raw), c(2L, 4L))
})

test_that("process_na_yz: kept_raw and omitted_raw partition idx_keep", {
  sc_na              <- sc
  sc_na$psa_level[1:5] <- NA_real_
  sc_data_na         <- make_sc_data(sc_na)
  res                <- process_na_yz(sc_data_na, y = "psa_level")
  all_idx            <- sort(c(res$na_info$kept_raw, res$na_info$omitted_raw))
  expect_equal(all_idx, sc_data_na$idx_keep)
})

test_that("process_na_yz: kept_raw and omitted_raw are disjoint", {
  sc_na              <- sc
  sc_na$psa_level[1:5] <- NA_real_
  sc_data_na         <- make_sc_data(sc_na)
  res                <- process_na_yz(sc_data_na, y = "psa_level")
  expect_length(intersect(res$na_info$kept_raw, res$na_info$omitted_raw), 0L)
})


# output subsetting alignment ----

test_that("process_na_yz: Y, X, w, sc all have row count equal to n_used", {
  sc_na                <- sc
  sc_na$psa_level[1:5] <- NA_real_
  sc_data_na           <- make_sc_data(sc_na)
  res                  <- process_na_yz(sc_data_na, y = "psa_level")
  n_used               <- res$na_info$n_used
  expect_length(res$Y, n_used)
  expect_equal(nrow(res$X),  n_used)
  expect_length(res$w,  n_used)
  expect_equal(nrow(res$sc), n_used)
})

test_that("process_na_yz: Y values equal the non-missing y entries in original order", {
  sc_na                  <- sc
  sc_na$psa_level[c(2, 4)] <- NA_real_
  sc_data_na             <- make_sc_data(sc_na)
  res                    <- process_na_yz(sc_data_na, y = "psa_level")
  expect_equal(res$Y, sc_na$psa_level[!is.na(sc_na$psa_level)])
})


# pass-through fields ----

test_that("process_na_yz: y_name and zcol are stored in the return list", {
  res <- process_na_yz(sc_data, y = "psa_level", zcol = "race")
  expect_equal(res$y_name, "psa_level")
  expect_equal(res$zcol,   "race")
})

test_that("process_na_yz: zcol is NULL in the return list when zcol = NULL", {
  res <- process_na_yz(sc_data, y = "psa_level")
  expect_null(res$zcol)
})


# domain integration with standardize_zcol ----

test_that("process_na_yz: zcol = NULL gives domain$mode = 'overall'", {
  res <- process_na_yz(sc_data, y = "psa_level")
  expect_equal(res$domain$mode, "overall")
})

test_that("process_na_yz: factor zcol gives domain$mode = 'factor' with matching labels", {
  res <- process_na_yz(sc_data, y = "psa_level", zcol = "race")
  expect_equal(res$domain$mode, "factor")
  expect_equal(
    sort(res$domain$labels),
    sort(paste0("race = ", levels(droplevels(factor(sc$race)))))
  )
})

test_that("process_na_yz: binary numeric zcol gives domain$mode = 'binary'", {
  sc_bin              <- sc
  sc_bin$flag         <- rep(c(0L, 1L), length.out = nrow(sc))
  sc_data_bin         <- make_sc_data(sc_bin)
  res                 <- process_na_yz(sc_data_bin, y = "psa_level", zcol = "flag")
  expect_equal(res$domain$mode, "binary")
})


# na_obj when no NAs are present ----

test_that("process_na_yz: na.omit with no NAs gives na_action = NULL", {
  res <- process_na_yz(sc_data, y = "psa_level", na.action = stats::na.omit)
  expect_null(res$na_info$na_action)
})

test_that("process_na_yz: na.exclude with no NAs gives na_action = NULL", {
  res <- process_na_yz(sc_data, y = "psa_level", na.action = stats::na.exclude)
  expect_null(res$na_info$na_action)
})

#' Finalize and assemble the `pw_fit` return object
#'
#' Attaches the reconstructed sample data frame, the pseudo-weight vector, and
#' NA-handling metadata to `result`, then assigns the `"pw_fit"`
#' class so that downstream S3 methods can dispatch on it.
#'
#' @param result A list accumulating outputs from the estimation pipeline.
#' @param sc_out The reconstructed sample data frame produced by
#'   `reconstruct_sc_output()`.
#' @param sc0 The original (pre-NA-removal) sample data frame, stored for
#'   diagnostic access via `result$internal$raw_sc`.
#' @param sc_wname Name of the pseudo-weight column in `sc_out`.
#' @param na_mode Character string describing how NAs were handled; stored in
#'   `result$internal$na`.
#' @param keep_sc Logical vector of retained rows; stored in
#'   `result$internal$na`.
#' @param na_action_obj The `na.action` attribute from NA removal; stored
#'   in `result$internal$na`.
#'
#' @return `result` with `sc_updated`, `pseudo_weights`, and
#'   `internal` fields populated, and class set to `"pw_fit"`.
#'
#' @keywords internal
finalize_pw_fit <- function(result, sc_out, sc0, sc_wname, na_mode, keep_sc, na_action_obj) {

  result$sc_updated     <- sc_out
  # For na_mode "exclude", this expands pseudo_weights to full length with NAs
  # for dropped rows; for other modes the value is unchanged from Step 5.
  result$pseudo_weights <- sc_out[[sc_wname]]

  result$internal$raw_sc <- sc0
  result$internal$na <- list(
    na_mode       = na_mode,
    keep_sc       = keep_sc,
    na_action_obj = na_action_obj
  )

  class(result) <- "pw_fit"

  result
}

# test-build_s7_finalize_output.R
# Tests for finalize_pw_fit()
#
# Function signature:
#   finalize_pw_fit(result, sc_out, sc0, sc_wname, na_mode, keep_sc, na_action_obj)
#
# The function:
#   - sets result$sc_updated  = sc_out
#   - sets result$pseudo_weights = sc_out[[sc_wname]]
#   - sets result$internal$raw_sc = sc0
#   - sets result$internal$na = list(na_mode, keep_sc, na_action_obj)
#   - assigns class "pw_fit"
#   - keeps any pre-existing fields in result

# setup ----

set.seed(2026)
n0 <- 10L

sc0 <- data.frame(
  id = seq_len(n0),
  x1 = c(1, 2, NA, 4, 5, 6, NA, 8, 9, 10),
  x2 = rnorm(n0)
)

keep_sc  <- !is.na(sc0$x1)        # 8 kept rows
excluded <- which(!keep_sc)
na_action_obj <- structure(excluded, class = "omit")
sc_wname <- "pseudo_wts"

# sc_out for "omit": 8 rows with weight column
set.seed(7)
w_fit_omit  <- runif(sum(keep_sc), 10, 100)
sc_out_omit <- sc0[keep_sc, ]
sc_out_omit[[sc_wname]] <- w_fit_omit

# sc_out for "exclude": all 10 rows, NA at excluded positions
w_full_excl <- rep(NA_real_, n0)
w_full_excl[keep_sc] <- runif(sum(keep_sc), 10, 100)
sc_out_excl <- sc0
sc_out_excl[[sc_wname]] <- w_full_excl

# sc_out for "fail" / "pass": all 10 rows, no NAs
w_fit_full  <- runif(n0, 10, 100)
sc_out_full <- sc0
sc_out_full[[sc_wname]] <- w_fit_full

# base result list representing earlier pipeline outputs
base_result <- list(
  method       = "calibration",
  coefficients = c("(Intercept)" = 0.5, x1 = 0.1),
  solver_diagnostics = list(termcd = 1L, fmax = 1e-10),
  internal     = list(Xc = matrix(1:6, 3, 2))
)


# class assignment ----

test_that("finalize_pw_fit: returned object has class 'pw_fit'", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_s3_class(out, "pw_fit")
})


# result$sc_updated ----

test_that("finalize_pw_fit: sc_updated equals sc_out passed in", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_identical(out$sc_updated, sc_out_omit)
})

test_that("finalize_pw_fit: sc_updated is a data frame", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_s3_class(out$sc_updated, "data.frame")
})


# result$pseudo_weights ----

test_that("finalize_pw_fit omit: pseudo_weights equals sc_out[[sc_wname]]", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_equal(out$pseudo_weights, sc_out_omit[[sc_wname]])
})

test_that("finalize_pw_fit omit: pseudo_weights has length sum(keep_sc)", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_length(out$pseudo_weights, sum(keep_sc))
})

test_that("finalize_pw_fit omit: pseudo_weights has no NA", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_false(anyNA(out$pseudo_weights))
})

test_that("finalize_pw_fit exclude: pseudo_weights has length nrow(sc0)", {
  out <- finalize_pw_fit(base_result, sc_out_excl, sc0, sc_wname,
                         "exclude", keep_sc, na_action_obj)
  expect_length(out$pseudo_weights, n0)
})

test_that("finalize_pw_fit exclude: pseudo_weights is NA at excluded rows", {
  out <- finalize_pw_fit(base_result, sc_out_excl, sc0, sc_wname,
                         "exclude", keep_sc, na_action_obj)
  expect_true(all(is.na(out$pseudo_weights[excluded])))
})

test_that("finalize_pw_fit exclude: pseudo_weights non-NA count equals sum(keep_sc)", {
  out <- finalize_pw_fit(base_result, sc_out_excl, sc0, sc_wname,
                         "exclude", keep_sc, na_action_obj)
  expect_equal(sum(!is.na(out$pseudo_weights)), sum(keep_sc))
})

test_that("finalize_pw_fit fail: pseudo_weights has length nrow(sc0) and no NA", {
  out <- finalize_pw_fit(base_result, sc_out_full, sc0, sc_wname,
                         "fail", rep(TRUE, n0), NULL)
  expect_length(out$pseudo_weights, n0)
  expect_false(anyNA(out$pseudo_weights))
})

test_that("finalize_pw_fit pass: pseudo_weights has length nrow(sc0) and no NA", {
  out <- finalize_pw_fit(base_result, sc_out_full, sc0, sc_wname,
                         "pass", rep(TRUE, n0), NULL)
  expect_length(out$pseudo_weights, n0)
  expect_false(anyNA(out$pseudo_weights))
})


# result$internal$raw_sc ----

test_that("finalize_pw_fit: internal$raw_sc equals sc0", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_identical(out$internal$raw_sc, sc0)
})

test_that("finalize_pw_fit: internal$raw_sc is the original unmodified sc0", {
  # raw_sc should have all n0 rows regardless of na_mode
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_equal(nrow(out$internal$raw_sc), n0)
})


# result$internal$na ----

test_that("finalize_pw_fit: internal$na contains na_mode, keep_sc, na_action_obj", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expected_names <- c("na_mode", "keep_sc", "na_action_obj")
  expect_true(all(expected_names %in% names(out$internal$na)))
})

test_that("finalize_pw_fit: internal$na$na_mode is stored correctly", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_equal(out$internal$na$na_mode, "omit")
})

test_that("finalize_pw_fit: internal$na$keep_sc is stored correctly", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_identical(out$internal$na$keep_sc, keep_sc)
})

test_that("finalize_pw_fit: internal$na$na_action_obj is stored correctly", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_identical(out$internal$na$na_action_obj, na_action_obj)
})

test_that("finalize_pw_fit: internal$na$na_action_obj is NULL when passed NULL", {
  out <- finalize_pw_fit(base_result, sc_out_full, sc0, sc_wname,
                         "pass", rep(TRUE, n0), NULL)
  expect_null(out$internal$na$na_action_obj)
})

test_that("finalize_pw_fit: internal$na$na_mode = 'exclude' stored correctly", {
  out <- finalize_pw_fit(base_result, sc_out_excl, sc0, sc_wname,
                         "exclude", keep_sc, na_action_obj)
  expect_equal(out$internal$na$na_mode, "exclude")
})


# pre-existing fields kept ----

test_that("finalize_pw_fit: pre-existing method field is kept", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_equal(out$method, base_result$method)
})

test_that("finalize_pw_fit: pre-existing coefficients field is kept", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_equal(out$coefficients, base_result$coefficients)
})

test_that("finalize_pw_fit: pre-existing solver_diagnostics field is kept", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_equal(out$solver_diagnostics, base_result$solver_diagnostics)
})

test_that("finalize_pw_fit: pre-existing internal$Xc is kept after adding raw_sc and na", {
  out <- finalize_pw_fit(base_result, sc_out_omit, sc0, sc_wname,
                         "omit", keep_sc, na_action_obj)
  expect_identical(out$internal$Xc, base_result$internal$Xc)
})

#' Process the participation model formula and build model matrices
#'
#' Converts `p_formula` and the cleaned data (`sc`, `sp`) into model
#' matrices ready for participation model estimation. Also performs
#' cumulative precalibration of reference survey weights in the
#' multi-reference case when `Pre.calibration = TRUE`.
#'
#' For both the one-reference and multi-reference cases the function:
#' \enumerate{
#'   \item Checks that all formula variables exist in `sc` and each `sp`.
#'   \item Verifies that factor levels of each variable are identical
#'     between `sc` and the corresponding `sp` (mismatched levels would
#'     silently produce different model matrix columns).
#'   \item Builds model matrices via `stats::model.matrix()` and drops
#'     the implicit intercept column.
#'   \item Checks that `sc` and `sp` produce the same number of model
#'     matrix columns, and aligns their column names.
#' }
#'
#' In the one-reference case `sc_new` contains only the model matrix
#' columns; `sp_new` contains the same columns plus the sampling weight
#' column.
#'
#' In the multi-reference case `sc_new` is the union of model matrix
#' columns across all reference surveys (duplicates removed by value
#' comparison). Each `sp_new[[j]]` contains the model matrix columns for
#' formula `j` plus the corresponding weight column. If
#' `Pre.calibration = TRUE` and there are at least two reference surveys,
#' `precal_cumulative_order()` is called to sequentially calibrate the
#' reference survey weights before estimation. The calibration variables
#' are the intersection of the variables appearing across all formulas in
#' `p_formula`, ensuring precalibration aligns reference surveys on
#' exactly the dimensions used in the participation models.
#'
#' @param sc A data frame. The nonprobability sample after NA processing.
#' @param sp A data frame (one-reference) or a named list of data frames
#'   (multi-reference). Reference survey analysis data after NA processing.
#' @param weight A single character string (one-reference) or character
#'   vector of the same length as `sp` (multi-reference) giving the
#'   sampling weight column name in each reference survey.
#' @param p_formula A one-sided formula (one-reference) or a list of
#'   one-sided formulas (multi-reference).
#' @param Pre.calibration Logical. Multi-reference case only. If `TRUE`,
#'   cumulative precalibration is applied to the reference survey weights
#'   via `precal_cumulative_order()`. Ignored in the one-reference case.
#' @param sp_order Character string, either `"size"` or `"given"`.
#'   Controls the processing order in `precal_cumulative_order()`.
#'   Ignored in the one-reference case.
#' @param verbose Logical. If `TRUE` and precalibration is skipped, a
#'   message is emitted recommending precalibration.
#'
#' @return A list with four elements:
#' \describe{
#'   \item{`sc`}{Data frame of model matrix columns for `sc`.}
#'   \item{`sp`}{Data frame (one-reference) or named list of data frames
#'     (multi-reference) of model matrix columns plus weight column(s)
#'     for each reference survey.}
#'   \item{`vars`}{Character vector of model matrix column names
#'     (one-reference), or a list of such vectors (multi-reference).}
#'   \item{`log_messages`}{Character vector of diagnostic messages from
#'     precalibration, or an empty vector in the one-reference case.}
#' }
#'
#' @keywords internal
#' @noRd
process_p_formula <- function(
    sc, sp, weight, p_formula,
    Pre.calibration = TRUE,
    sp_order = "size",
    verbose = FALSE
) {
  # -------------------------------------------------------------------
  # One-reference case
  # -------------------------------------------------------------------
  if (is.data.frame(sp)) {

    if (!inherits(p_formula, "formula")) {
      stop("'p_formula' must be a formula for one-reference case.", call. = FALSE)
    }

    vars_in_formula <- all.vars(p_formula)
    missing_sc <- setdiff(vars_in_formula, names(sc))
    missing_sp <- setdiff(vars_in_formula, names(sp))
    if (length(missing_sc) > 0 || length(missing_sp) > 0) {
      stop(paste0(
        "Missing variable(s): ",
        paste(unique(c(missing_sc, missing_sp)), collapse = ", ")
      ), call. = FALSE)
    }

    .check_factor_levels(sc, sp, vars_in_formula, ref_label = "sp")

    # Build model matrices (drop implicit intercept)
    Xc <- stats::model.matrix(p_formula, data = sc)
    Xp <- stats::model.matrix(p_formula, data = sp)
    if ("(Intercept)" %in% colnames(Xc)) Xc <- Xc[, -1, drop = FALSE]
    if ("(Intercept)" %in% colnames(Xp)) Xp <- Xp[, -1, drop = FALSE]

    vars <- colnames(Xc)
    if (ncol(Xp) != length(vars)) {
      stop(
        "sc and sp produced different model matrix columns -- check factor levels in p_formula variables.",
        call. = FALSE
      )
    }
    colnames(Xp) <- vars

    sc_new <- as.data.frame(Xc)

    if (!(weight %in% names(sp))) stop("Weight column not found in sp.", call. = FALSE)
    sp_new <- as.data.frame(Xp)
    sp_new[[weight]] <- sp[[weight]]


    return(list(
      sc = sc_new,
      sp = sp_new,
      vars = vars,
      log_messages = character()
    ))
  }

  # -------------------------------------------------------------------
  # Multi-reference case
  # -------------------------------------------------------------------
  if (!is.list(sp)) {
    stop("process_p_formula(): sp must be a data.frame or a list of data.frames.", call. = FALSE)
  }

  if (!(is.list(p_formula) && all(vapply(p_formula, inherits, logical(1), "formula")))) {
    stop("For multi-reference, 'p_formula' must be a list of formulas.", call. = FALSE)
  }
  if (length(p_formula) != length(sp)) {
    stop("For multi-reference, 'p_formula' must have the same length as 'sp'.", call. = FALSE)
  }
  if (!(is.character(weight) && length(weight) == length(sp))) {
    stop("For multi-reference, 'weight' must be a character vector with length = length(sp).", call. = FALSE)
  }


  sp_new <- vector("list", length(sp))
  vars_list <- vector("list", length(sp))
  sc_new <- NULL

  for (j in seq_along(sp)) {

    fml <- p_formula[[j]]
    vars_in_formula <- all.vars(fml)

    missing_sc <- setdiff(vars_in_formula, names(sc))
    missing_sp <- setdiff(vars_in_formula, names(sp[[j]]))
    if (length(missing_sc) > 0 || length(missing_sp) > 0) {
      stop(paste0(
        "For reference ", j, ", missing vars: ",
        paste(unique(c(missing_sc, missing_sp)), collapse = ", ")
      ), call. = FALSE)
    }

    .check_factor_levels(sc, sp[[j]], vars_in_formula, ref_label = paste0("sp[[", j, "]]"))

    Xc <- stats::model.matrix(fml, data = sc)
    Xp <- stats::model.matrix(fml, data = sp[[j]])
    if ("(Intercept)" %in% colnames(Xc)) Xc <- Xc[, -1, drop = FALSE]
    if ("(Intercept)" %in% colnames(Xp)) Xp <- Xp[, -1, drop = FALSE]

    if (ncol(Xp) != ncol(Xc)) {
      stop(sprintf(
        "Reference %d: sc and sp[[%d]] produced different model matrix columns -- check factor levels in p_formula variables.",
        j, j
      ), call. = FALSE)
    }
    colnames(Xp) <- colnames(Xc)

    # Merge covariates into sc_new without duplicates
    if (j == 1L) {
      sc_new <- as.data.frame(Xc)
    } else {
      for (nm in colnames(Xc)) {
        new_col <- Xc[, nm]
        duplicate <- any(vapply(sc_new, function(old_col) {
          if (is.numeric(old_col) && is.numeric(new_col)) {
            isTRUE(all.equal(old_col, new_col, tolerance = 1e-12))
          } else {
            identical(old_col, new_col)
          }
        }, logical(1)))
        if (!duplicate) sc_new[[nm]] <- new_col
      }
    }

    # Build sp_new[[j]]
    wj <- weight[j]

    if (!(wj %in% names(sp[[j]]))) {
      stop(sprintf("Weight column '%s' not found in sp[%d].", wj, j), call. = FALSE)
    }

    spj <- as.data.frame(Xp)
    spj[[wj]] <- sp[[j]][[wj]]

    sp_new[[j]] <- spj
    vars_list[[j]] <- colnames(Xp)
  }

  names(sp_new) <- if (!is.null(names(sp)) && all(nzchar(names(sp)))) {
    names(sp)
  } else {
    paste0("sp[[", seq_along(sp), "]]")
  }

  log_messages <- character()

  if (Pre.calibration && length(sp_new) > 1) {
    shared_vars <- Reduce(intersect, lapply(p_formula, all.vars))
    sp_for_precal <- lapply(seq_along(sp), function(j) {
      sp[[j]][, c(shared_vars, weight[j]), drop = FALSE]
    })

    out <- precal_cumulative_order(
      sp_raw   = sp_for_precal,
      sp_new   = sp_new,
      weight   = weight,
      sp_order = sp_order
    )
    sp_new <- out$sp_new
    log_messages <- c(log_messages, out$log_messages)
  } else {
    msg <- "Precalibration is recommended."
    log_messages <- c(log_messages, msg)
    if (verbose) message(msg)
  }

  return(list(
    sc = sc_new,
    sp = sp_new,
    vars = vars_list,
    log_messages = log_messages
  ))
}

# Setup ----

data(sc)
data(sp1)
data(sp2)


# one-reference: return structure ----

test_that("process_p_formula (one-ref): returns sc, sp, vars, log_messages", {
  res <- process_p_formula(sc, sp1, weight = "wts_sp1", p_formula = ~agecat)

  expect_named(res, c("sc", "sp", "vars", "log_messages"))
  expect_s3_class(res$sc, "data.frame")
  expect_s3_class(res$sp, "data.frame")
  expect_type(res$vars, "character")
  expect_type(res$log_messages, "character")
})

test_that("process_p_formula (one-ref): sc and sp have same column names", {
  res <- process_p_formula(sc, sp1, weight = "wts_sp1", p_formula = ~agecat)

  expect_equal(setdiff(colnames(res$sc), colnames(res$sp)), character(0))
})

test_that("process_p_formula (one-ref): intercept not in output columns", {
  res <- process_p_formula(sc, sp1, weight = "wts_sp1", p_formula = ~agecat)

  expect_false("(Intercept)" %in% colnames(res$sc))
  expect_false("(Intercept)" %in% colnames(res$sp))
})

test_that("process_p_formula (one-ref): weight column attached to sp only", {
  res <- process_p_formula(sc, sp1, weight = "wts_sp1", p_formula = ~agecat)

  expect_true("wts_sp1" %in% colnames(res$sp))
  expect_false("wts_sp1" %in% colnames(res$sc))
})

test_that("process_p_formula (one-ref): sc row count unchanged", {
  res <- process_p_formula(sc, sp1, weight = "wts_sp1", p_formula = ~agecat)

  expect_equal(nrow(res$sc), nrow(sc))
})

test_that("process_p_formula (one-ref): sp row count unchanged", {
  res <- process_p_formula(sc, sp1, weight = "wts_sp1", p_formula = ~agecat)

  expect_equal(nrow(res$sp), nrow(sp1))
})


# one-reference: input validation ----

test_that("process_p_formula (one-ref): errors when p_formula is not a formula", {
  expect_error(
    process_p_formula(sc, sp1, weight = "wts_sp1", p_formula = "agecat"),
    "'p_formula' must be a formula for one-reference case.",
    fixed = TRUE
  )
})

test_that("process_p_formula (one-ref): errors when variable missing from sc", {
  expect_error(
    process_p_formula(sc, sp1, weight = "wts_sp1", p_formula = ~only_in_sp),
    "Missing variable(s)",
    fixed = TRUE
  )
})

test_that("process_p_formula (one-ref): errors when variable missing from sp", {
  sc_extra <- sc
  sc_extra$only_in_sc <- 1L

  expect_error(
    process_p_formula(sc_extra, sp1, weight = "wts_sp1", p_formula = ~only_in_sc),
    "Missing variable(s)",
    fixed = TRUE
  )
})

test_that("process_p_formula (one-ref): errors when weight column not in sp", {
  expect_error(
    process_p_formula(sc, sp1, weight = "not_exist", p_formula = ~agecat),
    "Weight column not found in sp.",
    fixed = TRUE
  )
})

test_that("process_p_formula (one-ref): errors when sp is not data.frame or list", {
  expect_error(
    process_p_formula(sc, "not_a_df", weight = "wts_sp1", p_formula = ~agecat),
    "sp must be a data.frame or a list of data.frames.",
    fixed = TRUE
  )
})

test_that("process_p_formula (one-ref): errors when factor levels differ between sc and sp", {
  sp_bad <- sp1
  sp_bad$agecat <- factor(sp_bad$agecat, levels = levels(sp_bad$agecat)[-1])

  expect_error(
    process_p_formula(sc, sp_bad, weight = "wts_sp1", p_formula = ~agecat),
    "Factor levels of 'agecat' differ between sc and sp",
    fixed = TRUE
  )
})

test_that("process_p_formula (one-ref): factor level error message names the variable", {
  sp_bad <- sp1
  sp_bad$agecat <- factor(sp_bad$agecat, levels = levels(sp_bad$agecat)[-1])

  err <- tryCatch(
    process_p_formula(sc, sp_bad, weight = "wts_sp1", p_formula = ~agecat),
    error = function(e) conditionMessage(e)
  )

  expect_match(err, "agecat")
  expect_match(err, "sc :")
  expect_match(err, "sp :")
})


# multi-reference: return structure ----

test_that("process_p_formula (multi-ref): returns sc, sp, vars, log_messages", {
  res <- process_p_formula(
    sc, list(sp1, sp2),
    weight        = c("wts_sp1", "wts_sp2"),
    p_formula     = list(~agecat, ~agecat),
    Pre.calibration = FALSE
  )

  expect_named(res, c("sc", "sp", "vars", "log_messages"))
  expect_s3_class(res$sc, "data.frame")
  expect_true(is.list(res$sp))
  expect_length(res$sp, 2)
  expect_true(is.list(res$vars))
  expect_length(res$vars, 2)
})

test_that("process_p_formula (multi-ref): each sp_new[[j]] has its weight column", {
  res <- process_p_formula(
    sc, list(sp1, sp2),
    weight        = c("wts_sp1", "wts_sp2"),
    p_formula     = list(~agecat, ~agecat),
    Pre.calibration = FALSE
  )

  expect_true("wts_sp1" %in% colnames(res$sp[[1]]))
  expect_true("wts_sp2" %in% colnames(res$sp[[2]]))
})

test_that("process_p_formula (multi-ref): sc_new deduplicates shared columns", {
  res <- process_p_formula(
    sc, list(sp1, sp2),
    weight        = c("wts_sp1", "wts_sp2"),
    p_formula     = list(~agecat, ~agecat),
    Pre.calibration = FALSE
  )

  agecat_cols <- grep("^agecat", colnames(res$sc), value = TRUE)
  expect_equal(length(agecat_cols), length(unique(agecat_cols)))
})

test_that("process_p_formula (multi-ref): sp names kept from named list", {
  res <- process_p_formula(
    sc, list(survey_a = sp1, survey_b = sp2),
    weight        = c("wts_sp1", "wts_sp2"),
    p_formula     = list(~agecat, ~agecat),
    Pre.calibration = FALSE
  )

  expect_equal(names(res$sp), c("survey_a", "survey_b"))
})

test_that("process_p_formula (multi-ref): Pre.calibration = FALSE adds log message", {
  res <- process_p_formula(
    sc, list(sp1, sp2),
    weight        = c("wts_sp1", "wts_sp2"),
    p_formula     = list(~agecat, ~agecat),
    Pre.calibration = FALSE
  )

  expect_true(any(grepl("Precalibration is recommended", res$log_messages)))
})


# multi-reference: input validation ----

test_that("process_p_formula (multi-ref): errors when p_formula is not a list of formulas", {
  expect_error(
    process_p_formula(
      sc, list(sp1, sp2),
      weight        = c("wts_sp1", "wts_sp2"),
      p_formula     = ~agecat,
      Pre.calibration = FALSE
    ),
    "For multi-reference, 'p_formula' must be a list of formulas.",
    fixed = TRUE
  )
})

test_that("process_p_formula (multi-ref): errors when p_formula length mismatches sp", {
  expect_error(
    process_p_formula(
      sc, list(sp1, sp2),
      weight        = c("wts_sp1", "wts_sp2"),
      p_formula     = list(~agecat),
      Pre.calibration = FALSE
    ),
    "For multi-reference, 'p_formula' must have the same length as 'sp'.",
    fixed = TRUE
  )
})

test_that("process_p_formula (multi-ref): errors when weight length mismatches sp", {
  expect_error(
    process_p_formula(
      sc, list(sp1, sp2),
      weight        = "wts_sp1",
      p_formula     = list(~agecat, ~agecat),
      Pre.calibration = FALSE
    ),
    "For multi-reference, 'weight' must be a character vector with length = length(sp).",
    fixed = TRUE
  )
})

test_that("process_p_formula (multi-ref): errors when weight column missing from sp[[j]]", {
  expect_error(
    process_p_formula(
      sc, list(sp1, sp2),
      weight        = c("wts_sp1", "not_exist"),
      p_formula     = list(~agecat, ~agecat),
      Pre.calibration = FALSE
    ),
    "Weight column 'not_exist' not found in sp[2].",
    fixed = TRUE
  )
})

test_that("process_p_formula (multi-ref): errors when variable missing from sp[[j]]", {
  sc_extra <- sc
  sc_extra$only_in_sc <- 1L

  expect_error(
    process_p_formula(
      sc_extra, list(sp1, sp2),
      weight        = c("wts_sp1", "wts_sp2"),
      p_formula     = list(~agecat, ~only_in_sc),
      Pre.calibration = FALSE
    ),
    "For reference 2, missing vars",
    fixed = TRUE
  )
})

test_that("process_p_formula (multi-ref): errors when factor levels differ in sp[[j]]", {
  sp2_bad <- sp2
  sp2_bad$agecat <- factor(sp2_bad$agecat, levels = levels(sp2_bad$agecat)[-1])

  expect_error(
    process_p_formula(
      sc, list(sp1, sp2_bad),
      weight        = c("wts_sp1", "wts_sp2"),
      p_formula     = list(~agecat, ~agecat),
      Pre.calibration = FALSE
    ),
    "Factor levels of 'agecat' differ between sc and sp[[2]]",
    fixed = TRUE
  )
})

#' Calibration estimating equations
#' @noRd
raking_fn <- function(beta, Xc, f_p, ...) {
  eta <- as.vector(Xc %*% beta)

  if (any(!is.finite(eta))) {
    stop(
      "Calibration failed: linear predictor became non-finite.",
      call. = FALSE
    )
  }

  if (any(eta < -log(.Machine$double.xmax))) {
    stop(
      paste(
        "Calibration failed: exp(-eta) overflow.",
        "This indicates an unstable coefficient value."
      ),
      call. = FALSE
    )
  }

  wts_sc <- as.vector(exp(-eta))

  if (any(!is.finite(wts_sc))) {
    stop(
      "Calibration failed: pseudo-weights became non-finite.",
      call. = FALSE
    )
  }

  f_c <- colSums(wts_sc * Xc)

  as.numeric(f_c - f_p)
}

#' Jacobian of the calibration estimating equations
#' @noRd
raking_jac <- function(beta, Xc, f_p,
                       label = "One_Ref_Calibration",
                       tol_singular = 1e-7,
                       ...) {
  eta <- as.vector(Xc %*% beta)

  if (any(!is.finite(eta))) {
    stop(
      sprintf("[%s] Calibration failed: linear predictor became non-finite when computing Jacobian.", label),
      call. = FALSE
    )
  }

  if (any(eta < -log(.Machine$double.xmax))) {
    stop(
      sprintf(
        paste(
          "[%s] Calibration failed: exp(-eta) overflow when computing Jacobian.",
          "This usually indicates an unstable coefficient value."
        ),
        label
      ),
      call. = FALSE
    )
  }

  wts_sc <- as.vector(exp(-eta))

  if (any(!is.finite(wts_sc))) {
    stop(
      sprintf("[%s] Calibration failed: pseudo-weights became non-finite when computing Jacobian.", label),
      call. = FALSE
    )
  }

  g <- colSums(wts_sc * Xc) - f_p
  J <- -t(wts_sc * Xc) %*% Xc

  check_estimating_system(
    J = J,
    g = g,
    label = label,
    tol_singular = tol_singular
  )

  J
}

#' Starting values for the coefficients of the calibration participation model
#' @noRd
raking_start <- function(Xc, f_p) {
  p_dim <- ncol(Xc)
  beta_start <- rep(0, p_dim)

  wts0 <- as.vector(exp(-Xc %*% beta_start))
  f_c0 <- colSums(wts0 * Xc)

  if (!is.finite(f_p[1]) || !is.finite(f_c0[1]) || f_p[1] <= 0 || f_c0[1] <= 0) {
    stop(
      "Calibration failed: cannot construct initial intercept because intercept totals are invalid.",
      call. = FALSE
    )
  }

  beta_start[1] <- -log(f_p[1] / f_c0[1])

  if (any(!is.finite(beta_start))) {
    stop(
      "Calibration failed: initial coefficient vector contains non-finite values.",
      call. = FALSE
    )
  }

  beta_start
}

# test-build_s5_one.R
# Tests for Step 5 one-reference builders:
#   raking_fn / raking_jac / raking_start / raking_build
#   alp_fn    / alp_jac    / alp_start    / alp_build
#   clw_fn    / clw_jac    / clw_start    / clw_build
#   ipwm_one_build  (dispatcher)

# Shared test data ----
# Plain numeric covariates: no factor expansion.
# n_sc = 200, n_sp = 300, p = 3 (intercept + x1 + x2).

set.seed(42)

n_sc <- 200L
n_sp <- 300L

sc_df <- data.frame(
  x1 = rnorm(n_sc),
  x2 = rnorm(n_sc)
)

sp_df <- data.frame(
  x1     = rnorm(n_sp),
  x2     = rnorm(n_sp),
  sp_wts = runif(n_sp, 10, 100)
)

vars   <- c("x1", "x2")
Xc     <- add_intercept(vars, sc_df)      # 200 x 3
Xp     <- add_intercept(vars, sp_df)      # 300 x 3
wts_sp <- sp_df$sp_wts
f_p    <- colSums(wts_sp * Xp)

sp_des <- survey::svydesign(ids = ~1, weights = ~sp_wts, data = sp_df)
ctrl   <- pw_solver_control()




# Generalized estimating equation: one reference sample ----
#
# SIM paper equation (2):
#
#   sum_sc w(x_i, beta) h(x_i, beta)
#     - sum_sp d_j h(x_j, beta) = 0
#
#
# Method-specific choices:
#
# Calibration:
#   w(x, beta) = exp(-x beta)
#   h(x, beta) = x
#
# ALP:
#   w(x, beta) = exp(-x beta)
#   h(x, beta) = [1 + w(x, beta)]^{-1} x
#              = expit(x beta) x
#
#   This gives:
#   sum_sc (1 - p_i) x_i - sum_sp d_j p_j x_j = 0
#
# CLW:
#   w(x, beta) = 1 + exp(-x beta)
#   h(x, beta) = [w(x, beta)]^{-1} x
#
#   This gives:
#   sum_sc x_i - sum_sp d_j expit(x_j beta) x_j = 0

general_ee_one_ref <- function(beta, Xc, Xp, wts_sp, method) {
  eta_c <- drop(Xc %*% beta)
  eta_p <- drop(Xp %*% beta)

  if (method == "calibration") {
    w_c <- exp(-eta_c)

    ee <- colSums(w_c * Xc) - colSums(wts_sp * Xp)

  } else if (method == "alp") {
    p_c <- expit(eta_c)
    p_p <- expit(eta_p)

    w_c <- exp(-eta_c)

    ee <- colSums(w_c * p_c * Xc) - colSums(wts_sp * p_p * Xp)

  } else if (method == "clw") {
    pi_p <- expit(eta_p)

    ee <- colSums(Xc) - colSums(wts_sp * pi_p * Xp)

  } else {
    stop("Unknown method.", call. = FALSE)
  }

  ee
}

# raking_fn ----

test_that("raking_fn: returns a finite numeric vector of length p at beta = 0", {
  beta0 <- rep(0, ncol(Xc))
  out   <- raking_fn(beta0, Xc = Xc, f_p = f_p)

  expect_type(out, "double")
  expect_length(out, ncol(Xc))
  expect_true(all(is.finite(out)))
})

test_that("raking_fn: non-intercept values differ from zero at beta = 0", {
  # At beta = 0, weights = 1 for all rows; f_c = colSums(Xc) != f_p in general
  beta0 <- rep(0, ncol(Xc))
  out   <- raking_fn(beta0, Xc = Xc, f_p = f_p)

  expect_true(any(out != 0))
})


# raking_jac ----

test_that("raking_jac: returns a square finite matrix at beta = 0", {
  beta0 <- rep(0, ncol(Xc))
  J     <- raking_jac(beta0, Xc = Xc, f_p = f_p)

  expect_true(is.matrix(J))
  expect_equal(nrow(J), ncol(Xc))
  expect_equal(ncol(J), ncol(Xc))
  expect_true(all(is.finite(J)))
})

test_that("raking_jac: diagonal values are all negative at beta = 0", {
  # J = -t(wts_sc * Xc) %*% Xc; diagonal is -colSums(wts_sc * Xc^2) < 0
  beta0 <- rep(0, ncol(Xc))
  J     <- raking_jac(beta0, Xc = Xc, f_p = f_p)

  expect_true(all(diag(J) < 0))
})


# raking_start ----

test_that("raking_start: returns a finite numeric vector of length p", {
  bs <- raking_start(Xc = Xc, f_p = f_p)

  expect_type(bs, "double")
  expect_length(bs, ncol(Xc))
  expect_true(all(is.finite(bs)))
})

test_that("raking_start: intercept initialized to -log(sum_wts / n_sc)", {
  # At beta = 0: wts0 = rep(1, n_sc), f_c0[1] = n_sc, f_p[1] = sum(wts_sp)
  bs       <- raking_start(Xc = Xc, f_p = f_p)
  expected <- -log(sum(wts_sp) / n_sc)

  expect_equal(bs[1], expected, tolerance = 1e-10)
})

test_that("raking_start: non-intercept values are zero", {
  bs <- raking_start(Xc = Xc, f_p = f_p)

  expect_equal(bs[-1], rep(0, ncol(Xc) - 1))
})


# raking_build ----

test_that("raking_build: return list has expected names", {
  out <- raking_build(vars, sc_df, sp_df, sp_des,
                      wts.col = "sp_wts", control = ctrl)

  expect_named(out, c("weights", "coefficients", "solver_diagnostics",
                      "log_messages", "internal"))
})

test_that("raking_build: pseudo-weights are positive and finite", {
  out <- raking_build(vars, sc_df, sp_df, sp_des,
                      wts.col = "sp_wts", control = ctrl)

  expect_true(all(is.finite(out$weights)))
  expect_true(all(out$weights > 0))
})

test_that("raking_build: pseudo-weights have length nrow(sc)", {
  out <- raking_build(vars, sc_df, sp_df, sp_des,
                      wts.col = "sp_wts", control = ctrl)

  expect_length(out$weights, n_sc)
})

test_that("raking_build: calibration equations satisfied (max|eq| < 1e-6)", {
  out <- raking_build(vars, sc_df, sp_df, sp_des,
                      wts.col = "sp_wts", control = ctrl)
  res <- raking_fn(out$coefficients, Xc = Xc, f_p = f_p)

  expect_lt(max(abs(res)), 1e-6)
})

test_that("raking_build: coefficients are named by Xc column names", {
  out <- raking_build(vars, sc_df, sp_df, sp_des,
                      wts.col = "sp_wts", control = ctrl)

  expect_named(out$coefficients, colnames(Xc))
})

test_that("raking_build: solver converged (termcd in 1 or 2)", {
  out <- raking_build(vars, sc_df, sp_df, sp_des,
                      wts.col = "sp_wts", control = ctrl)

  expect_true(out$solver_diagnostics$termcd %in% c(1L, 2L))
})

test_that("raking_build: internal contains Xc, Xp, S_beta, D", {
  out <- raking_build(vars, sc_df, sp_df, sp_des,
                      wts.col = "sp_wts", control = ctrl)

  expect_true(all(c("Xc", "Xp", "S_beta", "D") %in% names(out$internal)))
})

test_that("raking_build: S_beta is a square p x p matrix", {
  out <- raking_build(vars, sc_df, sp_df, sp_des,
                      wts.col = "sp_wts", control = ctrl)
  p   <- ncol(Xc)

  expect_equal(dim(out$internal$S_beta), c(p, p))
})

test_that("raking_build: D is a square matrix", {
  out <- raking_build(vars, sc_df, sp_df, sp_des,
                      wts.col = "sp_wts", control = ctrl)
  D   <- out$internal$D

  expect_true(is.matrix(D))
  expect_equal(nrow(D), ncol(D))
})

test_that("raking_build: errors when weight column not in sp", {
  expect_error(
    raking_build(vars, sc_df, sp_df, sp_des,
                 wts.col = "not_exist", control = ctrl),
    "weight column 'not_exist' was not found",
    fixed = TRUE
  )
})

test_that("raking_build: errors when sp weights contain a negative value", {
  sp_bad           <- sp_df
  sp_bad$sp_wts[1] <- -1

  expect_error(
    raking_build(vars, sc_df, sp_bad, sp_des,
                 wts.col = "sp_wts", control = ctrl),
    "reference survey weights must be positive finite numbers",
    fixed = TRUE
  )
})

test_that("raking_build: errors when sp weights contain NA", {
  sp_bad           <- sp_df
  sp_bad$sp_wts[1] <- NA_real_

  expect_error(
    raking_build(vars, sc_df, sp_bad, sp_des,
                 wts.col = "sp_wts", control = ctrl),
    "reference survey weights must be positive finite numbers",
    fixed = TRUE
  )
})


# alp_fn ----

test_that("alp_fn: returns a finite numeric vector of length p at beta = 0", {
  beta0 <- rep(0, ncol(Xc))
  out   <- alp_fn(beta0, Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_type(out, "double")
  expect_length(out, ncol(Xc))
  expect_true(all(is.finite(out)))
})

test_that("alp_fn: non-intercept values differ from zero at beta = 0", {
  beta0 <- rep(0, ncol(Xc))
  out   <- alp_fn(beta0, Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_true(any(out != 0))
})


# alp_jac ----

test_that("alp_jac: returns a square finite matrix at beta = 0", {
  beta0 <- rep(0, ncol(Xc))
  J     <- alp_jac(beta0, Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_true(is.matrix(J))
  expect_equal(nrow(J), ncol(Xc))
  expect_equal(ncol(J), ncol(Xc))
  expect_true(all(is.finite(J)))
})

test_that("alp_jac: diagonal values are all negative at beta = 0", {
  # J = -t(p*(1-p)*Xc)%*%Xc - t(wts*p*(1-p)*Xp)%*%Xp; both terms are negative
  beta0 <- rep(0, ncol(Xc))
  J     <- alp_jac(beta0, Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_true(all(diag(J) < 0))
})


# alp_start ----

test_that("alp_start: returns a finite numeric vector of length p", {
  bs <- alp_start(Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_type(bs, "double")
  expect_length(bs, ncol(Xc))
  expect_true(all(is.finite(bs)))
})

test_that("alp_start: intercept initialized to -log(sum_wts / n_sc)", {
  # At beta = 0: expit(0) = 0.5 everywhere.
  # f_p0[1] = 0.5 * sum(wts_sp), f_c0[1] = 0.5 * n_sc -> ratio = sum(wts_sp)/n_sc
  bs       <- alp_start(Xc = Xc, Xp = Xp, wts_sp = wts_sp)
  expected <- -log(sum(wts_sp) / n_sc)

  expect_equal(bs[1], expected, tolerance = 1e-10)
})

test_that("alp_start: non-intercept values are zero", {
  bs <- alp_start(Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_equal(bs[-1], rep(0, ncol(Xc) - 1))
})


# alp_build ----

test_that("alp_build: return list has expected names", {
  out <- alp_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_named(out, c("weights", "coefficients", "solver_diagnostics",
                      "log_messages", "internal"))
})

test_that("alp_build: pseudo-weights are positive and finite", {
  out <- alp_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_true(all(is.finite(out$weights)))
  expect_true(all(out$weights > 0))
})

test_that("alp_build: pseudo-weights have length nrow(sc)", {
  out <- alp_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_length(out$weights, n_sc)
})

test_that("alp_build: calibration equations satisfied (max|eq| < 1e-6)", {
  out <- alp_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)
  res <- alp_fn(out$coefficients, Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_lt(max(abs(res)), 1e-6)
})

test_that("alp_build: coefficients are named by Xc column names", {
  out <- alp_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_named(out$coefficients, colnames(Xc))
})

test_that("alp_build: solver converged (termcd in 1 or 2)", {
  out <- alp_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_true(out$solver_diagnostics$termcd %in% c(1L, 2L))
})

test_that("alp_build: internal contains Xc, Xp, p_sc, p_sp, S_beta, D", {
  out <- alp_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_true(all(c("Xc", "Xp", "p_sc", "p_sp", "S_beta", "D") %in%
                    names(out$internal)))
})

test_that("alp_build: estimated probabilities p_sc are strictly in (0, 1)", {
  out <- alp_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_true(all(out$internal$p_sc > 0 & out$internal$p_sc < 1))
})

test_that("alp_build: estimated probabilities p_sp are strictly in (0, 1)", {
  out <- alp_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_true(all(out$internal$p_sp > 0 & out$internal$p_sp < 1))
})

test_that("alp_build: S_beta is a square p x p matrix", {
  out <- alp_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)
  p   <- ncol(Xc)

  expect_equal(dim(out$internal$S_beta), c(p, p))
})

test_that("alp_build: errors when weight column not in sp", {
  expect_error(
    alp_build(vars, sc_df, sp_df, sp_des,
              wts.col = "not_exist", control = ctrl),
    "weight column 'not_exist' was not found",
    fixed = TRUE
  )
})

test_that("alp_build: errors when sp weights contain a negative value", {
  sp_bad           <- sp_df
  sp_bad$sp_wts[1] <- -1

  expect_error(
    alp_build(vars, sc_df, sp_bad, sp_des,
              wts.col = "sp_wts", control = ctrl),
    "reference survey weights must be positive finite numbers",
    fixed = TRUE
  )
})

test_that("alp_build: errors when sp weights contain NA", {
  sp_bad           <- sp_df
  sp_bad$sp_wts[1] <- NA_real_

  expect_error(
    alp_build(vars, sc_df, sp_bad, sp_des,
              wts.col = "sp_wts", control = ctrl),
    "reference survey weights must be positive finite numbers",
    fixed = TRUE
  )
})


# clw_fn ----

test_that("clw_fn: returns a finite numeric vector of length p at beta = 0", {
  beta0 <- rep(0, ncol(Xc))
  out   <- clw_fn(beta0, Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_type(out, "double")
  expect_length(out, ncol(Xc))
  expect_true(all(is.finite(out)))
})

test_that("clw_fn: non-intercept values differ from zero at beta = 0", {
  beta0 <- rep(0, ncol(Xc))
  out   <- clw_fn(beta0, Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_true(any(out != 0))
})


# clw_jac ----

test_that("clw_jac: returns a square finite matrix at beta = 0", {
  beta0 <- rep(0, ncol(Xc))
  J     <- clw_jac(beta0, Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_true(is.matrix(J))
  expect_equal(nrow(J), ncol(Xc))
  expect_equal(ncol(J), ncol(Xc))
  expect_true(all(is.finite(J)))
})

test_that("clw_jac: diagonal values are all negative at beta = 0", {
  # J = -t(wts_sp * pi*(1-pi) * Xp) %*% Xp; diagonal < 0
  beta0 <- rep(0, ncol(Xc))
  J     <- clw_jac(beta0, Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_true(all(diag(J) < 0))
})


# clw_start ----

test_that("clw_start: returns a finite numeric vector of length p", {
  bs <- clw_start(Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_type(bs, "double")
  expect_length(bs, ncol(Xc))
  expect_true(all(is.finite(bs)))
})

test_that("clw_start: intercept initialized to -log(0.5 * sum_wts / n_sc)", {
  # At beta = 0: f_p0[1] = 0.5 * sum(wts_sp), f_c0[1] = colSums(Xc)[1] = n_sc
  bs       <- clw_start(Xc = Xc, Xp = Xp, wts_sp = wts_sp)
  expected <- -log(0.5 * sum(wts_sp) / n_sc)

  expect_equal(bs[1], expected, tolerance = 1e-10)
})

test_that("clw_start: non-intercept values are zero", {
  bs <- clw_start(Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_equal(bs[-1], rep(0, ncol(Xc) - 1))
})


# clw_build ----

test_that("clw_build: return list has expected names", {
  out <- clw_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_named(out, c("weights", "coefficients", "solver_diagnostics",
                      "log_messages", "internal"))
})

test_that("clw_build: pseudo-weights are positive and finite", {
  out <- clw_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_true(all(is.finite(out$weights)))
  expect_true(all(out$weights > 0))
})

test_that("clw_build: pseudo-weights have length nrow(sc)", {
  out <- clw_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_length(out$weights, n_sc)
})

test_that("clw_build: calibration equations satisfied (max|eq| < 1e-6)", {
  out <- clw_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)
  res <- clw_fn(out$coefficients, Xc = Xc, Xp = Xp, wts_sp = wts_sp)

  expect_lt(max(abs(res)), 1e-6)
})

test_that("clw_build: coefficients are named by Xc column names", {
  out <- clw_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_named(out$coefficients, colnames(Xc))
})

test_that("clw_build: solver converged (termcd in 1 or 2)", {
  out <- clw_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_true(out$solver_diagnostics$termcd %in% c(1L, 2L))
})

test_that("clw_build: internal contains Xc, Xp, pi_sc, pi_sp, S_beta, D", {
  out <- clw_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_true(all(c("Xc", "Xp", "pi_sc", "pi_sp", "S_beta", "D") %in%
                    names(out$internal)))
})

test_that("clw_build: estimated probabilities pi_sc are strictly in (0, 1)", {
  out <- clw_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_true(all(out$internal$pi_sc > 0 & out$internal$pi_sc < 1))
})

test_that("clw_build: estimated probabilities pi_sp are strictly in (0, 1)", {
  out <- clw_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)

  expect_true(all(out$internal$pi_sp > 0 & out$internal$pi_sp < 1))
})

test_that("clw_build: S_beta is a square p x p matrix", {
  out <- clw_build(vars, sc_df, sp_df, sp_des,
                   wts.col = "sp_wts", control = ctrl)
  p   <- ncol(Xc)

  expect_equal(dim(out$internal$S_beta), c(p, p))
})

test_that("clw_build: errors when weight column not in sp", {
  expect_error(
    clw_build(vars, sc_df, sp_df, sp_des,
              wts.col = "not_exist", control = ctrl),
    "weight column 'not_exist' was not found",
    fixed = TRUE
  )
})

test_that("clw_build: errors when sp weights contain a negative value", {
  sp_bad           <- sp_df
  sp_bad$sp_wts[1] <- -1

  expect_error(
    clw_build(vars, sc_df, sp_bad, sp_des,
              wts.col = "sp_wts", control = ctrl),
    "reference survey weights must be positive finite numbers",
    fixed = TRUE
  )
})

test_that("clw_build: errors when sp weights contain NA", {
  sp_bad           <- sp_df
  sp_bad$sp_wts[1] <- NA_real_

  expect_error(
    clw_build(vars, sc_df, sp_bad, sp_des,
              wts.col = "sp_wts", control = ctrl),
    "reference survey weights must be positive finite numbers",
    fixed = TRUE
  )
})


test_that("one-reference builders satisfy the generalized estimating equation", {
  out_raking <- raking_build(
    vars, sc_df, sp_df, sp_des,
    wts.col = "sp_wts",
    control = ctrl
  )

  ee_raking <- general_ee_one_ref(
    beta = out_raking$coefficients,
    Xc = Xc,
    Xp = Xp,
    wts_sp = wts_sp,
    method = "calibration"
  )

  expect_lt(max(abs(ee_raking)), 1e-6)


  out_alp <- alp_build(
    vars, sc_df, sp_df, sp_des,
    wts.col = "sp_wts",
    control = ctrl
  )

  ee_alp <- general_ee_one_ref(
    beta = out_alp$coefficients,
    Xc = Xc,
    Xp = Xp,
    wts_sp = wts_sp,
    method = "alp"
  )

  expect_lt(max(abs(ee_alp)), 1e-6)


  out_clw <- clw_build(
    vars, sc_df, sp_df, sp_des,
    wts.col = "sp_wts",
    control = ctrl
  )

  ee_clw <- general_ee_one_ref(
    beta = out_clw$coefficients,
    Xc = Xc,
    Xp = Xp,
    wts_sp = wts_sp,
    method = "clw"
  )

  expect_lt(max(abs(ee_clw)), 1e-6)
})












# ipwm_one_build ----

test_that("ipwm_one_build: return list has expected names", {
  out <- ipwm_one_build(
    sc = sc_df, sp = sp_df, sp_des = sp_des,
    vars = vars, weight = "sp_wts",
    method = "calibration", control = ctrl
  )

  expect_named(out, c("pseudo_weights", "coefficients", "method",
                      "solver_diagnostics", "internal"))
})

test_that("ipwm_one_build (calibration): pseudo_weights are positive with correct length", {
  out <- ipwm_one_build(
    sc = sc_df, sp = sp_df, sp_des = sp_des,
    vars = vars, weight = "sp_wts",
    method = "calibration", control = ctrl
  )

  expect_length(out$pseudo_weights, n_sc)
  expect_true(all(out$pseudo_weights > 0))
  expect_equal(out$method, "calibration")
})

test_that("ipwm_one_build (alp): pseudo_weights are positive with correct length", {
  out <- ipwm_one_build(
    sc = sc_df, sp = sp_df, sp_des = sp_des,
    vars = vars, weight = "sp_wts",
    method = "alp", control = ctrl
  )

  expect_length(out$pseudo_weights, n_sc)
  expect_true(all(out$pseudo_weights > 0))
  expect_equal(out$method, "alp")
})

test_that("ipwm_one_build (clw): pseudo_weights are positive with correct length", {
  out <- ipwm_one_build(
    sc = sc_df, sp = sp_df, sp_des = sp_des,
    vars = vars, weight = "sp_wts",
    method = "clw", control = ctrl
  )

  expect_length(out$pseudo_weights, n_sc)
  expect_true(all(out$pseudo_weights > 0))
  expect_equal(out$method, "clw")
})

test_that("ipwm_one_build: log_messages is added inside internal", {
  out <- ipwm_one_build(
    sc = sc_df, sp = sp_df, sp_des = sp_des,
    vars = vars, weight = "sp_wts",
    method = "calibration", control = ctrl
  )

  expect_true("log_messages" %in% names(out$internal))
})

test_that("ipwm_one_build: errors on unknown method", {
  expect_error(
    ipwm_one_build(
      sc = sc_df, sp = sp_df, sp_des = sp_des,
      vars = vars, weight = "sp_wts",
      method = "bad_method", control = ctrl
    ),
    "Unknown method",
    fixed = TRUE
  )
})

#' Estimate Pseudo-Weighted Means, Prevalences, and Standard Errors
#'
#' @description
#' Computes pseudo-weighted means and standard errors using a fitted
#' pseudo-weight object of class \code{"pw_fit"} returned by
#' \code{\link{est_pw}}. The function applies second-layer missing-data
#' handling for the outcome and optional domain variable, and then estimates
#' overall or domain-specific means or prevalences using the pseudo-weighting
#' method stored in \code{object}.
#'
#' @details
#' \strong{Missing data handling (layer 2).}
#' After pseudo-weights are constructed by \code{est_pw()}, estimation of the
#' mean requires complete cases for the outcome \code{y} and, if supplied, the
#' domain variable \code{zcol}. The argument \code{na.action} controls how
#' these missing values are handled at the outcome-estimation step.
#'
#' \strong{Input \code{object}.}
#' The \code{object} argument should be an object of class \code{"pw_fit"}
#' returned by \code{\link{est_pw}}. It stores the estimated pseudo-weights,
#' participation model information, and design-based quantities required for point
#' and variance estimation.
#'
#' \strong{Categorical outcomes.}
#' When \code{y} is a categorical variable (defined as a factor in R),
#' \code{pwmean()} estimates the prevalence (proportion) of each category.
#' To do so, each category is internally converted into a 0/1 indicator
#' variable, and the pseudo-weighted mean estimator is then computed for each
#' indicator.
#'
#' @usage
#' pwmean(object, y, zcol = NULL, na.action = stats::na.omit)
#'
#' @param object An object of class \code{"pw_fit"} returned by
#'   \code{\link{est_pw}}.
#' @param y A character string specifying the name of the outcome variable in the
#'   nonprobability sample stored in \code{object}. The outcome must be numeric
#'   for mean estimation, including binary 0/1 outcomes for prevalence
#'   estimation, or a factor for category prevalence estimation.
#' @param zcol Optional character string giving the name of a categorical domain
#'   variable in the nonprobability sample stored in the \code{object}. If
#'   \code{NULL}, the overall mean is estimated. If supplied, estimates are
#'   computed within domains defined by this variable. The following column
#'   types are supported: \code{logical} (must contain both \code{TRUE} and
#'   \code{FALSE}); \code{numeric} or \code{integer} containing only \code{0}
#'   and \code{1} after removing missing values; \code{character} (empty
#'   strings are treated as missing values); and \code{factor} (unused levels
#'   are dropped).
#' @param na.action Function specifying how missing values in \code{y} and
#'   \code{zcol} should be handled. Default is \code{stats::na.omit}.
#'
#' @return
#' An object of class \code{"pwmean"} containing unweighted and pseudo-weighted
#' estimates, standard errors, and confidence intervals. For categorical outcomes,
#' the estimate columns contain category prevalences.
#'
#' \describe{
#'   \item{\code{method}}{
#'     Character. The pseudo-weighting method used.
#'   }
#'
#'   \item{\code{estimates}}{
#'     A data frame containing the unweighted and pseudo-weighted estimates.
#'
#'     For numeric outcomes, the first column is \code{domain}. If
#'     \code{zcol = NULL}, \code{domain} is \code{"Overall"}. If \code{zcol}
#'     is a \code{logical} variable or a \code{numeric}/\code{integer}
#'     variable containing only \code{0} and \code{1}, there is one row with
#'     \code{domain} labeled \code{"<zcol> = 1"}. If \code{zcol} is a
#'     \code{factor} or \code{character} variable, there is one row per
#'     \code{zcol} level, with \code{domain} labeled
#'     \code{"<zcol> = <level>"}.
#'
#'     For categorical outcomes, the first two columns are
#'     \code{category} and \code{domain}. \code{category} identifies the
#'     outcome level as \code{"<y> = <level>"}. If \code{zcol = NULL},
#'     \code{domain} is \code{"Overall"} for each outcome level. If
#'     \code{zcol} is supplied, the rows are formed by each outcome category
#'     within each domain, and \code{domain} follows the same labels described
#'     above for \code{zcol}.
#'
#'     The columns are:
#'     \describe{
#'       \item{\code{category}}{Category label for categorical outcomes only.}
#'       \item{\code{domain}}{Domain label.}
#'       \item{\code{unweighted_mean}, \code{unweighted_se}}{Unweighted
#'         mean of \code{y} and its standard error.}
#'       \item{\code{unweighted_lower}, \code{unweighted_upper}}{Bounds of the
#'         95\% confidence interval for the unweighted mean, based on the
#'         normal approximation.}
#'       \item{\code{adjusted_mean}, \code{adjusted_se}}{Pseudo-weighted mean
#'         of \code{y} and its standard error.}
#'       \item{\code{adjusted_lower}, \code{adjusted_upper}}{Bounds of the
#'         95\% confidence interval for the pseudo-weighted mean, based on the
#'         normal approximation.}
#'     }
#'   }
#'
#'   \item{\code{na.action}}{
#'     Integer vector of row indices omitted at the outcome-estimation step,
#'     with class \code{"omit"} or \code{"exclude"} matching the
#'     \code{na.action} argument, or \code{NULL} if no observations were
#'     omitted. The indices refer to the nonprobability sample available to
#'     \code{pwmean()} after missing-data handling in
#'     \code{est_pw()}.
#'   }
#'
#'   \item{\code{call}}{
#'     The matched function call.
#'   }
#' }
#'
#' @seealso
#' \code{\link{est_pw}},
#' \code{\link{summary.pwmean}},
#' \code{\link{print.pwmean}}
#'
#' @examples
#' \donttest{
#' data(sc)
#' data(sp1)
#'
#' ref1_design <- survey::svydesign(
#'   ids     = ~psu_sp1,
#'   strata  = ~strata_sp1,
#'   weights = ~wts_sp1,
#'   data    = sp1,
#'   nest    = TRUE
#' )
#'
#' fit <- est_pw(
#'   data      = list(sc, ref1_design),
#'   p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
#'   method    = "calibration",
#'   control   = pw_solver_control(ftol=1e-6)
#' )
#'
#' out <- pwmean(fit, y = "psa_level", zcol = "BMI")
#'
#' print(out)
#'
#' summary(out)
#' }
#'
#' @export
#' @name pwmean
NULL

#' Cumulative precalibration across multiple probability samples
#'
#' Performs a cumulative precalibration across multiple probability samples by
#' sequentially calibrating each sample's weights to align with the marginal
#' totals of the previous samples. Called internally by \code{est_pw()}.
#'
#' @param sp_raw A list of data frames (raw reference samples).
#' @param sp_new A list of data frames (working reference samples) with the same
#'   structure as \code{sp_raw}. Calibrated weights are written back in place.
#' @param weight A list of weight column name strings, one per sample.
#' @param sp_order \code{"size"} to process largest sample first; otherwise
#'   the original list order is used.
#' @return A list with components \code{sp_new}, \code{total_vector},
#'   \code{log_messages}, and \code{order_used}.
#'
#' @importFrom survey svydesign svytotal calibrate
#' @importFrom stats model.matrix as.formula weights coef
#' @keywords internal

precal_cumulative_order <- function(sp_raw, sp_new, weight, sp_order) {

  sort_by_size <- (sp_order == "size")

  log_messages <- character()

  # ---- order control ----
  sizes <- vapply(sp_raw, nrow, integer(1))
  ord <- if (sort_by_size) order(sizes, decreasing = TRUE) else seq_along(sp_raw)

  sp_raw_ord <- sp_raw[ord]
  sp_new_ord <- sp_new[ord]
  w_ord      <- weight[ord]
  # ---- helpers ----
  .get_wname <- function(w) {
    if (length(w) != 1 || !is.character(w) || !nzchar(w)) {
      stop("Each weight[[i]] must be a single non-empty character string.")
    }
    w
  }

  .expand_raw <- function(df, wname) {
    vars <- setdiff(names(df), wname)
    if (length(vars) == 0) {
      return(matrix(, nrow = nrow(df), ncol = 0,
                    dimnames = list(NULL, character(0))))
    }
    X <- stats::model.matrix(~ ., data = df[, vars, drop = FALSE])
    if ("(Intercept)" %in% colnames(X)) X <- X[, -1, drop = FALSE]
    X
  }

  .svy_design_from_X <- function(X, w) {
    dat <- as.data.frame(X)
    dat$.w <- w
    survey::svydesign(ids = ~1, weights = ~.w, data = dat)
  }

  .population_vector <- function(ds, cols) {
    total_w <- sum(stats::weights(ds))
    if (length(cols) == 0) return(c("(Intercept)" = total_w))
    fml  <- stats::as.formula(paste0("~", paste(cols, collapse = " + ")))
    marg <- stats::coef(survey::svytotal(fml, ds))
    c("(Intercept)" = total_w, marg)
  }

  .pretty_cols <- function(cols, max_show = 30) {
    cols <- sort(unique(cols))
    if (length(cols) == 0) return("survey weights total only")
    if (length(cols) <= max_show) {
      return(paste0("survey weights total, ", paste(cols, collapse = ", ")))
    }
    paste0("survey weights total, ",
           paste(cols[1:max_show], collapse = ", "),
           ", ... (", length(cols), " cols)")
  }

  # ---- Step 0: init target totals from reference RAW sample ----
  ref_raw   <- sp_raw_ord[[1]]
  ref_new   <- sp_new_ord[[1]]
  ref_wname <- .get_wname(w_ord[[1]])

  X_ref  <- .expand_raw(ref_raw, ref_wname)
  ds_ref <- .svy_design_from_X(X_ref, ref_new[[ref_wname]])
  total_vec <- .population_vector(ds_ref, colnames(X_ref))

  ref_label <- paste0("sp[[", ord[1], "]]")

  s <- if (sort_by_size) "largest" else "first"

  msg <- sprintf(
    "\nPrecalibration summary:\nNon-calibrated sample (%s): %s\n",
    s, ref_label
  )

  log_messages <- c(log_messages, msg)

  # ---- Step 1..K: calibrate remaining samples ----
  if (length(sp_raw_ord) >= 2) {
    for (k in 2:length(sp_raw_ord)) {

      df_raw <- sp_raw_ord[[k]]
      df_new <- sp_new_ord[[k]]
      wname  <- .get_wname(w_ord[[k]])

      X_k <- .expand_raw(df_raw, wname)

      shared_cols <- intersect(colnames(X_k), setdiff(names(total_vec), "(Intercept)"))
      shared_cols <- sort(shared_cols)

      ds_k <- .svy_design_from_X(X_k, df_new[[wname]])

      pops <- c("(Intercept)" = unname(total_vec["(Intercept)"]))
      if (length(shared_cols) > 0) pops <- c(pops, total_vec[shared_cols])

      fml <- if (length(shared_cols) > 0) {
        stats::as.formula(paste0("~", paste(shared_cols, collapse = " + ")))
      } else {
        ~1
      }

      ds_k_cal <- survey::calibrate(
        design     = ds_k,
        formula    = fml,
        population = pops
      )

      sp_new_ord[[k]][[wname]] <- as.numeric(stats::weights(ds_k_cal))

      # append totals for NEW columns (using calibrated weights)
      current_cols <- setdiff(names(total_vec), "(Intercept)")
      new_cols <- setdiff(colnames(X_k), current_cols)
      new_cols <- sort(new_cols)

      if (length(new_cols) > 0) {
        ds_k2 <- .svy_design_from_X(X_k, sp_new_ord[[k]][[wname]])
        add_vec <- .population_vector(ds_k2, new_cols)
        add_vec <- add_vec[names(add_vec) != "(Intercept)"]
        total_vec <- c(total_vec, add_vec)
      }

      k_label <- paste0("sp[[", ord[k], "]]")
      msg <- sprintf(
        "Calibrated sample: %s\n  Calibration variables: %s\n",
        k_label, .pretty_cols(shared_cols)
      )
      log_messages <- c(log_messages, msg)
    }
  }

  # ---- restore original order only if we reordered ----
  sp_back <- if (sort_by_size) {
    out <- vector("list", length(sp_new))
    out[ord] <- sp_new_ord
    names(out) <- names(sp_new)
    out
  } else {
    sp_new_ord
  }

  list(
    sp_new       = sp_back,
    total_vector = total_vec,
    log_messages = log_messages,
    order_used   = ord
  )
}

# Setup ----

set.seed(123)

# sp1: factor a (shared), numeric x (unique to sp1), wt1 — 150 rows (larger)
# sp2: factor a (shared), factor b (unique to sp2), wt2 — 100 rows (smaller)
# Shared model-matrix columns after expansion: aB, aC
# sp1-only column: x
# sp2-only column: b

sp1_raw <- data.frame(
  a   = factor(sample(c("A","B","C"), 150, replace = TRUE), levels = c("A","B","C")),
  x   = rnorm(150),
  wt1 = runif(150, 10, 100)
)

sp2_raw <- data.frame(
  a   = factor(sample(c("A","B","C"), 100, replace = TRUE), levels = c("A","B","C")),
  b   = factor(sample(c("P","Q"),     100, replace = TRUE), levels = c("P","Q")),
  wt2 = runif(100, 10, 100)
)

sp1_new <- sp1_raw
sp2_new <- sp2_raw


# Return structure ----

test_that("precal_cumulative_order: returns correct list names", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_named(out, c("sp_new", "total_vector", "log_messages", "order_used"))
})

test_that("precal_cumulative_order: sp_new has same length as input", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_length(out$sp_new, 2)
})

test_that("precal_cumulative_order: total_vector starts with (Intercept)", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_equal(names(out$total_vector)[1], "(Intercept)")
})

test_that("precal_cumulative_order: log_messages is a character vector", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_type(out$log_messages, "character")
})


# order_used ----

test_that("precal_cumulative_order: sp_order = 'size' keeps larger sample first when already first", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_equal(out$order_used[1], 1L)
})

test_that("precal_cumulative_order: sp_order = 'size' reorders when smaller is listed first", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp2_raw, sp1_raw),
    sp_new   = list(sp2_new, sp1_new),
    weight   = c("wt2", "wt1"),
    sp_order = "size"
  )
  expect_equal(out$order_used[1], 2L)
})

test_that("precal_cumulative_order: sp_order = 'given' uses list order", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "given"
  )
  expect_equal(out$order_used, 1:2)
})

test_that("precal_cumulative_order: order_used length equals number of samples", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_length(out$order_used, 2L)
})


# Reference sample unchanged ----

test_that("precal_cumulative_order: reference sample weights are not modified", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_equal(out$sp_new[[1]][["wt1"]], sp1_raw[["wt1"]])
})


# Calibration correctness ----

test_that("precal_cumulative_order: calibrated total weight matches reference total (intercept)", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  wt2_cal <- out$sp_new[[2]][["wt2"]]
  expect_equal(sum(wt2_cal), unname(out$total_vector["(Intercept)"]), tolerance = 1e-5)
})

test_that("precal_cumulative_order: calibrated weighted totals match target for shared columns", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  wt2_cal <- out$sp_new[[2]][["wt2"]]
  X2 <- model.matrix(~ ., data = sp2_raw[, c("a", "b")])[, -1, drop = FALSE]

  for (col in c("aB", "aC")) {
    expect_equal(
      sum(wt2_cal * X2[, col]),
      unname(out$total_vector[col]),
      tolerance = 1e-5
    )
  }
})


# total_vector grows with new columns ----

test_that("precal_cumulative_order: total_vector includes unique columns from sp2", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_true("bQ" %in% names(out$total_vector))
})

test_that("precal_cumulative_order: total_vector contains all columns from both samples", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_true(all(c("(Intercept)", "aB", "aC", "x", "bQ") %in% names(out$total_vector)))
})


# No shared columns ----

test_that("precal_cumulative_order: calibrates intercept only when no columns are shared", {
  set.seed(99)
  sp_a <- data.frame(x = rnorm(100), wt1 = runif(100, 10, 100))
  sp_b <- data.frame(y = rnorm(80),  wt2 = runif(80,  10, 100))

  out <- precal_cumulative_order(
    sp_raw   = list(sp_a, sp_b),
    sp_new   = list(sp_a, sp_b),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )

  wt_b_cal <- out$sp_new[[2]][["wt2"]]
  expect_equal(sum(wt_b_cal), unname(out$total_vector["(Intercept)"]), tolerance = 1e-5)
})

test_that("precal_cumulative_order: total_vector contains columns from both samples when no overlap", {
  set.seed(99)
  sp_a <- data.frame(x = rnorm(100), wt1 = runif(100, 10, 100))
  sp_b <- data.frame(y = rnorm(80),  wt2 = runif(80,  10, 100))

  out <- precal_cumulative_order(
    sp_raw   = list(sp_a, sp_b),
    sp_new   = list(sp_a, sp_b),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )

  expect_true(all(c("x", "y") %in% names(out$total_vector)))
})


# List names kept through reordering ----

test_that("precal_cumulative_order: sp_new names kept with sp_order = 'given'", {
  out <- precal_cumulative_order(
    sp_raw   = list(survey_a = sp1_raw, survey_b = sp2_raw),
    sp_new   = list(survey_a = sp1_new, survey_b = sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "given"
  )
  expect_equal(names(out$sp_new), c("survey_a", "survey_b"))
})

test_that("precal_cumulative_order: sp_new names kept after size reorder", {
  out <- precal_cumulative_order(
    sp_raw   = list(survey_a = sp2_raw, survey_b = sp1_raw),
    sp_new   = list(survey_a = sp2_new, survey_b = sp1_new),
    weight   = c("wt2", "wt1"),
    sp_order = "size"
  )
  expect_equal(names(out$sp_new), c("survey_a", "survey_b"))
})


# log_messages content ----

test_that("precal_cumulative_order: log_messages has one entry per sample", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_length(out$log_messages, 2)
})

test_that("precal_cumulative_order: log_messages mentions non-calibrated sample", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_true(any(grepl("Non-calibrated sample", out$log_messages)))
})

test_that("precal_cumulative_order: log_messages mentions calibrated sample", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_true(any(grepl("Calibrated sample", out$log_messages)))
})

test_that("precal_cumulative_order: log_messages notes 'largest' when sp_order = 'size'", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "size"
  )
  expect_true(any(grepl("largest", out$log_messages)))
})

test_that("precal_cumulative_order: log_messages notes 'first' when sp_order = 'given'", {
  out <- precal_cumulative_order(
    sp_raw   = list(sp1_raw, sp2_raw),
    sp_new   = list(sp1_new, sp2_new),
    weight   = c("wt1", "wt2"),
    sp_order = "given"
  )
  expect_true(any(grepl("first", out$log_messages)))
})

#' Auto-build a default participation model formula
#'
#' Constructs a one-sided participation model formula from the variables
#' shared between the nonprobability sample (`sc`) and each reference
#' survey, excluding the sampling weight column. Called by `est_pw()` when
#' `p_formula = NULL`.
#'
#' For each reference survey, the candidate covariates are the column names
#' present in both `sc` and the reference survey data frame after dropping
#' the sampling weight column named by `weight`. A one-sided formula of the
#' form `~ var1 + var2 + ...` is built from the remaining shared variables.
#' An error is raised if no shared covariates remain after excluding the
#' weight column.
#'
#' The function distinguishes the one-reference case from the multi-reference
#' case by the type of `sp`: a plain data frame triggers the one-reference
#' path; a list of data frames triggers the multi-reference path. Note that
#' in R a `data.frame` is also a `list`, so the `is.data.frame()` check is
#' performed first.
#'
#' @param sc A data frame. The nonprobability sample.
#' @param sp A data frame (one-reference case) or a named list of data frames
#'   (multi-reference case). Each data frame contains the analysis variables
#'   of one reference survey, with design variables removed and a sampling
#'   weight column appended (as produced by `extract_analysis_data()`).
#' @param weight A single character string (one-reference case) or a
#'   character vector of the same length as `sp` (multi-reference case)
#'   giving the name of the sampling weight column in each reference survey
#'   data frame. This column is excluded from the candidate covariate set.
#'
#' @return A list with two elements:
#' \describe{
#'   \item{`p_formula`}{A one-sided formula (one-reference case) or a named
#'     list of one-sided formulas (multi-reference case).}
#'   \item{`log_messages`}{A character vector of messages describing the
#'     auto-generated formula(s), used for downstream printing when
#'     `verbose = TRUE`.}
#' }
#'
#' @keywords internal
p_formula_construction <- function(sc, sp, weight) {

  log_messages <- character(0)
  #-------------------------------#
  # one-reference case
  #-------------------------------#
  if (is.data.frame(sp)) {
    res <- build_one_formula(
      sc       = sc,
      sp_i     = sp,
      weight_i = weight
    )

    if (!is.null(res$log_messages) && length(res$log_messages) > 0) {
      log_messages <- c(log_messages, res$log_messages)
    }

    return(list(
      p_formula    = res$p_formula,
      log_messages = log_messages
    ))
  }

  #-------------------------------#
  # multi-reference case
  #-------------------------------#
  if (is.list(sp)) {
    if (!is.character(weight) || length(weight) != length(sp)) {
      stop(
        "For multi-reference case, `weight` must be a character vector with same length as `sp`.",
        call. = FALSE
      )
    }

    p_formula_list <- vector("list", length(sp))

    for (i in seq_along(sp)) {
      sp_name <- if (!is.null(names(sp)) && nzchar(names(sp)[i])) {
        names(sp)[i]
      } else {
        paste0("sp[[", i, "]]")
      }

      res_i <- build_one_formula(
        sc       = sc,
        sp_i     = sp[[i]],
        weight_i = weight[i],
        sp_name  = sp_name
      )

      p_formula_list[[i]] <- res_i$p_formula

      if (!is.null(res_i$log_messages) && length(res_i$log_messages) > 0) {
        log_messages <- c(log_messages, res_i$log_messages)
      }
    }

    names(p_formula_list) <- if (!is.null(names(sp)) && all(nzchar(names(sp)))) {
      names(sp)
    } else {
      paste0("sp[[", seq_along(sp), "]]")
    }

    return(list(
      p_formula    = p_formula_list,
      log_messages = log_messages
    ))
  }

  stop("`sp` must be either a data.frame or a list of data.frames.", call. = FALSE)
}

# Setup ----

data(sc)
data(sp1)
data(sp2)


# one-reference ----

test_that("p_formula_construction (one-ref): returns a formula", {
  res <- p_formula_construction(sc, sp1, weight = "sp_wts")

  expect_true(inherits(res$p_formula, "formula"))
})

test_that("p_formula_construction (one-ref): formula uses only shared variables", {
  res  <- p_formula_construction(sc, sp1, weight = "sp_wts")
  vars <- all.vars(res$p_formula)

  shared <- intersect(colnames(sc), colnames(sp1))
  expect_true(all(vars %in% shared))
})

test_that("p_formula_construction (one-ref): weight column excluded from formula", {
  res  <- p_formula_construction(sc, sp1, weight = "wts_sp1")
  vars <- all.vars(res$p_formula)

  expect_false("wts_sp1" %in% vars)
})

test_that("p_formula_construction (one-ref): returns log_messages as character", {
  res <- p_formula_construction(sc, sp1, weight = "sp_wts")

  expect_type(res$log_messages, "character")
  expect_gt(length(res$log_messages), 0)
})

test_that("p_formula_construction (one-ref): errors when no shared covariates", {
  sc_only <- data.frame(only_in_sc = 1:5)
  sp_only <- data.frame(sp_wts = rep(1, 5))

  expect_error(
    p_formula_construction(sc_only, sp_only, weight = "sp_wts"),
    "No shared covariates found to build default p_formula",
    fixed = TRUE
  )
})

test_that("p_formula_construction (one-ref): errors when only shared column is weight", {
  sc_w <- data.frame(wts_sp1 = 1:5, x = letters[1:5])
  sp_w <- data.frame(wts_sp1 = rep(1, 5))

  expect_error(
    p_formula_construction(sc_w, sp_w, weight = "wts_sp1"),
    "No shared covariates found to build default p_formula",
    fixed = TRUE
  )
})

test_that("p_formula_construction (one-ref): errors when sp is not a data.frame or list", {
  expect_error(
    p_formula_construction(sc, "not_a_df", weight = "sp_wts"),
    "`sp` must be either a data.frame or a list of data.frames.",
    fixed = TRUE
  )
})


# multi-reference ----

test_that("p_formula_construction (multi-ref): returns a list of formulas", {
  res <- p_formula_construction(
    sc, list(sp1, sp2),
    weight  = c("wts_sp1", "wts_sp2")
  )

  expect_true(is.list(res$p_formula))
  expect_length(res$p_formula, 2)
  expect_true(all(vapply(res$p_formula, inherits, logical(1), "formula")))
})

test_that("p_formula_construction (multi-ref): each formula excludes its own weight", {
  res <- p_formula_construction(
    sc, list(sp1, sp2),
    weight  = c("wts_sp1", "wts_sp2")
  )

  expect_false("wts_sp1" %in% all.vars(res$p_formula[[1]]))
  expect_false("wts_sp2" %in% all.vars(res$p_formula[[2]]))
})

test_that("p_formula_construction (multi-ref): list names assigned correctly", {
  sp_named <- list(survey_a = sp1, survey_b = sp2)

  res <- p_formula_construction(
    sc, sp_named,
    weight  = c("wts_sp1", "wts_sp2")
  )

  expect_equal(names(res$p_formula), c("survey_a", "survey_b"))
})

test_that("p_formula_construction (multi-ref): unnamed sp gets default names", {
  res <- p_formula_construction(
    sc, list(sp1, sp2),
    weight  = c("wts_sp1", "wts_sp2")
  )

  expect_equal(names(res$p_formula), c("sp[[1]]", "sp[[2]]"))
})

test_that("p_formula_construction (multi-ref): log_messages has one entry per sp", {
  res <- p_formula_construction(
    sc, list(sp1, sp2),
    weight  = c("wts_sp1", "wts_sp2")
  )

  expect_length(res$log_messages, 2)
})

test_that("p_formula_construction (multi-ref): errors when weight length mismatches sp", {
  expect_error(
    p_formula_construction(
      sc, list(sp1, sp2),
      weight  = "wts_sp1"
    ),
    "For multi-reference case, `weight` must be a character vector with same length as `sp`.",
    fixed = TRUE
  )
})

test_that("p_formula_construction (multi-ref): errors when no shared covariates for one sp", {
  sp_no_shared <- data.frame(wts_sp2 = rep(1, 5))

  expect_error(
    p_formula_construction(
      sc, list(sp1, sp_no_shared),
      weight  = c("wts_sp1", "wts_sp2")
    ),
    "No shared covariates found to build default p_formula",
    fixed = TRUE
  )
})

#' Build pseudo-weights for the multi-reference calibration method
#'
#' @param sc Data frame. The nonprobability sample.
#' @param sp List of data frames. Each element is one probability reference
#'   sample (already NA-processed and p_formula-processed).
#' @param vars List of character vectors, one per reference sample, giving the
#'   predictor variable names shared with `sc`.
#' @param weight Character vector, one per reference sample, giving the
#'   survey-weight column name in each element of `sp`.
#' @param sp_des List of `survey.design2` or `svyrep.design` objects,
#'   one per reference sample.
#' @param sp_order Character scalar, either `"size"` (sort reference samples
#'   largest-first) or `"given"` (keep user order).
#' @param control A list created by `pw_solver_control()`.
#' @param verbose Logical. If `TRUE`, progress messages are printed.
#' @param log_messages Character vector of messages accumulated in earlier
#'   steps, forwarded into the returned object.
#'
#' @return A list with components:
#'   \item{pseudo_weights}{Numeric vector of estimated pseudo-weights for `sc`.}
#'   \item{coefficients}{Named numeric vector of participation model coefficients.}
#'   \item{method}{Character scalar, always `"multi"`.}
#'   \item{solver_diagnostics}{List of solver convergence diagnostics.}
#'   \item{internal}{List of intermediate objects (block design matrices,
#'     reference weighted totals, block column indices, sandwich components,
#'     and sorted metadata) needed by the estimation stage.}
#'
#' @keywords internal
ipwm_multi_build <- function(
    sc, sp, vars, weight,
    sp_des, sp_order,
    control,
    verbose = FALSE,
    log_messages = NULL
) {

  if (is.null(log_messages)) {
    log_messages <- character(0)
  }

  step_try <- function(step_num, step_name, expr) {
    tryCatch(
      expr,
      error = function(e) {
        stop(
          sprintf(
            "Step %s (%s) failed: %s",
            step_num, step_name, conditionMessage(e)
          ),
          call. = FALSE
        )
      }
    )
  }


  sorted <- step_try("1", "sort reference samples", {
    sort_by_sp_size(
      sp       = sp,
      vars     = vars,
      weight   = weight,
      design   = sp_des,
      sp_order = sp_order,
      verbose  = verbose
    )
  })

  sp_sorted     <- sorted$sp
  vars_sorted   <- sorted$vars
  weight_sorted <- sorted$weight
  design_sorted <- sorted$design
  log_messages  <- c(log_messages, sorted$log)

  valid <- step_try("2", "validate inputs", {
    check_input_multi(
      sc        = sc,
      sp_list   = sp_sorted,
      vars_list = vars_sorted,
      wts_cols  = weight_sorted,
      verbose   = verbose
    )
  })

  sc_work      <- valid$sc
  sp_list      <- valid$sp_list
  vars_XC      <- valid$vars_XC
  xcol         <- valid$xcol
  wts_cols     <- valid$wts_cols
  log_messages <- c(log_messages, valid$log)

  DM <- step_try("3", "construct design matrices", {
    multi_matrix_construction(
      vars_XC  = vars_XC,
      sc       = sc_work,
      sp_list  = sp_list,
      xcol     = xcol,
      wts_cols = wts_cols
    )
  })

  Xc       <- DM$Xc
  Xp_list  <- DM$Xp_list
  wts_list <- DM$wts_list

  if (verbose) {
    message(sprintf(
      "Fitting multi-reference calibration model: %d reference surveys, %d variables total.",
      length(Xp_list), ncol(Xc)
    ))
  }

  block_cols <- step_try("4", "match multi-reference blocks", {
    multi_raking_block_cols(
      Xc = Xc,
      Xp_list = Xp_list,
      label = "Multi_Calibration"
    )
  })

  f_p <- step_try("5", "construct reference weighted totals", {
    multi_raking_fp(
      Xp_list = Xp_list,
      wts_list = wts_list,
      label = "Multi_Calibration"
    )
  })

  beta_start <- step_try("6", "construct starting values", {
    multi_raking_start(
      Xc = Xc,
      Xp_list = Xp_list,
      block_cols = block_cols,
      f_p = f_p,
      label = "Multi_Calibration"
    )
  })

  sol <- step_try("7", "solve multi-reference calibration equations", {
    solve_participation_model(
      beta_start = beta_start,
      fn = multi_raking_fn,
      jac = multi_raking_jac,
      label = "Multi_Calibration",
      control = control,
      Xc = Xc,
      f_p = f_p,
      block_cols = block_cols
    )
  })

  beta <- sol$coefficients

  wts_sc <- step_try("8", "compute pseudo-weights", {
    eta <- as.vector(Xc %*% beta)

    if (any(!is.finite(eta))) {
      stop(
        "Multi-reference calibration failed: final linear predictor contains non-finite values.",
        call. = FALSE
      )
    }

    if (any(eta < -log(.Machine$double.xmax))) {
      stop(
        "Multi-reference calibration failed: final pseudo-weights overflow.",
        call. = FALSE
      )
    }

    out <- as.vector(exp(-eta))

    if (any(!is.finite(out))) {
      stop(
        "Multi-reference calibration failed: final pseudo-weights contain non-finite values.",
        call. = FALSE
      )
    }

    out
  })

  f_final <- step_try("9", "check final estimating equations", {
    out <- multi_raking_fn(
      beta = beta,
      Xc = Xc,
      f_p = f_p,
      block_cols = block_cols,
      label = "Multi_Calibration"
    )

    if (any(!is.finite(out))) {
      stop(
        "Multi-reference calibration failed: final estimating equations contain non-finite values.",
        call. = FALSE
      )
    }

    out
  })

  max_abs_eq <- max(abs(f_final))

  if (verbose) {
    if (sol$termcd == 1L || sol$termcd == 2L) {
      message(sprintf("Converged after %d iterations (max |eq| = %.3e).",
                      sol$iterations, max_abs_eq))
    } else {
      message(sprintf(
        "WARNING: solver did not converge (termcd = %d). Check solver_diagnostics for details.",
        sol$termcd
      ))
    }
  }

  D <- step_try("10", "construct D matrix", {
    make_block_D_multi(
      sp_des_list = design_sorted,
      Xp_list     = Xp_list
    )
  })

  S_beta_full <- step_try("11", "construct S_beta", {
    t(wts_sc * Xc) %*% Xc
  })

  out <- list(
    pseudo_weights = wts_sc,

    coefficients   = stats::setNames(as.numeric(beta), colnames(Xc)),

    method         = "multi",

    solver_diagnostics = list(
      solver  = sol$solver,
      termcd  = sol$termcd,
      message = sol$message,
      method  = sol$solver_method,
      iter    = sol$iterations,
      fmax    = max_abs_eq
    ),

    internal = list(
      Xc         = Xc,
      Xp_list    = Xp_list,
      wts_list   = wts_list,
      f_p        = f_p,
      block_cols = block_cols,
      D          = D,
      S_beta     = S_beta_full,

      xcol         = xcol,
      vars_XC      = vars_XC,
      wts_cols     = wts_cols,
      design       = design_sorted,
      sp_order     = sp_order,
      log_messages = log_messages
    )
  )

  return(out)
}

#' Extract and align sc data from a build object
#'
#' Reads the raw sc sample, the build-stage complete-case index
#' (`keep_sc`), the covariate matrix `Xc`, and the pseudo-weights from
#' `build`. Remove NA placeholders inserted by `na.exclude` so that all
#' returned vectors and matrices have the same length (the number of
#' build-stage complete cases).
#'
#' @param build A `pw_fit` object returned by the build step.
#'
#' @return A list with components:
#'
#' - `sc`: data frame of build-stage complete cases.
#' - `X`: covariate matrix `build$internal$Xc`.
#' - `w`: pseudo-weight vector with NA placeholders removed.
#' - `idx_keep`: integer index of which rows of `raw_sc` were retained.
#'
#' @keywords internal
prepare_sc_data <- function(build) {

  raw_sc  <- build$internal$raw_sc
  keep_sc <- build$internal$na$keep_sc

  if (!is.logical(keep_sc))
    stop("prepare_sc_data: keep_sc must be a logical vector.", call. = FALSE)
  idx_keep <- which(keep_sc)
  sc_keep  <- raw_sc[idx_keep, , drop = FALSE]

  X <- build$internal$Xc
  w <- build$pseudo_weights

  # for na.exclude builds, pseudo_weights has NA placeholders for dropped rows;
  # remove them here so w aligns with the complete-case rows in X and sc_keep
  w <- w[!is.na(w)]

  nx <- nrow(X)
  if (is.null(nx)) stop("prepare_sc_data: nrow(build$Xc) is NULL (Xc is not a matrix/data.frame).", call. = FALSE)

  if (nx != length(w)) {
    stop("prepare_sc_data: mismatch between nrow(build$Xc) and length(build$pseudo_weights).", call. = FALSE)
  }
  if (nrow(sc_keep) != length(w)) {
    stop("prepare_sc_data: mismatch between nrow(sc_keep) and length(build$pseudo_weights).", call. = FALSE)
  }

  list(sc = sc_keep, X = X, w = w, idx_keep = idx_keep)
}

# Setup ----

data(sc)

# helper: build a minimal pw_fit around a data frame
make_pw_fit <- function(raw_sc, weights = NULL, keep_sc = NULL, Xc = NULL) {
  n      <- nrow(raw_sc)
  if (is.null(keep_sc)) keep_sc <- rep(TRUE, n)
  n_keep <- sum(keep_sc)
  if (is.null(weights)) weights <- rep(1.0, n_keep)
  if (is.null(Xc))      Xc      <- matrix(1.0, nrow = n_keep, ncol = 1L)
  structure(
    list(
      pseudo_weights = weights,
      internal = list(
        raw_sc = raw_sc,
        na     = list(keep_sc = keep_sc),
        Xc     = Xc
      )
    ),
    class = "pw_fit"
  )
}

fit <- make_pw_fit(sc)


# clean path ----

test_that("prepare_sc_data: returns named list with sc, X, w, idx_keep", {
  res <- prepare_sc_data(fit)
  expect_named(res, c("sc", "X", "w", "idx_keep"))
})

test_that("prepare_sc_data: sc has same columns as raw_sc", {
  res <- prepare_sc_data(fit)
  expect_equal(names(res$sc), names(sc))
})

test_that("prepare_sc_data: sc row count equals sum(keep_sc)", {
  keep         <- rep(TRUE, nrow(sc))
  keep[1:10]   <- FALSE
  fit_subset   <- make_pw_fit(sc, keep_sc = keep)
  res          <- prepare_sc_data(fit_subset)
  expect_equal(nrow(res$sc), sum(keep))
})

test_that("prepare_sc_data: idx_keep matches which(keep_sc)", {
  keep         <- rep(TRUE, nrow(sc))
  keep[c(1, 5, 10)] <- FALSE
  fit_subset   <- make_pw_fit(sc, keep_sc = keep)
  res          <- prepare_sc_data(fit_subset)
  expect_equal(res$idx_keep, which(keep))
})

test_that("prepare_sc_data: sc rows correspond to keep_sc indices", {
  keep         <- rep(c(TRUE, FALSE), length.out = nrow(sc))
  fit_subset   <- make_pw_fit(sc, keep_sc = keep)
  res          <- prepare_sc_data(fit_subset)
  expect_equal(res$sc, sc[keep, , drop = FALSE], ignore_attr = TRUE)
})

test_that("prepare_sc_data: X is returned unchanged from build$internal$Xc", {
  Xc           <- matrix(seq_len(2 * nrow(sc)), nrow = nrow(sc), ncol = 2L)
  fit_with_Xc  <- make_pw_fit(sc, Xc = Xc)
  res          <- prepare_sc_data(fit_with_Xc)
  expect_equal(res$X, Xc)
})

test_that("prepare_sc_data: w length equals nrow(X) when no NA weights", {
  res <- prepare_sc_data(fit)
  expect_equal(length(res$w), nrow(res$X))
})


# na.exclude: NA weights removed ----

test_that("prepare_sc_data: NA weights from na.exclude builds are removed", {
  n             <- nrow(sc)
  keep          <- rep(TRUE, n)
  keep[1:5]     <- FALSE
  w_full        <- rep(1.0, n)
  w_full[!keep] <- NA_real_
  fit_excl      <- make_pw_fit(sc, weights = w_full, keep_sc = keep,
                               Xc = matrix(1.0, nrow = sum(keep), ncol = 1L))
  res           <- prepare_sc_data(fit_excl)
  expect_false(anyNA(res$w))
  expect_equal(length(res$w), sum(keep))
})


# type guards ----

test_that("prepare_sc_data: rejects non-logical keep_sc", {
  bad                      <- fit
  bad$internal$na$keep_sc  <- seq_len(nrow(sc))
  expect_error(
    prepare_sc_data(bad),
    "prepare_sc_data: keep_sc must be a logical vector.",
    fixed = TRUE
  )
})

test_that("prepare_sc_data: rejects vector Xc (nrow is NULL)", {
  bad              <- fit
  bad$internal$Xc  <- rep(1.0, nrow(sc))
  expect_error(
    prepare_sc_data(bad),
    "prepare_sc_data: nrow(build$Xc) is NULL",
    fixed = TRUE
  )
})


# length mismatch checks ----

test_that("prepare_sc_data: errors on nrow(X) vs length(w) mismatch", {
  n      <- nrow(sc)
  bad    <- make_pw_fit(sc,
                        weights = rep(1.0, n),
                        Xc      = matrix(1.0, nrow = n + 5L, ncol = 1L))
  expect_error(
    prepare_sc_data(bad),
    "prepare_sc_data: mismatch between nrow(build$Xc) and length(build$pseudo_weights).",
    fixed = TRUE
  )
})

test_that("prepare_sc_data: errors on nrow(sc_keep) vs length(w) mismatch", {
  n      <- nrow(sc)
  keep   <- rep(TRUE, n)
  keep[1:5] <- FALSE
  # weights and Xc sized to n (not n_keep), so sc_keep length differs from w
  bad    <- make_pw_fit(sc,
                        weights = rep(1.0, n),
                        keep_sc = keep,
                        Xc      = matrix(1.0, nrow = n, ncol = 1L))
  expect_error(
    prepare_sc_data(bad),
    "prepare_sc_data: mismatch between nrow(sc_keep) and length(build$pseudo_weights).",
    fixed = TRUE
  )
})

#' Resolve na.action to a standard mode string
#'
#' Converts the `na.action` argument -- which may be a function, a character
#' string, or `NULL` -- into one of four canonical mode strings: `"omit"`,
#' `"exclude"`, `"fail"`, or `"pass"`. When `NULL` is supplied the function
#' reads `getOption("na.action")`, following the same convention as `lm()`.
#'
#' Accepted function inputs: `stats::na.omit`, `stats::na.exclude`,
#' `stats::na.fail`, `stats::na.pass`.
#'
#' Accepted string inputs (case-insensitive): `"na.omit"` / `"omit"`,
#' `"na.exclude"` / `"exclude"`, `"na.fail"` / `"fail"`,
#' `"na.pass"` / `"pass"`.
#'
#' @param na.action A function, a single character string, or `NULL`.
#'
#' @return A single character string: one of `"omit"`, `"exclude"`,
#'   `"fail"`, or `"pass"`.
#'
#' @keywords internal
#' @noRd
resolve_na_action <- function(na.action) {

  if (is.null(na.action)) {
    na.action <- getOption("na.action")
  }

  # 1) function input
  if (is.function(na.action)) {
    if (identical(na.action, stats::na.omit))    return("omit")
    if (identical(na.action, stats::na.exclude)) return("exclude")
    if (identical(na.action, stats::na.fail))    return("fail")
    if (identical(na.action, stats::na.pass))    return("pass")
    stop("Unsupported na.action function. Use na.omit/na.exclude/na.fail/na.pass.", call. = FALSE)
  }

  # 2) character input
  if (is.character(na.action) && length(na.action) == 1L) {
    x <- tolower(na.action)

    if (x %in% c("na.omit", "omit"))       return("omit")
    if (x %in% c("na.exclude", "exclude")) return("exclude")
    if (x %in% c("na.fail", "fail"))       return("fail")
    if (x %in% c("na.pass", "pass"))       return("pass")

    stop("Invalid na.action string. Use 'na.omit','na.exclude','na.fail','na.pass' (or omit/exclude/fail/pass).",
         call. = FALSE)
  }

  stop("Invalid na.action type. Use a function (na.omit/na.exclude/na.fail/na.pass), a string, or NULL.",
       call. = FALSE)
}


#' Core NA filtering for the IPWM build stage
#'
#' Identifies rows with missing values in the participation model variables,
#' applies the chosen NA handling mode to `sc` and `sp`, and constructs an
#' lm-style `na.action` attribute for `sc`. Called by `process_na_build()`.
#'
#' For the one-reference case `sp` is a plain data frame and `p_formula` is
#' a single formula; the same set of variables is used to check NAs in both
#' `sc` and `sp`. For the multi-reference case `sp` is a list of data frames
#' and `p_formula` is a list of formulas; `sc` is checked against the union
#' of all formula variables, while each `sp[[j]]` is checked only against the
#' variables in `p_formula[[j]]`.
#'
#' @param sc A data frame. The nonprobability sample.
#' @param sp A data frame (one-reference) or a named list of data frames
#'   (multi-reference). Reference survey analysis data.
#' @param p_formula A one-sided formula (one-reference) or a list of
#'   one-sided formulas (multi-reference).
#' @param na_mode Character string. One of `"omit"`, `"exclude"`,
#'   `"fail"`, or `"pass"`, as returned by `resolve_na_action()`.
#'
#' @return A list with five elements:
#' \describe{
#'   \item{`sc`}{Cleaned nonprobability sample (rows with NAs removed
#'     when `na_mode` is `"omit"`, `"exclude"`, or `"fail"`).}
#'   \item{`sp`}{Cleaned reference survey data frame(s).}
#'   \item{`keep_sc`}{Logical vector of length `nrow(sc)` indicating
#'     retained rows.}
#'   \item{`keep_sp`}{Logical vector (one-reference) or list of logical
#'     vectors (multi-reference) indicating retained rows in each `sp`.}
#'   \item{`na_action`}{An lm-style `na.action` attribute (class `"omit"`
#'     or `"exclude"`) recording the dropped row indices of `sc`, or
#'     `NULL` if no rows were dropped.}
#'   \item{`log`}{Character vector of per-variable NA detail messages,
#'     propagated to `log_messages` in `est_pw()`.}
#' }
#'
#' @keywords internal
#' @noRd
handle_na_for_ipwm <- function(sc, sp, p_formula, na_mode = c("omit", "exclude", "fail", "pass")) {

  na_mode <- match.arg(na_mode)
  log <- character(0)

  #----------------------------------------------------------
  # helper: NA summary logging -- returns messages, no <<-
  #----------------------------------------------------------
  .na_summary <- function(df, vars, label) {
    if (length(vars) == 0L) return(character(0))
    na_ct <- colSums(is.na(df[, vars, drop = FALSE]))
    na_ct <- na_ct[na_ct > 0]
    if (length(na_ct) == 0L) return(character(0))
    msgs <- c(
      "\nMissing value summary:\n",
      sprintf("%d NA detected in p_formula variables in %s:", sum(na_ct), label)
    )
    for (v in names(na_ct)) msgs <- c(msgs, sprintf("  - %s", v))
    msgs
  }

  #----------------------------------------------------------
  # 1. sc
  #----------------------------------------------------------
  if (is.list(p_formula)) {
    vars_sc <- unique(unlist(lapply(p_formula, all.vars)))
  } else {
    vars_sc <- all.vars(p_formula)
  }

  if (length(vars_sc) > 0) {
    log <- c(log, .na_summary(sc, vars_sc, "sc"))
    keep_sc <- stats::complete.cases(sc[, vars_sc, drop = FALSE])
  } else {
    keep_sc <- rep(TRUE, nrow(sc))
  }
  removed_sc <- sum(!keep_sc)

  if (na_mode == "fail" && removed_sc > 0) {
    stop("na.fail: NA found in p_formula variables in sc.", call. = FALSE)
  }

  if (na_mode == "pass") {
    keep_sc[] <- TRUE
  } else if (na_mode %in% c("omit", "exclude") && removed_sc > 0) {
    log <- c(log, sprintf("\nRemoved %d rows from sc due to NA in p_formula variables.\n", removed_sc))
  }

  sc_clean <- if (na_mode %in% c("omit", "exclude", "fail")) {
    sc[keep_sc, , drop = FALSE]
  } else {
    sc
  }

  #----------------------------------------------------------
  # 2. sp: one-reference
  #----------------------------------------------------------
  if (is.data.frame(sp)) {

    if (length(vars_sc) > 0) {
      log <- c(log, .na_summary(sp, vars_sc, "sp"))
      keep_sp <- stats::complete.cases(sp[, vars_sc, drop = FALSE])
    } else {
      keep_sp <- rep(TRUE, nrow(sp))
    }
    removed_sp <- sum(!keep_sp)

    if (na_mode == "fail" && removed_sp > 0) {
      stop("na.fail: NA found in p_formula variables in sp.", call. = FALSE)
    }

    if (na_mode == "pass") {
      keep_sp[] <- TRUE
    } else if (na_mode %in% c("omit", "exclude") && removed_sp > 0) {
      log <- c(log, sprintf("\nRemoved %d rows from sp due to NA in p_formula variables.\n", removed_sp))
    }

    sp_clean <- if (na_mode %in% c("omit", "exclude", "fail")) {
      sp[keep_sp, , drop = FALSE]
    } else {
      sp
    }

    keep_sp_out <- keep_sp

    #----------------------------------------------------------
    # 2. sp: multi-reference
    #----------------------------------------------------------
  } else if (is.list(sp) && all(vapply(sp, is.data.frame, logical(1)))) {

    sp_clean     <- vector("list", length(sp))
    keep_sp_list <- vector("list", length(sp))

    sp_names  <- names(sp)
    use_names <- if (!is.null(sp_names) && all(nzchar(sp_names))) {
      sp_names
    } else {
      paste0("sp[[", seq_along(sp), "]]")
    }

    for (j in seq_along(sp)) {

      if (!is.list(p_formula) || length(p_formula) < j) {
        stop("For multi-reference sp, p_formula must be a list with same length as sp.", call. = FALSE)
      }

      vars_j <- all.vars(p_formula[[j]])

      if (length(vars_j) > 0) {
        log <- c(log, .na_summary(sp[[j]], vars_j, use_names[j]))
        keep_spj <- stats::complete.cases(sp[[j]][, vars_j, drop = FALSE])
      } else {
        keep_spj <- rep(TRUE, nrow(sp[[j]]))
      }
      removed_spj <- sum(!keep_spj)

      if (na_mode == "fail" && removed_spj > 0) {
        stop(sprintf("na.fail: NA found in p_formula variables in %s.", use_names[j]), call. = FALSE)
      }

      if (na_mode == "pass") {
        keep_spj[] <- TRUE
      } else if (na_mode %in% c("omit", "exclude") && removed_spj > 0) {
        log <- c(log, sprintf("\nRemoved %d rows from %s due to NA in p_formula variables.\n",
                              removed_spj, use_names[j]))
      }

      sp_clean[[j]] <- if (na_mode %in% c("omit", "exclude", "fail")) {
        sp[[j]][keep_spj, , drop = FALSE]
      } else {
        sp[[j]]
      }
      keep_sp_list[[j]] <- keep_spj
    }

    names(sp_clean)  <- use_names
    keep_sp_out      <- keep_sp_list

  } else {
    stop("'sp' must be a data.frame (one-reference) or a list of data.frames (multi-reference).",
         call. = FALSE)
  }

  #----------------------------------------------------------
  # 3. lm-style na.action object for sc mapping
  #----------------------------------------------------------
  na_action_obj <- NULL
  if (na_mode %in% c("omit", "exclude")) {
    idx_drop <- which(!keep_sc)
    if (length(idx_drop) > 0) {
      cls           <- if (na_mode == "exclude") "exclude" else "omit"
      na_action_obj <- structure(idx_drop, class = cls)
    }
  }

  return(list(
    sc        = sc_clean,
    sp        = sp_clean,
    keep_sc   = keep_sc,
    keep_sp   = keep_sp_out,
    na_action = na_action_obj,
    log       = log
  ))
}



#' Summarise NA exclusions into a pw_na_summary object
#'
#' Computes per-dataset row counts (original, used, excluded) from the
#' `keep_sc` and `keep_sp` logical vectors produced by
#' `handle_na_for_ipwm()`. If `verbose = TRUE`, prints a one-line
#' message per dataset via `message()`. Returns `NULL` when no rows
#' were excluded in any dataset, or a `pw_na_summary` object otherwise.
#'
#' @param res The intermediate result list from `process_na_build()`,
#'   which must contain `keep_sc`, `keep_sp`, and `n_sp_orig`.
#' @param n_sc_orig Integer. Original row count of `sc` before NA removal.
#' @param n_ref Integer. Number of reference surveys.
#' @param verbose Logical. If `TRUE`, prints exclusion counts via
#'   `message()`.
#'
#' @return A `pw_na_summary` object (a list with elements `sc`, `sp`,
#'   and `n_ref`, classed `"pw_na_summary"`), or `NULL` if no rows were
#'   excluded.
#'
#' @keywords internal
#' @noRd
.report_na_exclusions <- function(res, n_sc_orig, n_ref, verbose = FALSE) {
  keep_sc   <- res$keep_sc
  keep_sp   <- res$keep_sp
  n_sp_orig <- res$n_sp_orig

  n_sc_used <- if (is.logical(keep_sc)) sum(keep_sc) else length(keep_sc)
  n_sc_excl <- n_sc_orig - n_sc_used
  sc_info   <- list(n_orig = n_sc_orig, n_used = n_sc_used, n_excluded = n_sc_excl)

  if (verbose) {
    if (n_sc_excl > 0) {
      message(sprintf(
        "sc: %d rows used, %d excluded due to missing participation model variables.",
        n_sc_used, n_sc_excl
      ))
    } else {
      message(sprintf("sc: %d rows, no exclusions.", n_sc_used))
    }
  }

  if (n_ref == 1) {
    n_sp_used <- if (is.logical(keep_sp)) sum(keep_sp) else n_sp_orig
    n_sp_excl <- n_sp_orig - n_sp_used
    sp_info   <- list(n_orig = n_sp_orig, n_used = n_sp_used, n_excluded = n_sp_excl)

    if (verbose) {
      if (n_sp_excl > 0) {
        message(sprintf(
          "sp: %d rows used, %d excluded due to missing participation model variables.",
          n_sp_used, n_sp_excl
        ))
      } else {
        message(sprintf("sp: %d rows, no exclusions.", n_sp_orig))
      }
    }
  } else {
    sp_info <- vector("list", length(keep_sp))
    for (.i in seq_along(keep_sp)) {
      n_sp_used_i <- if (is.logical(keep_sp[[.i]])) sum(keep_sp[[.i]]) else n_sp_orig[.i]
      n_sp_excl_i <- n_sp_orig[.i] - n_sp_used_i
      sp_info[[.i]] <- list(n_orig = n_sp_orig[.i], n_used = n_sp_used_i, n_excluded = n_sp_excl_i)

      if (verbose) {
        if (n_sp_excl_i > 0) {
          message(sprintf(
            "sp[[%d]]: %d rows used, %d excluded due to missing participation model variables.",
            .i, n_sp_used_i, n_sp_excl_i
          ))
        } else {
          message(sprintf("sp[[%d]]: %d rows, no exclusions.", .i, n_sp_orig[.i]))
        }
      }
    }
  }

  any_excl <- n_sc_excl > 0 || {
    if (n_ref == 1) {
      sp_info$n_excluded > 0
    } else {
      any(vapply(sp_info, function(x) x$n_excluded > 0, logical(1)))
    }
  }

  if (!any_excl) return(NULL)

  out <- list(sc = sc_info, sp = sp_info, n_ref = n_ref)
  class(out) <- "pw_na_summary"
  out
}



#' Print method for pw_na_summary
#'
#' Prints a formatted table showing the original row count, rows used,
#' and rows excluded due to missing participation model variables, for
#' each dataset (`sc` and each reference survey).
#'
#' @param x A `pw_na_summary` object returned by `.report_na_exclusions()`.
#' @param ... Further arguments passed to or from other methods (unused).
#'
#' @return Invisibly returns `x`.
#'
#' @method print pw_na_summary
#' @export
print.pw_na_summary <- function(x, ...) {
  cat("NA processing summary:\n")

  if (x$n_ref == 1) {
    rows <- list(sc = x$sc, sp = x$sp)
  } else {
    rows <- c(list(sc = x$sc), x$sp)
    sp_labels <- if (!is.null(names(x$sp)) && all(nzchar(names(x$sp)))) {
      names(x$sp)
    } else {
      paste0("sp[[", seq_along(x$sp), "]]")
    }
    names(rows) <- c("sc", sp_labels)
  }

  df <- do.call(rbind, lapply(rows, function(r) {
    data.frame(n_orig = r$n_orig, n_used = r$n_used, n_excluded = r$n_excluded)
  }))

  print(df)
  invisible(x)
}

# Setup ----

data(sc)
data(sp1)
data(sp2)

des1 <- survey::svydesign(
  ids     = ~psu_sp1,
  strata  = ~strata_sp1,
  weights = ~wts_sp1,
  data    = sp1,
  nest    = TRUE
)

des2 <- survey::svydesign(
  ids     = ~psu_sp2,
  strata  = ~strata_sp2,
  weights = ~wts_sp2,
  data    = sp2,
  nest    = TRUE
)


# resolve_na_action ----

test_that("resolve_na_action handles function inputs", {
  expect_equal(resolve_na_action(stats::na.omit),    "omit")
  expect_equal(resolve_na_action(stats::na.exclude), "exclude")
  expect_equal(resolve_na_action(stats::na.fail),    "fail")
  expect_equal(resolve_na_action(stats::na.pass),    "pass")
})

test_that("resolve_na_action handles character inputs", {
  expect_equal(resolve_na_action("omit"),       "omit")
  expect_equal(resolve_na_action("na.omit"),    "omit")
  expect_equal(resolve_na_action("exclude"),    "exclude")
  expect_equal(resolve_na_action("na.exclude"), "exclude")
  expect_equal(resolve_na_action("fail"),       "fail")
  expect_equal(resolve_na_action("na.fail"),    "fail")
  expect_equal(resolve_na_action("pass"),       "pass")
  expect_equal(resolve_na_action("na.pass"),    "pass")
})

test_that("resolve_na_action rejects unsupported function", {
  expect_error(
    resolve_na_action(mean),
    "Unsupported na.action function",
    fixed = TRUE
  )
})

test_that("resolve_na_action rejects invalid string", {
  expect_error(
    resolve_na_action("drop"),
    "Invalid na.action string",
    fixed = TRUE
  )
})

test_that("resolve_na_action rejects invalid type", {
  expect_error(
    resolve_na_action(1L),
    "Invalid na.action type",
    fixed = TRUE
  )
})


# handle_na_for_ipwm ----

test_that("handle_na_for_ipwm (omit): removes NA rows from sc and sp", {
  sc_na       <- sc;  sc_na$agecat[1:3] <- NA
  sp_na       <- sp1; sp_na$agecat[1:2] <- NA

  res <- handle_na_for_ipwm(sc_na, sp_na, ~agecat, na_mode = "omit")

  expect_equal(nrow(res$sc), nrow(sc) - 3)
  expect_equal(nrow(res$sp), nrow(sp1) - 2)
  expect_equal(sum(!res$keep_sc), 3)
  expect_equal(sum(!res$keep_sp), 2)
})

test_that("handle_na_for_ipwm (omit): na_action has class 'omit'", {
  sc_na <- sc; sc_na$agecat[1:2] <- NA

  res <- handle_na_for_ipwm(sc_na, sp1, ~agecat, na_mode = "omit")

  expect_equal(class(res$na_action), "omit")
  expect_equal(as.integer(res$na_action), 1:2)
})

test_that("handle_na_for_ipwm (exclude): na_action has class 'exclude'", {
  sc_na <- sc; sc_na$agecat[1] <- NA

  res <- handle_na_for_ipwm(sc_na, sp1, ~agecat, na_mode = "exclude")

  expect_equal(class(res$na_action), "exclude")
  expect_equal(as.integer(res$na_action), 1L)
})

test_that("handle_na_for_ipwm (fail): errors when NA in sc", {
  sc_na <- sc; sc_na$agecat[1] <- NA

  expect_error(
    handle_na_for_ipwm(sc_na, sp1, ~agecat, na_mode = "fail"),
    "na.fail: NA found in p_formula variables in sc.",
    fixed = TRUE
  )
})

test_that("handle_na_for_ipwm (fail): errors when NA in sp", {
  sp_na <- sp1; sp_na$agecat[1] <- NA

  expect_error(
    handle_na_for_ipwm(sc, sp_na, ~agecat, na_mode = "fail"),
    "na.fail: NA found in p_formula variables in sp.",
    fixed = TRUE
  )
})

test_that("handle_na_for_ipwm (pass): keeps all rows despite NA", {
  sc_na <- sc;  sc_na$agecat[1:5] <- NA
  sp_na <- sp1; sp_na$agecat[1:3] <- NA

  res <- handle_na_for_ipwm(sc_na, sp_na, ~agecat, na_mode = "pass")

  expect_equal(nrow(res$sc), nrow(sc))
  expect_equal(nrow(res$sp), nrow(sp1))
  expect_true(all(res$keep_sc))
  expect_true(all(res$keep_sp))
})

test_that("handle_na_for_ipwm: no NA -> all rows kept, na_action NULL", {
  res <- handle_na_for_ipwm(sc, sp1, ~agecat, na_mode = "omit")

  expect_true(all(res$keep_sc))
  expect_true(all(res$keep_sp))
  expect_null(res$na_action)
})

test_that("handle_na_for_ipwm (multi-ref): each sp handled independently", {
  sp_na1 <- sp1; sp_na1$agecat[1:2] <- NA
  sp_list <- list(sp_na1, sp2)
  fml_list <- list(~agecat, ~agecat)

  res <- handle_na_for_ipwm(sc, sp_list, fml_list, na_mode = "omit")

  expect_equal(sum(!res$keep_sp[[1]]), 2)
  expect_true(all(res$keep_sp[[2]]))
  expect_equal(nrow(res$sp[[1]]), nrow(sp1) - 2)
  expect_equal(nrow(res$sp[[2]]), nrow(sp2))
})


# process_na_build ----

test_that("process_na_build: na_summary is NULL when no NAs", {
  res <- process_na_build(
    sc        = sc,
    sp        = sp1,
    sp_des    = des1,
    p_formula = ~agecat,
    na.action = stats::na.omit,
    n_ref     = 1L,
    verbose   = FALSE
  )

  expect_null(res$na_summary)
})

test_that("process_na_build: na_summary has correct class and counts", {
  sc_na <- sc;  sc_na$agecat[1:3] <- NA
  sp_na <- sp1; sp_na$agecat[1:2] <- NA
  des_na <- survey::svydesign(
    ids     = ~psu_sp1,
    strata  = ~strata_sp1,
    weights = ~wts_sp1,
    data    = sp_na,
    nest    = TRUE
  )

  res <- process_na_build(
    sc        = sc_na,
    sp        = sp_na,
    sp_des    = des_na,
    p_formula = ~agecat,
    na.action = stats::na.omit,
    n_ref     = 1L,
    verbose   = FALSE
  )

  expect_s3_class(res$na_summary, "pw_na_summary")
  expect_equal(res$na_summary$sc$n_orig,     nrow(sc_na))
  expect_equal(res$na_summary$sc$n_excluded, 3)
  expect_equal(res$na_summary$sp$n_orig,     nrow(sp_na))
  expect_equal(res$na_summary$sp$n_excluded, 2)
})

test_that("process_na_build: sp_des is subsetted after NA removal", {
  sp_na <- sp1; sp_na$agecat[1:4] <- NA
  des_na <- survey::svydesign(
    ids     = ~psu_sp1,
    strata  = ~strata_sp1,
    weights = ~wts_sp1,
    data    = sp_na,
    nest    = TRUE
  )

  res <- process_na_build(
    sc        = sc,
    sp        = sp_na,
    sp_des    = des_na,
    p_formula = ~agecat,
    na.action = stats::na.omit,
    n_ref     = 1L,
    verbose   = FALSE
  )

  expect_equal(nrow(res$sp_des$variables), nrow(sp1) - 4)
})

test_that("process_na_build: na.fail errors when NA present", {
  sc_na <- sc; sc_na$agecat[1] <- NA

  expect_error(
    process_na_build(
      sc        = sc_na,
      sp        = sp1,
      sp_des    = des1,
      p_formula = ~agecat,
      na.action = stats::na.fail,
      n_ref     = 1L,
      verbose   = FALSE
    ),
    "na.fail: NA found in p_formula variables in sc.",
    fixed = TRUE
  )
})

test_that("process_na_build: verbose = TRUE sends messages", {
  sc_na <- sc; sc_na$agecat[1] <- NA

  expect_message(
    process_na_build(
      sc        = sc_na,
      sp        = sp1,
      sp_des    = des1,
      p_formula = ~agecat,
      na.action = stats::na.omit,
      n_ref     = 1L,
      verbose   = TRUE
    )
  )
})

#' Nonprobability Sample (sc)
#'
#' This dataset represents a synthetic nonprobability sample generated via
#' Poisson sampling from a finite population constructed from the National
#' Health and Nutrition Examination Survey (NHANES) cycles 1999--2010. It is
#' intended to illustrate the pseudo-weighting methods implemented in the
#' \code{nonprobsampling} package.
#'
#' @format A data frame with 2404 observations and 8 variables:
#' \describe{
#'   \item{psa_level}{Outcome variable: serum prostate-specific antigen level (numeric)}
#'   \item{BMI}{Body mass index category (factor with 4 levels: "Normal", "Overweight", "Obese", "Morbidly Obese")}
#'   \item{race}{Race category (factor with 4 levels: 1 = White, 2 = Black, 3 = Hispanic, 4 = Other)}
#'   \item{agecat}{Age category (factor with 4 levels: 1 = 55--59, 2 = 60--64, 3 = 65--69, 4 = 70+)}
#'   \item{education}{Education level (factor with 5 levels: 1 = Less Than 8 Years, 2 = 8--11 Years, 3 = 12 Years Or Completed High School, 4 = College Graduate, 5 = Postgraduate)}
#'   \item{pros_enlarged}{Prostate enlargement indicator (factor with 2 levels: 0 = No, 1 = Yes)}
#'   \item{comorbidity}{General comorbidity indicator (factor with 2 levels: 0 = No, 1 = Yes)}
#'   \item{diabetes}{Diabetes diagnosis indicator (factor with 2 levels: 0 = No, 1 = Yes)}
#' }
#'
#' @details
#' The dataset has 2,404 complete-case observations, with
#' \code{psa_level} serving as the outcome variable. Auxiliary variables
#' shared with the probability reference surveys \code{sp1} and \code{sp2}
#' are used to construct pseudo-weights aimed at correcting for participation
#' bias.
#'
#' @source
#' Synthetic data generated by the package authors. The underlying finite
#' population was constructed from the National Health and Nutrition
#' Examination Survey (NHANES), 1999--2010 cycles, conducted by the U.S.
#' National Center for Health Statistics (NCHS).
#'
#' @usage data(sc)
#'
#' @examples
#' data(sc)
#' str(sc)
#' summary(sc)
"sc"

#' Compute design-based covariance matrix D for CLW
#'
#' Internal helper to compute
#' \deqn{D = Var_p\left( \sum_{j \in s_p} d_j \pi_j x_j \right)}
#' from an existing survey design object created by
#' survey::svydesign() or survey::svrepdesign().
#'
#' In the CLW method, the probability sample contribution in the
#' estimating equation is \eqn{\pi_j d_j x_j}, where `pi_j` corresponds
#' to the inverse of the estimated pseudo-weight. Thus the design-based
#' covariance matrix is obtained by treating \eqn{h_j = \pi_j x_j}
#' as the survey total variable.
#'
#' @param sp_des A survey design object of class `"survey.design2"` or
#'   `"svyrep.design"`.
#' @param pi_sp Numeric vector of estimated values multiplying `X_p`
#'   in the CLW estimating equation. Length must equal the number of
#'   rows in `sp_des$variables`.
#' @param Xp Numeric matrix of dimension `n_p x p`. Each row is the
#'   covariate vector `x_j` used in the CLW estimating equation.
#'
#' @return A `p x p` covariance matrix `D`.
#' @keywords internal
compute_D_CLW <- function(sp_des, pi_sp, Xp) {

  if (!inherits(sp_des, c("survey.design2", "svyrep.design"))) {
    stop(
      "'sp_des' must be a survey design object from svydesign() or svrepdesign().",
      call. = FALSE
    )
  }

  if (!is.matrix(Xp)) {
    Xp <- as.matrix(Xp)
  }

  n_sp <- nrow(sp_des$variables)

  if (length(pi_sp) != n_sp) {
    stop(
      sprintf(
        "Length mismatch: length(pi_sp) = %d but the survey design object contains %d observations.",
        length(pi_sp), n_sp
      ),
      call. = FALSE
    )
  }

  if (nrow(Xp) != n_sp) {
    stop(
      sprintf(
        "Row mismatch: nrow(Xp) = %d but the survey design object contains %d observations.",
        nrow(Xp), n_sp
      ),
      call. = FALSE
    )
  }

  if (!is.numeric(pi_sp)) {
    stop("'pi_sp' must be numeric.", call. = FALSE)
  }

  if (anyNA(pi_sp)) {
    stop("'pi_sp' contains NA values.", call. = FALSE)
  }

  if (anyNA(Xp)) {
    stop("'Xp' contains NA values.", call. = FALSE)
  }

  # safe column names for formula parsing
  xp_names <- colnames(Xp)
  if (is.null(xp_names)) {
    xp_names <- paste0("x", seq_len(ncol(Xp)))
  }
  xp_names <- make.names(xp_names, unique = TRUE)

  # CLW h_j = pi_j * x_j
  H <- sweep(Xp, 1, pi_sp, "*")

  H_df <- as.data.frame(H)
  colnames(H_df) <- paste0("h_", xp_names)

  # attach h_j columns to existing survey design object
  sp_des_h <- sp_des
  sp_des_h$variables <- cbind(sp_des_h$variables, H_df)

  # no intercept: H already contains all columns of h_j
  fml <- stats::reformulate(colnames(H_df), intercept = FALSE)

  tot <- survey::svytotal(fml, design = sp_des_h)
  D   <- stats::vcov(tot)

  return(D)
}

# Runs the pwmean estimation pipeline for a numeric outcome.
pwmean_numeric <- function(
    object,
    y,
    zcol = NULL,
    na.action = stats::na.omit,
    mc = match.call(),
    check_inputs = TRUE
) {

  #------------------------------#
  # Step 1: input checking
  #------------------------------#
  if (isTRUE(check_inputs)) {
    tryCatch(
      check_ipwm_inputs_estimate(object, y, zcol),
      error = function(e)
        stop("Step 1 (input check) failed: ", e$message, call. = FALSE)
    )
  }

  #------------------------------#
  # Step 2: prepare sc data
  #------------------------------#
  sc_data <- tryCatch(
    prepare_sc_data(object),
    error = function(e)
      stop("Step 2 (prepare_sc_data) failed: ", e$message, call. = FALSE)
  )

  #------------------------------#
  # Step 3: NA processing
  #------------------------------#
  yz_data <- tryCatch(
    process_na_yz(
      sc_data = sc_data,
      y = y,
      zcol = zcol,
      na.action = na.action
    ),
    error = function(e)
      stop("Step 3 (process_na_yz) failed: ", e$message, call. = FALSE)
  )

  #------------------------------#
  # Step 4: dispatch estimator
  #------------------------------#
  est <- tryCatch(
    dispatch_estimator(
      build = object,
      yz_data = yz_data
    ),
    error = function(e)
      stop("Step 4 (dispatch_estimator) failed: ", e$message, call. = FALSE)
  )

  #------------------------------#
  # Step 5: naive estimator
  #------------------------------#
  naive <- tryCatch(
    naive_mean(
      object$internal$raw_sc,
      domain_var = zcol,
      y = y
    ),
    error = function(e)
      stop("Step 5 (naive_mean) failed: ", e$message, call. = FALSE)
  )

  #------------------------------#
  # Step 6: assemble output
  #------------------------------#
  result <- tryCatch(
    assemble_output(
      build = object,
      est = est,
      naive = naive,
      na_info = yz_data$na_info
    ),
    error = function(e)
      stop("Step 6 (assemble_output) failed: ", e$message, call. = FALSE)
  )

  stopifnot(is.list(result))

  result$call <- mc

  class(result) <- "pwmean"

  return(result)
}

#' Validate inputs at the estimation stage
#'
#' Checks that `build` is a valid `pw_fit` object with all required internal
#' fields, that `y` names a numeric or factor column in the raw convenience
#' sample, and
#' that `zcol` (if supplied) is a supported domain variable with at least two
#' non-missing levels in both the full sample and the build-stage complete cases.
#'
#' @param build A `pw_fit` object returned by the build step.
#' @param y Single character string naming the numeric or categorical outcome
#'   variable in `build$internal$raw_sc`.
#' @param zcol Single character string naming the domain variable in
#'   `build$internal$raw_sc`, or NULL for the overall mean.
#'
#' @return `invisible(TRUE)` on success; stops with an informative message
#'   on failure.
#'
#' @keywords internal
check_ipwm_inputs_estimate <- function(build, y, zcol = NULL) {

  #----------------------------------#
  # Step 1: required fields in build
  #----------------------------------#
  if (!inherits(build, "pw_fit")) {
    stop("'object' must be of class \"pw_fit\".", call. = FALSE)
  }

  if (is.null(build$internal$raw_sc)) {
    stop("build$internal$raw_sc is missing.", call. = FALSE)
  }

  if (is.null(build$internal$na$keep_sc)) {
    stop("build$internal$na$keep_sc is missing.", call. = FALSE)
  }

  if (is.null(build$pseudo_weights)) {
    stop("build$pseudo_weights is missing.", call. = FALSE)
  }

  if (is.null(build$internal$Xc)) {
    stop("build$internal$Xc is missing.", call. = FALSE)
  }

  if (is.null(build$method) || !is.character(build$method) || length(build$method) != 1L) {
    stop("build$method must be a single character string.", call. = FALSE)
  }

  if (is.null(build$internal$D)) {
    stop("build$internal$D is missing.", call. = FALSE)
  }

  if (is.null(build$internal$S_beta)) {
    stop("build$internal$S_beta is missing.", call. = FALSE)
  }

  sc <- build$internal$raw_sc

  #----------------------------------#
  # Step 2: check y
  #----------------------------------#
  if (!is.character(y) || length(y) != 1L || is.na(y) || !nzchar(y)) {
    stop("'y' must be a single non-empty character string.", call. = FALSE)
  }

  if (!(y %in% names(sc))) {
    stop(sprintf("Outcome variable '%s' not found in sc.", y), call. = FALSE)
  }

  if (!(is.numeric(sc[[y]]) || is.factor(sc[[y]]))) {
    stop(sprintf("Outcome variable '%s' must be numeric or factor.", y), call. = FALSE)
  }

  if (is.factor(sc[[y]])) {
    nonmiss_levels <- levels(droplevels(sc[[y]]))
    if (length(nonmiss_levels) < 1L) {
      stop(
        sprintf("Outcome variable '%s' must contain at least one non-missing level.", y),
        call. = FALSE
      )
    }
  }

  #----------------------------------#
  # Step 3: check zcol
  #----------------------------------#
  if (!is.null(zcol)) {

    if (!is.character(zcol) || length(zcol) != 1L || is.na(zcol) || !nzchar(zcol)) {
      stop("'zcol' must be a single non-empty character string or NULL.", call. = FALSE)
    }

    if (!(zcol %in% names(sc))) {
      stop(sprintf("Domain variable '%s' not found in sc.", zcol), call. = FALSE)
    }

    z <- sc[[zcol]]

    # validate type and level count in raw_sc; use if/else if so execution
    # falls through to Step 4 (sc_keep level check) after a passing branch
    if (is.logical(z)) {
      if (length(unique(stats::na.omit(z))) < 2L) {
        stop(
          sprintf("Domain variable '%s' must contain both TRUE and FALSE.", zcol),
          call. = FALSE
        )
      }

    } else if (is.factor(z)) {
      nonmiss_levels <- levels(droplevels(z))
      if (length(nonmiss_levels) < 2L) {
        stop(
          sprintf(
            "Domain variable '%s' must contain at least two non-missing levels.",
            zcol
          ),
          call. = FALSE
        )
      }

    } else if (is.character(z)) {
      # trimws() and blank-to-NA follow what standardize_zcol() does before
      # building the factor, so the level count here matches what estimation sees
      z_trim <- trimws(z)
      z_trim[z_trim == ""] <- NA_character_
      nonmiss_levels <- unique(stats::na.omit(z_trim))

      if (length(nonmiss_levels) < 2L) {
        stop(
          sprintf(
            "Domain variable '%s' must contain at least two non-missing levels.",
            zcol
          ),
          call. = FALSE
        )
      }

    } else if (is.numeric(z)) {
      # allow numeric/integer only if binary (ignoring NA)
      uniq_vals <- sort(unique(stats::na.omit(z)))

      if (!all(uniq_vals %in% c(0, 1))) {
        stop(
          sprintf(
            paste0(
              "Domain variable '%s' is numeric but not binary. ",
              "Allowed numeric domain variables must contain only {0, 1} ",
              "after removing missing values. Found values: %s"
            ),
            zcol,
            paste(uniq_vals, collapse = ", ")
          ),
          call. = FALSE
        )
      }

      if (length(uniq_vals) < 2L) {
        stop(
          sprintf(
            "Domain variable '%s' must contain both 0 and 1.",
            zcol
          ),
          call. = FALSE
        )
      }

    } else {
      # unsupported type
      stop(
        sprintf(
          paste0(
            "Domain variable '%s' must be one of the following: ",
            "logical, binary numeric/integer (0/1), character, factor, or NULL."
          ),
          zcol
        ),
        call. = FALSE
      )
    }
  }

  #----------------------------------#
  # Step 4: check zcol levels in sc_keep
  #
  # Notice: Step 3 confirms zcol has >= 2 non-missing levels in raw_sc.
  # but dispatch_estimator in step 4 operates on sc_keep. If keep_sc
  # removes all rows belonging to a level, est$labels will differ from
  # naive$labels (built from raw_sc), causing assemble_output to error
  # with label mismatch message. We catch this here so the user gets
  # an  error at input-check step instead.
  #----------------------------------#
  if (!is.null(zcol)) {

    sc_keep <- sc[which(build$internal$na$keep_sc), , drop = FALSE]
    z_keep  <- sc_keep[[zcol]]

    # follow the same level-counting logic as Step 3 for each type
    n_levels <- if (is.factor(z_keep)) {
      length(levels(droplevels(z_keep)))
    } else if (is.character(z_keep)) {
      z_tr <- trimws(z_keep)
      z_tr[z_tr == ""] <- NA_character_
      length(unique(stats::na.omit(z_tr)))
    } else {
      # logical and numeric: count distinct non-missing values
      length(unique(stats::na.omit(z_keep)))
    }

    if (n_levels < 2L) {
      stop(
        sprintf(
          paste0(
            "Domain variable '%s' has fewer than two non-missing levels in ",
            "the build-stage complete cases. Build-stage NA removal may have ",
            "eliminated an entire level."
          ),
          zcol
        ),
        call. = FALSE
      )
    }
  }

  invisible(TRUE)
}

# Setup ----

data(sc)

# helper: build a minimal pw_fit around a data frame
make_pw_fit <- function(raw_sc, weights = NULL, keep_sc = NULL, Xc = NULL,
                        method = "alp", D = NULL, S_beta = NULL) {
  n      <- nrow(raw_sc)
  if (is.null(keep_sc)) keep_sc <- rep(TRUE, n)
  n_keep <- sum(keep_sc)
  if (is.null(weights)) weights <- rep(1.0, n_keep)
  if (is.null(Xc))      Xc      <- matrix(1.0, nrow = n_keep, ncol = 1L)
  if (is.null(D))       D       <- matrix(0.0, nrow = 1L,     ncol = 1L)
  if (is.null(S_beta))  S_beta  <- matrix(1.0, nrow = 1L,     ncol = 1L)
  structure(
    list(
      pseudo_weights = weights,
      method         = method,
      internal = list(
        raw_sc = raw_sc,
        na     = list(keep_sc = keep_sc),
        Xc     = Xc,
        D      = D,
        S_beta = S_beta
      )
    ),
    class = "pw_fit"
  )
}

fit <- make_pw_fit(sc)


# valid inputs ----

test_that("check_ipwm_inputs_estimate: valid input without zcol returns invisible TRUE", {
  expect_invisible(check_ipwm_inputs_estimate(fit, y = "psa_level"))
})

test_that("check_ipwm_inputs_estimate: valid input with factor zcol returns invisible TRUE", {
  expect_invisible(check_ipwm_inputs_estimate(fit, y = "psa_level", zcol = "race"))
})

test_that("check_ipwm_inputs_estimate: valid input with logical zcol returns invisible TRUE", {
  sc_log          <- sc
  sc_log$is_obese <- sc_log$BMI == "Obese"
  fit_log         <- make_pw_fit(sc_log)
  expect_invisible(check_ipwm_inputs_estimate(fit_log, y = "psa_level", zcol = "is_obese"))
})

test_that("check_ipwm_inputs_estimate: rejects logical zcol with only one unique value", {
  sc_log          <- sc
  sc_log$all_true <- rep(TRUE, nrow(sc))
  fit_log         <- make_pw_fit(sc_log)
  expect_error(
    check_ipwm_inputs_estimate(fit_log, y = "psa_level", zcol = "all_true"),
    "must contain both TRUE and FALSE.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: valid input with character zcol returns invisible TRUE", {
  sc_chr          <- sc
  sc_chr$race_chr <- as.character(sc_chr$race)
  fit_chr         <- make_pw_fit(sc_chr)
  expect_invisible(check_ipwm_inputs_estimate(fit_chr, y = "psa_level", zcol = "race_chr"))
})

test_that("check_ipwm_inputs_estimate: valid input with binary numeric zcol returns invisible TRUE", {
  sc_bin            <- sc
  sc_bin$has_diabetes <- as.integer(sc_bin$diabetes == "1")
  fit_bin           <- make_pw_fit(sc_bin)
  expect_invisible(check_ipwm_inputs_estimate(fit_bin, y = "psa_level", zcol = "has_diabetes"))
})

test_that("check_ipwm_inputs_estimate: valid input with factor y returns invisible TRUE", {
  expect_invisible(check_ipwm_inputs_estimate(fit, y = "diabetes"))
})


# build object checks ----

test_that("check_ipwm_inputs_estimate: rejects non-pw_fit object", {
  expect_error(
    check_ipwm_inputs_estimate(list(), y = "psa_level"),
    "'object' must be of class \"pw_fit\".",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects missing raw_sc", {
  bad                    <- fit
  bad$internal$raw_sc   <- NULL
  expect_error(
    check_ipwm_inputs_estimate(bad, y = "psa_level"),
    "build$internal$raw_sc is missing.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects missing keep_sc", {
  bad                       <- fit
  bad$internal$na$keep_sc  <- NULL
  expect_error(
    check_ipwm_inputs_estimate(bad, y = "psa_level"),
    "build$internal$na$keep_sc is missing.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects missing pseudo_weights", {
  bad                  <- fit
  bad$pseudo_weights   <- NULL
  expect_error(
    check_ipwm_inputs_estimate(bad, y = "psa_level"),
    "build$pseudo_weights is missing.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects missing Xc", {
  bad              <- fit
  bad$internal$Xc  <- NULL
  expect_error(
    check_ipwm_inputs_estimate(bad, y = "psa_level"),
    "build$internal$Xc is missing.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects NULL method", {
  bad         <- fit
  bad$method  <- NULL
  expect_error(
    check_ipwm_inputs_estimate(bad, y = "psa_level"),
    "build$method must be a single character string.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects non-character method", {
  bad         <- fit
  bad$method  <- 1L
  expect_error(
    check_ipwm_inputs_estimate(bad, y = "psa_level"),
    "build$method must be a single character string.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects missing D", {
  bad              <- fit
  bad$internal$D   <- NULL
  expect_error(
    check_ipwm_inputs_estimate(bad, y = "psa_level"),
    "build$internal$D is missing.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects missing S_beta", {
  bad                  <- fit
  bad$internal$S_beta  <- NULL
  expect_error(
    check_ipwm_inputs_estimate(bad, y = "psa_level"),
    "build$internal$S_beta is missing.",
    fixed = TRUE
  )
})


# y checks ----

test_that("check_ipwm_inputs_estimate: rejects non-character y", {
  expect_error(
    check_ipwm_inputs_estimate(fit, y = 1L),
    "'y' must be a single non-empty character string.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects NA y", {
  expect_error(
    check_ipwm_inputs_estimate(fit, y = NA_character_),
    "'y' must be a single non-empty character string.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects empty string y", {
  expect_error(
    check_ipwm_inputs_estimate(fit, y = ""),
    "'y' must be a single non-empty character string.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects y not found in sc", {
  expect_error(
    check_ipwm_inputs_estimate(fit, y = "not_a_column"),
    "Outcome variable 'not_a_column' not found in sc.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects unsupported y type", {
  sc_chr          <- sc
  sc_chr$race_chr <- as.character(sc_chr$race)
  fit_chr         <- make_pw_fit(sc_chr)
  expect_error(
    check_ipwm_inputs_estimate(fit_chr, y = "race_chr"),
    "Outcome variable 'race_chr' must be numeric or factor.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects factor y with no non-missing level", {
  sc_empty        <- sc
  sc_empty$empty  <- factor(rep(NA_character_, nrow(sc)))
  fit_empty       <- make_pw_fit(sc_empty)
  expect_error(
    check_ipwm_inputs_estimate(fit_empty, y = "empty"),
    "must contain at least one non-missing level.",
    fixed = TRUE
  )
})


# zcol checks ----

test_that("check_ipwm_inputs_estimate: rejects non-character zcol", {
  expect_error(
    check_ipwm_inputs_estimate(fit, y = "psa_level", zcol = 1L),
    "'zcol' must be a single non-empty character string or NULL.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects NA zcol", {
  expect_error(
    check_ipwm_inputs_estimate(fit, y = "psa_level", zcol = NA_character_),
    "'zcol' must be a single non-empty character string or NULL.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects empty string zcol", {
  expect_error(
    check_ipwm_inputs_estimate(fit, y = "psa_level", zcol = ""),
    "'zcol' must be a single non-empty character string or NULL.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects zcol not found in sc", {
  expect_error(
    check_ipwm_inputs_estimate(fit, y = "psa_level", zcol = "not_a_column"),
    "Domain variable 'not_a_column' not found in sc.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects factor zcol with only one level", {
  sc_one             <- sc
  sc_one$single_lev  <- factor(rep("A", nrow(sc)))
  fit_one            <- make_pw_fit(sc_one)
  expect_error(
    check_ipwm_inputs_estimate(fit_one, y = "psa_level", zcol = "single_lev"),
    "must contain at least two non-missing levels.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects character zcol with only one non-missing level", {
  sc_one      <- sc
  sc_one$grp  <- rep("A", nrow(sc))
  fit_one     <- make_pw_fit(sc_one)
  expect_error(
    check_ipwm_inputs_estimate(fit_one, y = "psa_level", zcol = "grp"),
    "must contain at least two non-missing levels.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: character zcol trimming treats ' A' and 'A' as same level", {
  sc_trim      <- sc
  sc_trim$grp  <- rep(c(" A", "A"), length.out = nrow(sc))
  fit_trim     <- make_pw_fit(sc_trim)
  expect_error(
    check_ipwm_inputs_estimate(fit_trim, y = "psa_level", zcol = "grp"),
    "must contain at least two non-missing levels.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects non-binary numeric zcol", {
  sc_num        <- sc
  sc_num$score  <- c(0, 1, 2, rep(0, nrow(sc) - 3))
  fit_num       <- make_pw_fit(sc_num)
  expect_error(
    check_ipwm_inputs_estimate(fit_num, y = "psa_level", zcol = "score"),
    "is numeric but not binary.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects numeric zcol containing only 0", {
  sc_zero           <- sc
  sc_zero$all_zeros <- rep(0L, nrow(sc))
  fit_zero          <- make_pw_fit(sc_zero)
  expect_error(
    check_ipwm_inputs_estimate(fit_zero, y = "psa_level", zcol = "all_zeros"),
    "must contain both 0 and 1.",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects unsupported zcol type", {
  sc_bad          <- sc
  sc_bad$bad_col  <- as.list(rep(1, nrow(sc)))
  fit_bad         <- make_pw_fit(sc_bad)
  expect_error(
    check_ipwm_inputs_estimate(fit_bad, y = "psa_level", zcol = "bad_col"),
    "must be one of the following",
    fixed = TRUE
  )
})


# zcol level check against sc_keep (build-stage filtered rows) ----
#
# Step 3 validates zcol against raw_sc. Step 4 repeats the level-count check
# against sc_keep so the user gets a clear error at input-check time rather
# than an uninformative label-mismatch error from assemble_output later.

test_that("check_ipwm_inputs_estimate: rejects factor zcol with one level after build-stage filtering", {
  n          <- nrow(sc)
  sc_grp     <- sc
  sc_grp$grp <- factor(rep(c("A", "B"), length.out = n))
  keep       <- sc_grp$grp != "A"   # remove all A rows from sc_keep
  fit_keep   <- make_pw_fit(sc_grp, keep_sc = keep)
  expect_error(
    check_ipwm_inputs_estimate(fit_keep, y = "psa_level", zcol = "grp"),
    "fewer than two non-missing levels in the build-stage complete cases",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects character zcol with one level after build-stage filtering", {
  n          <- nrow(sc)
  sc_grp     <- sc
  sc_grp$grp <- rep(c("A", "B"), length.out = n)
  keep       <- sc_grp$grp != "A"
  fit_keep   <- make_pw_fit(sc_grp, keep_sc = keep)
  expect_error(
    check_ipwm_inputs_estimate(fit_keep, y = "psa_level", zcol = "grp"),
    "fewer than two non-missing levels in the build-stage complete cases",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects logical zcol with one value after build-stage filtering", {
  n           <- nrow(sc)
  sc_log      <- sc
  sc_log$flag <- rep(c(TRUE, FALSE), length.out = n)
  keep        <- !sc_log$flag   # remove all TRUE rows from sc_keep
  fit_keep    <- make_pw_fit(sc_log, keep_sc = keep)
  expect_error(
    check_ipwm_inputs_estimate(fit_keep, y = "psa_level", zcol = "flag"),
    "fewer than two non-missing levels in the build-stage complete cases",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: rejects binary numeric zcol with one value after build-stage filtering", {
  n        <- nrow(sc)
  sc_bin   <- sc
  sc_bin$flag <- rep(c(0L, 1L), length.out = n)
  keep     <- sc_bin$flag == 0L   # remove all 1-rows from sc_keep
  fit_keep <- make_pw_fit(sc_bin, keep_sc = keep)
  expect_error(
    check_ipwm_inputs_estimate(fit_keep, y = "psa_level", zcol = "flag"),
    "fewer than two non-missing levels in the build-stage complete cases",
    fixed = TRUE
  )
})

test_that("check_ipwm_inputs_estimate: passes when zcol retains >= 2 levels after build-stage filtering", {
  n          <- nrow(sc)
  sc_grp     <- sc
  sc_grp$grp <- factor(rep(c("A", "B"), length.out = n))
  keep       <- rep(TRUE, n)
  keep[1L]   <- FALSE   # drop one row only; both levels remain
  fit_keep   <- make_pw_fit(sc_grp, keep_sc = keep)
  expect_invisible(check_ipwm_inputs_estimate(fit_keep, y = "psa_level", zcol = "grp"))
})

#--------------------------------------------------#
# helper: extract variables from formula
#--------------------------------------------------#
get_formula_vars <- function(fml) {
  vars <- all.vars(fml)
  vars <- setdiff(vars, c(".", "1"))
  vars
}

#--------------------------------------------------#
# helper: check variables in one formula
#--------------------------------------------------#
check_formula_vars_exist <- function(fml, sc, sp, label = "p_formula") {
  vars <- get_formula_vars(fml)

  miss_sc <- setdiff(vars, names(sc))
  miss_sp <- setdiff(vars, names(sp))

  if (length(miss_sc) > 0L) {
    stop(
      sprintf(
        "Variables in %s not found in `sc`: %s",
        label, paste(miss_sc, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  if (length(miss_sp) > 0L) {
    stop(
      sprintf(
        "Variables in %s not found in corresponding `sp`: %s",
        label, paste(miss_sp, collapse = ", ")
      ),
      call. = FALSE
    )
  }
}

#--------------------------------------------------#
# helper: check single psu problem
#--------------------------------------------------#
check_lonely_psu <- function(des, name = "sp_des") {

  if (!inherits(des, "survey.design2")) {
    return(invisible(NULL))
  }

  # both strata and cluster must exist
  if (is.null(des$strata) || is.null(des$cluster)) {
    return(invisible(NULL))
  }

  strata <- des$strata[[1]]
  psu    <- des$cluster[[1]]

  ok <- !(is.na(strata) | is.na(psu))
  strata <- strata[ok]
  psu    <- psu[ok]

  if (length(strata) == 0L) {
    return(invisible(NULL))
  }

  # number of unique PSUs in each stratum
  n_psu <- tapply(psu, strata, function(x) length(unique(x)))
  lonely_strata <- names(n_psu)[n_psu == 1L]

  if (length(lonely_strata) == 0L) {
    return(invisible(NULL))
  }

  max_show <- 5L
  show_strata <- utils::head(lonely_strata, max_show)

  more_flag <- if (length(lonely_strata) > max_show) {
    paste0(" ... (", length(lonely_strata) - max_show, " more)")
  } else {
    ""
  }

  lonely_opt <- getOption("survey.lonely.psu")
  if (is.null(lonely_opt)) lonely_opt <- "fail"
  lonely_opt <- tolower(trimws(as.character(lonely_opt)))

  msg <- paste0(
    "Lonely PSU detected in ", name, ".\n",
    "Strata with only one PSU: ",
    paste(show_strata, collapse = ", "),
    more_flag, ".\n",
    "Current setting survey.lonely.psu = \"", lonely_opt, "\".\n",
    "Please either:\n",
    "  (1) combine these strata with nearby strata, or\n",
    "  (2) set options(survey.lonely.psu = \"adjust\"), ",
    "\"average\", \"certainty\", or \"remove\"."
  )

  if (identical(lonely_opt, "fail")) {
    stop(msg, call. = FALSE)
  }

  invisible(NULL)
}

pwmean <- function(
    object,
    y,
    zcol = NULL,
    na.action = stats::na.omit
) {

  mc <- match.call()

  #------------------------------#
  # Step 1: input checking
  #------------------------------#
  tryCatch(
    check_ipwm_inputs_estimate(object, y, zcol),
    error = function(e)
      stop("Step 1 (input check) failed: ", e$message, call. = FALSE)
  )

  if (is.factor(object$internal$raw_sc[[y]])) {
    return(pwmean_factor(
      object = object,
      y = y,
      zcol = zcol,
      na.action = na.action,
      mc = mc
    ))
  }

  pwmean_numeric(
    object = object,
    y = y,
    zcol = zcol,
    na.action = na.action,
    mc = mc,
    check_inputs = FALSE
  )
}

# test-pwmean.R
# Integration tests for pwmean() — the main exported function.
#
# These tests cover the full workflow through the public API.
# They complement the step-level unit tests by checking that the pipeline
# is connected correctly and that the documented behaviour is preserved.
#
# Covered scenarios:
#   A. Domain modes: overall / binary / factor
#   B. Methods: calibration / alp / clw / multi all produce valid pwmean objects
#   C. Return-object structure (class, fields, estimates data.frame layout)
#   D. CI invariants: lower < mean < upper; width = 2 * z975 * se
#   E. NA handling: na.omit / na.exclude
#   F. Errors are tagged with the correct pipeline step


# Setup ----

data(sc)
data(sp1)
data(sp2)

des1 <- survey::svydesign(
  ids     = ~psu_sp1,
  strata  = ~strata_sp1,
  weights = ~wts_sp1,
  data    = sp1,
  nest    = TRUE
)

des2 <- survey::svydesign(
  ids     = ~psu_sp2,
  strata  = ~strata_sp2,
  weights = ~wts_sp2,
  data    = sp2,
  nest    = TRUE
)

# Build the four pw_fit variants once and reuse across tests.
fit_cali  <- suppressMessages(est_pw(list(sc, des1), method = "calibration",
                                     control = pw_solver_control(ftol = 1e-6)))
fit_alp   <- suppressMessages(est_pw(list(sc, des1), method = "alp",
                                     control = pw_solver_control(ftol = 1e-6)))
fit_clw   <- suppressMessages(est_pw(list(sc, des1), method = "clw",
                                     control = pw_solver_control(ftol = 1e-6)))
fit_multi <- suppressMessages(est_pw(
  data      = list(sc, des1, des2),
  p_formula = list(~agecat, ~race),
  control   = pw_solver_control(ftol = 1e-6)
))

z975 <- stats::qnorm(0.975)

DOM_COLS <- c(
  "domain",
  "unweighted_mean", "unweighted_se", "unweighted_lower", "unweighted_upper",
  "adjusted_mean",   "adjusted_se",   "adjusted_lower",   "adjusted_upper"
)

FAC_COLS <- c("category", DOM_COLS)


# A. Domain modes ----

test_that("pwmean overall (zcol = NULL): returns a single-row pwmean object", {
  out <- pwmean(fit_cali, y = "psa_level")

  expect_s3_class(out, "pwmean")
  expect_equal(nrow(out$estimates), 1L)
  expect_equal(out$estimates$domain, "Overall")
})

test_that("pwmean binary zcol: returns a single-row domain with 'name = 1' label", {
  sc_bin            <- sc
  sc_bin$has_diab   <- as.integer(sc_bin$diabetes == "1")
  fit_bin           <- suppressMessages(est_pw(
    list(sc_bin, des1), method = "calibration",
    control = pw_solver_control(ftol = 1e-6)
  ))
  out <- pwmean(fit_bin, y = "psa_level", zcol = "has_diab")

  expect_s3_class(out, "pwmean")
  expect_equal(nrow(out$estimates), 1L)
  expect_equal(out$estimates$domain, "has_diab = 1")
})

test_that("pwmean factor zcol (race, 4 levels): returns 4 domain rows in factor-level order", {
  out <- pwmean(fit_cali, y = "psa_level", zcol = "race")

  expect_equal(nrow(out$estimates), 4L)
  expect_equal(out$estimates$domain, paste0("race = ", levels(sc$race)))
})

test_that("pwmean factor y: returns one prevalence row per outcome level", {
  out <- pwmean(fit_cali, y = "diabetes")

  expect_s3_class(out, "pwmean")
  expect_equal(nrow(out$estimates), length(levels(sc$diabetes)))
  expect_equal(out$estimates$category, paste0("diabetes = ", levels(sc$diabetes)))
  expect_equal(out$estimates$domain, rep("Overall", length(levels(sc$diabetes))))
  expect_equal(sum(out$estimates$unweighted_mean), 1, tolerance = 1e-12)
  expect_equal(sum(out$estimates$adjusted_mean), 1, tolerance = 1e-12)
})

test_that("pwmean factor y matches manual 0/1 outcome estimates", {
  out <- pwmean(fit_cali, y = "diabetes")

  for (level in levels(sc$diabetes)) {
    fit_level <- fit_cali
    tmp_y <- paste0("diabetes_", level)
    fit_level$internal$raw_sc[[tmp_y]] <-
      as.numeric(fit_level$internal$raw_sc$diabetes == level)

    manual <- pwmean(fit_level, y = tmp_y)
    idx <- which(out$estimates$category == paste0("diabetes = ", level))

    expect_equal(out$estimates$unweighted_mean[idx], manual$estimates$unweighted_mean)
    expect_equal(out$estimates$adjusted_mean[idx], manual$estimates$adjusted_mean)
    expect_equal(out$estimates$adjusted_se[idx], manual$estimates$adjusted_se)
  }
})

test_that("pwmean 4-level factor y matches separate 0/1 BMI outcome estimates", {
  out <- pwmean(fit_cali, y = "BMI")

  bmi_level_outputs <- lapply(seq_along(levels(sc$BMI)), function(level_index) {
    level <- levels(sc$BMI)[[level_index]]
    fit_level <- fit_cali
    tmp_y <- paste0("BMI_", make.names(level))

    fit_level$internal$raw_sc[[tmp_y]] <-
      as.numeric(fit_level$internal$raw_sc$BMI == level)

    manual <- pwmean(fit_level, y = tmp_y)
    manual$estimates$category <- paste0("BMI = ", level)
    manual$estimates <- manual$estimates[, c("category", names(manual$estimates)[names(manual$estimates) != "category"])]
    manual$estimates
  })
  manual_domains <- do.call(rbind, bmi_level_outputs)
  row.names(manual_domains) <- NULL

  expect_equal(out$estimates, manual_domains)
})

test_that("pwmean factor y with factor zcol returns category-by-domain prevalences", {
  out <- pwmean(fit_cali, y = "diabetes", zcol = "race")

  n_y <- length(levels(sc$diabetes))
  n_z <- length(levels(sc$race))

  expect_equal(nrow(out$estimates), n_y * n_z)
  expect_equal(
    out$estimates$category,
    rep(paste0("diabetes = ", levels(sc$diabetes)), each = n_z)
  )
  expect_equal(
    out$estimates$domain,
    rep(paste0("race = ", levels(sc$race)), times = n_y)
  )

  adjusted <- matrix(out$estimates$adjusted_mean, nrow = n_z, ncol = n_y)
  expect_equal(rowSums(adjusted), rep(1, n_z), tolerance = 1e-12)
})


# B. Methods ----

test_that("pwmean works for all four methods (calibration / alp / clw / multi)", {
  for (fit in list(fit_cali, fit_alp, fit_clw, fit_multi)) {
    out <- pwmean(fit, y = "psa_level")
    expect_s3_class(out, "pwmean")
    expect_equal(out$method, fit$method)
  }
})

test_that("pwmean: adjusted means are finite for every method", {
  for (fit in list(fit_cali, fit_alp, fit_clw, fit_multi)) {
    out <- pwmean(fit, y = "psa_level")
    expect_true(all(is.finite(out$estimates$adjusted_mean)))
    expect_true(all(is.finite(out$estimates$adjusted_se)))
    expect_true(all(out$estimates$adjusted_se >= 0))
  }
})


# C. Return-object structure ----

test_that("pwmean object has fields: method, estimates, na.action, call", {
  out <- pwmean(fit_cali, y = "psa_level")

  expect_named(out, c("method", "estimates", "na.action", "call"))
})

test_that("pwmean$call is a captured call to pwmean()", {
  out <- pwmean(fit_cali, y = "psa_level")

  expect_true(is.call(out$call))
  expect_equal(as.character(out$call[[1L]]), "pwmean")
})

test_that("pwmean$estimates has all nine documented columns in order", {
  out <- pwmean(fit_cali, y = "psa_level", zcol = "race")

  expect_s3_class(out$estimates, "data.frame")
  expect_equal(names(out$estimates), DOM_COLS)
})

test_that("pwmean$estimates for factor y has category and domain columns first", {
  out <- pwmean(fit_cali, y = "diabetes", zcol = "race")

  expect_s3_class(out$estimates, "data.frame")
  expect_equal(names(out$estimates), FAC_COLS)
})

test_that("pwmean$method propagates from the input pw_fit", {
  expect_equal(pwmean(fit_cali,  y = "psa_level")$method, "calibration")
  expect_equal(pwmean(fit_alp,   y = "psa_level")$method, "alp")
  expect_equal(pwmean(fit_clw,   y = "psa_level")$method, "clw")
  expect_equal(pwmean(fit_multi, y = "psa_level")$method, "multi")
})


# D. CI invariants ----

test_that("pwmean CIs are symmetric: width equals 2 * qnorm(0.975) * se", {
  out <- pwmean(fit_cali, y = "psa_level", zcol = "race")
  d   <- out$estimates

  expect_equal(d$adjusted_upper   - d$adjusted_lower,   2 * z975 * d$adjusted_se)
  expect_equal(d$unweighted_upper - d$unweighted_lower, 2 * z975 * d$unweighted_se)
})

test_that("pwmean CIs bracket the point estimates (lower <= mean <= upper)", {
  out <- pwmean(fit_cali, y = "psa_level", zcol = "race")
  d   <- out$estimates

  expect_true(all(d$adjusted_lower   <= d$adjusted_mean))
  expect_true(all(d$adjusted_mean    <= d$adjusted_upper))
  expect_true(all(d$unweighted_lower <= d$unweighted_mean))
  expect_true(all(d$unweighted_mean  <= d$unweighted_upper))
})


# E. NA handling ----

test_that("pwmean na.omit: drops missing-y rows silently and proceeds", {
  sc_na              <- sc
  sc_na$psa_level[1:5] <- NA_real_
  fit_na             <- suppressMessages(est_pw(
    list(sc_na, des1), method = "calibration",
    control = pw_solver_control(ftol = 1e-6)
  ))
  out <- pwmean(fit_na, y = "psa_level", na.action = stats::na.omit)

  expect_s3_class(out, "pwmean")
  expect_true(is.finite(out$estimates$adjusted_mean))
})

test_that("pwmean na.exclude: produces a valid pwmean object", {
  sc_na              <- sc
  sc_na$psa_level[1:5] <- NA_real_
  fit_na             <- suppressMessages(est_pw(
    list(sc_na, des1), method = "calibration",
    control = pw_solver_control(ftol = 1e-6)
  ))
  out <- pwmean(fit_na, y = "psa_level", na.action = stats::na.exclude)

  expect_s3_class(out, "pwmean")
  expect_true(is.finite(out$estimates$adjusted_mean))
})


# F. Step error tagging ----
#
# pwmean() wraps each pipeline step in tryCatch and prefixes the failing
# step's error with "Step N (<fn>) failed:". These tests verify that an
# error raised inside a step is correctly tagged with its step number.

test_that("pwmean tags Step 1 errors with 'Step 1 (input check) failed:'", {
  expect_error(
    pwmean(list(), y = "psa_level"),
    "Step 1 (input check) failed:",
    fixed = TRUE
  )
})

test_that("pwmean tags Step 1 errors when y is not in the convenience sample", {
  expect_error(
    pwmean(fit_cali, y = "not_a_column"),
    "Step 1 (input check) failed:",
    fixed = TRUE
  )
})

# test-pwmean_methods.R
# Tests for the S3 methods on pwmean objects:
#   print.pwmean, summary.pwmean, na.action.pwmean.


# Setup ----

data(sc)
data(sp1)

des1 <- survey::svydesign(ids = ~psu_sp1, strata = ~strata_sp1,
                          weights = ~wts_sp1, data = sp1, nest = TRUE)

ctrl <- pw_solver_control(ftol = 1e-6)

fit_cali <- suppressMessages(est_pw(list(sc, des1), method = "calibration",
                                    p_formula = ~ agecat + race, control = ctrl))
fit_alp  <- suppressMessages(est_pw(list(sc, des1), method = "alp",
                                    p_formula = ~ agecat + race, control = ctrl))

out_overall <- pwmean(fit_cali, y = "psa_level")
out_race    <- pwmean(fit_cali, y = "psa_level", zcol = "race")
out_factor  <- pwmean(fit_cali, y = "BMI")


# print.pwmean ----

test_that("print.pwmean shows the estimator header and per-domain fields", {
  expect_output(print(out_overall), "Pseudo-weighted", fixed = TRUE)
  expect_output(print(out_overall), "Estimators:",     fixed = TRUE)
  expect_output(print(out_overall), "Domain:",         fixed = TRUE)
  expect_output(print(out_overall), "Mean:",           fixed = TRUE)
  expect_output(print(out_overall), "Std. Error:",     fixed = TRUE)
  expect_output(print(out_overall), "95% CI:",         fixed = TRUE)
})

test_that("print.pwmean prints one 'Domain:' block per domain level", {
  out <- capture.output(print(out_race))
  expect_equal(sum(grepl("Domain:", out, fixed = TRUE)), nrow(out_race$estimates))
})

test_that("print.pwmean uppercases ALP in the header", {
  out_alp <- pwmean(fit_alp, y = "psa_level")
  expect_output(print(out_alp), "Pseudo-weighted (ALP)", fixed = TRUE)
})

test_that("print.pwmean returns its argument invisibly", {
  txt <- capture.output(rv <- withVisible(print(out_overall)))
  expect_false(rv$visible)
  expect_identical(rv$value, out_overall)
})

test_that("print.pwmean_factor uses category and prevalence labels", {
  expect_s3_class(out_factor, "pwmean_factor")

  out <- capture.output(print(out_factor))

  expect_true(any(grepl("Category:", out, fixed = TRUE)))
  expect_true(any(grepl("Domain:", out, fixed = TRUE)))
  expect_true(any(grepl("Prevalence:", out, fixed = TRUE)))
  expect_false(any(grepl("Mean:", out, fixed = TRUE)))
})


# summary.pwmean ----

test_that("summary.pwmean shows call, method and estimator sections (overall)", {
  expect_output(summary(out_overall), "Call:",                            fixed = TRUE)
  expect_output(summary(out_overall), "Method: One reference calibration", fixed = TRUE)
  expect_output(summary(out_overall), "Domain: Overall",                  fixed = TRUE)
  expect_output(summary(out_overall), "Unweighted estimators:",           fixed = TRUE)
  expect_output(summary(out_overall), "Pseudo-weighted",                  fixed = TRUE)
})

test_that("summary.pwmean prints a multi-row table for a factor domain", {
  expect_output(summary(out_race), "Unweighted estimators:", fixed = TRUE)
  expect_output(summary(out_race), "domain",                 fixed = TRUE)
})

test_that("summary.pwmean returns its argument invisibly", {
  txt <- capture.output(rv <- withVisible(summary(out_overall)))
  expect_false(rv$visible)
  expect_identical(rv$value, out_overall)
})

test_that("summary.pwmean_factor uses category and prevalence labels", {
  out <- capture.output(summary(out_factor))

  expect_true(any(grepl("category", out, fixed = TRUE)))
  expect_true(any(grepl("domain", out, fixed = TRUE)))
  expect_true(any(grepl("prevalence", out, fixed = TRUE)))
  expect_false(any(grepl("Mean:", out, fixed = TRUE)))
  expect_false(any(grepl(" mean ", out, fixed = TRUE)))
})


# na.action.pwmean ----

test_that("na.action(pwmean) reports no omitted rows when y has no NA", {
  na_obj <- na.action(out_overall)
  expect_true(is.null(na_obj) || length(na_obj) == 0L)
})

test_that("na.action(pwmean) returns the omitted index object when y has NA", {
  sc_y <- sc
  sc_y$psa_level[1:5] <- NA_real_
  fit_y <- suppressMessages(est_pw(list(sc_y, des1), method = "calibration",
                                   p_formula = ~ agecat + race, control = ctrl))
  out_y <- pwmean(fit_y, y = "psa_level", na.action = stats::na.exclude)
  expect_false(is.null(na.action(out_y)))
})

#' Validate the Jacobian and estimating equation vector
#'
#' Checks that the Jacobian `J` and estimating equation vector `g`
#' are numerically valid and that `J` is square and full rank before
#' a Newton-Raphson step is taken.
#'
#' @param J Square numeric Jacobian matrix of dimension \eqn{p \times p}.
#' @param g Numeric estimating equation vector of length \eqn{p}.
#' @param label Character string used as a prefix in error messages to
#'   identify the calling method.
#' @param tol_singular Tolerance passed to `qr()` for detecting rank
#'   deficiency. Default is `1e-7`, matching R's `qr()` default.
#'
#' @return `invisible(TRUE)` if all checks pass; otherwise stops with
#'   an informative error message.
#'
#' @keywords internal
check_estimating_system <- function(J, g, label = "method",
                                    tol_singular = 1e-7) {
  # J: Jacobian / derivative matrix
  # g: estimating equation vector

  if (!is.matrix(J)) {
    stop(
      sprintf("[%s] Internal error: Jacobian is not a matrix.", label),
      call. = FALSE
    )
  }

  if (!is.numeric(J) || !is.numeric(g)) {
    stop(
      sprintf(
        "[%s] Internal error: Jacobian and estimating equation vector must be numeric.",
        label
      ),
      call. = FALSE
    )
  }

  if (nrow(J) != ncol(J)) {
    stop(
      sprintf(
        paste0(
          "[%s] Estimating system failed: Jacobian is not square ",
          "(%d x %d). The number of estimating equations ",
          "does not match the number of parameters."
        ),
        label, nrow(J), ncol(J)
      ),
      call. = FALSE
    )
  }

  if (length(g) != nrow(J)) {
    stop(
      sprintf(
        paste0(
          "[%s] Estimating system failed: estimating equation vector length (%d) ",
          "does not match Jacobian dimension (%d)."
        ),
        label, length(g), nrow(J)
      ),
      call. = FALSE
    )
  }

  if (any(!is.finite(J))) {
    stop(
      sprintf(
        paste0(
          "[%s] Estimating system failed: Jacobian contains NA, NaN, or Inf. ",
          "This may happen when the fitted weights become extremely large, ",
          "the linear predictor overflows, or the input data contain invalid values."
        ),
        label
      ),
      call. = FALSE
    )
  }

  if (any(!is.finite(g))) {
    stop(
      sprintf(
        paste0(
          "[%s] Estimating system failed: estimating equation vector contains ",
          "NA, NaN, or Inf. This may happen when the fitted weights become ",
          "extremely large, the linear predictor overflows, or the input data ",
          "contain invalid values."
        ),
        label
      ),
      call. = FALSE
    )
  }

  qJ <- qr(J, tol = tol_singular)
  rankJ <- qJ$rank
  p <- ncol(J)

  if (rankJ < p) {
    stop(
      sprintf(
        paste0(
          "[%s] Estimating system failed because the Jacobian is singular ",
          "or not identifiable (Jacobian rank = %d < %d).\n",
          "This means the model cannot determine a unique set of coefficients.\n",
          "Common reasons:\n",
          "  1. Redundant or perfectly collinear variables in p_formula.\n",
          "  2. Some categories are rare or absent in one of the samples.\n",
          "  3. There are too many variables or categories for the available data.\n",
          "  4. The variables almost perfectly distinguish the nonprobability sample ",
          "from the reference survey.\n",
          "Try simplifying p_formula, removing overlapping variables, or combining ",
          "rare factor levels."
        ),
        label, rankJ, p
      ),
      call. = FALSE
    )
  }

  invisible(TRUE)
}


#' Assemble arguments for nleqslv
#'
#' Extracts and assembles the solver settings from a control object created
#' by `pw_solver_control()` into the format expected by `nleqslv::nleqslv()`.
#'
#' @param control A list created by `pw_solver_control()`.
#'
#' @return A list with components `method`, `global`, `xscalm`, and `control`
#'   (containing `ftol`, `xtol`, `maxit`, `trace`, and any user-supplied
#'   extras from `nleqslv_control`).
#'
#' @keywords internal
prepare_nleqslv_args <- function(control) {

  reserved <- c("x", "fn", "jac")
  bad <- intersect(names(control$nleqslv_control), reserved)

  if (length(bad) > 0L) {
    stop(
      sprintf(
        "`nleqslv_control` must not contain reserved argument(s): %s.",
        paste(bad, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  list(
    method  = control$method,
    global  = control$global,
    xscalm  = control$xscalm,
    control = c(
      list(
        ftol  = control$ftol,
        xtol  = control$xtol,
        maxit = control$maxit,
        trace = as.integer(control$trace)
      ),
      control$nleqslv_control
    )
  )
}


#' Check the result returned by nleqslv
#'
#' Inspects the termination code (`termcd`) from `nleqslv::nleqslv()` and
#' either returns silently on success, issues a warning for partial
#' convergence, or stops with an informative error message on failure.
#'
#' @param root The list returned by `nleqslv::nleqslv()`.
#' @param label Character string used as a prefix in messages to identify
#'   the calling method.
#' @param ftol The function-value convergence tolerance, used to assess
#'   whether partial convergence is acceptable.
#'
#' @return `invisible(TRUE)` on success or acceptable convergence;
#'   otherwise stops or warns.
#'
#' @keywords internal
check_nleqslv_result <- function(root, label, ftol) {
  if (is.null(root$fvec)) {
    stop(
      sprintf("[%s] nleqslv failed: result does not contain `fvec`.", label),
      call. = FALSE
    )
  }

  if (is.null(root$termcd)) {
    stop(
      sprintf("[%s] nleqslv failed: result does not contain `termcd`.", label),
      call. = FALSE
    )
  }

  if (is.null(root$message)) {
    root$message <- ""
  }

  fmax <- max(abs(root$fvec))

  if (!is.finite(fmax)) {
    stop(
      sprintf(
        "[%s] nleqslv failed: final estimating equations contain non-finite values.",
        label
      ),
      call. = FALSE
    )
  }

  termcd <- as.integer(root$termcd)

  if (termcd == 1L) {
    return(invisible(TRUE))
  }

  if (termcd == 2L) {
    if (fmax <= ftol) {
      return(invisible(TRUE))
    }

    warning(
      sprintf(
        paste0(
          "[%s] nleqslv returned termcd = 2: the coefficient updates became very small, ",
          "but the estimating equations are not sufficiently close to zero.\n",
          "max|fvec| = %.3e; ftol = %.3e.\n",
          "The coefficients and pseudo-weights are returned, but users should decide ",
          "whether this remaining equation imbalance is acceptable."
        ),
        label, fmax, ftol
      ),
      call. = FALSE
    )

    return(invisible(TRUE))
  }

  if (termcd == 3L) {
    if (fmax <= ftol) {
      warning(
        sprintf(
          paste0(
            "[%s] nleqslv returned termcd = 3: no better point was found, ",
            "but the estimating equations are acceptably small.\n",
            "max|fvec| = %.3e; ftol = %.3e."
          ),
          label, fmax, ftol
        ),
        call. = FALSE
      )
      return(invisible(TRUE))
    }

    stop(
      sprintf(
        paste0(
          "[%s] nleqslv failed: no better point was found.\n",
          "termcd = 3; message = %s; max|fvec| = %.3e.\n",
          "The algorithm stopped before the estimating equations became small enough. ",
          "Try changing the nleqslv global strategy, ",
          "using xscalm = 'auto', simplifying p_formula, or combining rare factor levels."
        ),
        label, root$message, fmax
      ),
      call. = FALSE
    )
  }

  if (termcd == 4L) {
    stop(
      sprintf(
        paste0(
          "[%s] nleqslv failed: iteration limit was reached.\n",
          "termcd = 4; message = %s; max|fvec| = %.3e.\n",
          "Try increasing the maximum number of iterations, ",
          "changing the nleqslv global strategy, or simplifying p_formula."
        ),
        label, root$message, fmax
      ),
      call. = FALSE
    )
  }

  if (termcd %in% c(5L, 6L, 7L)) {
    jac_reason <- switch(
      as.character(termcd),
      "5" = "Jacobian is too ill-conditioned.",
      "6" = "Jacobian is singular.",
      "7" = "Jacobian is unusable."
    )

    stop(
      sprintf(
        paste0(
          "[%s] nleqslv failed: %s\n",
          "termcd = %s; message = %s; max|fvec| = %.3e.\n",
          "This means the estimating system is not numerically stable at the final iterate.\n",
          "Common reasons:\n",
          "  1. Some variables in p_formula are highly correlated (collinear).\n",
          "  2. Some categories are rare or absent in one of the samples.\n",
          "  3. There are too many variables or categories for the available data.\n",
          "  4. The variables almost perfectly distinguish the nonprobability sample ",
          "from the reference survey.\n",
          "Try simplifying p_formula, removing overlapping variables, combining rare factor levels, ",
          "or changing the global strategy in pw_solver_control()."
        ),
        label, jac_reason, termcd, root$message, fmax
      ),
      call. = FALSE
    )
  }

  if (termcd == -10L) {
    stop(
      sprintf(
        paste0(
          "[%s] nleqslv solver failed due to an internal numerical issue.\n",
          "termcd = -10; message = %s; max|fvec| = %.3e.\n",
          "This is likely caused by numerical instability in the model.\n",
          "Try simplifying p_formula, removing highly correlated variables, ",
          "or combining rare factor levels."
        ),
        label, root$message, fmax
      ),
      call. = FALSE
    )
  }

  stop(
    sprintf(
      paste0(
        "[%s] nleqslv failed with unrecognized termination code.\n",
        "termcd = %s; message = %s; max|fvec| = %.3e."
      ),
      label, termcd, root$message, fmax
    ),
    call. = FALSE
  )
}

#' Solve the participation model estimating equations
#'
#' A unified wrapper that calls `nleqslv::nleqslv()` to solve the
#' system of estimating equations \eqn{g(\beta) = 0}, then validates the
#' result via `check_nleqslv_result()`.
#'
#' @param beta_start Numeric vector of starting values for the coefficient
#'   vector \eqn{\beta}.
#' @param fn Function returning the estimating equation vector \eqn{g(\beta)}.
#' @param jac Function returning the Jacobian matrix \eqn{J(\beta)}.
#' @param label Character string used as a prefix in error messages to
#'   identify the calling method.
#' @param control A list created by `pw_solver_control()`.
#' @param ... Additional arguments passed to `fn` and `jac`
#'   (e.g., design matrices and weights).
#'
#' @return A list with components `coefficients`, `iterations`, `solver`,
#'   `solver_result`, `solver_method`, `fvec`, `termcd`, and `message`.
#'
#' @keywords internal
solve_participation_model <- function(
    beta_start,
    fn,
    jac,
    label,
    control = NULL,
    ...
) {

  if (!is.list(control) || is.null(control$solver)) {
    stop(
      "`control` must be created by pw_solver_control().",
      call. = FALSE
    )
  }

  if (tolower(control$solver) != "nleqslv") {
    stop(
      "Only solver = 'nleqslv' is currently supported.",
      call. = FALSE
    )
  }

  if (!is.numeric(beta_start) || any(!is.finite(beta_start))) {
    stop(
      "`beta_start` must be a finite numeric vector.",
      call. = FALSE
    )
  }

  if (!is.function(fn)) {
    stop("`fn` must be a function.", call. = FALSE)
  }

  if (!is.function(jac)) {
    stop("`jac` must be a function.", call. = FALSE)
  }

  args <- prepare_nleqslv_args(control)

  root <- do.call(
    nleqslv::nleqslv,
    c(
      list(
        x = beta_start,
        fn = fn,
        jac = jac
      ),
      args,
      list(...)
    )
  )

  check_nleqslv_result(
    root = root,
    label = label,
    ftol = args$control$ftol
  )

  list(
    coefficients = as.numeric(root$x),
    iterations = root$iter,
    solver = "nleqslv",
    solver_result = root,
    solver_method = args$method,
    fvec = root$fvec,
    termcd = root$termcd,
    message = root$message
  )
}

# -------------------------------------------------------------------
# Helper: check factor levels consistency between sc and sp_i
# -------------------------------------------------------------------
.check_factor_levels <- function(sc, sp_i, vars_in_formula, ref_label = "sp") {
  for (v in vars_in_formula) {
    if (is.factor(sc[[v]]) || is.factor(sp_i[[v]])) {
      lev_sc <- levels(as.factor(sc[[v]]))
      lev_sp <- levels(as.factor(sp_i[[v]]))
      if (!identical(lev_sc, lev_sp)) {
        stop(sprintf(
          "Factor levels of '%s' differ between sc and %s.\n  sc : %s\n  %s : %s",
          v,
          ref_label,
          paste(lev_sc, collapse = ", "),
          ref_label,
          paste(lev_sp, collapse = ", ")
        ), call. = FALSE)
      }
    }
  }
}

# Estimates prevalence for each level of a categorical outcome.
pwmean_factor <- function(
    object,
    y,
    zcol = NULL,
    na.action = stats::na.omit,
    mc = match.call()
) {

  y_factor <- droplevels(object$internal$raw_sc[[y]])
  y_levels <- levels(y_factor)

  if (length(y_levels) < 1L) {
    stop(
      sprintf("Outcome variable '%s' must contain at least one non-missing level.", y),
      call. = FALSE
    )
  }

  level_outputs <- lapply(seq_along(y_levels), function(j) {
    pwmean_one_factor_level(
      object = object,
      y = y,
      level = y_levels[[j]],
      level_index = j,
      zcol = zcol,
      na.action = na.action,
      mc = mc
    )
  })

  result <- level_outputs[[1L]]
  result$estimates <- do.call(rbind, lapply(level_outputs, function(x) x$estimates))
  row.names(result$estimates) <- NULL
  result$call <- mc
  class(result) <- c("pwmean_factor", "pwmean")

  result
}

# Helper: convert one factor level to 0/1 and estimates its prevalence.
pwmean_one_factor_level <- function(
    object,
    y,
    level,
    level_index,
    zcol = NULL,
    na.action = stats::na.omit,
    mc = match.call()
) {

  tmp_object <- object
  tmp_y <- make_factor_outcome_name(
    y = y,
    level_index = level_index,
    existing_names = names(tmp_object$internal$raw_sc)
  )

  tmp_object$internal$raw_sc[[tmp_y]] <-
    as.numeric(tmp_object$internal$raw_sc[[y]] == level)

  out <- pwmean_numeric(
    object = tmp_object,
    y = tmp_y,
    zcol = zcol,
    na.action = na.action,
    mc = mc,
    check_inputs = FALSE
  )

  out$estimates <- add_factor_outcome_columns(
    estimates = out$estimates,
    y = y,
    level = level,
    has_domain = !is.null(zcol)
  )

  out
}

# Helper: creates a temporary internal outcome name that does not overwrite user data.
make_factor_outcome_name <- function(y, level_index, existing_names) {
  base_name <- paste0("..pwmean_", make.names(y), "_", level_index)
  out <- base_name
  suffix <- 1L

  while (out %in% existing_names) {
    suffix <- suffix + 1L
    out <- paste0(base_name, "_", suffix)
  }

  out
}

# Helper: adds user-facing category and domain columns for categorical outcomes.
add_factor_outcome_columns <- function(estimates, y, level, has_domain) {
  category_label <- paste0(y, " = ", level)
  domain_labels <- estimates$domain

  if (!isTRUE(has_domain)) {
    domain_labels <- rep("Overall", nrow(estimates))
  }

  estimates$category <- rep(category_label, nrow(estimates))
  estimates$domain <- domain_labels
  estimates[, c(
    "category", "domain",
    setdiff(names(estimates), c("category", "domain"))
  )]
}

#' Sort reference samples by size
#'
#' @param sp List of reference sample data frames.
#' @param vars List of character vectors of predictor variable names,
#'   one per element of `sp`.
#' @param weight Character vector of survey-weight column names,
#'   one per element of `sp`.
#' @param design List of `survey.design2` or `svyrep.design` objects,
#'   one per element of `sp`.
#' @param sp_order Character scalar. `"size"` reorders reference samples
#'   largest-first; `"given"` keeps the user-supplied order.
#' @param verbose Logical. If `TRUE`, a summary of the ordering is printed.
#'
#' @return A list with components:
#'   \item{sp}{Reordered list of reference sample data frames.}
#'   \item{vars}{Reordered list of predictor variable name vectors.}
#'   \item{weight}{Reordered character vector of weight column names.}
#'   \item{design}{Reordered list of survey design objects.}
#'   \item{order_used}{Integer vector giving the reordering index.}
#'   \item{log}{Character vector of log messages describing the ordering.}
#'
#' @keywords internal
sort_by_sp_size <- function(sp, vars, weight, design,
                            sp_order, verbose = FALSE) {

  log_messages <- character()
  order_by_size <- identical(sp_order, "size")

  n <- length(sp)

  # strict checks
  if (!is.list(sp) || n == 0) {
    stop("'sp' must be a non-empty list.", call. = FALSE)
  }

  if (length(weight) != n) {
    stop("'weight' must have the same length as 'sp'.", call. = FALSE)
  }

  if (!is.list(vars) || length(vars) != n) {
    stop("'vars' must be a list with the same length as 'sp'.", call. = FALSE)
  }

  # design is REQUIRED and must be a list of survey design objects
  if (missing(design) || is.null(design)) {
    stop("'design' is required for the multi-reference case.", call. = FALSE)
  }

  if (!is.list(design) || length(design) != n) {
    stop("'design' must be a list of survey design objects with length equal to length(sp).",
         call. = FALSE)
  }

  bad_i <- which(
    !vapply(design,
            function(d) inherits(d, c("survey.design2", "svyrep.design")),
            logical(1))
  )

  if (length(bad_i) > 0) {
    stop(
      sprintf(
        "Each design[[i]] must be a survey design object of class 'survey.design2' or 'svyrep.design'. Problem at i = %s.",
        paste(bad_i, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  # optional consistency check: each sp[[i]] should align with design[[i]]
  bad_n <- which(
    !vapply(seq_len(n), function(i) {
      NROW(sp[[i]]) == nrow(design[[i]]$variables)
    }, logical(1))
  )

  if (length(bad_n) > 0) {
    stop(
      sprintf(
        "Row mismatch between sp[[i]] and design[[i]]$variables at i = %s.",
        paste(bad_n, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  # compute row counts
  sizes <- vapply(sp, NROW, integer(1))

  # ordering index
  idx <- if (order_by_size) order(sizes, decreasing = TRUE) else seq_along(sp)

  # reorder
  sp_sorted     <- sp[idx]
  vars_sorted   <- vars[idx]
  weight_sorted <- weight[idx]
  design_sorted <- design[idx]

  # names: always use positional indices for display
  disp_names   <- paste0("sp[[", seq_along(sp), "]]")
  sorted_names <- disp_names[idx]
  sorted_sizes <- sizes[idx]

  # message
  msg_lines <- paste0("  ", sorted_names, " (n = ", sorted_sizes, ")")
  msg <- if (order_by_size) {
    paste0(
      "\nReference samples summary:\n",
      "Order of samples by size (largest to smallest):\n",
      paste(msg_lines, collapse = "\n"), "\n"
    )
  } else {
    paste0(
      "\nReference samples summary:\n",
      "Order of samples kept as provided:\n",
      paste(msg_lines, collapse = "\n"), "\n"
    )
  }

  log_messages <- c(log_messages, msg)

  # assign names
  names(sp_sorted)     <- sorted_names
  names(vars_sorted)   <- sorted_names
  names(weight_sorted) <- sorted_names
  names(design_sorted) <- sorted_names

  list(
    sp         = sp_sorted,
    vars       = vars_sorted,
    weight     = weight_sorted,
    design     = design_sorted,
    order_used = idx,
    log        = log_messages
  )
}

# utils.R
# General-purpose utility functions shared across build and estimation stages.


# Logistic (inverse-logit) transformation.
# Used by ALP, CLW, and raking solver helpers.
#
# Implemented via stats::plogis() rather than the algebraically equivalent
# exp(x) / (1 + exp(x)). The naive form overflows to Inf once x exceeds about
# 710, so exp(x) / (1 + exp(x)) evaluates to NaN in that range; that NaN then
# trips the "non-finite fitted probabilities" guards in the ALP/CLW solvers and
# turns a merely large linear predictor into a hard error. stats::plogis() is
# numerically stable across the whole real line and saturates to 0 or 1 in the
# tails instead of returning NaN.
expit <- function(x) {
  stats::plogis(x)
}


# Null-coalescing operator: return `a` if non-NULL, otherwise `b`.
# Used in S3 print/summary methods.
`%||%` <- function(a, b) if (!is.null(a)) a else b


# Check that a design matrix X is full column rank; stop with a clear message
# if rank-deficient. Used by all one-reference and multi-reference builders.
check_design_identifiability <- function(X, label = "method", tol = 1e-7) {
  qx <- qr(X, tol = tol)
  if (qx$rank < ncol(X)) {
    stop(
      sprintf(
        paste0("[%s] Model matrix is rank-deficient before NR iteration ",
               "(rank = %d < %d). This suggests collinearity or redundant terms ",
               "in p_formula. Please simplify the model."),
        label, qx$rank, ncol(X)
      ),
      call. = FALSE
    )
  }
  invisible(TRUE)
}


# Construct a design matrix from `vars` columns of `data`, optionally prepending
# an intercept column. Used by all one-reference and multi-reference builders.
add_intercept <- function(vars, data, intercept = TRUE) {
  if (length(vars) == 0) {
    stop("No variables provided for design matrix.")
  }

  miss <- setdiff(vars, colnames(data))
  if (length(miss) > 0) {
    stop(
      sprintf("Columns not found in data: %s", paste(miss, collapse = ", "))
    )
  }

  X <- as.matrix(data[, vars, drop = FALSE])

  if (intercept) {
    X <- cbind("(Intercept)" = 1, X)
  }

  X
}

#' Probability Reference Sample 1 (sp1)
#'
#' This dataset represents a probability sample derived from the
#' 1999--2010 cycles of the National Health and Nutrition
#' Examination Survey (NHANES). It is used as a probability
#' reference survey to support the pseudo-weighting methods
#' implemented in the \code{nonprobsampling} package.
#'
#'
#' @format A data frame with 3494 observations and 14 variables:
#' \describe{
#'   \item{agecat}{Age category (factor with 4 levels: 1 = 55--59, 2 = 60--64, 3 = 65--69, 4 = 70+)}
#'   \item{marital}{Marital status (factor with 4 levels: 1 = Married Or Living As Married, 2 = Widowed, 3 = Divorced or Separated, 4 = Never Married)}
#'   \item{race}{Race category (factor with 4 levels: 1 = White, 2 = Black, 3 = Hispanic, 4 = Other)}
#'   \item{education}{Education level (factor with 5 levels: 1 = Less Than 8 Years, 2 = 8--11 Years, 3 = 12 Years Or Completed High School, 4 = College Graduate, 5 = Postgraduate)}
#'   \item{employment}{Employment status (factor with 2 levels: 0 = Not Working, 1 = Working)}
#'   \item{smoking}{Smoking status (factor with 3 levels: 1 = Never Smoker, 2 = Former Smoker, 3 = Current Smoker)}
#'   \item{comorbidity}{General comorbidity indicator (factor with 2 levels: 0 = No, 1 = Yes)}
#'   \item{psa_level}{Serum prostate-specific antigen level (numeric)}
#'   \item{BMI}{Body mass index category (factor with 4 levels: "Normal", "Overweight", "Obese", "Morbidly Obese")}
#'   \item{diabetes}{Diabetes diagnosis indicator (factor with 2 levels: 0 = No, 1 = Yes)}
#'   \item{pros_enlarged}{Prostate enlargement indicator (factor with 2 levels: 0 = No, 1 = Yes)}
#'   \item{strata_sp1}{Stratum identifier for complex survey design (numeric)}
#'   \item{psu_sp1}{Primary sampling unit identifier for complex survey design (numeric)}
#'   \item{wts_sp1}{10-year interview sampling weights (numeric)}
#' }
#'
#' @details
#' The dataset includes auxiliary variables shared with the nonprobability
#' sample \code{sc}, enabling the construction of pseudo-weights to adjust
#' for participation bias. Survey design variables and sampling weights are
#' provided to support design-consistent estimation.
#'
#' The \code{sp1} dataset
#' contains the outcome variable \code{psa_level}, which is also observed in
#' \code{sc}, allowing for the evaluation of pseudo-weighted estimators against
#' estimates based on true sampling weights. It may also be incorporated into
#' the participation model, potentially enhancing bias reduction when
#' participation depends on the outcome.
#'
#' @source
#' Derived from the National Health and Nutrition Examination Survey (NHANES),
#' 1999--2010 cycles, conducted by the U.S. National Center for Health
#' Statistics (NCHS).
#'
#' @usage data(sp1)
#'
#' @examples
#' data(sp1)
#' str(sp1)
#' summary(sp1)
"sp1"

#' Parse the data list passed to est_pw
#'
#' Splits the input list into its two parts: the nonprobability sample
#' (`sc`) and one or more reference survey designs (`sp_des`).
#' For each design it calls `extract_analysis_data()` to remove design
#' variables and attach a `sp_wts` column, giving back a plain data
#' frame ready for modeling. If the list elements have no names, or if
#' any element name is an empty string (partially named list), default
#' names `"sp[[1]]"`, `"sp[[2]]"`, ... are assigned to all elements.
#'
#' @param data A named or unnamed list. The first element is the
#'   nonprobability sample as a data frame. Every remaining element is a
#'   survey design object (`survey.design2`, `survey.design`, or
#'   `svyrep.design`).
#'
#' @return A list with four elements:
#'
#' - `sc`: the nonprobability sample (first element of `data`).
#' - `sp_des`: named list of the survey design objects.
#' - `sp_vars`: named list of plain data frames, one per reference survey,
#'   with design variables removed and a `sp_wts` column added.
#' - `n_ref`: integer giving the number of reference surveys.
#'
#' @keywords internal
parse_ipwm_data <- function(data) {

  sc     <- data[[1]]
  sp_des <- data[-1]

  ref_names <- names(sp_des)
  if (is.null(ref_names) || any(!nzchar(ref_names))) {
    ref_names <- paste0("sp[[", seq_along(sp_des), "]]")
  }

  sp_vars <- lapply(seq_along(sp_des), function(i) {
    tryCatch(
      extract_analysis_data(sp_des[[i]]),
      error = function(e) {
        stop(
          sprintf("Failed to parse reference survey %d: %s", i, e$message),
          call. = FALSE
        )
      }
    )
  })

  names(sp_des)  <- ref_names
  names(sp_vars) <- ref_names

  list(
    sc      = sc,
    sp_des  = sp_des,
    sp_vars = sp_vars,
    n_ref   = length(sp_des)
  )
}

# Setup ----

data(sc)
data(sp1)
data(sp2)

des1 <- survey::svydesign(
  ids = ~psu_sp1,
  strata = ~strata_sp1,
  weights = ~wts_sp1,
  data = sp1,
  nest = TRUE
)

des2 <- survey::svydesign(
  ids = ~psu_sp2,
  strata = ~strata_sp2,
  weights = ~wts_sp2,
  data = sp2,
  nest = TRUE
)

# check_ipwm_inputs_build ----

test_that("check_ipwm_inputs_build validates correct one-reference input", {
  expect_invisible(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = NULL,
      method = "alp"
    )
  )

  expect_invisible(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = NULL,
      method = "clw"
    )
  )

  expect_invisible(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = NULL,
      method = "calibration"
    )
  )

  expect_invisible(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = NULL,
      method = "cali"
    )
  )
})


test_that("check_ipwm_inputs_build validates correct multi-reference input", {
  expect_invisible(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = NULL,
      method = "multi",
      sp_order = "size",
      precali = TRUE
    )
  )

  expect_invisible(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = NULL,
      method = "multi",
      sp_order = "given",
      precali = FALSE
    )
  )
})


test_that("check_ipwm_inputs_build rejects invalid data structure", {
  expect_error(
    check_ipwm_inputs_build(
      data = 42L,
      p_formula = NULL,
      method = "alp"
    ),
    "`data` must be a list",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc),
      p_formula = NULL,
      method = "alp"
    ),
    "`data` must be a list",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(as.matrix(sc), sc),
      p_formula = NULL,
      method = "alp"
    ),
    "The first element of `data` must be a data.frame",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build rejects invalid survey design objects", {
  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, sp1),
      p_formula = NULL,
      method = "alp"
    ),
    "not valid survey design objects",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build rejects zero-row sc", {
  sc_empty <- sc[0, ]

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc_empty, des1),
      p_formula = NULL,
      method = "alp"
    ),
    "`sc` has zero rows.",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build rejects duplicated sc column names", {
  sc_dup <- sc
  names(sc_dup)[2] <- names(sc_dup)[1]

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc_dup, des1),
      p_formula = NULL,
      method = "alp"
    ),
    "`sc` has duplicated column names.",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build rejects invalid method", {
  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = NULL,
      method = "bad_method"
    ),
    "Invalid method",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = NULL,
      method = c("alp", "clw")
    ),
    "`method` must be a single character string or NULL.",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build checks method-reference consistency", {
  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = NULL,
      method = "multi"
    ),
    "`method = 'multi'` is only allowed when multiple reference samples are provided.",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = NULL,
      method = "alp",
      sp_order = "size",
      precali = TRUE
    ),
    "only valid for a single reference sample",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build checks p_formula in one-reference case", {
  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = list(~ agecat),
      method = "alp"
    ),
    "For one-reference case, `p_formula` must be a formula or NULL.",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = ~ not_a_variable,
      method = "alp"
    ),
    "Variables in `p_formula` not found in `sc`",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = agecat ~ marital,
      method = "alp"
    ),
    "`p_formula` must be a one-sided formula (e.g., ~ x + y), not two-sided.",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build rejects two-sided formula in multi-reference case", {
  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = list(~ agecat, agecat ~ marital),
      method = "multi",
      sp_order = "size",
      precali = TRUE
    ),
    "All formulas in `p_formula` must be one-sided (e.g., ~ x + y), not two-sided.",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build checks sp_order and precali in multi-reference case", {
  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = NULL,
      method = "multi",
      precali = TRUE
    ),
    "`sp_order` must be one of 'size' or 'given'.",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = NULL,
      method = "multi",
      sp_order = "wrong",
      precali = TRUE
    ),
    "`sp_order` must be one of 'size' or 'given'.",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = NULL,
      method = "multi",
      sp_order = "size"
    ),
    "`precali` must be a single TRUE/FALSE value.",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = NULL,
      method = "multi",
      sp_order = "size",
      precali = "TRUE"
    ),
    "`precali` must be a single TRUE/FALSE value.",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build checks p_formula in multi-reference case", {
  p_formula <- ~ agecat + marital + race + educat + empstat +
    smoking + comorbidity

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = p_formula,
      method = "multi",
      sp_order = "size",
      precali = TRUE
    ),
    "For multi-reference case, `p_formula` must be a list",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = list(p_formula),
      method = "multi",
      sp_order = "size",
      precali = TRUE
    ),
    "For multi-reference case, `p_formula` must be a list of 2 formulas.",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = list(p_formula, "not_formula"),
      method = "multi",
      sp_order = "size",
      precali = TRUE
    ),
    "All elements of `p_formula` must be formulas.",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = list(~ agecat, ~ not_a_variable),
      method = "multi",
      sp_order = "size",
      precali = TRUE
    ),
    "Variables in `p_formula[[2]]` not found in `sc`",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build accepts NULL method and NULL p_formula", {
  expect_invisible(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = NULL,
      method = NULL
    )
  )

  expect_invisible(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = NULL,
      method = NULL,
      sp_order = "size",
      precali = TRUE
    )
  )
})


test_that("check_ipwm_inputs_build rejects method = NA", {
  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1),
      p_formula = NULL,
      method = NA_character_
    ),
    "`method` must be a single character string or NULL.",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build rejects p_formula variables missing from sp", {
  sc_extra <- sc
  sc_extra$only_in_sc <- 1L

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc_extra, des1),
      p_formula = ~ only_in_sc,
      method = "alp"
    ),
    "Variables in `p_formula` not found in corresponding `sp`",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc_extra, des1, des2),
      p_formula = list(~ agecat, ~ only_in_sc),
      method = "multi",
      sp_order = "size",
      precali = TRUE
    ),
    "Variables in `p_formula[[2]]` not found in corresponding `sp`",
    fixed = TRUE
  )
})


test_that("check_ipwm_inputs_build rejects invalid precali values", {
  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = NULL,
      method = "multi",
      sp_order = "size",
      precali = NA
    ),
    "`precali` must be a single TRUE/FALSE value.",
    fixed = TRUE
  )

  expect_error(
    check_ipwm_inputs_build(
      data = list(sc, des1, des2),
      p_formula = NULL,
      method = "multi",
      sp_order = "size",
      precali = c(TRUE, FALSE)
    ),
    "`precali` must be a single TRUE/FALSE value.",
    fixed = TRUE
  )
})


# extract_analysis_data ----

test_that("extract_analysis_data: returns a data.frame", {
  out <- extract_analysis_data(des1)

  expect_s3_class(out, "data.frame")
})

test_that("extract_analysis_data: weight column 'sp_wts' is added", {
  out <- extract_analysis_data(des1)

  expect_true("sp_wts" %in% names(out))
})

test_that("extract_analysis_data: all weights are positive finite numbers", {
  out <- extract_analysis_data(des1)

  expect_true(all(is.finite(out$sp_wts)))
  expect_true(all(out$sp_wts > 0))
})

test_that("extract_analysis_data: design variables (psu_sp1, strata_sp1, wts_sp1) are removed", {
  out <- extract_analysis_data(des1)

  expect_false("psu_sp1"     %in% names(out))
  expect_false("strata_sp1"  %in% names(out))
  expect_false("wts_sp1" %in% names(out))
})

test_that("extract_analysis_data: row count matches original data", {
  out <- extract_analysis_data(des1)

  expect_equal(nrow(out), nrow(sp1))
})

test_that("extract_analysis_data: custom weight_name is used", {
  out <- extract_analysis_data(des1, weight_name = "my_wts")

  expect_true("my_wts"  %in% names(out))
  expect_false("sp_wts" %in% names(out))
})

test_that("extract_analysis_data: errors when des is not a survey design object", {
  expect_error(
    extract_analysis_data(sp1),
    "`des` must be a survey design object.",
    fixed = TRUE
  )
})

test_that("extract_analysis_data: errors when weight_name already exists in analysis vars", {
  # Build a design where a covariate is named "sp_wts"
  df_clash <- sp1
  df_clash$sp_wts <- 1L
  des_clash <- survey::svydesign(
    ids     = ~psu_sp1,
    strata  = ~strata_sp1,
    weights = ~wts_sp1,
    data    = df_clash,
    nest    = TRUE
  )

  expect_error(
    extract_analysis_data(des_clash, weight_name = "sp_wts"),
    "already exists",
    fixed = TRUE
  )
})


# parse_ipwm_data ----

test_that("parse_ipwm_data: returns list with sc, sp_des, sp_vars, n_ref", {
  out <- parse_ipwm_data(list(sc, des1))

  expect_named(out, c("sc", "sp_des", "sp_vars", "n_ref"))
})

test_that("parse_ipwm_data: sc is the first element unchanged", {
  out <- parse_ipwm_data(list(sc, des1))

  expect_identical(out$sc, sc)
})

test_that("parse_ipwm_data: n_ref is 1 for one reference survey", {
  out <- parse_ipwm_data(list(sc, des1))

  expect_equal(out$n_ref, 1L)
})

test_that("parse_ipwm_data: n_ref is 2 for two reference surveys", {
  out <- parse_ipwm_data(list(sc, des1, des2))

  expect_equal(out$n_ref, 2L)
})

test_that("parse_ipwm_data: sp_des is a named list of design objects", {
  out <- parse_ipwm_data(list(sc, des1, des2))

  expect_true(is.list(out$sp_des))
  expect_length(out$sp_des, 2L)
  expect_true(all(vapply(out$sp_des, inherits, logical(1), "survey.design2")))
})

test_that("parse_ipwm_data: sp_vars are plain data.frames with 'sp_wts' column", {
  out <- parse_ipwm_data(list(sc, des1))

  expect_s3_class(out$sp_vars[[1]], "data.frame")
  expect_true("sp_wts" %in% names(out$sp_vars[[1]]))
})

test_that("parse_ipwm_data: unnamed list gets default names sp[[1]], sp[[2]]", {
  out <- parse_ipwm_data(list(sc, des1, des2))

  expect_equal(names(out$sp_des),  c("sp[[1]]", "sp[[2]]"))
  expect_equal(names(out$sp_vars), c("sp[[1]]", "sp[[2]]"))
})

test_that("parse_ipwm_data: named list keeps names", {
  out <- parse_ipwm_data(list(sc, survey_a = des1, survey_b = des2))

  expect_equal(names(out$sp_des),  c("survey_a", "survey_b"))
  expect_equal(names(out$sp_vars), c("survey_a", "survey_b"))
})

test_that("parse_ipwm_data: partially named list falls back to default names", {
  out <- parse_ipwm_data(list(sc, survey_a = des1, des2))

  expect_equal(names(out$sp_des),  c("sp[[1]]", "sp[[2]]"))
  expect_equal(names(out$sp_vars), c("sp[[1]]", "sp[[2]]"))
})

#' CLW estimating equations
#' @noRd
clw_fn <- function(beta, Xc, Xp, wts_sp, ...) {
  eta_sp <- as.vector(Xp %*% beta)

  if (any(!is.finite(eta_sp))) {
    stop(
      "CLW failed: reference-sample linear predictor became non-finite.",
      call. = FALSE
    )
  }

  pi_sp <- as.vector(expit(eta_sp))

  if (any(!is.finite(pi_sp))) {
    stop(
      paste(
        "CLW failed: fitted probabilities became non-finite.",
        "This usually indicates numerical overflow or unstable coefficient values."
      ),
      call. = FALSE
    )
  }

  colSums(Xc) - colSums(wts_sp * pi_sp * Xp)
}


#' Jacobian of the CLW estimating equations
#' @noRd
clw_jac <- function(beta, Xc, Xp, wts_sp,
                    label = "One_Ref_CLW",
                    tol_singular = 1e-7,
                    ...) {
  eta_sp <- as.vector(Xp %*% beta)

  if (any(!is.finite(eta_sp))) {
    stop(
      sprintf("[%s] CLW failed: reference-sample linear predictor became non-finite when computing Jacobian.", label),
      call. = FALSE
    )
  }

  pi_sp <- as.vector(expit(eta_sp))

  if (any(!is.finite(pi_sp))) {
    stop(
      sprintf(
        paste(
          "[%s] CLW failed: fitted probabilities became non-finite when computing Jacobian.",
          "This usually indicates numerical overflow or unstable coefficient values."
        ),
        label
      ),
      call. = FALSE
    )
  }

  g <- colSums(Xc) - colSums(wts_sp * pi_sp * Xp)

  J <- -t(wts_sp * pi_sp * (1 - pi_sp) * Xp) %*% Xp

  check_estimating_system(
    J = J,
    g = g,
    label = label,
    tol_singular = tol_singular
  )

  J
}


#' Starting values for the coefficients of the CLW participation model
#' @noRd
clw_start <- function(Xc, Xp, wts_sp) {
  p_dim <- ncol(Xc)
  beta_start <- rep(0, p_dim)

  pi_sp0 <- as.vector(expit(Xp %*% beta_start))

  if (any(!is.finite(pi_sp0))) {
    stop(
      "CLW failed: cannot construct initial values because initial fitted probabilities are non-finite.",
      call. = FALSE
    )
  }

  f_p0 <- colSums(wts_sp * pi_sp0 * Xp)
  f_c0 <- colSums(Xc)

  if (!is.finite(f_p0[1]) || !is.finite(f_c0[1]) ||
      f_p0[1] <= 0 || f_c0[1] <= 0) {
    stop(
      "CLW failed: cannot construct initial intercept because intercept totals are invalid.",
      call. = FALSE
    )
  }

  beta_start[1] <- -log(f_p0[1] / f_c0[1])

  if (any(!is.finite(beta_start))) {
    stop(
      "CLW failed: initial coefficient vector contains non-finite values.",
      call. = FALSE
    )
  }

  beta_start
}

#' Probability Reference Sample 2 (sp2)
#'
#' This dataset represents a probability survey derived from the
#' 1997--2008 cycles of the National Health Interview Survey (NHIS).
#' It is intended for use
#' alongside \code{sc} and \code{sp1} to illustrate the multi-reference
#' calibration method implemented in the \code{nonprobsampling} package.
#'
#' @format A data frame with 35525 observations and 11 variables:
#' \describe{
#'   \item{agecat}{Age category (factor with 4 levels: 1 = 55--59, 2 = 60--64, 3 = 65--69, 4 = 70+)}
#'   \item{marital}{Marital status (factor with 4 levels: 1 = Married Or Living As Married, 2 = Widowed, 3 = Divorced or Separated, 4 = Never Married)}
#'   \item{race}{Race category (factor with 4 levels: 1 = White, 2 = Black, 3 = Hispanic, 4 = Other)}
#'   \item{employment}{Employment status (factor with 2 levels: 0 = Not Working, 1 = Working)}
#'   \item{diabetes}{Diabetes diagnosis indicator (factor with 2 levels: 0 = No, 1 = Yes)}
#'   \item{BMI}{Body mass index category (factor with 4 levels: "Normal", "Overweight", "Obese", "Morbidly Obese")}
#'   \item{smoking}{Smoking status (factor with 3 levels: 1 = Never Smoker, 2 = Former Smoker, 3 = Current Smoker)}
#'   \item{comorbidity}{General comorbidity indicator (factor with 2 levels: 0 = No, 1 = Yes)}
#'   \item{wts_sp2}{Sampling weights (numeric)}
#'   \item{strata_sp2}{Stratum identifier for complex survey design (numeric)}
#'   \item{psu_sp2}{Primary sampling unit identifier for complex survey design (numeric)}
#' }
#'
#' @details
#' The dataset includes auxiliary variables shared with the nonprobability
#' sample \code{sc}, enabling the construction of pseudo-weights to adjust
#' for participation bias. Survey design variables and sampling weights are
#' provided to support design-consistent estimation.
#'
#' @source
#' Derived from the National Health Interview Survey (NHIS), 1997--2008 cycles,
#' conducted by the U.S. National Center for Health Statistics (NCHS).
#'
#' @usage data(sp2)
#'
#' @examples
#' data(sp2)
#' str(sp2)
#' summary(sp2)
"sp2"

#' Standardize a domain variable into a common internal format
#'
#' Converts `zcol` in `data` to a standard form used throughout the
#' estimation pipeline. Logical and binary numeric (0/1) variables become a
#' single integer indicator column; character variables are trimmed and
#' coerced to a factor; factor variables are dropped of unused levels and
#' expanded to one integer indicator column per level. The returned list
#' always has the same structure so downstream code can branch on `mode`
#' alone.
#'
#' @param data A data frame containing the column named by `zcol`.
#' @param zcol Single character string naming the domain variable in `data`,
#'   or NULL for the overall (no-domain) case.
#'
#' @return A list with components:
#'
#' - `mode`: one of `"overall"`, `"binary"`, or `"factor"`.
#' - `z_name`: the value of `zcol`, or NULL when `zcol` is NULL.
#' - `labels`: character vector of domain labels shown to the user.
#' - `indicators`: data frame of integer 0/1 indicator columns (one column
#'   per domain level), or NULL in the overall case.
#'
#' @keywords internal
standardize_zcol <- function(data, zcol = NULL) {
  #----------------------------------#
  # Step 1: overall case
  #----------------------------------#
  if (is.null(zcol)) {
    return(list(
      mode = "overall",      # one of: overall / binary / factor
      z_name = NULL,         # original variable name
      labels = "Overall",    # domain labels shown to user
      indicators = NULL      # data.frame of internal 0/1 indicators
    ))
  }

  #----------------------------------#
  # Step 2: basic input checks
  #----------------------------------#
  if (!is.data.frame(data)) {
    stop("`data` must be a data.frame.", call. = FALSE)
  }

  if (!is.character(zcol) || length(zcol) != 1L) {
    stop("`zcol` must be NULL or a single character string.", call. = FALSE)
  }

  if (!zcol %in% names(data)) {
    stop("`zcol` was not found in `data`.", call. = FALSE)
  }

  z <- data[[zcol]]

  #----------------------------------#
  # Step 3: logical -> binary
  #----------------------------------#
  if (is.logical(z)) {
    ind_df <- data.frame(.z_domain = as.integer(z))

    return(list(
      mode = "binary",
      z_name = zcol,
      labels = paste0(zcol, " = 1"),
      indicators = ind_df
    ))
  }

  #----------------------------------#
  # Step 4: numeric/integer binary
  #----------------------------------#
  if (is.numeric(z) || is.integer(z)) {
    uz <- sort(unique(stats::na.omit(z)))

    if (length(uz) == 2L && all(uz %in% c(0, 1))) {
      ind_df <- data.frame(.z_domain = as.integer(z))

      return(list(
        mode = "binary",
        z_name = zcol,
        labels = paste0(zcol, " = 1"),
        indicators = ind_df
      ))
    }

    stop(
      paste0(
        "`zcol` is numeric/integer but not binary. ",
        "Allowed numeric domain variables must contain only 0/1 (ignoring NA)."
      ),
      call. = FALSE
    )
  }

  #----------------------------------#
  # Step 5: character -> factor
  #----------------------------------#
  if (is.character(z)) {
    z <- trimws(z)

    # treat empty string as missing
    z[z == ""] <- NA_character_

    z <- factor(z)
  }

  #----------------------------------#
  # Step 6: factor -> one indicator per level
  #----------------------------------#
  if (is.factor(z)) {
    z <- droplevels(z)
    levs <- levels(z)

    if (length(levs) < 2L) {
      stop(
        "`zcol` factor must contain at least two non-missing levels.",
        call. = FALSE
      )
    }

    ind_list <- lapply(levs, function(lv) as.integer(z == lv))
    ind_df <- as.data.frame(ind_list, stringsAsFactors = FALSE)

    # internal safe names only for internal use
    safe_names <- paste0(".z_", make.names(levs, unique = TRUE))
    names(ind_df) <- safe_names

    return(list(
      mode = "factor",
      z_name = zcol,
      labels = paste0(zcol, " = ", levs),
      indicators = ind_df
    ))
  }

  #----------------------------------#
  # Step 7: unsupported type
  #----------------------------------#
  stop(
    paste0(
      "`zcol` must be one of the following: ",
      "NULL, logical, binary numeric/integer (0/1), character, or factor."
    ),
    call. = FALSE
  )
}

#' Control Solver Settings for Pseudo-Weight Estimation
#'
#' @description
#' \code{pw_solver_control()} creates a solver control object used by
#' \code{\link{est_pw}} to manage numerical settings for pseudo-weight
#' estimation.
#'
#' @details
#' The control object stores solver settings used by pseudo-weight estimation
#' step. It is passed to \code{\link{est_pw}} through the \code{control}
#' argument.
#'
#' Currently, only \code{solver = "nleqslv"} is supported. The arguments
#' \code{method}, \code{global}, \code{xscalm}, \code{ftol}, \code{xtol}, and
#' \code{maxit} correspond to options used by \code{nleqslv::nleqslv()}.
#' They are collected internally and passed to \code{nleqslv::nleqslv()} at the
#' pseudo-weight estimation step.
#'
#' The argument \code{ftol} is the function-value convergence tolerance. It
#' controls convergence based on the size of the estimating function. The
#' argument \code{xtol} is the parameter-step convergence tolerance. It controls
#' convergence based on changes in the parameter vector. The argument
#' \code{maxit} controls the maximum number of solver iterations.
#'
#' Additional, less commonly used \code{nleqslv} control options can be supplied
#' through \code{nleqslv_control}. To avoid ambiguity, do not supply
#' \code{ftol}, \code{xtol}, \code{maxit}, or \code{trace} inside
#' \code{nleqslv_control}; use the main arguments instead.
#'
#' @param solver Character string specifying the numerical solver used for
#'   solving the estimating equations. Currently, only \code{"nleqslv"} is
#'   supported. Default is \code{"nleqslv"}.
#'
#' @param maxit Positive finite numeric value passed to
#'   \code{nleqslv::nleqslv()} as the maximum number of solver iterations. The
#'   value is converted to an integer before being stored in the returned
#'   control object. Default is \code{150} when a global strategy is specified
#'   (i.e., \code{global != "none"}), and \code{20} when no global strategy is
#'   used (\code{global = "none"}), matching \code{nleqslv}'s own defaults.
#'   Since the default global strategy is \code{"dbldog"}, the effective default
#'   is \code{150} unless \code{global = "none"} is explicitly specified.
#'
#' @param trace Logical. If \code{TRUE}, tracing or solver progress information
#'   may be requested from the underlying numerical routine when supported.
#'   Default is \code{FALSE}.
#'
#' @param method Character string specifying the numerical method passed to
#'   \code{nleqslv::nleqslv()}. Supported values are \code{"Newton"} and
#'   \code{"Broyden"}. Default is \code{"Newton"}.
#'
#' @param global Character string specifying the global strategy passed to
#'   \code{nleqslv::nleqslv()}. Supported values are \code{"dbldog"},
#'   \code{"cline"}, \code{"pwldog"}, \code{"qline"}, \code{"gline"},
#'   \code{"hook"}, and \code{"none"}. Default is \code{"dbldog"}.
#'
#' @param xscalm Character string specifying the scaling method passed to
#'   \code{nleqslv::nleqslv()}. Supported values are \code{"fixed"} and
#'   \code{"auto"}. Default is \code{"fixed"}.
#'
#' @param ftol Positive finite numeric value passed to
#'   \code{nleqslv::nleqslv()} as the function-value convergence tolerance. This
#'   controls convergence based on the size of the estimating function. Default
#'   is \code{1e-8}.
#'
#' @param xtol Positive finite numeric value passed to
#'   \code{nleqslv::nleqslv()} as the parameter-step convergence tolerance. This
#'   controls convergence based on changes in the parameter vector. Default is
#'   \code{1e-8}.
#'
#' @param nleqslv_control A list of additional control options
#'   passed to \code{nleqslv::nleqslv()}. This can include less commonly used
#'   control options, such as \code{btol}, \code{cndtol},
#'   \code{sigma}, and \code{scalex}. See
#'   \code{\link[nleqslv]{nleqslv}} for details.
#'
#' @return
#' A flat list containing all solver control settings for pseudo-weight
#' estimation:
#'
#' \describe{
#'   \item{\code{solver}}{The selected numerical solver.}
#'   \item{\code{method}}{The nleqslv numerical method.}
#'   \item{\code{global}}{The nleqslv global strategy.}
#'   \item{\code{xscalm}}{The nleqslv scaling method.}
#'   \item{\code{ftol}}{The function-value convergence tolerance.}
#'   \item{\code{xtol}}{The parameter-step convergence tolerance.}
#'   \item{\code{maxit}}{The maximum number of solver iterations, stored as an integer.
#'     \code{150} if a global strategy is used; \code{20} if \code{global = "none"}.
#'     Since the default global strategy is \code{"dbldog"}, the effective default
#'     is \code{150} unless \code{global = "none"} is explicitly specified.}
#'   \item{\code{trace}}{Logical value indicating whether tracing information is requested.}
#'   \item{\code{nleqslv_control}}{A list of additional options passed to
#'     \code{nleqslv::nleqslv()}.}
#' }
#'
#' @seealso
#' \code{\link{est_pw}}
#'
#' @examples
#' ## Default solver control settings
#' ctrl <- pw_solver_control()
#'
#' ## Custom nleqslv solver settings
#' ctrl <- pw_solver_control(
#'   maxit  = 20,
#'   trace  = FALSE,
#'   method = "Newton",
#'   global = "cline",
#'   xscalm = "auto",
#'   ftol   = 1e-8,
#'   xtol   = 1e-10
#' )
#'
#' ## Additional nleqslv control options
#' ctrl <- pw_solver_control(
#'   method = "Newton",
#'   global = "dbldog",
#'   nleqslv_control = list(
#'     btol = 1e-3
#'   )
#' )
#'
#' @export
pw_solver_control <- function(
    solver = "nleqslv",
    maxit = NULL,
    trace = FALSE,
    method = c("Newton", "Broyden"),
    global = c("dbldog", "cline", "pwldog", "qline", "gline", "hook", "none"),
    xscalm = c("fixed", "auto"),
    ftol = 1e-8,
    xtol = 1e-8,
    nleqslv_control = list()
) {

  if (!is.character(solver) || length(solver) != 1L || is.na(solver)) {
    stop("`solver` must be a single character string.", call. = FALSE)
  }

  solver <- tolower(trimws(solver))

  if (solver != "nleqslv") {
    stop("Only solver = 'nleqslv' is currently supported.", call. = FALSE)
  }

  if (!is.logical(trace) || length(trace) != 1L || is.na(trace)) {
    stop("`trace` must be TRUE or FALSE.", call. = FALSE)
  }

  method <- match.arg(method)
  global <- match.arg(global)
  xscalm <- match.arg(xscalm)

  if (is.null(maxit)) {
    maxit <- if (global == "none") 20L else 150L
  }

  if (!is.numeric(maxit) || length(maxit) != 1L || !is.finite(maxit) || maxit <= 0) {
    stop("`maxit` must be a positive finite number.", call. = FALSE)
  }

  if (!is.numeric(ftol) || length(ftol) != 1L || !is.finite(ftol) || ftol <= 0) {
    stop("`ftol` must be a positive finite number.", call. = FALSE)
  }

  if (!is.numeric(xtol) || length(xtol) != 1L || !is.finite(xtol) || xtol <= 0) {
    stop("`xtol` must be a positive finite number.", call. = FALSE)
  }

  if (!is.list(nleqslv_control)) {
    stop("`nleqslv_control` must be a list.", call. = FALSE)
  }

  duplicated_control_names <- intersect(
    names(nleqslv_control),
    c("ftol", "xtol", "maxit", "trace")
  )

  if (length(duplicated_control_names) > 0L) {
    stop(
      paste0(
        "Do not supply ",
        paste(sprintf("`%s`", duplicated_control_names), collapse = ", "),
        " inside `nleqslv_control`."
      ),
      call. = FALSE
    )
  }

  list(
    solver          = solver,
    method          = method,
    global          = global,
    xscalm          = xscalm,
    ftol            = ftol,
    xtol            = xtol,
    maxit           = as.integer(maxit),
    trace           = trace,
    nleqslv_control = nleqslv_control
  )
}

# test-est_pw.R
# Integration tests for est_pw() — the main exported function.
#
# These tests cover the full workflow through the public API.
# They complement the step-level unit tests by checking that the pipeline
# is connected correctly and that the documented behaviour is preserved.
#
# Covered scenarios:
#   A. One-reference: all four methods (calibration / alp / clw / cali alias)
#   B. One-reference: explicit p_formula
#   C. One-reference: custom sc_wname
#   D. Multi-reference: default (precali = TRUE, sp_order = "size")
#   E. Multi-reference: precali = FALSE and sp_order = "given"
#   F. NA handling: na.omit / na.exclude / na.fail
#   G. Return-object structure (class, fields, pseudo-weight properties)
#   H. Errors are assigned to the correct estimation step


# Setup ----

data(sc)
data(sp1)
data(sp2)

des1 <- survey::svydesign(
  ids     = ~psu_sp1,
  strata  = ~strata_sp1,
  weights = ~wts_sp1,
  data    = sp1,
  nest    = TRUE
)

des2 <- survey::svydesign(
  ids     = ~psu_sp2,
  strata  = ~strata_sp2,
  weights = ~wts_sp2,
  data    = sp2,
  nest    = TRUE
)

# Convenience wrappers ----
run_one   <- function(...) suppressMessages(
  est_pw(data = list(sc, des1),
         control = pw_solver_control(ftol=1e-6),
         ...)
)

# Multi-ref requires each survey to introduce new variables.
# sc shares {race, agecat, comorbidity} with both sp1 and sp2.
# Assigning ~agecat to sp1 and ~race to sp2 ensures sp1 covers {agecat} and
# sp2 contributes the new variable {race}, so the calibration equations have
# one distinct block of auxiliary variables from each reference survey.
run_multi <- function(...) suppressMessages(
  est_pw(
    data      = list(sc, des1, des2),
    p_formula = list(~agecat, ~race),
    control   = pw_solver_control(ftol = 1e-6),
    ...
  )
)


# A. One-reference: methods ----

test_that("est_pw one-ref (method = NULL): runs without error and defaults to calibration", {
  out <- run_one()

  expect_s3_class(out, "pw_fit")
  expect_equal(out$method, "calibration")
})

test_that("est_pw one-ref (method = 'calibration'): produces positive finite pseudo-weights", {
  out <- run_one(method = "calibration")

  expect_true(all(is.finite(out$pseudo_weights)))
  expect_true(all(out$pseudo_weights > 0))
})

test_that("est_pw one-ref (method = 'alp'): runs and sets method correctly", {
  out <- run_one(method = "alp")

  expect_equal(out$method, "alp")
  expect_true(all(out$pseudo_weights > 0))
  expect_true(all(is.finite(out$pseudo_weights)))
})

test_that("est_pw one-ref (method = 'clw'): runs and sets method correctly", {
  out <- run_one(method = "clw")

  expect_equal(out$method, "clw")
  expect_true(all(out$pseudo_weights > 0))
  expect_true(all(is.finite(out$pseudo_weights)))
})

test_that("est_pw one-ref (method = 'cali'): alias resolves to 'calibration'", {
  out <- run_one(method = "cali")

  expect_equal(out$method, "calibration")
})


# B. One-reference: explicit p_formula ----

test_that("est_pw one-ref: explicit p_formula runs without error", {
  out <- run_one(p_formula = ~agecat)

  expect_s3_class(out, "pw_fit")
  expect_true(all(out$pseudo_weights > 0))
})

test_that("est_pw one-ref: multi-variable p_formula runs without error", {
  out <- run_one(p_formula = ~agecat + race)

  expect_s3_class(out, "pw_fit")
  expect_true(all(is.finite(out$pseudo_weights)))
})

test_that("est_pw one-ref: two-sided p_formula raises a Step 1 error", {
  expect_error(
    run_one(p_formula = agecat ~ race),
    "Step 1",
    fixed = TRUE
  )
})


# C. One-reference: custom sc_wname ----

test_that("est_pw one-ref: custom sc_wname appears as column in sc_updated", {
  out <- run_one(sc_wname = "my_wts")

  expect_true("my_wts" %in% names(out$sc_updated))
  expect_false("pseudo_wts" %in% names(out$sc_updated))
})


# D. Multi-reference: default settings ----

test_that("est_pw multi-ref: runs without error with default settings", {
  out <- run_multi()

  expect_s3_class(out, "pw_fit")
  expect_equal(out$method, "multi")
})

test_that("est_pw multi-ref: pseudo-weights are positive and finite", {
  out <- run_multi()

  expect_true(all(out$pseudo_weights > 0))
  expect_true(all(is.finite(out$pseudo_weights)))
})

test_that("est_pw multi-ref: pseudo-weights have length nrow(sc)", {
  out <- run_multi()

  expect_length(out$pseudo_weights, nrow(sc))
})

test_that("est_pw multi-ref: method is always 'multi' regardless of method argument", {
  # Multi-reference always uses method = "multi" regardless of what is passed
  out <- run_multi()

  expect_equal(out$method, "multi")
})

test_that("est_pw multi-ref: sp_order = 'given' runs without error", {
  out <- run_multi(sp_order = "given")

  expect_s3_class(out, "pw_fit")
  expect_true(all(out$pseudo_weights > 0))
})


# E. Multi-reference: precali = FALSE ----

# precali = FALSE with sp_order = "given" produces a singular Jacobian for
# this dataset because the uncalibrated weights create an ill-conditioned
# system.  sp_order = "size" (the default) conditions the problem correctly
# and allows convergence even without precalibration.

test_that("est_pw multi-ref (precali = FALSE, sp_order = 'size'): runs without error", {
  out <- run_multi(precali = FALSE, sp_order = "size")

  expect_s3_class(out, "pw_fit")
  expect_equal(out$method, "multi")
})

test_that("est_pw multi-ref (precali = FALSE, sp_order = 'size'): pseudo-weights are positive", {
  out <- run_multi(precali = FALSE, sp_order = "size")

  expect_true(all(out$pseudo_weights > 0))
})


# F. NA handling ----

# Introduce NAs into sc in a p_formula variable
sc_na <- sc
sc_na$agecat[1:3] <- NA

test_that("est_pw one-ref (na.omit): sc_updated drops NA rows", {
  out <- suppressMessages(
    est_pw(
      data      = list(sc_na, des1),
      p_formula = ~agecat,
      na.action = stats::na.omit,
      control   = pw_solver_control(ftol = 1e-4)
    )
  )

  expect_equal(nrow(out$sc_updated), nrow(sc_na) - 3L)
  expect_equal(length(out$pseudo_weights), nrow(sc_na) - 3L)
  expect_false(anyNA(out$pseudo_weights))
})

test_that("est_pw one-ref (na.omit): na_summary reports excluded rows", {
  out <- suppressMessages(
    est_pw(
      data      = list(sc_na, des1),
      p_formula = ~agecat,
      na.action = stats::na.omit,
      control   = pw_solver_control(ftol = 1e-4)
    )
  )

  expect_s3_class(out$na_summary, "pw_na_summary")
  expect_equal(out$na_summary$sc$n_excluded, 3L)
})

test_that("est_pw one-ref (na.exclude): sc_updated has nrow(sc) rows; NA at excluded", {
  out <- suppressMessages(
    est_pw(
      data      = list(sc_na, des1),
      p_formula = ~agecat,
      na.action = stats::na.exclude,
      control   = pw_solver_control(ftol = 1e-4)
    )
  )

  expect_equal(nrow(out$sc_updated), nrow(sc_na))
  expect_true(all(is.na(out$pseudo_weights[1:3])))
  expect_false(anyNA(out$pseudo_weights[-(1:3)]))
})

test_that("est_pw one-ref (na.fail): errors when NA present", {
  expect_error(
    suppressMessages(
      est_pw(
        data      = list(sc_na, des1),
        p_formula = ~agecat,
        na.action = stats::na.fail
      )
    ),
    "Step 3",
    fixed = TRUE
  )
})

test_that("est_pw one-ref (no NA): na_summary is NULL", {
  out <- run_one(p_formula = ~agecat)

  expect_null(out$na_summary)
})


# G. Return-object structure ----

test_that("est_pw one-ref: returned object has class 'pw_fit'", {
  out <- run_one()
  expect_s3_class(out, "pw_fit")
})

test_that("est_pw one-ref: returned object has required top-level fields", {
  out <- run_one()

  # na_summary is always present (NULL when no NAs, set via list-assignment)
  required <- c("pseudo_weights", "coefficients", "method",
                "solver_diagnostics", "sc_updated", "internal",
                "call", "na_summary")
  expect_true(all(required %in% names(out)))
})

test_that("est_pw one-ref: pseudo_weights has length nrow(sc)", {
  out <- run_one()

  expect_length(out$pseudo_weights, nrow(sc))
})

test_that("est_pw one-ref: sc_updated has the default sc_wname column", {
  out <- run_one()

  expect_true("pseudo_wts" %in% names(out$sc_updated))
})

test_that("est_pw one-ref: sc_updated has nrow(sc) rows when no NAs", {
  out <- run_one()

  expect_equal(nrow(out$sc_updated), nrow(sc))
})

test_that("est_pw one-ref: call is stored as a call object", {
  out <- run_one()

  expect_true(inherits(out$call, "call"))
})

test_that("est_pw one-ref: coefficients are named", {
  out <- run_one()

  expect_false(is.null(names(out$coefficients)))
  expect_true("(Intercept)" %in% names(out$coefficients))
})

test_that("est_pw one-ref: solver_diagnostics has expected fields", {
  out <- run_one()

  expected <- c("solver", "termcd", "message", "method", "iter", "fmax")
  expect_true(all(expected %in% names(out$solver_diagnostics)))
})

test_that("est_pw one-ref: solver converged (termcd 1 or 2)", {
  out <- run_one()

  expect_true(out$solver_diagnostics$termcd %in% c(1L, 2L))
})

test_that("est_pw multi-ref: returned object has required top-level fields", {
  out <- run_multi()

  required <- c("pseudo_weights", "coefficients", "method",
                "solver_diagnostics", "sc_updated", "internal",
                "call", "na_summary")
  expect_true(all(required %in% names(out)))
})

test_that("est_pw multi-ref: na_summary is NULL when no NAs", {
  out <- run_multi()

  expect_true("na_summary" %in% names(out))
  expect_null(out$na_summary)
})


# H. Error attributed correctly ----

test_that("est_pw: invalid data (not a list) raises a Step 1 error", {
  expect_error(
    est_pw(data = 42L),
    "Step 1",
    fixed = TRUE
  )
})

test_that("est_pw: invalid method raises a Step 1 error", {
  expect_error(
    run_one(method = "bad_method"),
    "Step 1",
    fixed = TRUE
  )
})

test_that("est_pw: method = 'multi' with one reference raises a Step 1 error", {
  expect_error(
    run_one(method = "multi"),
    "Step 1",
    fixed = TRUE
  )
})

test_that("est_pw: p_formula with variable absent from sc raises a Step 1 error", {
  expect_error(
    run_one(p_formula = ~no_such_column),
    "Step 1",
    fixed = TRUE
  )
})

# test-print_pw_fit.R
# Tests for the print.pw_fit S3 method and na.action.pw_fit.
#
# The output checks use stable labels rather than exact numeric values,
# since example data may change over time.


# Setup ----

data(sc)
data(sp1)
data(sp2)

des1 <- survey::svydesign(ids = ~psu_sp1, strata = ~strata_sp1,
                          weights = ~wts_sp1, data = sp1, nest = TRUE)
des2 <- survey::svydesign(ids = ~psu_sp2, strata = ~strata_sp2,
                          weights = ~wts_sp2, data = sp2, nest = TRUE)

ctrl <- pw_solver_control(ftol = 1e-6)

fit_cali  <- suppressMessages(est_pw(list(sc, des1), method = "calibration",
                                     p_formula = ~ agecat + race, control = ctrl))
fit_alp   <- suppressMessages(est_pw(list(sc, des1), method = "alp",
                                     p_formula = ~ agecat + race, control = ctrl))
fit_clw   <- suppressMessages(est_pw(list(sc, des1), method = "clw",
                                     p_formula = ~ agecat + race, control = ctrl))
fit_multi <- suppressMessages(est_pw(list(sc, des1, des2),
                                     p_formula = list(~ agecat, ~ race), control = ctrl))


# Common sections ----

test_that("print.pw_fit shows header, call, method, model, convergence, weights", {
  expect_output(print(fit_cali), "Pseudo-weight fit",      fixed = TRUE)
  expect_output(print(fit_cali), "Call:",                  fixed = TRUE)
  expect_output(print(fit_cali), "Method:",                fixed = TRUE)
  expect_output(print(fit_cali), "Participation model:",   fixed = TRUE)
  expect_output(print(fit_cali), "Convergence:",           fixed = TRUE)
  expect_output(print(fit_cali), "Pseudo-weights",         fixed = TRUE)
  expect_output(print(fit_cali), "Sum:",                   fixed = TRUE)
  expect_output(print(fit_cali), "summary()",              fixed = TRUE)
})

test_that("print.pw_fit returns its argument invisibly", {
  txt <- capture.output(rv <- withVisible(print(fit_cali)))
  expect_false(rv$visible)
  expect_identical(rv$value, fit_cali)
})


# Method labels ----

test_that("print.pw_fit labels one-reference calibration", {
  expect_output(print(fit_cali), "One reference calibration", fixed = TRUE)
})

test_that("print.pw_fit labels one-reference ALP", {
  expect_output(print(fit_alp), "One reference ALP", fixed = TRUE)
})

test_that("print.pw_fit labels one-reference CLW", {
  expect_output(print(fit_clw), "One reference CLW", fixed = TRUE)
})

test_that("print.pw_fit labels multi-reference calibration", {
  expect_output(print(fit_multi), "Multi-reference calibration", fixed = TRUE)
})


# Participation model parameter count ----

test_that("print.pw_fit reports the number of participation model parameters", {
  p <- length(fit_cali$coefficients)
  expect_output(print(fit_cali), sprintf("%d parameters", p), fixed = TRUE)
})

test_that("print.pw_fit notes '(incl. intercept)' for one-reference but not multi", {
  expect_output(print(fit_cali), "(incl. intercept)", fixed = TRUE)

  out <- capture.output(print(fit_multi))
  expect_false(any(grepl("incl. intercept", out, fixed = TRUE)))
})


# Convergence ----

test_that("print.pw_fit reports a converged solver as 'converged'", {
  expect_output(print(fit_cali), "converged", fixed = TRUE)
})


# Pseudo-weight summary ----

test_that("print.pw_fit weight summary reports n and a six-number summary", {
  n <- sum(!is.na(fit_cali$pseudo_weights))
  expect_output(print(fit_cali), sprintf("n = %d", n), fixed = TRUE)
  expect_output(print(fit_cali), "Min.",   fixed = TRUE)
  expect_output(print(fit_cali), "Median", fixed = TRUE)
  expect_output(print(fit_cali), "Max.",   fixed = TRUE)
})


# na.exclude path ----

test_that("print.pw_fit (na.exclude) reports NA weights and a deletion note", {
  sc_na <- sc
  sc_na$agecat[1:6] <- NA
  fit_ex <- suppressMessages(est_pw(list(sc_na, des1), method = "calibration",
                                    p_formula = ~ agecat + race, control = ctrl,
                                    na.action = stats::na.exclude))
  expect_output(print(fit_ex), "NA", fixed = TRUE)
  expect_output(print(fit_ex), "deleted due to missingness in sc", fixed = TRUE)
})


# na.action.pw_fit ----

test_that("na.action(pw_fit) reports no omitted rows for a complete-case fit", {
  na_obj <- na.action(fit_cali)
  expect_true(is.null(na_obj) || length(na_obj) == 0L)
})

test_that("na.action(pw_fit) returns the omitted-row index object when rows drop", {
  sc_na <- sc
  sc_na$agecat[1:6] <- NA
  fit_om <- suppressMessages(est_pw(list(sc_na, des1), method = "calibration",
                                    p_formula = ~ agecat + race, control = ctrl,
                                    na.action = stats::na.omit))
  na_obj <- na.action(fit_om)
  expect_false(is.null(na_obj))
  expect_s3_class(na_obj, "omit")
})

# test-summary_pw_fit.R
# Tests for the summary.pw_fit S3 method (full call / method / solver
# diagnostics / coefficient table / NA summary).


# Setup ----

data(sc)
data(sp1)
data(sp2)

des1 <- survey::svydesign(ids = ~psu_sp1, strata = ~strata_sp1,
                          weights = ~wts_sp1, data = sp1, nest = TRUE)
des2 <- survey::svydesign(ids = ~psu_sp2, strata = ~strata_sp2,
                          weights = ~wts_sp2, data = sp2, nest = TRUE)

ctrl <- pw_solver_control(ftol = 1e-6)

fit_cali  <- suppressMessages(est_pw(list(sc, des1), method = "calibration",
                                     p_formula = ~ agecat + race, control = ctrl))
fit_alp   <- suppressMessages(est_pw(list(sc, des1), method = "alp",
                                     p_formula = ~ agecat + race, control = ctrl))
fit_clw   <- suppressMessages(est_pw(list(sc, des1), method = "clw",
                                     p_formula = ~ agecat + race, control = ctrl))
fit_multi <- suppressMessages(est_pw(list(sc, des1, des2),
                                     p_formula = list(~ agecat, ~ race), control = ctrl))


# Common sections ----

test_that("summary.pw_fit prints call, method, diagnostics and coefficients", {
  expect_output(summary(fit_cali), "Call:",                            fixed = TRUE)
  expect_output(summary(fit_cali), "Method: One reference calibration", fixed = TRUE)
  expect_output(summary(fit_cali), "Solver diagnostics:",              fixed = TRUE)
  expect_output(summary(fit_cali), "Participation model coefficients:", fixed = TRUE)
})

test_that("summary.pw_fit returns its argument invisibly", {
  txt <- capture.output(rv <- withVisible(summary(fit_cali)))
  expect_false(rv$visible)
  expect_identical(rv$value, fit_cali)
})


# Method labels ----

test_that("summary.pw_fit labels one-reference ALP", {
  expect_output(summary(fit_alp), "Method: One reference ALP", fixed = TRUE)
})

test_that("summary.pw_fit labels one-reference CLW", {
  expect_output(summary(fit_clw), "Method: One reference CLW", fixed = TRUE)
})

test_that("summary.pw_fit labels multi-reference calibration", {
  expect_output(summary(fit_multi), "Multi-reference calibration", fixed = TRUE)
})


# Solver diagnostics block ----

test_that("summary.pw_fit prints solver name and termination code", {
  expect_output(summary(fit_cali), "Solver:",           fixed = TRUE)
  expect_output(summary(fit_cali), "Termination code:", fixed = TRUE)
  expect_output(summary(fit_cali), "Iterations:",       fixed = TRUE)
})


# NA summary ----

test_that("summary.pw_fit reports the NA deletion note when rows are dropped", {
  sc_na <- sc
  sc_na$agecat[1:6] <- NA
  fit_na <- suppressMessages(est_pw(list(sc_na, des1), method = "calibration",
                                    p_formula = ~ agecat + race, control = ctrl))
  expect_output(summary(fit_na), "deleted due to missingness in sc", fixed = TRUE)
})

# test-print_pw_na_summary.R
# Tests for the print.pw_na_summary S3 method. The pw_na_summary object is
# produced by est_pw() and stored in fit$na_summary whenever rows are dropped
# because of missing participation model variables.


# Setup ----

data(sc)
data(sp1)
data(sp2)

des1 <- survey::svydesign(ids = ~psu_sp1, strata = ~strata_sp1,
                          weights = ~wts_sp1, data = sp1, nest = TRUE)
des2 <- survey::svydesign(ids = ~psu_sp2, strata = ~strata_sp2,
                          weights = ~wts_sp2, data = sp2, nest = TRUE)

ctrl <- pw_solver_control(ftol = 1e-6)

# Exclusions by introducing NA in a participation model variable in sc.
sc_na <- sc
sc_na$agecat[1:6] <- NA

fit_one <- suppressMessages(est_pw(list(sc_na, des1), method = "calibration",
                                   p_formula = ~ agecat + race, control = ctrl))
nas_one <- fit_one$na_summary

fit_multi <- suppressMessages(est_pw(list(sc_na, des1, des2),
                                     p_formula = list(~ agecat, ~ race), control = ctrl))
nas_multi <- fit_multi$na_summary


# Object class ----

test_that("est_pw stores a pw_na_summary object when rows are excluded", {
  expect_s3_class(nas_one, "pw_na_summary")
})


# print.pw_na_summary ----

test_that("print.pw_na_summary prints the header and the count columns", {
  expect_output(print(nas_one), "NA processing summary:", fixed = TRUE)
  expect_output(print(nas_one), "n_orig",                 fixed = TRUE)
  expect_output(print(nas_one), "n_used",                 fixed = TRUE)
  expect_output(print(nas_one), "n_excluded",             fixed = TRUE)
})

test_that("print.pw_na_summary lists sc and the reference survey rows (one-ref)", {
  out <- capture.output(print(nas_one))
  expect_true(any(grepl("^sc", out)))
  expect_true(any(grepl("sp", out)))
})

test_that("print.pw_na_summary returns its argument invisibly", {
  txt <- capture.output(rv <- withVisible(print(nas_one)))
  expect_false(rv$visible)
  expect_identical(rv$value, nas_one)
})

test_that("print.pw_na_summary handles the multi-reference layout", {
  skip_if(is.null(nas_multi), "no exclusions recorded in the multi-reference fit")
  expect_s3_class(nas_multi, "pw_na_summary")
  expect_output(print(nas_multi), "NA processing summary:", fixed = TRUE)
  out <- capture.output(print(nas_multi))
  expect_gte(sum(grepl("sp", out)), 1L)
})

#' Build pseudo-weights using the CLW method
#'
#' @param vars Character vector of predictor variable names.
#' @param sc Data frame. The nonprobability sample.
#' @param sp Data frame. The probability reference sample.
#' @param sp_des A `survey.design2` or `svyrep.design` object for `sp`.
#' @param wts.col Character scalar. Name of the survey-weight column in `sp`.
#' @param control A list created by `pw_solver_control()`.
#' @param verbose Logical. If `TRUE`, convergence messages are printed.
#' @param log_messages Character vector of messages accumulated in earlier
#'   steps, appended to the returned object unchanged.
#'
#' @return A list with components:
#'   \item{weights}{Numeric vector of CLW pseudo-weights for `sc`.}
#'   \item{coefficients}{Named numeric vector of participation model coefficients.}
#'   \item{solver_diagnostics}{List of solver convergence diagnostics.}
#'   \item{log_messages}{Updated character vector of log messages.}
#'   \item{internal}{List containing design matrices (`Xc`, `Xp`), fitted
#'     participation probabilities (`pi_sc`, `pi_sp`), and sandwich components
#'     (`S_beta`, `D`) needed by the estimation stage.}
#'
#' @keywords internal
clw_build <- function(vars, sc, sp, sp_des, wts.col,
                      control,
                      verbose = FALSE,
                      log_messages = NULL) {

  if (is.null(log_messages)) {
    log_messages <- character(0)
  }

  Xc <- add_intercept(vars = vars, data = sc)
  Xp <- add_intercept(vars = vars, data = sp)

  check_design_identifiability(Xc, label = "One_Ref_CLW")
  check_design_identifiability(Xp, label = "One_Ref_CLW")

  if (ncol(Xc) != ncol(Xp)) {
    stop(
      "CLW failed: Xc and Xp must have the same number of columns.",
      call. = FALSE
    )
  }

  if (!wts.col %in% names(sp)) {
    stop(
      sprintf("CLW failed: weight column '%s' was not found in `sp`.", wts.col),
      call. = FALSE
    )
  }

  wts_sp <- sp[[wts.col]]

  if (!is.numeric(wts_sp) || any(!is.finite(wts_sp)) || any(wts_sp <= 0)) {
    stop(
      "CLW failed: reference survey weights must be positive finite numbers.",
      call. = FALSE
    )
  }

  if (verbose) {
    message(sprintf("Fitting CLW participation model with %d variables.", ncol(Xc)))
  }

  beta_start <- clw_start(
    Xc = Xc,
    Xp = Xp,
    wts_sp = wts_sp
  )

  sol <- solve_participation_model(
    beta_start = beta_start,
    fn = clw_fn,
    jac = clw_jac,
    label = "One_Ref_CLW",
    control = control,
    Xc = Xc,
    Xp = Xp,
    wts_sp = wts_sp
  )

  beta <- sol$coefficients

  eta_sc <- as.vector(Xc %*% beta)
  eta_sp <- as.vector(Xp %*% beta)

  if (any(!is.finite(eta_sc)) || any(!is.finite(eta_sp))) {
    stop(
      "CLW failed: final linear predictor contains non-finite values.",
      call. = FALSE
    )
  }

  pi_sc <- as.vector(expit(eta_sc))
  pi_sp <- as.vector(expit(eta_sp))

  if (any(!is.finite(pi_sc)) || any(!is.finite(pi_sp))) {
    stop(
      "CLW failed: final fitted probabilities contain non-finite values.",
      call. = FALSE
    )
  }

  if (any(pi_sc <= 0 | pi_sc >= 1)) {
    stop(
      "CLW failed: final fitted probabilities in sc must be strictly between 0 and 1.",
      call. = FALSE
    )
  }

  wts_sc <- 1 / pi_sc

  if (any(!is.finite(wts_sc))) {
    stop(
      "CLW failed: final pseudo-weights contain non-finite values.",
      call. = FALSE
    )
  }

  f_final <- clw_fn(
    beta = beta,
    Xc = Xc,
    Xp = Xp,
    wts_sp = wts_sp
  )

  max_abs_eq <- max(abs(f_final))

  if (!is.finite(max_abs_eq)) {
    stop(
      "CLW failed: final estimating equations contain non-finite values.",
      call. = FALSE
    )
  }

  if (verbose) {
    if (sol$termcd == 1L || sol$termcd == 2L) {
      message(sprintf("Converged after %d iterations (max |eq| = %.3e).",
                      sol$iterations, max_abs_eq))
    } else {
      message(sprintf(
        "WARNING: solver did not converge (termcd = %d). Check solver_diagnostics for details.",
        sol$termcd
      ))
    }
  }

  S_beta_full <- t(wts_sp * pi_sp * (1 - pi_sp) * Xp) %*% Xp
  D <- compute_D_CLW(sp_des, pi_sp, Xp)

  out <- list(
    weights      = wts_sc,
    coefficients = stats::setNames(as.numeric(beta), colnames(Xc)),

    solver_diagnostics = list(
      solver  = sol$solver,
      termcd  = sol$termcd,
      message = sol$message,
      method  = sol$solver_method,
      iter    = sol$iterations,
      fmax    = max_abs_eq
    ),

    log_messages = log_messages,
    internal = list(
      Xc     = Xc,
      Xp     = Xp,
      pi_sc  = pi_sc,
      pi_sp  = pi_sp,
      S_beta = S_beta_full,
      D      = D
    )
  )

  return(out)
}

#' Construct block-diagonal D matrix for multi-reference raking
#'
#' Internal helper to assemble the block-diagonal covariance matrix
#' \eqn{D} for the multi-reference raking estimator. Each reference survey
#' contributes one design-based covariance block computed from
#' `compute_D_raking()`.
#'
#' @param sp_des_list A non-empty list of survey design objects, each of class
#'   `"survey.design2"` or `"svyrep.design"`.
#' @param Xp_list A list of reference-sample design matrices, one for each
#'   reference survey. Each matrix must have the same number of rows as the
#'   corresponding survey design object in `sp_des_list`.
#'
#' @return A block-diagonal covariance matrix \eqn{D}.
#' @keywords internal
make_block_D_multi <- function(sp_des_list, Xp_list) {

  n <- length(sp_des_list)

  if (!is.list(sp_des_list) || n == 0) {
    stop("'sp_des_list' must be a non-empty list of survey design objects.",
         call. = FALSE)
  }

  if (!is.list(Xp_list) || length(Xp_list) != n) {
    stop("'Xp_list' must be a list with the same length as 'sp_des_list'.",
         call. = FALSE)
  }

  ok_des <- vapply(
    sp_des_list,
    function(x) inherits(x, c("survey.design2", "svyrep.design")),
    logical(1)
  )
  if (!all(ok_des)) {
    bad <- which(!ok_des)
    stop(
      sprintf(
        "Elements %s of 'sp_des_list' are not valid survey design objects.",
        paste(bad, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  block_sizes <- integer(n)

  for (i in seq_len(n)) {
    Xp_i <- Xp_list[[i]]

    if (!is.matrix(Xp_i)) {
      Xp_i <- as.matrix(Xp_i)
    }

    n_sp <- nrow(sp_des_list[[i]]$variables)
    if (nrow(Xp_i) != n_sp) {
      stop(
        sprintf(
          "Reference survey %d: nrow(Xp_list[[%d]]) = %d but design has %d observations.",
          i, i, nrow(Xp_i), n_sp
        ),
        call. = FALSE
      )
    }

    Xp_list[[i]] <- Xp_i
    block_sizes[i] <- ncol(Xp_i)
  }

  total_dim <- sum(block_sizes)
  starts <- cumsum(c(1, utils::head(block_sizes, -1)))
  ends   <- cumsum(block_sizes)

  D <- matrix(0, total_dim, total_dim)

  for (i in seq_len(n)) {
    D_i <- compute_D_raking(
      sp_des     = sp_des_list[[i]],
      Xp         = Xp_list[[i]]
    )

    D_i <- as.matrix(D_i)
    idx <- starts[i]:ends[i]

    if (!all(dim(D_i) == c(length(idx), length(idx)))) {
      stop(
        sprintf(
          "Reference survey %d: dimension mismatch. D_i is %d x %d, but block expects %d x %d.",
          i, nrow(D_i), ncol(D_i), length(idx), length(idx)
        ),
        call. = FALSE
      )
    }

    D[idx, idx] <- D_i
  }

  D
}

#' Extract NA action from a pw_fit object
#'
#' Returns the \code{na.action} component recorded during the build step.
#'
#' @param object An object of class \code{"pw_fit"} returned by \code{\link{est_pw}}.
#' @param ... Additional arguments (not used).
#'
#' @return The \code{na.action} object recorded by \code{\link{est_pw}} during
#'   the build step: an integer vector of the nonprobability-sample rows omitted
#'   because of missing participation model variables (of class \code{"omit"} or
#'   \code{"exclude"}), or \code{NULL} if no rows were omitted.
#'
#' @method na.action pw_fit
#' @importFrom stats na.action
#' @export
na.action.pw_fit <- function(object, ...) {
  object$internal$na$na_action_obj
}


#' Extract NA action from a pwmean object
#'
#' Returns the \code{na.action} component recorded during estimation,
#' mimicking \code{\link[stats]{na.action}} behavior for fitted model objects.
#'
#' @param object An object of class \code{"pwmean"} returned by \code{\link{pwmean}}.
#' @param ... Additional arguments (not used).
#'
#' @return The \code{na.action} object recorded by \code{\link{pwmean}}: an
#'   integer vector of rows omitted because of missing outcome or domain values
#'   (of class \code{"omit"} or \code{"exclude"}), or \code{NULL} if no rows were
#'   omitted.
#'
#' @method na.action pwmean
#' @importFrom stats na.action
#' @export
na.action.pwmean <- function(object, ...) {
  object$na.action
}

#' Build the estimates summary data frame
#'
#' Constructs the `estimates` data frame that is stored in a `pwmean` object.
#' Each row corresponds to one estimate. The 95% confidence intervals are
#' computed using the 0.975 normal quantile (approximately 1.96).
#'
#' @param labels Character vector of domain labels, one entry per domain.
#' @param mean_unw Numeric vector of unweighted (naive) domain means.
#' @param se_unw Numeric vector of standard errors for the unweighted means.
#' @param mean_adj Numeric vector of pseudo-weighted (adjusted) domain means.
#' @param se_adj Numeric vector of standard errors for the adjusted means.
#'
#' @return A data frame with one row per domain and columns `domain`,
#'   `unweighted_mean`, `unweighted_se`, `unweighted_lower`,
#'   `unweighted_upper`, `adjusted_mean`, `adjusted_se`, `adjusted_lower`,
#'   and `adjusted_upper`.
#'
#' @keywords internal
build_domains_df <- function(labels, mean_unw, se_unw, mean_adj, se_adj) {
  z975 <- stats::qnorm(0.975)
  df <- data.frame(
    domain          = labels,
    unweighted_mean = mean_unw,
    unweighted_se   = se_unw,
    adjusted_mean   = mean_adj,
    adjusted_se     = se_adj,
    stringsAsFactors = FALSE
  )
  df$unweighted_lower <- df$unweighted_mean - z975 * df$unweighted_se
  df$unweighted_upper <- df$unweighted_mean + z975 * df$unweighted_se
  df$adjusted_lower   <- df$adjusted_mean   - z975 * df$adjusted_se
  df$adjusted_upper   <- df$adjusted_mean   + z975 * df$adjusted_se
  df[, c(
    "domain",
    "unweighted_mean", "unweighted_se", "unweighted_lower", "unweighted_upper",
    "adjusted_mean",   "adjusted_se",   "adjusted_lower",   "adjusted_upper"
  )]
}


#' Assemble the final output object
#'
#' Combines the pseudo-weighted estimates and naive (unweighted) estimates
#' into a plain list to be returned by `pwmean()`.
#'
#' @param build A `pw_fit` object returned by the build step.
#' @param est A list returned by `dispatch_estimator()`, with components
#'   `type`, `labels`, and `estimates`.
#' @param naive A list returned by `naive_mean()`, with the same structure
#'   as `est`.
#' @param na_info NA-handling information returned by `process_na_yz()`.
#'
#' @return A list with components `method`, `estimates`, and `na.action`.
#'   The S3 class is assigned by the calling function `pwmean()`.
#'
#' @keywords internal
assemble_output <- function(build, est, naive, na_info) {

  #----------------------------------#
  # Step 1: basic checks
  #----------------------------------#
  if (is.null(est$type) || is.null(naive$type)) {
    stop("`est` and `naive` must both contain a `type` field.", call. = FALSE)
  }

  if (!identical(est$type, naive$type)) {
    stop("`est$type` and `naive$type` do not match.", call. = FALSE)
  }

  #----------------------------------#
  # Step 2: single-result case
  #----------------------------------#
  if (identical(est$type, "single")) {

    estimates_df <- build_domains_df(
      labels   = naive$labels,
      mean_unw = naive$estimates$mean,
      se_unw   = sqrt(naive$estimates$variance),
      mean_adj = est$estimates$mean,
      se_adj   = sqrt(est$estimates$variance)
    )

    return(list(method = build$method, estimates = estimates_df, na.action = na_info$na_action))
  }

  #----------------------------------#
  # Step 3: multi-result case
  #----------------------------------#
  if (identical(est$type, "multi")) {

    if (length(est$labels) != length(est$estimates)) {
      stop("Length mismatch between `est$labels` and `est$estimates`.", call. = FALSE)
    }

    if (length(naive$labels) != length(naive$estimates)) {
      stop("Length mismatch between `naive$labels` and `naive$estimates`.", call. = FALSE)
    }

    if (!identical(est$labels, naive$labels)) {
      stop("`est$labels` and `naive$labels` do not match.", call. = FALSE)
    }

    estimates_df <- build_domains_df(
      labels   = est$labels,
      mean_unw = vapply(naive$estimates, function(x) x$mean,            numeric(1)),
      se_unw   = vapply(naive$estimates, function(x) sqrt(x$variance),  numeric(1)),
      mean_adj = vapply(est$estimates,   function(x) x$mean,            numeric(1)),
      se_adj   = vapply(est$estimates,   function(x) sqrt(x$variance),  numeric(1))
    )

    return(list(method = build$method, estimates = estimates_df, na.action = na_info$na_action))
  }

  #----------------------------------#
  # Step 4: unsupported type
  #----------------------------------#
  stop("Unsupported result type in `est$type` / `naive$type`.", call. = FALSE)
}

# helpers ----
#
# Synthetic inputs with known analytical results.
#
# z975 = qnorm(0.975) is used for 95% CI half-width.
#
# Single-domain inputs:
#   naive: mean = 3.5, variance = 1.75  -> se = sqrt(1.75)
#   est  : mean = 4.0, variance = 0.25  -> se = 0.5
#
# Multi-domain inputs (labels = c("A", "B")):
#   naive A: mean = 2.5, variance = 2.25  -> se = 1.5
#   naive B: mean = 4.5, variance = 6.25  -> se = 2.5
#   est   A: mean = 3.0, variance = 1.00  -> se = 1.0
#   est   B: mean = 5.0, variance = 4.00  -> se = 2.0

z975 <- stats::qnorm(0.975)

make_build <- function(method = "alp") {
  list(method = method)
}

make_na_info <- function(action_class = "omit") {
  list(na_action = structure(integer(0), class = action_class))
}

make_single_est <- function(mean = 4.0, variance = 0.25) {
  list(type = "single", labels = "Overall",
       estimates = list(mean = mean, variance = variance))
}

make_single_naive <- function(mean = 3.5, variance = 1.75) {
  list(type = "single", labels = "Overall",
       estimates = list(mean = mean, variance = variance))
}

make_multi_est <- function() {
  list(
    type   = "multi",
    labels = c("A", "B"),
    estimates = list(
      list(mean = 3.0, variance = 1.0),
      list(mean = 5.0, variance = 4.0)
    )
  )
}

make_multi_naive <- function() {
  list(
    type   = "multi",
    labels = c("A", "B"),
    estimates = list(
      list(mean = 2.5, variance = 2.25),
      list(mean = 4.5, variance = 6.25)
    )
  )
}


# build_domains_df: output structure ----

test_that("build_domains_df: returns a data.frame", {
  res <- build_domains_df(
    labels   = "Overall",
    mean_unw = 3.5, se_unw = sqrt(1.75),
    mean_adj = 4.0, se_adj = 0.5
  )
  expect_s3_class(res, "data.frame")
})

test_that("build_domains_df: column names and order match the spec", {
  res <- build_domains_df(
    labels   = "Overall",
    mean_unw = 3.5, se_unw = sqrt(1.75),
    mean_adj = 4.0, se_adj = 0.5
  )
  expect_equal(
    names(res),
    c("domain",
      "unweighted_mean", "unweighted_se", "unweighted_lower", "unweighted_upper",
      "adjusted_mean",   "adjusted_se",   "adjusted_lower",   "adjusted_upper")
  )
})

test_that("build_domains_df: nrow equals length(labels)", {
  res <- build_domains_df(
    labels   = c("A", "B"),
    mean_unw = c(2.5, 4.5), se_unw = c(1.5, 2.5),
    mean_adj = c(3.0, 5.0), se_adj = c(1.0, 2.0)
  )
  expect_equal(nrow(res), 2L)
})

test_that("build_domains_df: domain column is character (not factor)", {
  res <- build_domains_df(
    labels   = c("A", "B"),
    mean_unw = c(2.5, 4.5), se_unw = c(1.5, 2.5),
    mean_adj = c(3.0, 5.0), se_adj = c(1.0, 2.0)
  )
  expect_type(res$domain, "character")
})


# build_domains_df: value correctness ----

test_that("build_domains_df: mean and se columns equal the inputs", {
  res <- build_domains_df(
    labels   = "Overall",
    mean_unw = 3.5, se_unw = sqrt(1.75),
    mean_adj = 4.0, se_adj = 0.5
  )
  expect_equal(res$unweighted_mean, 3.5)
  expect_equal(res$unweighted_se,   sqrt(1.75))
  expect_equal(res$adjusted_mean,   4.0)
  expect_equal(res$adjusted_se,     0.5)
})

test_that("build_domains_df: unweighted CI equals mean +/- qnorm(0.975) * se", {
  res <- build_domains_df(
    labels   = "Overall",
    mean_unw = 3.5, se_unw = sqrt(1.75),
    mean_adj = 4.0, se_adj = 0.5
  )
  expect_equal(res$unweighted_lower, 3.5 - z975 * sqrt(1.75))
  expect_equal(res$unweighted_upper, 3.5 + z975 * sqrt(1.75))
})

test_that("build_domains_df: adjusted CI equals mean +/- qnorm(0.975) * se", {
  res <- build_domains_df(
    labels   = "Overall",
    mean_unw = 3.5, se_unw = sqrt(1.75),
    mean_adj = 4.0, se_adj = 0.5
  )
  expect_equal(res$adjusted_lower, 4.0 - z975 * 0.5)
  expect_equal(res$adjusted_upper, 4.0 + z975 * 0.5)
})

test_that("build_domains_df: CIs vectorize correctly over multiple rows", {
  res <- build_domains_df(
    labels   = c("A", "B"),
    mean_unw = c(2.5, 4.5), se_unw = c(1.5, 2.5),
    mean_adj = c(3.0, 5.0), se_adj = c(1.0, 2.0)
  )
  expect_equal(res$unweighted_lower, c(2.5, 4.5) - z975 * c(1.5, 2.5))
  expect_equal(res$unweighted_upper, c(2.5, 4.5) + z975 * c(1.5, 2.5))
  expect_equal(res$adjusted_lower,   c(3.0, 5.0) - z975 * c(1.0, 2.0))
  expect_equal(res$adjusted_upper,   c(3.0, 5.0) + z975 * c(1.0, 2.0))
})

test_that("build_domains_df: NA in se propagates to lower and upper", {
  res <- build_domains_df(
    labels   = "Overall",
    mean_unw = 3.5, se_unw = NA_real_,
    mean_adj = 4.0, se_adj = 0.5
  )
  expect_true(is.na(res$unweighted_lower))
  expect_true(is.na(res$unweighted_upper))
  expect_false(is.na(res$adjusted_lower))
  expect_false(is.na(res$adjusted_upper))
})

test_that("build_domains_df: domain column preserves the label vector", {
  res <- build_domains_df(
    labels   = c("A", "B"),
    mean_unw = c(2.5, 4.5), se_unw = c(1.5, 2.5),
    mean_adj = c(3.0, 5.0), se_adj = c(1.0, 2.0)
  )
  expect_equal(res$domain, c("A", "B"))
})


# assemble_output: single-result case ----

test_that("assemble_output: single-result returns list with method, estimates, na.action", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_single_est(),
    naive   = make_single_naive(),
    na_info = make_na_info("omit")
  )
  expect_named(res, c("method", "estimates", "na.action"))
})

test_that("assemble_output: single-result method propagates from build", {
  res <- assemble_output(
    build   = make_build("calibration"),
    est     = make_single_est(),
    naive   = make_single_naive(),
    na_info = make_na_info("omit")
  )
  expect_equal(res$method, "calibration")
})

test_that("assemble_output: single-result na.action propagates from na_info", {
  na_info <- make_na_info("exclude")
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_single_est(),
    naive   = make_single_naive(),
    na_info = na_info
  )
  expect_identical(res$na.action, na_info$na_action)
})

test_that("assemble_output: single-result estimates is a one-row data.frame", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_single_est(),
    naive   = make_single_naive(),
    na_info = make_na_info("omit")
  )
  expect_s3_class(res$estimates, "data.frame")
  expect_equal(nrow(res$estimates), 1L)
})

test_that("assemble_output: single-result domain label equals naive$labels", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_single_est(),
    naive   = make_single_naive(),
    na_info = make_na_info("omit")
  )
  expect_equal(res$estimates$domain, "Overall")
})

test_that("assemble_output: single-result unweighted mean and se come from naive", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_single_est(),
    naive   = make_single_naive(mean = 3.5, variance = 1.75),
    na_info = make_na_info("omit")
  )
  expect_equal(res$estimates$unweighted_mean, 3.5)
  expect_equal(res$estimates$unweighted_se,   sqrt(1.75))
})

test_that("assemble_output: single-result adjusted mean and se come from est", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_single_est(mean = 4.0, variance = 0.25),
    naive   = make_single_naive(),
    na_info = make_na_info("omit")
  )
  expect_equal(res$estimates$adjusted_mean, 4.0)
  expect_equal(res$estimates$adjusted_se,   0.5)
})

test_that("assemble_output: single-result CIs equal mean +/- qnorm(0.975) * se", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_single_est(mean = 4.0, variance = 0.25),
    naive   = make_single_naive(mean = 3.5, variance = 1.75),
    na_info = make_na_info("omit")
  )
  expect_equal(res$estimates$unweighted_lower, 3.5 - z975 * sqrt(1.75))
  expect_equal(res$estimates$unweighted_upper, 3.5 + z975 * sqrt(1.75))
  expect_equal(res$estimates$adjusted_lower,   4.0 - z975 * 0.5)
  expect_equal(res$estimates$adjusted_upper,   4.0 + z975 * 0.5)
})


# assemble_output: multi-result case ----

test_that("assemble_output: multi-result returns list with method, estimates, na.action", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_multi_est(),
    naive   = make_multi_naive(),
    na_info = make_na_info("omit")
  )
  expect_named(res, c("method", "estimates", "na.action"))
})

test_that("assemble_output: multi-result estimates has one row per domain", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_multi_est(),
    naive   = make_multi_naive(),
    na_info = make_na_info("omit")
  )
  expect_equal(nrow(res$estimates), 2L)
})

test_that("assemble_output: multi-result domain column equals est$labels", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_multi_est(),
    naive   = make_multi_naive(),
    na_info = make_na_info("omit")
  )
  expect_equal(res$estimates$domain, c("A", "B"))
})

test_that("assemble_output: multi-result unweighted columns come from naive estimates", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_multi_est(),
    naive   = make_multi_naive(),
    na_info = make_na_info("omit")
  )
  expect_equal(res$estimates$unweighted_mean, c(2.5, 4.5))
  expect_equal(res$estimates$unweighted_se,   c(1.5, 2.5))
})

test_that("assemble_output: multi-result adjusted columns come from est estimates", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_multi_est(),
    naive   = make_multi_naive(),
    na_info = make_na_info("omit")
  )
  expect_equal(res$estimates$adjusted_mean, c(3.0, 5.0))
  expect_equal(res$estimates$adjusted_se,   c(1.0, 2.0))
})

test_that("assemble_output: multi-result CIs equal mean +/- qnorm(0.975) * se", {
  res <- assemble_output(
    build   = make_build("alp"),
    est     = make_multi_est(),
    naive   = make_multi_naive(),
    na_info = make_na_info("omit")
  )
  expect_equal(res$estimates$unweighted_lower, c(2.5, 4.5) - z975 * c(1.5, 2.5))
  expect_equal(res$estimates$unweighted_upper, c(2.5, 4.5) + z975 * c(1.5, 2.5))
  expect_equal(res$estimates$adjusted_lower,   c(3.0, 5.0) - z975 * c(1.0, 2.0))
  expect_equal(res$estimates$adjusted_upper,   c(3.0, 5.0) + z975 * c(1.0, 2.0))
})

test_that("assemble_output: multi-result method propagates from build", {
  res <- assemble_output(
    build   = make_build("multi"),
    est     = make_multi_est(),
    naive   = make_multi_naive(),
    na_info = make_na_info("omit")
  )
  expect_equal(res$method, "multi")
})


# assemble_output: input validation ----

test_that("assemble_output: errors when est$type is NULL", {
  bad_est        <- make_single_est()
  bad_est$type   <- NULL
  expect_error(
    assemble_output(
      build = make_build(), est = bad_est, naive = make_single_naive(),
      na_info = make_na_info()
    ),
    "`est` and `naive` must both contain a `type` field.",
    fixed = TRUE
  )
})

test_that("assemble_output: errors when naive$type is NULL", {
  bad_naive       <- make_single_naive()
  bad_naive$type  <- NULL
  expect_error(
    assemble_output(
      build = make_build(), est = make_single_est(), naive = bad_naive,
      na_info = make_na_info()
    ),
    "`est` and `naive` must both contain a `type` field.",
    fixed = TRUE
  )
})

test_that("assemble_output: errors when est$type and naive$type differ", {
  expect_error(
    assemble_output(
      build = make_build(), est = make_single_est(), naive = make_multi_naive(),
      na_info = make_na_info()
    ),
    "`est$type` and `naive$type` do not match.",
    fixed = TRUE
  )
})

test_that("assemble_output: errors when est$labels and est$estimates differ in length", {
  bad_est            <- make_multi_est()
  bad_est$estimates  <- bad_est$estimates[1]   # length 1, labels still length 2
  expect_error(
    assemble_output(
      build = make_build(), est = bad_est, naive = make_multi_naive(),
      na_info = make_na_info()
    ),
    "Length mismatch between `est$labels` and `est$estimates`.",
    fixed = TRUE
  )
})

test_that("assemble_output: errors when naive$labels and naive$estimates differ in length", {
  bad_naive             <- make_multi_naive()
  bad_naive$estimates   <- bad_naive$estimates[1]
  expect_error(
    assemble_output(
      build = make_build(), est = make_multi_est(), naive = bad_naive,
      na_info = make_na_info()
    ),
    "Length mismatch between `naive$labels` and `naive$estimates`.",
    fixed = TRUE
  )
})

test_that("assemble_output: errors when multi est$labels and naive$labels do not match", {
  bad_naive          <- make_multi_naive()
  bad_naive$labels   <- c("X", "Y")
  expect_error(
    assemble_output(
      build = make_build(), est = make_multi_est(), naive = bad_naive,
      na_info = make_na_info()
    ),
    "`est$labels` and `naive$labels` do not match.",
    fixed = TRUE
  )
})

test_that("assemble_output: errors on unsupported type value", {
  bad_est         <- make_single_est()
  bad_est$type    <- "weird"
  bad_naive       <- make_single_naive()
  bad_naive$type  <- "weird"
  expect_error(
    assemble_output(
      build = make_build(), est = bad_est, naive = bad_naive,
      na_info = make_na_info()
    ),
    "Unsupported result type in `est$type` / `naive$type`.",
    fixed = TRUE
  )
})

#' Probability Reference Sample 1 with Bootstrap Replicate Weights (sp1_bootstrap)
#'
#' A replicate-weight version of \code{sp1}, including the main survey weight
#' and 500 bootstrap replicate weights. It is provided to illustrate
#' design-based variance estimation with \code{\link[survey]{svrepdesign}}.
#' The original primary sampling unit and stratum identifiers are not included.
#'
#' @format A data frame with 3494 rows and 512 columns. The first 12 columns
#'   are the substantive survey variables from \code{sp1} (\code{agecat},
#'   \code{marital}, \code{race}, \code{education}, \code{employment},
#'   \code{smoking}, \code{comorbidity}, \code{psa_level}, \code{BMI},
#'   \code{diabetes}, \code{pros_enlarged}, \code{wts_sp1}). The remaining
#'   500 columns are bootstrap replicate weights named \code{bw1} through
#'   \code{bw500} (numeric).
#'
#' @details
#' The bootstrap replicate weights were constructed from the original
#' stratified cluster design of \code{sp1}, using the Rao-Wu rescaling
#' bootstrap method. A total of \code{R = 500} replicates were
#' produced with \code{seed = 2026}.
#'
#' The variables \code{psu_sp1} and \code{strata_sp1} are not included in this
#' dataset because they are not needed when using replicate weights for
#' variance estimation. The main survey weight \code{wts_sp1} and the replicate
#' weight columns \code{bw1}--\code{bw500} are sufficient for constructing a
#' replicate-weight survey design object via \code{survey::svrepdesign()}.
#'
#' @source
#' Derived from \code{sp1} (National Health and Nutrition Examination Survey,
#' NHANES, 1999--2010 cycles), with bootstrap replicate weights added by the
#' package authors using the Rao-Wu rescaling bootstrap.
#'
#' @usage data(sp1_bootstrap)
#'
#' @examples
#' data(sp1_bootstrap)
#'
#' # Example: create replicate-weight survey design object
#' des_boot <- survey::svrepdesign(
#'   data    = sp1_bootstrap,
#'   weights = ~wts_sp1,
#'   repweights = "bw[0-9]+",
#'   type    = "bootstrap",
#'   combined.weights = FALSE
#' )
#'
#' summary(des_boot)
"sp1_bootstrap"

#' Validate IPWM build-stage inputs
#'
#' Performs structural validation of all arguments passed to
#' `est_pw()` before any computation begins. Checks cover
#' `verbose`, `sc_wname`, the `data` list structure,
#' the survey design objects, reference-sample row counts, `method`
#' consistency with the number of reference samples, and
#' `p_formula` variable availability.
#'
#' Specifically, this function checks that:
#'
#' - `verbose` is a single TRUE/FALSE value;
#' - `sc_wname` is a single non-empty character string that does
#'   not already exist as a column in `sc`;
#' - `data` is a list of the form `list(sc, sp1_design, sp2_design, ...)`;
#' - the first element of `data` is a non-empty data frame for
#'   `sc` with no duplicated column names;
#' - all remaining elements of `data` are valid survey design
#'   objects, either `survey.design2` or `svyrep.design`;
#' - each reference sample has at least one row;
#' - lonely PSU problems are detected for standard survey design objects;
#' - `method`, if supplied, is one of `"alp"`, `"clw"`,
#'   `"multi"`, `"cali"`, or `"calibration"`;
#' - `method = "multi"` is only used with multiple reference samples;
#' - one-reference methods are only used with a single reference sample;
#' - in the one-reference case, `p_formula` is either NULL or a
#'   one-sided formula; a two-sided formula raises an error; a warning
#'   is issued if `sp_order` differs from its default (`"size"`) or if
#'   `precali` is FALSE, since both are ignored in the one-reference case;
#' - in the multi-reference case, `sp_order` must be either
#'   `"size"` or `"given"`;
#' - in the multi-reference case, `precali` must be a single
#'   TRUE or FALSE value;
#' - in the multi-reference case, `p_formula`, if supplied, is a
#'   list of one-sided formulas with one formula per reference sample;
#'   any two-sided formula in the list raises an error;
#' - all variables used in each participation model formula are present
#'   in both `sc` and the corresponding reference sample.
#'
#' This function does not check outcome variables or domain variables. Those
#' are validated later at the estimation stage by
#' `check_ipwm_inputs_estimate()`.
#'
#' `method` is expected to be normalized early (`tolower` +
#' `trimws`, with `"cali"` already converted to `"calibration"`)
#' by the caller (`est_pw()`).
#'
#' @param data A list of input data objects. The first element must be a
#'   non-empty data frame with unique column names (`sc`). All remaining
#'   elements must be `"survey.design2"` or `"svyrep.design"`
#'   objects with at least one row in their analysis data.
#' @param p_formula Participation model formula. Must be one-sided (no
#'   left-hand side); a two-sided formula raises an error. For a single
#'   reference sample, supply a formula or NULL. For multiple reference
#'   samples, supply a list of one-sided formulas with one formula per
#'   reference sample, or NULL. When non-NULL, all variables in each formula
#'   must be present in both `sc` and the corresponding reference sample data.
#' @param method Pseudo-weighting method, or NULL. Expected to be
#'   pre-normalised by the caller. Supported values are `"alp"`,
#'   `"clw"`, `"multi"`, `"cali"`, and `"calibration"`.
#'   Must be consistent with the number of reference samples: `"multi"`
#'   requires more than one reference, while `"alp"`, `"clw"`, and
#'   `"calibration"` require exactly one.
#' @param sp_order Reference-sample ordering rule for the multi-reference case.
#'   Must be either `"size"` or `"given"` when multiple reference
#'   samples are provided. Ignored in the one-reference case; a warning is
#'   issued if a value other than `"size"` (the default) is supplied.
#' @param precali Single logical value indicating whether cumulative
#'   precalibration is used in the multi-reference case. Required when
#'   multiple reference samples are provided. Ignored in the one-reference
#'   case; a warning is issued if FALSE is supplied.
#' @param sc_wname Single non-empty character string giving the intended name
#'   of the pseudo-weight column. An error is raised if the name already
#'   exists in `sc`.
#' @param verbose Single TRUE/FALSE controlling diagnostic output. Validated
#'   before any data checks are performed.
#'
#' @return Invisibly returns TRUE if all checks pass. Otherwise, the
#'   function stops with an informative error message.
#'
#' @keywords internal
#'
#' @noRd
check_ipwm_inputs_build <- function(data, p_formula, method,
                                    sp_order = NULL, precali = NULL,
                                    sc_wname = "pseudo_wts",
                                    verbose  = FALSE) {

  #--------------------------------------------------#
  # verbose
  #--------------------------------------------------#
  if (!is.logical(verbose) || length(verbose) != 1L || is.na(verbose)) {
    stop("`verbose` must be a single TRUE/FALSE value.", call. = FALSE)
  }

  #--------------------------------------------------#
  # sc_wname
  #--------------------------------------------------#
  if (!is.character(sc_wname) || length(sc_wname) != 1L ||
      is.na(sc_wname) || !nzchar(sc_wname)) {
    stop("`sc_wname` must be a single non-empty character string.", call. = FALSE)
  }

  #--------------------------------------------------#
  # data: basic structure
  #--------------------------------------------------#
  if (!is.list(data) || length(data) < 2L) {
    stop("`data` must be a list like list(sc, sp1.des, sp2.des, ...).",
         call. = FALSE)
  }

  sc <- data[[1]]
  sp_des <- data[-1]
  n_ref <- length(sp_des)

  if (!is.data.frame(sc)) {
    stop("The first element of `data` must be a data.frame for `sc`.",
         call. = FALSE)
  }

  if (nrow(sc) == 0L) {
    stop("`sc` has zero rows.", call. = FALSE)
  }

  if (anyDuplicated(names(sc))) {
    stop("`sc` has duplicated column names.", call. = FALSE)
  }

  if (sc_wname %in% names(sc)) {
    stop(
      sprintf("Column '%s' already exists in `sc`. Choose a different `sc_wname`.",
              sc_wname),
      call. = FALSE
    )
  }

  ok_des <- vapply(
    sp_des,
    function(x) inherits(x, c("survey.design2", "svyrep.design")),
    logical(1)
  )

  if (!all(ok_des)) {
    bad <- which(!ok_des)
    stop(
      sprintf(
        "Elements %s of `data` are not valid survey design objects.",
        paste(bad + 1L, collapse = ", ")
      ),
      call. = FALSE
    )
  }
  # stop when lonely psu + getOption("survey.lonely.psu") = "fail"
  for (i in seq_along(sp_des)) {
    if (inherits(sp_des[[i]], "survey.design2")) {
      check_lonely_psu(
        sp_des[[i]],
        name = paste0("sp_des[[", i, "]]")
      )
    }
  }

  # extract analysis data from survey design objects
  sp <- lapply(sp_des, extract_analysis_data)

  # check for zero-row reference samples
  for (i in seq_along(sp)) {
    if (nrow(sp[[i]]) == 0L) {
      stop(sprintf("Reference sample %d has zero rows.", i), call. = FALSE)
    }
  }

  #--------------------------------------------------#
  # method
  #--------------------------------------------------#
  valid_methods <- c("alp", "clw", "multi", "cali", "calibration")

  if (!is.null(method)) {

    if (!is.character(method) || length(method) != 1L || is.na(method)) {
      stop("`method` must be a single character string or NULL.",
           call. = FALSE)
    }

    method_lc <- tolower(trimws(method))

    if (!(method_lc %in% valid_methods)) {
      stop(
        sprintf(
          paste0(
            "Invalid method '%s'.\n",
            "For one-reference settings, valid methods are: ",
            "'alp', 'clw', 'calibration' (or 'cali').\n",
            "For multi-reference settings, the valid method is: 'multi'."
          ),
          method
        ),
        call. = FALSE
      )
    }

    #--------------------------------------------------#
    # consistency check with number of references
    #--------------------------------------------------#
    if (n_ref == 1 && method_lc == "multi") {
      stop(
        "`method = 'multi'` is only allowed when multiple reference samples are provided.",
        call. = FALSE
      )
    }

    if (n_ref > 1 && method_lc %in% c("alp", "clw", "calibration")) {
      stop(
        paste0(
          "method = '", method, "' is only valid for a single reference sample. ",
          "Use `method = 'multi'` when multiple reference samples are provided."
        ),
        call. = FALSE
      )
    }
  }

  #--------------------------------------------------#
  # one-reference case
  #--------------------------------------------------#
  if (n_ref == 1L) {

    if (!is.null(sp_order) && sp_order != "size") {
      warning("`sp_order` is ignored in the one-reference case.", call. = FALSE)
    }

    if (!is.null(precali) && !isTRUE(precali)) {
      warning("`precali` is ignored in the one-reference case.", call. = FALSE)
    }

    if (!is.null(p_formula)) {
      if (!inherits(p_formula, "formula")) {
        stop("For one-reference case, `p_formula` must be a formula or NULL.",
             call. = FALSE)
      }

      if (length(p_formula) == 3L) {
        stop(
          "`p_formula` must be a one-sided formula (e.g., ~ x + y), not two-sided.",
          call. = FALSE
        )
      }

      check_formula_vars_exist(
        fml   = p_formula,
        sc    = sc,
        sp    = sp[[1]],
        label = "`p_formula`"
      )
    }

    return(invisible(TRUE))
  }

  #--------------------------------------------------#
  # multi-reference case
  #--------------------------------------------------#
  if (n_ref >= 2L) {

    if (!is.character(sp_order) || length(sp_order) != 1L || is.na(sp_order) ||
        !(sp_order %in% c("size", "given"))) {
      stop("`sp_order` must be one of 'size' or 'given'.",
           call. = FALSE)
    }

    if (!is.logical(precali) || length(precali) != 1L || is.na(precali)) {
      stop("`precali` must be a single TRUE/FALSE value.",
           call. = FALSE)
    }

    if (!is.null(p_formula)) {
      if (!is.list(p_formula) || length(p_formula) != n_ref) {
        stop(
          sprintf(
            "For multi-reference case, `p_formula` must be a list of %d formulas.",
            n_ref
          ),
          call. = FALSE
        )
      }

      ok_formula <- vapply(
        p_formula,
        function(x) inherits(x, "formula"),
        logical(1)
      )

      if (!all(ok_formula)) {
        stop("All elements of `p_formula` must be formulas.",
             call. = FALSE)
      }

      two_sided <- vapply(p_formula, function(x) length(x) == 3L, logical(1))
      if (any(two_sided)) {
        stop(
          "All formulas in `p_formula` must be one-sided (e.g., ~ x + y), not two-sided.",
          call. = FALSE
        )
      }

      for (i in seq_len(n_ref)) {
        check_formula_vars_exist(
          fml   = p_formula[[i]],
          sc    = sc,
          sp    = sp[[i]],
          label = sprintf("`p_formula[[%d]]`", i)
        )
      }
    }

    return(invisible(TRUE))
  }
}

# Simulation study based on Landsman et al. (2026), Statistics in Medicine,
# doi:10.1002/sim.70403.
#
# This script reproduces selected continuous-outcome results from the paper:
# Table 1, Case 0, for the one-reference ALP, CLW, and calibration methods;
# and Table 2, Case PC.1, for the multi-reference calibration method.
#
# The reported metrics follow the paper: RB(%), v(*100), MSE(*100), SER, and CP.

library(nonprobsampling)
library(sampling)
library(survey)
library(knitr)

# Temporarily set display options for reproducible printed output.
# These options are restored at the end of the script.
old_options <- options(
  scipen = 999,
  digits = 4
)

# -------- Simulation rounds and seeds ------------
# Reduce n_sim for a quick trial run; set to 4000 to reproduce the paper.
set.seed(123456)
n_sim <- 4000
seeds_sc <- sample.int(1e6, n_sim)
seeds_sp <- sample.int(1e6, n_sim)

# -------- Helper functions ------------

# True-weight estimator using known participation probabilities.
# sc must have columns 'y' and 'wt_sc' (= 1 / pi_c).
# Returns c(mean, se) under Poisson sampling.
TW.sc <- function(sc) {
  w <- sc$wt_sc
  y <- sc$y
  m <- sum(w * y) / sum(w)
  v <- sum(w * (w - 1) * (y - m)^2) / sum(w)^2
  c(m, sqrt(v))
}

# Simulation evaluation metrics.
#
# mu:      numeric vector of point estimates across simulation replications,
#          i.e., mu_hat^(1), ..., mu_hat^(B).
# se:      numeric vector of estimated standard errors across replications.
# mu_true: true population mean.
#
# Returns:
#   RB(%)      relative bias: 100 * mean((mu_hat - mu_true) / mu_true).
#   v(*100)    average estimated analytic variance * 100: mean(se^2) * 100.
#   MSE(*100)  mean squared error * 100: mean((mu_hat - mu_true)^2) * 100.
#   SER        ratio of estimated SE to simulated SE: mean(se) / sd(mu).
#   CP         empirical 95% Wald confidence interval coverage probability.

evaluation <- function(mu, se, mu_true) {
  RB  <- mean((mu - mu_true) / mu_true)
  MSE <- mean((mu - mu_true)^2)
  ESE <- mean(se)
  SSE <- sd(mu)
  SER <- ESE / SSE
  CP  <- mean(mu_true >= mu - 1.96 * se & mu_true <= mu + 1.96 * se)
  return(c(RB * 100, v = mean(se^2) * 100, MSE * 100, SER, CP))
}

# -------- Finite population ------------
N <- 500000

v1 <- rbinom(N, size = 1, prob = 0.5)
v2 <- runif(N, min = 0, max = 2)
v3 <- rexp(N, rate = 1)
v4 <- rchisq(N, df = 4)

x1 <- v1
x2 <- v2 + 0.3 * x1
x3 <- v3 + 0.2 * (x1 + x2)
x4 <- v4 + 0.1 * (x1 + x2 + x3)

mu  <- -x1 - x2 + x3 + x4
y   <- mu + rnorm(N)
mu_true <- mean(y)

# -------- Sampling designs ------------

# sp1: Poisson sampling (n ≈ 12500)
np1 <- 12500
a <- x3 + 0.03 * y
rng <- range(a)
cnst_sp1 <- (rng[2] - 20 * rng[1]) / 19
q <- cnst_sp1 + a
pi_p1 <- np1 * q / sum(q)
di1   <- 1 / pi_p1

# sp2: randomized systematic PPS (n = 25000)
np2    <- 25000
const2 <- 0.05
z      <- const2 + x2
pi_p2  <- np2 * z / sum(z)
di2    <- 1 / pi_p2

# Nonprobability sample sc: Poisson sampling (n ≈ 2500)
nc  <- 2500
eta <- 0.18 * x1 + 0.18 * x2 - 0.27 * x3 - 0.27 * x4
beta0 <- log(nc / sum(exp(eta)))
pi_c  <- exp(eta + beta0)

# Finite population frame
fp <- data.frame(
  x1     = x1, x2 = x2, x3 = x3, x4 = x4,
  y      = y,
  pi_sp1 = pi_p1, wt_sp1 = di1,
  pi_sp2 = pi_p2, wt_sp2 = di2,
  pi_c   = pi_c,  wt_sc  = 1 / pi_c
)


########################################
# Simulation 1: sp1 (Poisson) as reference
########################################

result_sp1 <- matrix(NA, nrow = n_sim, ncol = 10)

for (i in seq_len(n_sim)) {

  # nonprobability sample
  set.seed(seeds_sc[i])
  sc <- fp[rbinom(N, 1, pi_c) == 1,
           c("x1", "x2", "x3", "x4", "y", "wt_sc")]

  # probability reference sample sp1 (Poisson)
  set.seed(seeds_sp[i])
  sp1 <- fp[rbinom(N, 1, pi_p1) == 1,
            c("x1", "x2", "x3", "x4", "wt_sp1")]
  sp1$pi_sp1 <- 1 / sp1$wt_sp1

  des_sp1 <- svydesign(ids = ~1, probs = ~pi_sp1, data = sp1)

  # true-weight estimator
  result_tw <- TW.sc(sc)

  # pseudo-weight estimation: ALP, CLW, Calibration
  fit_alp  <- est_pw(data = list(sc, des_sp1), method = "alp")
  fit_clw  <- est_pw(data = list(sc, des_sp1), method = "clw")
  fit_cali <- est_pw(data = list(sc, des_sp1), method = "calibration")

  res_alp  <- pwmean(fit_alp,  y = "y")
  res_clw  <- pwmean(fit_clw,  y = "y")
  res_cali <- pwmean(fit_cali, y = "y")

  # naive (from any fit)
  result_sp1[i, 1:2]  <- c(res_alp$estimates$unweighted_mean,
                            res_alp$estimates$unweighted_se)
  # true weight
  result_sp1[i, 3:4]  <- result_tw
  # ALP
  result_sp1[i, 5:6]  <- c(res_alp$estimates$adjusted_mean,
                            res_alp$estimates$adjusted_se)
  # CLW
  result_sp1[i, 7:8]  <- c(res_clw$estimates$adjusted_mean,
                            res_clw$estimates$adjusted_se)
  # Calibration
  result_sp1[i, 9:10] <- c(res_cali$estimates$adjusted_mean,
                            res_cali$estimates$adjusted_se)

  if (i %% 100 == 0) cat(i, " ")
}

table_sp1 <- matrix(NA, nrow = 5, ncol = 5)
colnames(table_sp1) <- c("RB(%)", "v(*100)", "MSE(*100)", "SER", "CP")
rownames(table_sp1) <- c("Naive", "True weight", "ALP", "CLW", "Calibration")

table_sp1[1, ] <- evaluation(result_sp1[, 1],  result_sp1[, 2],  mu_true)
table_sp1[2, ] <- evaluation(result_sp1[, 3],  result_sp1[, 4],  mu_true)
table_sp1[3, ] <- evaluation(result_sp1[, 5],  result_sp1[, 6],  mu_true)
table_sp1[4, ] <- evaluation(result_sp1[, 7],  result_sp1[, 8],  mu_true)
table_sp1[5, ] <- evaluation(result_sp1[, 9],  result_sp1[, 10], mu_true)


########################################
# Simulation 2: sp2 (systematic PPS) as reference
########################################

result_sp2 <- matrix(NA, nrow = n_sim, ncol = 10)

for (i in seq_len(n_sim)) {

  # nonprobability sample
  set.seed(seeds_sc[i])
  sc <- fp[rbinom(N, 1, pi_c) == 1,
           c("x1", "x2", "x3", "x4", "y", "wt_sc")]

  # probability reference sample sp2 (randomized systematic PPS)
  set.seed(seeds_sp[i])
  s2  <- UPrandomsystematic(pi_p2)
  sp2 <- fp[s2 == 1, c("x1", "x2", "x3", "x4", "wt_sp2", "pi_sp2")]

  des_sp2 <- svydesign(ids = ~1, fpc = ~pi_sp2, data = sp2, pps = "brewer")

  # true-weight estimator
  result_tw <- TW.sc(sc)

  # pseudo-weight estimation: ALP, CLW, Calibration
  fit_alp  <- est_pw(data = list(sc, des_sp2), method = "alp")
  fit_clw  <- est_pw(data = list(sc, des_sp2), method = "clw")
  fit_cali <- est_pw(data = list(sc, des_sp2), method = "calibration")

  res_alp  <- pwmean(fit_alp,  y = "y")
  res_clw  <- pwmean(fit_clw,  y = "y")
  res_cali <- pwmean(fit_cali, y = "y")

  # naive
  result_sp2[i, 1:2]  <- c(res_alp$estimates$unweighted_mean,
                            res_alp$estimates$unweighted_se)
  # true weight
  result_sp2[i, 3:4]  <- result_tw
  # ALP
  result_sp2[i, 5:6]  <- c(res_alp$estimates$adjusted_mean,
                            res_alp$estimates$adjusted_se)
  # CLW
  result_sp2[i, 7:8]  <- c(res_clw$estimates$adjusted_mean,
                            res_clw$estimates$adjusted_se)
  # Calibration
  result_sp2[i, 9:10] <- c(res_cali$estimates$adjusted_mean,
                            res_cali$estimates$adjusted_se)

  if (i %% 100 == 0) cat(i, " ")
}

table_sp2 <- matrix(NA, nrow = 5, ncol = 5)
colnames(table_sp2) <- c("RB(%)", "v(*100)", "MSE(*100)", "SER", "CP")
rownames(table_sp2) <- c("Naive", "True weight", "ALP", "CLW", "Calibration")

table_sp2[1, ] <- evaluation(result_sp2[, 1],  result_sp2[, 2],  mu_true)
table_sp2[2, ] <- evaluation(result_sp2[, 3],  result_sp2[, 4],  mu_true)
table_sp2[3, ] <- evaluation(result_sp2[, 5],  result_sp2[, 6],  mu_true)
table_sp2[4, ] <- evaluation(result_sp2[, 7],  result_sp2[, 8],  mu_true)
table_sp2[5, ] <- evaluation(result_sp2[, 9],  result_sp2[, 10], mu_true)


########################################
# Simulation 3: sp1 + sp2 (multi-reference)
# sp1 covers x1, x2; sp2 covers x3, x4
########################################

result_both <- matrix(NA, nrow = n_sim, ncol = 4)
colnames(result_both) <- c("mean_precali_TRUE",  "se_precali_TRUE",
                           "mean_precali_FALSE", "se_precali_FALSE")

for (i in seq_len(n_sim)) {

  # nonprobability sample
  set.seed(seeds_sc[i])
  sc <- fp[rbinom(N, 1, pi_c) == 1,
           c("x1", "x2", "x3", "x4", "y", "wt_sc")]

  # sp1: Poisson, covers x1 and x2
  set.seed(seeds_sp[i])
  sp1 <- fp[rbinom(N, 1, pi_p1) == 1, c("x1", "x2", "wt_sp1")]
  sp1$pi_sp1 <- 1 / sp1$wt_sp1
  des_sp1 <- svydesign(ids = ~1, probs = ~pi_sp1, data = sp1)

  # sp2: systematic PPS, covers x3 and x4
  # sp2 continues from the same random stream as sp1 within each replication,
  # matching the paper's simulation design.
  s2  <- UPrandomsystematic(pi_p2)
  sp2 <- fp[s2 == 1, c("x3", "x4", "wt_sp2", "pi_sp2")]
  des_sp2 <- svydesign(ids = ~1, fpc = ~pi_sp2, data = sp2, pps = "brewer")

  # multi-reference with precalibration
  fit_true <- est_pw(
    data      = list(sc, des_sp1, des_sp2),
    precali   = TRUE,
    p_formula = list(~ x1 + x2, ~ x3 + x4)
  )
  res_true <- pwmean(fit_true, y = "y")

  # multi-reference without precalibration
  fit_false <- est_pw(
    data      = list(sc, des_sp1, des_sp2),
    precali   = FALSE,
    p_formula = list(~ x1 + x2, ~ x3 + x4)
  )
  res_false <- pwmean(fit_false, y = "y")

  result_both[i, 1:2] <- c(res_true$estimates$adjusted_mean,
                            res_true$estimates$adjusted_se)
  result_both[i, 3:4] <- c(res_false$estimates$adjusted_mean,
                            res_false$estimates$adjusted_se)

  if (i %% 100 == 0) cat(i, " ")
}

table_both <- matrix(NA, nrow = 2, ncol = 5)
colnames(table_both) <- c("RB(%)", "v(*100)", "MSE(*100)", "SER", "CP")
rownames(table_both) <- c("Multi (precali = TRUE)", "Multi (precali = FALSE)")

table_both[1, ] <- evaluation(result_both[, "mean_precali_TRUE"],
                              result_both[, "se_precali_TRUE"],  mu_true)
table_both[2, ] <- evaluation(result_both[, "mean_precali_FALSE"],
                              result_both[, "se_precali_FALSE"], mu_true)


########################################
# Results summary
########################################
# To export results, uncomment the write.csv() lines.

cat("\n\n--- Table 1, Case 0: one-reference (sp1, Poisson) ---\n")
print(kable(round(table_sp1, 2), align = "c"))

cat("\n--- Table 1, Case 0: one-reference (sp2, PPS) ---\n")
print(kable(round(table_sp2, 2), align = "c"))

cat("\n--- Table 2, Case PC.1: multi-reference (sp1 + sp2) ---\n")
print(kable(round(table_both, 2), align = "c"))

# write.csv(round(table_sp1,  2), "table_sp1.csv")
# write.csv(round(table_sp2,  2), "table_sp2.csv")
# write.csv(round(table_both, 2), "table_both.csv")

options(old_options)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----load-data----------------------------------------------------------------
library(nonprobsampling)

data("sc")
data("sp1")
data("sp1_bootstrap")
data("sp2")

str(sc)
str(sp1)
str(sp2)

## ----peek-data-sc-------------------------------------------------------------
head(sc)

## ----PSU-survey-designs-------------------------------------------------------
ref1_design <- survey::svydesign(
  ids     = ~psu_sp1,
  strata  = ~strata_sp1,
  weights = ~wts_sp1,
  data    = sp1,
  nest    = TRUE
)

ref2_design <- survey::svydesign(
  ids     = ~psu_sp2,
  strata  = ~strata_sp2,
  weights = ~wts_sp2,
  data    = sp2,
  nest    = TRUE
)

## ----replicate-wts-survey-design----------------------------------------------
ref1_design_rep_wts <- survey::svrepdesign(
   data    = sp1_bootstrap,
   weights = ~wts_sp1,
   repweights = "bw[0-9]+",
   type    = "bootstrap"
)


## ----one-ref-calibration-classic----------------------------------------------

fit_cali <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "calibration",
  control   = pw_solver_control(ftol = 1e-6)
)

print(fit_cali)

summary(fit_cali)

## ----one-ref-calibration-autopfml---------------------------------------------
fit_cali_auto <- est_pw(
  data   = list(sc, ref1_design),
  method = "calibration",
  control   = pw_solver_control(ftol = 1e-6)
)
summary(fit_cali_auto)

## ----one-ref-calibration-int--------------------------------------------------

fit_cali_interaction <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ BMI * psa_level + diabetes + pros_enlarged + I(psa_level^2),
  method    = "calibration",
  control   = pw_solver_control(ftol = 1e-6)
)

summary(fit_cali_interaction)

## ----one-ref-calibration-verbose----------------------------------------------
fit_cali <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "calibration",
  verbose   = TRUE,
  control   = pw_solver_control(ftol = 1e-6)
)

## ----one-ref-calibration-sc_updated-------------------------------------------
head(fit_cali$sc_updated)

## ----one-ref-calibration-pseudo_weights---------------------------------------
head(fit_cali$pseudo_weights)

## ----one-ref-alp--------------------------------------------------------------
fit_alp <- est_pw(
  data      = list(sc, ref1_design_rep_wts),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "alp"
)

print(fit_alp)

## ----one-ref-clw--------------------------------------------------------------
fit_clw <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "clw"
)

print(fit_clw)

## ----multi-ref-by-size--------------------------------------------------------
fit_multi_order_by_size <- est_pw(
  data      = list(sc, ref1_design, ref2_design),
  p_formula = list(
    ~ agecat + race + education + psa_level  + pros_enlarged + comorbidity + BMI + diabetes,
    ~ agecat + race + diabetes  + BMI + comorbidity
  ),
  sp_order  = "size",
  precali   = TRUE,
  control   = pw_solver_control(ftol=1e-6)
)

summary(fit_multi_order_by_size)

## ----multi-ref-as-given-------------------------------------------------------
fit_multi_as_given <- est_pw(
  data      = list(sc, ref1_design, ref2_design),
  p_formula = list(
    ~ agecat + race + education + psa_level + pros_enlarged + comorbidity + diabetes,
    ~ agecat + race + BMI + diabetes + comorbidity
  ),
  sp_order  = "given",
  precali   = TRUE
)

summary(fit_multi_as_given)

## ----pwmean-numeric-----------------------------------------------------------
est_cali_numeric <- pwmean(fit_cali, y = "psa_level")
print(est_cali_numeric)
summary(est_cali_numeric)

## ----pwmean-factor------------------------------------------------------------
est_cali_factor <- pwmean(fit_cali, y = "BMI")
print(est_cali_factor)
summary(est_cali_factor)

## ----pwmean-domain-numeric----------------------------------------------------
est_by_bmi_numeric <- pwmean(fit_cali, y = "psa_level", zcol = "BMI")
print(est_by_bmi_numeric)
summary(est_by_bmi_numeric)

## ----pwmean-domain-factor-----------------------------------------------------
est_by_bmi_factor <- pwmean(fit_cali, y = "agecat", zcol = "BMI")
print(est_by_bmi_factor)
summary(est_by_bmi_factor)

## ----na-setup-----------------------------------------------------------------
sc_with_na <- sc
sc_with_na$agecat[c(1, 2, 4)] <- NA
sc_with_na$psa_level[c(6, 7, 8, 9)] <- NA

sp1_with_na <- sp1
sp1_with_na$education[c(3, 4)] <- NA

ref1_design_with_na <- survey::svydesign(
  ids     = ~psu_sp1,
  strata  = ~strata_sp1,
  weights = ~wts_sp1,
  data    = sp1_with_na,
  nest    = TRUE
)

## ----fit-with-na--------------------------------------------------------------
fit_na_exclude <- est_pw(
  data      = list(sc_with_na, ref1_design_with_na),
  method    = "calibration",
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  na.action = stats::na.exclude,
  control   = pw_solver_control(ftol = 1e-6)
)
summary(fit_na_exclude)


## ----na_summary---------------------------------------------------------------
fit_na_exclude$na_summary

## ----na.action----------------------------------------------------------------
na.action(fit_na_exclude)

## ----na-head------------------------------------------------------------------
head(fit_na_exclude$pseudo_weights)
head(fit_na_exclude$sc_updated)

## ----na-2nd-layer-------------------------------------------------------------
result <- pwmean(fit_na_exclude, "psa_level")
summary(result)

## ----na-info-2nd--------------------------------------------------------------
result$na.action
na.action(result)

## ----fit-na-omit--------------------------------------------------------------
fit_na_omit <- est_pw(
  data      = list(sc_with_na, ref1_design_with_na),
  method    = "calibration",
  na.action = stats::na.omit,
  control   = pw_solver_control(ftol = 1e-6)
)

## ----na-omit-na.action--------------------------------------------------------
na.action(fit_na_omit)

## ----na-omit-head-------------------------------------------------------------
head(fit_na_omit$pseudo_weights)
head(fit_na_omit$sc_updated)

## ----solver-control-----------------------------------------------------------
ctrl <- pw_solver_control(
  method   = "Broyden",
  xtol     = 1e-10,
  ftol     = 1e-4,
  maxit    = 30,
  trace    = TRUE
)

fit_ctrl <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI,
  method    = "calibration",
  control   = ctrl
)

#' Dispatch one-reference pseudo-weight estimation
#'
#' @param sc Data frame. The nonprobability sample.
#' @param sp Data frame. The single probability reference sample (already
#'   NA-processed and p_formula-processed).
#' @param sp_des A `survey.design2` or `svyrep.design` object for `sp`.
#' @param vars Character vector of predictor variable names to use in the
#'   participation model design matrices.
#' @param weight Character scalar. Name of the survey-weight column in `sp`.
#' @param method Character scalar. One of `"alp"`, `"clw"`, `"calibration"`,
#'   or `"cali"`.
#' @param control A list created by `pw_solver_control()`.
#' @param verbose Logical. If `TRUE`, progress messages are printed.
#' @param log_messages Character vector of messages accumulated in earlier
#'   steps, forwarded into the returned object.
#'
#' @return A list with components:
#'   \item{pseudo_weights}{Numeric vector of estimated pseudo-weights for `sc`.}
#'   \item{coefficients}{Named numeric vector of participation model coefficients.}
#'   \item{method}{Character scalar identifying the method used.}
#'   \item{solver_diagnostics}{List of solver convergence diagnostics.}
#'   \item{internal}{List of intermediate objects (design matrices, fitted
#'     probabilities, sandwich components) needed by the estimation stage.}
#'
#' @keywords internal
ipwm_one_build <- function(
    sc,
    sp,
    sp_des,
    vars = NULL,
    weight,
    method,
    control,
    verbose = FALSE,
    log_messages = NULL
) {

  # ---------------------------------------------------------------
  # Dispatch to *_build
  # ---------------------------------------------------------------
  m_lower <- tolower(method)

  if (m_lower == "alp") {

    built <- alp_build(
      sc = sc,
      sp = sp,
      sp_des = sp_des,
      vars = vars,
      wts.col = weight,
      control = control,
      verbose = verbose,
      log_messages = log_messages
    )

  } else if (m_lower == "clw") {

    built <- clw_build(
      sc = sc,
      sp = sp,
      sp_des = sp_des,
      vars = vars,
      wts.col = weight,
      control = control,
      verbose = verbose,
      log_messages = log_messages
    )

  } else if (m_lower == "calibration") {

    built <- raking_build(
      sc = sc,
      sp = sp,
      sp_des = sp_des,
      vars = vars,
      wts.col = weight,
      control = control,
      verbose = verbose,
      log_messages = log_messages
    )

  } else {
    stop("Unknown method. Use method = 'ALP', 'CLW', or 'calibration'.",
         call. = FALSE)
  }


  # ---------------------------------------------------------------
  # 4. Return object
  # ---------------------------------------------------------------
  internal_out <- built$internal
  internal_out[["log_messages"]] <- built$log_messages

  out <- list(
    pseudo_weights     = built$weights,
    coefficients       = built$coefficients,
    method             = method,
    solver_diagnostics = built$solver_diagnostics,
    internal           = internal_out
  )
  return(out)
}

#--------------------------------------------------#
# Helper: build a default one-sided p_formula from shared variables
#--------------------------------------------------#
build_one_formula <- function(sc, sp_i, weight_i, sp_name = NULL) {
  log_messages <- character(0)

  shared     <- intersect(colnames(sc), colnames(sp_i))
  drop_these <- weight_i
  vars       <- setdiff(shared, drop_these)

  if (length(vars) == 0L) {
    stop(
      if (is.null(sp_name)) {
        "No shared covariates found to build default p_formula."
      } else {
        sprintf("No shared covariates found to build default p_formula for %s.", sp_name)
      },
      call. = FALSE
    )
  }

  fml <- stats::as.formula(paste("~", paste(vars, collapse = " + ")))

  label <- if (is.null(sp_name)) "Generated default p_formula" else paste0("Generated default p_formula for ", sp_name)

  msg <- paste0(
    label, ": ",
    paste(deparse(fml), collapse = ""),
    "\n"
  )

  log_messages <- c(log_messages, msg)

  return(list(
    p_formula    = fml,
    log_messages = log_messages
  ))
}

#' Run the estimator for a single domain
#'
#' Selects and calls the specific estimator function based on
#' `build$method`, passing the relevant matrices and vectors
#' for one domain at a time.
#'
#' @param build A `pw_fit` object returned by the build step.
#' @param Y Outcome vector of length `n`.
#' @param zvec Domain indicator vector of length `n`.
#' @param w Pseudo-weight vector of length `n`.
#' @param X Design matrix for the convenience sample, with dimension
#'  `n x p`.
#'
#' @return A list with components `mean` and `variance`.
#'
#' @keywords internal
dispatch_estimator_one_domain <- function(build, Y, zvec, w, X) {

  if (length(zvec) != length(Y)) {
    stop("Length mismatch between `Y` and `zvec`.", call. = FALSE)
  }

  method <- build$method
  D <- build$internal$D
  S_beta <- build$internal$S_beta

  method_key <- tolower(method)

  out <- switch(
    method_key,
    "calibration" = raking_estimate(Y, zvec, w, X, D, S_beta),
    "cali"        = raking_estimate(Y, zvec, w, X, D, S_beta),

    "alp"         = alp_estimate(Y, zvec, w, X, D, S_beta),
    "clw"         = clw_estimate(Y, zvec, w, X, D, S_beta),

    "multi"       = multi_estimate(Y, zvec, w, X, D, S_beta),

    stop(
      sprintf(
        "Unknown method '%s'. Must be one of: 'alp', 'clw', 'calibration'/'cali', 'multi'.",
        method
      ),
      call. = FALSE
    )
  )

  out
}

#' Dispatch the estimator across domains
#'
#' Determines the domain mode (overall, binary, or factor) from
#' `yz_data` and calls `dispatch_estimator_one_domain()` for each domain,
#' returning a unified result list.
#'
#' @param build A `pw_fit` object returned by the build step.
#' @param yz_data A list returned by `process_na_yz()`, containing
#'   `Y`, `w`, `X`, and `domain`.
#'
#' @return A list with components:
#'
#' - `type`: either `"single"` or `"multi"`.
#' - `labels`: character vector of domain labels.
#' - `estimates`: a list of per-domain results, each with `mean` and `variance`.
#'
#' @keywords internal
dispatch_estimator <- function(build, yz_data) {

  Y <- yz_data$Y
  w <- yz_data$w
  X <- yz_data$X

  domain <- yz_data$domain
  mode <- domain$mode

  #----------------------------------#
  # overall mean
  #----------------------------------#
  if (identical(mode, "overall")) {
    zvec <- rep.int(1L, length(Y))

    est <- dispatch_estimator_one_domain(
      build = build,
      Y = Y,
      zvec = zvec,
      w = w,
      X = X
    )

    return(list(
      type = "single",
      labels = "Overall",
      estimates = est
    ))
  }

  #----------------------------------#
  # binary domain variable
  #----------------------------------#
  if (identical(mode, "binary")) {
    zvec <- domain$indicators[[1]]

    est <- dispatch_estimator_one_domain(
      build = build,
      Y = Y,
      zvec = zvec,
      w = w,
      X = X
    )

    return(list(
      type = "single",
      labels = domain$labels,
      estimates = est
    ))
  }

  #----------------------------------#
  # factor / character domain variable
  #----------------------------------#
  if (identical(mode, "factor")) {
    ind_df <- domain$indicators
    labs <- domain$labels

    est_list <- lapply(seq_along(labs), function(j) {
      dispatch_estimator_one_domain(
        build = build,
        Y = Y,
        zvec = ind_df[[j]],
        w = w,
        X = X
      )
    })

    return(list(
      type = "multi",
      labels = labs,
      estimates = est_list
    ))
  }

  stop("Unsupported domain mode in `yz_data$domain`.", call. = FALSE)
}

# test-estimate.R

# helpers ----

make_est_inputs <- function() {
  n      <- 4L
  Y      <- c(1, 2, 4, 7)
  Z      <- rep(1L, n)
  w      <- rep(2.0, n)
  X      <- cbind("(Intercept)" = 1, x1 = seq_len(n))

  S_beta <- t(w * X) %*% X
  D      <- matrix(0.0, nrow = 2L, ncol = 2L)

  list(
    Y = Y,
    Z = Z,
    w = w,
    X = X,
    D = D,
    S_beta = S_beta
  )
}

make_est_inputs_nonzero_D <- function() {
  inp <- make_est_inputs()

  inp$D <- matrix(
    c(0.20, 0.05,
      0.05, 0.30),
    nrow = 2L,
    byrow = TRUE
  )

  inp
}

make_dispatch_build <- function(method = "alp", inp = make_est_inputs()) {
  structure(
    list(
      method = method,
      internal = list(
        D = inp$D,
        S_beta = inp$S_beta
      )
    ),
    class = "pw_fit"
  )
}

domain_overall <- list(
  mode       = "overall",
  z_name     = NULL,
  labels     = "Overall",
  indicators = NULL
)

domain_binary <- list(
  mode       = "binary",
  z_name     = "grp",
  labels     = "grp = 1",
  indicators = data.frame(.z_domain = c(1L, 1L, 0L, 0L))
)

domain_factor <- list(
  mode       = "factor",
  z_name     = "grp",
  labels     = c("A", "B"),
  indicators = data.frame(
    .z_A = c(1L, 0L, 1L, 0L),
    .z_B = c(0L, 1L, 0L, 1L)
  )
)

make_yz_data <- function(domain_obj) {
  inp <- make_est_inputs()

  list(
    Y = inp$Y,
    w = inp$w,
    X = inp$X,
    domain = domain_obj
  )
}

estimate_funs <- list(
  alp    = alp_estimate,
  clw    = clw_estimate,
  raking = raking_estimate,
  multi  = multi_estimate
)


# output structure ----

test_that("all estimate functions return mean and variance", {
  inp <- make_est_inputs()

  for (fn in estimate_funs) {
    res <- fn(inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta)

    expect_type(res, "list")
    expect_named(res, c("mean", "variance"))

    expect_length(res$mean, 1L)
    expect_length(res$variance, 1L)

    expect_true(is.finite(res$mean))
    expect_true(is.finite(res$variance))
    expect_gte(res$variance, 0)
  }
})


# mean correctness ----

test_that("all estimate functions compute the pseudo-weighted Hajek mean", {
  inp <- make_est_inputs()

  expected <- sum(inp$Y * inp$Z * inp$w) / sum(inp$Z * inp$w)

  for (fn in estimate_funs) {
    res <- fn(inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta)
    expect_equal(res$mean, expected, tolerance = 1e-12)
  }
})

test_that("all estimate functions compute domain-specific pseudo-weighted mean", {
  inp   <- make_est_inputs()
  Z_dom <- c(1L, 1L, 0L, 0L)

  expected <- sum(inp$Y * Z_dom * inp$w) / sum(Z_dom * inp$w)

  for (fn in estimate_funs) {
    res <- fn(inp$Y, Z_dom, inp$w, inp$X, inp$D, inp$S_beta)
    expect_equal(res$mean, expected, tolerance = 1e-12)
  }
})


# variance correctness when D = 0 ----

test_that("raking_estimate variance matches analytical value when D = 0", {
  inp <- make_est_inputs()

  res <- raking_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )

  expect_equal(res$variance, 2 / 64, tolerance = 1e-12)
})

test_that("multi_estimate variance matches raking in one-block D = 0 case", {
  inp <- make_est_inputs()

  res_multi <- multi_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )

  res_raking <- raking_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )

  expect_equal(res_multi$variance, res_raking$variance, tolerance = 1e-12)
  expect_equal(res_multi$variance, 2 / 64, tolerance = 1e-12)
})

test_that("alp_estimate variance matches analytical value when D = 0", {
  inp <- make_est_inputs()

  res <- alp_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )

  expect_equal(res$variance, 89 / 288, tolerance = 1e-12)
})

test_that("clw_estimate variance matches analytical value when D = 0", {
  inp <- make_est_inputs()

  res <- clw_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )

  expect_equal(res$variance, 49 / 128, tolerance = 1e-12)
})


# variance correctness when D is nonzero ----
# For this synthetic example:
# b_raking = b_alp = c(-5, 2)
# b_clw    = c(-2.5, 1)
# with D = matrix(c(.20, .05, .05, .30), 2, 2, byrow = TRUE)

test_that("raking_estimate variance includes nonzero D component", {
  inp <- make_est_inputs_nonzero_D()

  res <- raking_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )

  # D = 0 variance is 2/64.
  # Reference component is b D b' = 5.2.
  # Total variance is (2 + 5.2) / 64 = 9/80.
  expect_equal(res$variance, 9 / 80, tolerance = 1e-12)
})

test_that("multi_estimate variance matches raking in one-block nonzero-D case", {
  inp <- make_est_inputs_nonzero_D()

  res_multi <- multi_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )

  res_raking <- raking_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )

  expect_equal(res_multi$variance, res_raking$variance, tolerance = 1e-12)
  expect_equal(res_multi$variance, 9 / 80, tolerance = 1e-12)
})

test_that("alp_estimate variance includes nonzero D component", {
  inp <- make_est_inputs_nonzero_D()

  res <- alp_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )

  # D = 0 variance is 89/288.
  # Reference component is 5.2 / 64.
  # Total variance is 281/720.
  expect_equal(res$variance, 281 / 720, tolerance = 1e-12)
})

test_that("clw_estimate variance includes nonzero D component", {
  inp <- make_est_inputs_nonzero_D()

  res <- clw_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )

  # D = 0 variance is 49/128.
  # Reference component is 1.3 / 64.
  # Total variance is 129/320.
  expect_equal(res$variance, 129 / 320, tolerance = 1e-12)
})


# method differences ----

test_that("estimators have expected variance relationships in the synthetic example", {
  inp <- make_est_inputs()

  v_raking <- raking_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )$variance

  v_multi <- multi_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )$variance

  v_alp <- alp_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )$variance

  v_clw <- clw_estimate(
    inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta
  )$variance

  expect_equal(v_raking, v_multi, tolerance = 1e-12)
  expect_false(isTRUE(all.equal(v_raking, v_alp)))
  expect_false(isTRUE(all.equal(v_raking, v_clw)))
  expect_false(isTRUE(all.equal(v_alp, v_clw)))
})


# dispatch_estimator_one_domain routing ----

test_that("dispatch_estimator_one_domain routes alp correctly", {
  inp   <- make_est_inputs()
  build <- make_dispatch_build("alp", inp)

  expect_equal(
    dispatch_estimator_one_domain(build, inp$Y, inp$Z, inp$w, inp$X),
    alp_estimate(inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta)
  )
})

test_that("dispatch_estimator_one_domain routes clw correctly", {
  inp   <- make_est_inputs()
  build <- make_dispatch_build("clw", inp)

  expect_equal(
    dispatch_estimator_one_domain(build, inp$Y, inp$Z, inp$w, inp$X),
    clw_estimate(inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta)
  )
})

test_that("dispatch_estimator_one_domain routes calibration correctly", {
  inp   <- make_est_inputs()
  build <- make_dispatch_build("calibration", inp)

  expect_equal(
    dispatch_estimator_one_domain(build, inp$Y, inp$Z, inp$w, inp$X),
    raking_estimate(inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta)
  )
})

test_that("dispatch_estimator_one_domain treats cali as calibration", {
  inp    <- make_est_inputs()
  build1 <- make_dispatch_build("calibration", inp)
  build2 <- make_dispatch_build("cali", inp)

  expect_equal(
    dispatch_estimator_one_domain(build1, inp$Y, inp$Z, inp$w, inp$X),
    dispatch_estimator_one_domain(build2, inp$Y, inp$Z, inp$w, inp$X)
  )
})

test_that("dispatch_estimator_one_domain routes multi correctly", {
  inp   <- make_est_inputs()
  build <- make_dispatch_build("multi", inp)

  expect_equal(
    dispatch_estimator_one_domain(build, inp$Y, inp$Z, inp$w, inp$X),
    multi_estimate(inp$Y, inp$Z, inp$w, inp$X, inp$D, inp$S_beta)
  )
})

test_that("dispatch_estimator_one_domain method matching is case-insensitive", {
  inp    <- make_est_inputs()
  build1 <- make_dispatch_build("alp", inp)
  build2 <- make_dispatch_build("ALP", inp)

  expect_equal(
    dispatch_estimator_one_domain(build1, inp$Y, inp$Z, inp$w, inp$X),
    dispatch_estimator_one_domain(build2, inp$Y, inp$Z, inp$w, inp$X)
  )
})

test_that("dispatch_estimator_one_domain errors for unknown method", {
  inp   <- make_est_inputs()
  build <- make_dispatch_build("unknown_method", inp)

  expect_error(
    dispatch_estimator_one_domain(build, inp$Y, inp$Z, inp$w, inp$X),
    "Unknown method 'unknown_method'",
    fixed = TRUE
  )
})

test_that("dispatch_estimator_one_domain errors when Y and zvec lengths differ", {
  inp   <- make_est_inputs()
  build <- make_dispatch_build("alp", inp)

  expect_error(
    dispatch_estimator_one_domain(
      build = build,
      Y = inp$Y,
      zvec = c(1L, 1L),
      w = inp$w,
      X = inp$X
    ),
    "Length mismatch between `Y` and `zvec`.",
    fixed = TRUE
  )
})


# dispatch_estimator domain routing ----

test_that("dispatch_estimator handles overall mode", {
  build   <- make_dispatch_build("alp")
  yz_data <- make_yz_data(domain_overall)

  res <- dispatch_estimator(build, yz_data)

  expect_equal(res$type, "single")
  expect_equal(res$labels, "Overall")
  expect_named(res$estimates, c("mean", "variance"))

  expect_equal(res$estimates$mean, 3.5, tolerance = 1e-12)
})

test_that("dispatch_estimator handles binary domain mode", {
  build   <- make_dispatch_build("alp")
  yz_data <- make_yz_data(domain_binary)

  res <- dispatch_estimator(build, yz_data)

  expect_equal(res$type, "single")
  expect_equal(res$labels, "grp = 1")
  expect_named(res$estimates, c("mean", "variance"))

  # Binary domain uses observations 1 and 2.
  expect_equal(res$estimates$mean, 1.5, tolerance = 1e-12)
})

test_that("dispatch_estimator handles factor domain mode", {
  build   <- make_dispatch_build("alp")
  yz_data <- make_yz_data(domain_factor)

  res <- dispatch_estimator(build, yz_data)

  expect_equal(res$type, "multi")
  expect_equal(res$labels, c("A", "B"))
  expect_length(res$estimates, 2L)

  expect_named(res$estimates[[1]], c("mean", "variance"))
  expect_named(res$estimates[[2]], c("mean", "variance"))

  # A uses observations 1 and 3.
  # B uses observations 2 and 4.
  expect_equal(res$estimates[[1]]$mean, 2.5, tolerance = 1e-12)
  expect_equal(res$estimates[[2]]$mean, 4.5, tolerance = 1e-12)
})

test_that("dispatch_estimator errors for unsupported domain mode", {
  build <- make_dispatch_build("alp")

  yz_data <- make_yz_data(
    list(
      mode = "unknown",
      labels = NULL,
      indicators = NULL
    )
  )

  expect_error(
    dispatch_estimator(build, yz_data),
    "Unsupported domain mode in `yz_data$domain`.",
    fixed = TRUE
  )
})

#' Summarize a Pseudo-Weight Fit
#'
#' @param object An object of class \code{"pw_fit"}, returned by \code{\link{est_pw}}.
#' @param ... Additional arguments, currently unused.
#'
#' @return Invisibly returns \code{object}.
#'
#' @export
summary.pw_fit <- function(object, ...) {

  cat("Call:\n")
  print(object$call)

  method <- object$method

  if (method %in% c("alp", "clw", "calibration")) {
    if (method == "alp")         method <- "ALP"
    if (method == "clw")         method <- "CLW"
    if (method == "calibration") method <- "calibration"

    cat("\nMethod: One reference", method, "\n")

    log <- object$internal$log_messages
    if (!is.null(log) && length(log) > 0) {
      cat(log, sep = "")
    }

  } else if (method == "multi") {

    cat("\nMethod: Multi-reference calibration\n")

    log <- object$internal$log_messages
    if (!is.null(log) && length(log) > 0) {
      cat(log, sep = "")
    }
  }

  # --- Model information ---
  if (!is.null(object$coefficients)) {
    cat("\nParticipation model involves the following variables:\n")
    cat(names(object$coefficients)[-1], "\n\n")
  }

  # --- Solver diagnostics ---
  diag <- object$solver_diagnostics

  if (!is.null(diag)) {
    cat("Solver diagnostics:\n")
    cat("  Solver:", diag$solver, "\n")

    if (!is.null(diag$method)) {
      cat("  Method:", diag$method, "\n")
    }

    cat("  Termination code:", diag$termcd, "\n")
    cat("  Iterations:", diag$iter, "\n")
    cat(
      "  Max |estimating equation|:",
      formatC(diag$fmax, format = "e", digits = 3),
      "\n"
    )

    if (!is.null(diag$message)) {
      cat("  Message:", diag$message, "\n")
    }
  }

  # --- Coefficients section ---
  if (!is.null(object$coefficients)) {
    cat("\nParticipation model coefficients:\n")

    df <- rbind(object$coefficients)
    rownames(df) <- NULL

    formatted_df <- format(round(df, 4), justify = "left", width = 10)
    print(as.data.frame(formatted_df), row.names = FALSE)
  }

  # --- na summary ---
  na_sum <- object$na_summary
  if (!is.null(na_sum)) {
    lines <- character(0)

    if (na_sum$sc$n_excluded > 0) {
      n <- na_sum$sc$n_excluded
      lines <- c(lines, sprintf("(%d observation%s deleted due to missingness in sc)",
                                n, if (n != 1L) "s" else ""))
    }

    if (na_sum$n_ref == 1) {
      if (na_sum$sp$n_excluded > 0) {
        n <- na_sum$sp$n_excluded
        lines <- c(lines, sprintf("(%d observation%s deleted due to missingness in sp)",
                                  n, if (n != 1L) "s" else ""))
      }
    } else {
      for (i in seq_along(na_sum$sp)) {
        if (na_sum$sp[[i]]$n_excluded > 0) {
          n <- na_sum$sp[[i]]$n_excluded
          lines <- c(lines, sprintf("(%d observation%s deleted due to missingness in sp[[%d]])",
                                    n, if (n != 1L) "s" else "", i))
        }
      }
    }

    if (length(lines) > 0) {
      cat("\n", paste(lines, collapse = "\n"), "\n", sep = "")
    }
  }

  invisible(object)
}

#' Reconstruct the sample data frame with fitted pseudo-weights
#'
#' Inserts the fitted pseudo-weights `w_fit` back into the original sample
#' data frame `sc0`, using the NA handling strategy used during
#' estimation. For `"omit"`, only the rows that were kept are returned.
#' For `"exclude"`, all original rows are returned with NA weights
#' for rows that were dropped. For `"fail"` or `"pass"`, all rows
#' are assumed to be present and weights are attached directly.
#'
#' @param sc0 The original sample data frame (before any NA removal).
#' @param w_fit Numeric vector of fitted pseudo-weights, one per kept row.
#' @param keep_sc Logical vector identifying which rows of `sc0` were retained
#'   for fitting.
#' @param na_mode Character string; one of `"omit"`, `"exclude"`,
#'   `"fail"`, or `"pass"`.
#' @param na_action_obj The `na.action` attribute produced during NA
#'   removal, or NULL.
#' @param sc_wname Name of the column in which pseudo-weights will be stored.
#'
#' @return A data frame derived from `sc0` with a pseudo-weight column
#'   named `sc_wname`.
#'
#' @keywords internal
reconstruct_sc_output <- function(sc0, w_fit, keep_sc, na_mode, na_action_obj, sc_wname) {

  n0     <- nrow(sc0)
  n_fit  <- length(w_fit)
  n_keep <- sum(keep_sc)

  if (n_fit != n_keep) {
    stop(
      sprintf(
        "Length mismatch: w_fit has %d values but keep_sc has %d indices.",
        n_fit, n_keep
      ),
      call. = FALSE
    )
  }

  if (na_mode == "omit") {
    sc_out <- sc0[keep_sc, , drop = FALSE]
    sc_out[[sc_wname]] <- w_fit

  } else if (na_mode == "exclude") {
    w_full <- rep(NA_real_, n0)
    w_full[keep_sc] <- w_fit
    sc_out <- sc0
    sc_out[[sc_wname]] <- w_full
    if (!is.null(na_action_obj)) {
      attr(sc_out, "na.action") <- na_action_obj
    }

  } else {
    # "fail" or "pass": no rows were removed
    if (n_fit != n0) {
      stop(
        sprintf(
          "For na_mode '%s', expected w_fit length %d to equal nrow(sc0) = %d.",
          na_mode, n_fit, n0
        ),
        call. = FALSE
      )
    }
    sc_out <- sc0
    sc_out[[sc_wname]] <- w_fit
  }

  sc_out
}

# test-build_s6_sc_update.R
# Tests for reconstruct_sc_output()
#
# Function signature:
#   reconstruct_sc_output(sc0, w_fit, keep_sc, na_mode, na_action_obj, sc_wname)
#
# Four na_mode branches are tested separately:
#   "omit"    -- only kept rows returned; no NA in weight column
#   "exclude" -- all rows returned; NA weight for dropped rows; na.action attr set
#   "fail"    -- all rows returned; w_fit must equal nrow(sc0)
#   "pass"    -- same as "fail"

# setup ----

set.seed(2026)
n0 <- 10L

# sc0 has two NA rows (rows 3 and 7) in x1
sc0 <- data.frame(
  id = seq_len(n0),
  x1 = c(1, 2, NA, 4, 5, 6, NA, 8, 9, 10),
  x2 = rnorm(n0)
)

keep_sc  <- !is.na(sc0$x1)           # TRUE for rows 1,2,4,5,6,8,9,10  (n_keep = 8)
n_keep   <- sum(keep_sc)             # 8
excluded <- which(!keep_sc)          # c(3, 7)

set.seed(7)
w_fit_kept <- runif(n_keep, 10, 100)  # weights for the 8 kept rows

# na_action_obj: structure produced by na.omit / na.exclude
na_action_obj <- structure(excluded, class = "omit")

sc_wname <- "pseudo_wts"

# Full-length weights for "fail" / "pass" tests (all rows kept)
keep_all   <- rep(TRUE, n0)
w_fit_full <- runif(n0, 10, 100)


# na_mode = "omit" ----

test_that("reconstruct_sc_output omit: result has sum(keep_sc) rows", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "omit", na_action_obj, sc_wname)
  expect_equal(nrow(out), n_keep)
})

test_that("reconstruct_sc_output omit: left rows match sc0[keep_sc, ]", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "omit", na_action_obj, sc_wname)
  expect_equal(out$id, sc0$id[keep_sc])
  expect_equal(out$x1, sc0$x1[keep_sc])
})

test_that("reconstruct_sc_output omit: weight column equals w_fit", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "omit", na_action_obj, sc_wname)
  expect_equal(out[[sc_wname]], w_fit_kept)
})

test_that("reconstruct_sc_output omit: weight column has no NA", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "omit", na_action_obj, sc_wname)
  expect_false(anyNA(out[[sc_wname]]))
})

test_that("reconstruct_sc_output omit: weight column is named sc_wname", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "omit", na_action_obj, sc_wname)
  expect_true(sc_wname %in% names(out))
})


# na_mode = "exclude" ----

test_that("reconstruct_sc_output exclude: result has nrow(sc0) rows", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "exclude", na_action_obj, sc_wname)
  expect_equal(nrow(out), n0)
})

test_that("reconstruct_sc_output exclude: kept rows have correct weights", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "exclude", na_action_obj, sc_wname)
  expect_equal(out[[sc_wname]][keep_sc], w_fit_kept)
})

test_that("reconstruct_sc_output exclude: excluded rows have NA weight", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "exclude", na_action_obj, sc_wname)
  expect_true(all(is.na(out[[sc_wname]][excluded])))
})

test_that("reconstruct_sc_output exclude: exactly n_keep non-NA weights", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "exclude", na_action_obj, sc_wname)
  expect_equal(sum(!is.na(out[[sc_wname]])), n_keep)
})

test_that("reconstruct_sc_output exclude: na.action attribute is set when na_action_obj is not NULL", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "exclude", na_action_obj, sc_wname)
  expect_false(is.null(attr(out, "na.action")))
  expect_equal(attr(out, "na.action"), na_action_obj)
})

test_that("reconstruct_sc_output exclude: no na.action attribute when na_action_obj is NULL", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "exclude", NULL, sc_wname)
  expect_null(attr(out, "na.action"))
})

test_that("reconstruct_sc_output exclude: original columns are kept unchanged", {
  out <- reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "exclude", na_action_obj, sc_wname)
  expect_equal(out$id, sc0$id)
  expect_equal(out$x1, sc0$x1)
})


# na_mode = "fail" ----

test_that("reconstruct_sc_output fail: result has nrow(sc0) rows", {
  out <- reconstruct_sc_output(sc0, w_fit_full, keep_all, "fail", NULL, sc_wname)
  expect_equal(nrow(out), n0)
})

test_that("reconstruct_sc_output fail: weight column equals w_fit", {
  out <- reconstruct_sc_output(sc0, w_fit_full, keep_all, "fail", NULL, sc_wname)
  expect_equal(out[[sc_wname]], w_fit_full)
})

test_that("reconstruct_sc_output fail: no NA in weight column", {
  out <- reconstruct_sc_output(sc0, w_fit_full, keep_all, "fail", NULL, sc_wname)
  expect_false(anyNA(out[[sc_wname]]))
})

test_that("reconstruct_sc_output fail: error when length(w_fit) != nrow(sc0)", {
  # The general length check passes because w_fit_kept and keep_sc
  # both correspond to the 8 left observations. For na_mode = "fail",
  # however, the fit must cover all 10 original observations, so the
  # fail-specific check should error.
  expect_error(
    reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "fail", NULL, sc_wname),
    "expected w_fit length",
    fixed = TRUE
  )
})


# na_mode = "pass" ----

test_that("reconstruct_sc_output pass: result has nrow(sc0) rows", {
  out <- reconstruct_sc_output(sc0, w_fit_full, keep_all, "pass", NULL, sc_wname)
  expect_equal(nrow(out), n0)
})

test_that("reconstruct_sc_output pass: weight column equals w_fit", {
  out <- reconstruct_sc_output(sc0, w_fit_full, keep_all, "pass", NULL, sc_wname)
  expect_equal(out[[sc_wname]], w_fit_full)
})

test_that("reconstruct_sc_output pass: error when length(w_fit) != nrow(sc0)", {
  # Same logic as the "fail" test above: n_fit == n_keep passes the first
  # check, but n_fit != n0 triggers the pass-specific error.
  expect_error(
    reconstruct_sc_output(sc0, w_fit_kept, keep_sc, "pass", NULL, sc_wname),
    "expected w_fit length",
    fixed = TRUE
  )
})


# length mismatch ----

test_that("reconstruct_sc_output: error when length(w_fit) != sum(keep_sc)", {
  w_wrong <- w_fit_kept[-1]   # one element too short
  expect_error(
    reconstruct_sc_output(sc0, w_wrong, keep_sc, "omit", na_action_obj, sc_wname),
    "Length mismatch",
    fixed = TRUE
  )
})

test_that("reconstruct_sc_output: error message reports actual counts", {
  w_wrong <- w_fit_kept[-1]
  expect_error(
    reconstruct_sc_output(sc0, w_wrong, keep_sc, "omit", na_action_obj, sc_wname),
    as.character(n_keep - 1L),   # reported length of w_fit
    fixed = TRUE
  )
})

#' Build pseudo-weights using the calibration (raking) method
#'
#' @param vars Character vector of predictor variable names.
#' @param sc Data frame. The nonprobability sample.
#' @param sp Data frame. The probability reference sample.
#' @param sp_des A `survey.design2` or `svyrep.design` object for `sp`.
#' @param wts.col Character scalar. Name of the survey-weight column in `sp`.
#' @param control A list created by `pw_solver_control()`.
#' @param verbose Logical. If `TRUE`, convergence messages are printed.
#' @param log_messages Character vector of messages accumulated in earlier
#'   steps, appended to the returned object unchanged.
#'
#' @return A list with components:
#'   \item{weights}{Numeric vector of calibration pseudo-weights for `sc`.}
#'   \item{coefficients}{Named numeric vector of participation model coefficients.}
#'   \item{solver_diagnostics}{List of solver convergence diagnostics.}
#'   \item{log_messages}{Updated character vector of log messages.}
#'   \item{internal}{List containing design matrices (`Xc`, `Xp`) and sandwich
#'     components (`S_beta`, `D`) needed by the estimation stage.}
#'
#' @keywords internal
raking_build <- function(vars, sc, sp, sp_des, wts.col,
                         control,
                         verbose = FALSE,
                         log_messages = NULL) {

  if (is.null(log_messages)) {
    log_messages <- character(0)
  }

  Xc <- add_intercept(vars = vars, data = sc)
  Xp <- add_intercept(vars = vars, data = sp)

  check_design_identifiability(Xc, label = "One_Ref_Calibration")
  check_design_identifiability(Xp, label = "One_Ref_Calibration")

  if (ncol(Xc) != ncol(Xp)) {
    stop(
      "Calibration failed: Xc and Xp must have the same number of columns.",
      call. = FALSE
    )
  }

  if (!wts.col %in% names(sp)) {
    stop(
      sprintf("Calibration failed: weight column '%s' was not found in `sp`.", wts.col),
      call. = FALSE
    )
  }

  wts_sp <- sp[[wts.col]]

  if (!is.numeric(wts_sp) || any(!is.finite(wts_sp)) || any(wts_sp <= 0)) {
    stop(
      "Calibration failed: reference survey weights must be positive finite numbers.",
      call. = FALSE
    )
  }

  f_p <- colSums(wts_sp * Xp)

  if (any(!is.finite(f_p))) {
    stop(
      "Calibration failed: reference weighted totals contain non-finite values.",
      call. = FALSE
    )
  }

  if (verbose) {
    message(sprintf("Fitting Calibration participation model with %d variables.", ncol(Xc)))
  }

  beta_start <- raking_start(Xc = Xc, f_p = f_p)

  sol <- solve_participation_model(
    beta_start = beta_start,
    fn = raking_fn,
    jac = raking_jac,
    label = "One_Ref_Calibration",
    control = control,
    Xc = Xc,
    f_p = f_p
  )

  beta <- sol$coefficients

  eta <- as.vector(Xc %*% beta)

  if (any(!is.finite(eta))) {
    stop(
      "Calibration failed: final linear predictor contains non-finite values.",
      call. = FALSE
    )
  }

  if (any(eta < -log(.Machine$double.xmax))) {
    stop(
      "Calibration failed: final pseudo-weights overflow.",
      call. = FALSE
    )
  }

  wts_sc <- as.vector(exp(-eta))

  if (any(!is.finite(wts_sc))) {
    stop(
      "Calibration failed: final pseudo-weights contain non-finite values.",
      call. = FALSE
    )
  }

  f_final <- raking_fn(beta, Xc = Xc, f_p = f_p)
  max_abs_eq <- max(abs(f_final))

  if (!is.finite(max_abs_eq)) {
    stop(
      "Calibration failed: final estimating equations contain non-finite values.",
      call. = FALSE
    )
  }

  if (verbose) {
    if (sol$termcd == 1L || sol$termcd == 2L) {
      message(sprintf("Converged after %d iterations (max |eq| = %.3e).",
                      sol$iterations, max_abs_eq))
    } else {
      message(sprintf(
        "WARNING: solver did not converge (termcd = %d). Check solver_diagnostics for details.",
        sol$termcd
      ))
    }
  }

  S_beta_full <- t(wts_sc * Xc) %*% Xc
  D <- compute_D_raking(sp_des, Xp)

  out <- list(
    weights      = wts_sc,
    coefficients = stats::setNames(as.numeric(beta), colnames(Xc)),

    solver_diagnostics = list(
      solver  = sol$solver,
      termcd  = sol$termcd,
      message = sol$message,
      method  = sol$solver_method,
      iter    = sol$iterations,
      fmax    = max_abs_eq
    ),


    log_messages = log_messages,
    internal = list(
      Xc = Xc,
      Xp = Xp,
      S_beta = S_beta_full,
      D = D
    )
  )

  return(out)
}

#' Print method for pw_fit objects
#'
#' Compact one-screen overview of a fitted pseudo-weight object: the call, the
#' pseudo-weighting method, the participation model size, solver convergence,
#' and a summary of the estimated pseudo-weights. For the full coefficient
#' table and detailed solver diagnostics, use \code{\link{summary.pw_fit}}.
#'
#' @param x An object of class \code{"pw_fit"}, returned by \code{\link{est_pw}}.
#' @param ... Additional arguments, currently unused.
#'
#' @return Invisibly returns \code{x}.
#'
#' @method print pw_fit
#' @export
print.pw_fit <- function(x, ...) {

  cat("Pseudo-weight fit (\"pw_fit\")\n\n")

  # --- Call ---
  cat("Call:\n")
  if (!is.null(x$call)) print(x$call) else cat("  (not available)\n")
  cat("\n")

  # --- Method ---
  m <- tolower(trimws(as.character(x$method %||% "")))
  method_label <- if (m == "alp") {
    "One reference ALP"
  } else if (m == "clw") {
    "One reference CLW"
  } else if (m %in% c("cali", "calibration")) {
    "One reference calibration"
  } else if (m == "multi") {
    "Multi-reference calibration"
  } else {
    x$method %||% "(unknown)"
  }
  cat(sprintf("%-21s %s\n", "Method:", method_label))

  # --- Participation model size (parameter count) ---
  if (!is.null(x$coefficients)) {
    p <- length(x$coefficients)
    cat(sprintf(
      "%-21s %d parameters%s\n",
      "Participation model:", p,
      if (m == "multi") "" else " (incl. intercept)"
    ))
  }

  # --- Convergence ---
  diag <- x$solver_diagnostics
  if (!is.null(diag)) {
    tc <- suppressWarnings(as.integer(diag$termcd))
    status <- if (!is.na(tc) && tc == 1L) {
      "converged"
    } else if (!is.na(tc) && tc %in% c(2L, 3L)) {
      "converged (tolerance warning)"
    } else {
      "not converged"
    }
    cat(sprintf(
      "%-21s %s  (%s, %s iter, max|EE| = %s)\n",
      "Convergence:", status,
      diag$solver %||% "?",
      format(as.integer(diag$iter %||% NA)),
      formatC(diag$fmax %||% NA_real_, format = "e", digits = 2)
    ))
  }

  # --- Pseudo-weights summary (six-number + sum) ---
  w <- x$pseudo_weights
  if (!is.null(w)) {
    n_na <- sum(is.na(w))
    w_ok <- w[!is.na(w)]
    n    <- length(w_ok)

    if (n > 0L) {
      cat(sprintf(
        "\nPseudo-weights (n = %d%s):\n",
        n, if (n_na > 0L) sprintf(", %d NA", n_na) else ""
      ))
      print(summary(w_ok))

      neg <- sum(w_ok < 0)
      cat(sprintf(
        "Sum: %s%s\n",
        format(round(sum(w_ok)), big.mark = ",", scientific = FALSE),
        if (neg > 0L) sprintf("   (%d negative weights)", neg) else ""
      ))
    }
  }

  # --- NA note (nonprobability sample only; sp details are in summary) ---
  na_sum <- x$na_summary
  if (!is.null(na_sum) && !is.null(na_sum$sc) && isTRUE(na_sum$sc$n_excluded > 0)) {
    k <- na_sum$sc$n_excluded
    cat(sprintf(
      "\n(%d observation%s deleted due to missingness in sc)\n",
      k, if (k != 1L) "s" else ""
    ))
  }

  cat("\nUse summary() for coefficients and diagnostics; pwmean() to estimate means.\n")

  invisible(x)
}

est_pw <- function(
    data,
    sp_order = c("size", "given"),
    precali = TRUE,
    p_formula = NULL,
    method = NULL,
    na.action = stats::na.omit,
    sc_wname = "pseudo_wts",
    control = pw_solver_control(),
    verbose = FALSE
) {

  sp_order <- match.arg(sp_order)

  if (!is.null(method)) {
    method <- tolower(trimws(method))
    if (method == "cali") method <- "calibration"
  }

  #--------------------------------------------------------------------------#
  # Step 1. Parse input and basic validation
  #--------------------------------------------------------------------------#
  step1 <- tryCatch({

    check_ipwm_inputs_build(
      data      = data,
      p_formula = p_formula,
      method    = method,
      sp_order  = sp_order,
      precali   = precali,
      sc_wname  = sc_wname,
      verbose   = verbose
    )

    parsed <- parse_ipwm_data(data)

    sc     <- parsed$sc
    sp_des <- parsed$sp_des
    sp     <- parsed$sp_vars
    n_ref  <- parsed$n_ref

    if (n_ref == 1L) {
      sp     <- sp[[1]]
      sp_des <- sp_des[[1]]
    }

    weight <- rep("sp_wts", n_ref)

    method <- if (n_ref > 1L) {
      "multi"
    } else if (is.null(method)) {
      "calibration"
    } else {
      method
    }

    list(
      sc     = sc,
      sp_des = sp_des,
      sp     = sp,
      n_ref  = n_ref,
      weight = weight,
      method = method
    )

  }, error = function(e) {
    stop("Step 1 (input parsing and validation) failed: ", e$message, call. = FALSE)
  })

  sc     <- step1$sc
  sp_des <- step1$sp_des
  sp     <- step1$sp
  n_ref  <- step1$n_ref
  weight <- step1$weight
  method <- step1$method

  sc0 <- sc

  #--------------------------------------------------------------------------#
  # Step 2. Auto-build and preprocess p_formula
  #--------------------------------------------------------------------------#
  step2 <- tryCatch({

    log_messages <- character(0)

    if (is.null(p_formula)) {
      built <- p_formula_construction(
        sc     = sc,
        sp     = sp,
        weight = weight
      )

      p_formula    <- built$p_formula
      log_messages <- c(log_messages, built$log_messages)
    }

    list(
      p_formula    = p_formula,
      log_messages = log_messages
    )

  }, error = function(e) {
    stop("Step 2 (p_formula construction) failed: ", e$message, call. = FALSE)
  })

  p_formula    <- step2$p_formula
  log_messages <- step2$log_messages

  #--------------------------------------------------------------------------#
  # Step 3. NA processing
  #--------------------------------------------------------------------------#
  step3 <- tryCatch({

    process_na_build(
      sc        = sc,
      sp        = sp,
      sp_des    = sp_des,
      p_formula = p_formula,
      na.action = na.action,
      n_ref     = n_ref,
      verbose   = verbose
    )

  }, error = function(e) {
    stop("Step 3 (NA processing) failed: ", e$message, call. = FALSE)
  })

  na_mode       <- step3$na_mode
  sc            <- step3$sc
  sp            <- step3$sp
  sp_des        <- step3$sp_des
  keep_sc       <- step3$keep_sc
  keep_sp       <- step3$keep_sp
  n_sp_orig     <- step3$n_sp_orig
  na_action_obj <- step3$na_action_obj
  na_summary    <- step3$na_summary

  #--------------------------------------------------------------------------#
  # Step 4. Process p_formula
  #--------------------------------------------------------------------------#
  step4 <- tryCatch({

     processed <- process_p_formula(
      sc              = sc,
      sp              = sp,
      weight          = weight,
      Pre.calibration = precali,
      p_formula       = p_formula,
      sp_order        = sp_order,
      verbose         = verbose
    )

    list(
      sc           = processed$sc,
      sp           = processed$sp,
      p_vars       = processed$vars,
      log_messages = processed$log_messages
    )

  }, error = function(e) {
    stop("Step 4 (process p_formula) failed: ", e$message, call. = FALSE)
  })

  sc     <- step4$sc
  sp     <- step4$sp
  p_vars <- step4$p_vars

  if (!is.null(step4$log_messages)) {
    log_messages <- c(log_messages, step4$log_messages)
  }

  #--------------------------------------------------------------------------#
  # Step 5. Estimate pseudo-weights
  #--------------------------------------------------------------------------#
  result <- tryCatch({

    if (method %in% c("alp", "clw", "calibration")) {

      ipwm_one_build(
        sc           = sc,
        sp           = sp,
        sp_des       = sp_des,
        vars         = p_vars,
        weight       = weight,
        method       = method,
        control      = control,
        verbose      = verbose,
        log_messages = log_messages
      )

    } else if (method == "multi") {

      out <- ipwm_multi_build(
        sc           = sc,
        sp           = sp,
        sp_des       = sp_des,
        vars         = p_vars,
        weight       = weight,
        sp_order     = sp_order,
        control      = control,
        verbose      = verbose,
        log_messages = log_messages
      )

      out$method <- "multi"
      out

    } else {
      stop("Unknown method: ", method, call. = FALSE)
    }

  }, error = function(e) {
    stop("Step 5 (pseudo-weight estimation) failed: ", e$message, call. = FALSE)
  })

  #--------------------------------------------------------------------------#
  # Step 6. Construct sc_updated based on na_mode
  #--------------------------------------------------------------------------#
  sc_out <- tryCatch({

    reconstruct_sc_output(
      sc0           = sc0,
      w_fit         = result$pseudo_weights,
      keep_sc       = keep_sc,
      na_mode       = na_mode,
      na_action_obj = na_action_obj,
      sc_wname      = sc_wname
    )

  }, error = function(e) {
    stop("Step 6 (reconstruct sc_updated) failed: ", e$message, call. = FALSE)
  })

  #--------------------------------------------------------------------------#
  # Step 7. Finalize and return
  #--------------------------------------------------------------------------#
  result <- tryCatch({

    result <- finalize_pw_fit(
      result        = result,
      sc_out        = sc_out,
      sc0           = sc0,
      sc_wname      = sc_wname,
      na_mode       = na_mode,
      keep_sc       = keep_sc,
      na_action_obj = na_action_obj
    )

    result$call              <- match.call()
    result["na_summary"]     <- list(na_summary)

    result

  }, error = function(e) {
    stop("Step 7 (finalize output) failed: ", e$message, call. = FALSE)
  })

  invisible(result)
}

#' @keywords internal
#'
#' @description
#' \pkg{nonprobsampling} implements pseudo-weighting methods for
#' finite population inference from nonprobability samples, such as
#' convenience samples, volunteer cohorts, and opt-in panels. Because the
#' participation mechanism in a nonprobability sample is unknown, unadjusted
#' estimates of population means and prevalences may be biased. The package
#' addresses this issue by leveraging auxiliary information from one or more
#' probability reference surveys to estimate participation probabilities and
#' using their inverses as pseudo-weights to obtain bias-corrected estimates
#' of finite population means and prevalences.
#'
#' @details
#' The package implements the generalized estimating equation framework of
#' Landsman et al. (2026). Pseudo-weights are obtained by solving estimating
#' equations that equate auxiliary variable totals estimated from the
#' nonprobability sample with the corresponding totals estimated from one or
#' more probability reference surveys. Totals from the nonprobability sample are
#' computed using estimated pseudo-weights, whereas totals from the reference
#' surveys are computed using the known survey sampling weights.
#'
#' Several one reference methods arise as special cases of this framework under
#' different choices of the weight and estimating functions:
#' \itemize{
#'   \item the raking ratio calibration method of
#'     Landsman et al. (2026);
#'   \item the adjusted logistic propensity (ALP) method of
#'     Wang, Valliant, and Li (2021);
#'   \item the Chen--Li--Wu (CLW) method of Chen, Li, and Wu (2020).
#' }
#'
#' A key feature of the package is the multi-reference extension of the
#' calibration method. This extension enables the integration of auxiliary
#' information across multiple surveys when no single reference survey
#' contains all variables relevant to participation (Landsman et al., 2026),
#' with an optional cumulative precalibration step to preserve information on
#' the relationships between overlapping and unique auxiliary variables across
#' reference surveys.
#'
#' Variance estimation is based on Taylor linearization. The resulting analytic
#' variance estimator accounts for uncertainty from pseudo-weight estimation and
#' for the sampling designs of the reference surveys through integration with
#' the \pkg{survey} package. The package also supports bootstrap-based variance
#' estimation when bootstrap weights are provided with the reference surveys.
#'
#' @section Typical workflow:
#' Estimation is carried out in two steps:
#' \enumerate{
#'   \item \code{\link{est_pw}()} estimates pseudo-weights using the
#'     nonprobability sample and one or more reference survey design objects.
#'     This step fits the participation model and stores the internal quantities
#'     needed for subsequent estimation. It does not require an outcome variable.
#'   \item \code{\link{pwmean}()} uses the object returned by \code{est_pw()}
#'     to estimate a pseudo-weighted mean or prevalence for an outcome, either
#'     overall or by category. It also returns the standard error and confidence
#'     interval.
#' }
#'
#' Numerical settings for solving the estimating equations can be specified with
#' \code{\link{pw_solver_control}()}.
#'
#' @section Datasets:
#' The package includes example datasets for demonstrating one reference and
#' multiple reference analyses: \code{\link{sc}} is a nonprobability sample,
#' \code{\link{sp1}} and \code{\link{sp2}} are probability reference surveys,
#' and \code{\link{sp1_bootstrap}} contains replicate weights for
#' \code{sp1}. See the package vignette for worked examples.
#'
#' @references
#' Chen, Y., Li, P., and Wu, C. (2020).
#' Doubly robust inference with nonprobability survey samples.
#' \emph{Journal of the American Statistical Association},
#' 115(532), 2011--2021.
#' doi:10.1080/01621459.2019.1677241
#'
#' Landsman, V., Wang, L., Carrillo-Garcia, I., Mitani, A. A.,
#' Smith, P. M., Graubard, B. I., Bui, T., and Carnide, N. (2026).
#' Correction for participation bias in nonprobability samples using multiple
#' reference surveys.
#' \emph{Statistics in Medicine}, 45(3--5).
#' doi:10.1002/sim.70403
#'
#' Wang, L., Valliant, R., and Li, Y. (2021).
#' Adjusted logistic propensity weighting methods for population inference
#' using nonprobability volunteer-based epidemiologic cohorts.
#' \emph{Statistics in Medicine}, 40(24), 5237--5250.
#' doi:10.1002/sim.9122
#'
#' @seealso
#' \code{\link{est_pw}}, \code{\link{pwmean}}, \code{\link{pw_solver_control}}
#'
"_PACKAGE"

#' Summary method for pwmean objects
#'
#' Provides console output for objects of class \code{"pwmean"}, including
#' unweighted and pseudo-weighted mean estimates, standard errors, confidence
#' intervals, and optional domain-level summaries.
#'
#' @param object An object of class \code{"pwmean"}, returned by
#'   \code{\link{pwmean}}.
#' @param ... Additional arguments, currently unused.
#'
#' @return Invisibly returns \code{object}.
#'
#' @method summary pwmean
#' @export
summary.pwmean <- function(object, ...) {
  summary_pwmean_impl(
    object,
    group_label = "Domain",
    group_col = "domain",
    estimate_label = "Mean",
    estimate_col = "mean"
  )
}

#' Summary method for pwmean objects with categorical outcomes
#'
#' Provides console output for objects of class \code{"pwmean_factor"},
#' including unweighted and pseudo-weighted prevalence estimates, standard
#' errors, and confidence intervals.
#'
#' @param object An object of class \code{"pwmean_factor"}, returned by
#'   \code{\link{pwmean}} when \code{y} is a factor.
#' @param ... Additional arguments, currently unused.
#'
#' @return Invisibly returns \code{object}.
#'
#' @method summary pwmean_factor
#' @export
summary.pwmean_factor <- function(object, ...) {
  summary_pwmean_impl(
    object,
    group_label = "Category",
    group_col = "category",
    estimate_label = "Prevalence",
    estimate_col = "prevalence"
  )
}

summary_pwmean_impl <- function(
    object,
    group_label,
    group_col,
    estimate_label,
    estimate_col
) {

  if (!is.null(object$call)) {
    cat("Call:\n")
    print(object$call)
  } else {
    cat("Call:\n  (not available)\n")
  }

  method_raw <- trimws(as.character(object$method %||% ""))
  method_key <- tolower(method_raw)

  if (method_key == "alp") {
    method_label <- "One reference ALP"
    method_short <- "ALP"
  } else if (method_key == "clw") {
    method_label <- "One reference CLW"
    method_short <- "CLW"
  } else if (method_key %in% c("cali", "calibration")) {
    method_label <- "One reference calibration"
    method_short <- "calibration"
  } else if (method_key == "multi") {
    method_label <- "Multi-reference calibration"
    method_short <- "calibration"
  } else {
    method_label <- method_raw
    method_short <- method_raw
  }

  cat("\nMethod:", method_label, "\n")

  d <- object$estimates

  has_category <- "category" %in% names(d)

  if (nrow(d) == 1L && !has_category) {

    cat(sprintf("\n%s: %s\n", group_label, d$domain))

    cat("\nUnweighted estimators:\n")
    cat(sprintf("  %-15s %10.6f\n", paste0(estimate_label, ":"), d$unweighted_mean))
    cat(sprintf("  %-15s %10.6f\n", "Std. Error:",                d$unweighted_se))
    cat(sprintf("  %-15s [%0.6f, %0.6f]\n",
                "95% CI:", d$unweighted_lower, d$unweighted_upper))

    cat("\nPseudo-weighted (", method_short, ") estimators:\n", sep = "")
    cat(sprintf("  %-15s %10.6f\n", paste0(estimate_label, ":"), d$adjusted_mean))
    cat(sprintf("  %-15s %10.6f\n", "Std. Error:",                d$adjusted_se))
    cat(sprintf("  %-15s [%0.6f, %0.6f]\n",
                "95% CI:", d$adjusted_lower, d$adjusted_upper))

  } else {

    if (has_category) {
      uw_tab <- data.frame(
        d$category,
        d$domain,
        d$unweighted_mean,
        d$unweighted_se,
        d$unweighted_lower,
        d$unweighted_upper,
        stringsAsFactors = FALSE
      )
      names(uw_tab) <- c("category", "domain", estimate_col, "se", "CI_lower", "CI_upper")
    } else {
      uw_tab <- data.frame(
        d$domain,
        d$unweighted_mean,
        d$unweighted_se,
        d$unweighted_lower,
        d$unweighted_upper,
        stringsAsFactors = FALSE
      )
      names(uw_tab) <- c(group_col, estimate_col, "se", "CI_lower", "CI_upper")
    }

    cat("\nUnweighted estimators:\n")
    print(uw_tab, row.names = FALSE)

    if (has_category) {
      ad_tab <- data.frame(
        d$category,
        d$domain,
        d$adjusted_mean,
        d$adjusted_se,
        d$adjusted_lower,
        d$adjusted_upper,
        stringsAsFactors = FALSE
      )
      names(ad_tab) <- c("category", "domain", estimate_col, "se", "CI_lower", "CI_upper")
    } else {
      ad_tab <- data.frame(
        d$domain,
        d$adjusted_mean,
        d$adjusted_se,
        d$adjusted_lower,
        d$adjusted_upper,
        stringsAsFactors = FALSE
      )
      names(ad_tab) <- c(group_col, estimate_col, "se", "CI_lower", "CI_upper")
    }

    cat("\nPseudo-weighted (", method_short, ") estimators:\n", sep = "")
    print(ad_tab, row.names = FALSE)
  }

  na_act <- na.action(object)
  if (!is.null(na_act))
    cat("\n(", stats::naprint(na_act), ")\n", sep = "")

  invisible(object)
}

#' Compute design-based covariance matrix D for calibration
#'
#' Internal helper to compute
#' \deqn{D = Var_p\left( \sum_{j \in s_p} d_j x_j \right)}
#' from an existing survey design object created by
#' survey::svydesign() or survey::svrepdesign().
#'
#' In the raking ratio calibration method, the probability sample
#' contribution in the estimating equation is \eqn{d_j x_j}, so the
#' design-based covariance matrix is obtained by treating
#' \eqn{h_j = x_j} as the survey total variable.
#'
#' @param sp_des A survey design object of class `"survey.design2"` or
#'   `"svyrep.design"`.
#' @param Xp Numeric matrix of dimension `n_p x p`. Each row is the
#'   covariate vector `x_j` used in the calibration estimating equation.
#'
#' @return A `p x p` covariance matrix `D`.
#' @keywords internal
compute_D_raking <- function(sp_des, Xp) {

  if (!inherits(sp_des, c("survey.design2", "svyrep.design"))) {
    stop("'sp_des' must be a survey design object from svydesign() or svrepdesign().",
         call. = FALSE)
  }

  if (!is.matrix(Xp)) {
    Xp <- as.matrix(Xp)
  }

  n_sp <- nrow(sp_des$variables)
  if (nrow(Xp) != n_sp) {
    stop(sprintf(
      "Row mismatch: nrow(Xp) = %d but the survey design object contains %d observations.",
      nrow(Xp), n_sp
    ), call. = FALSE)
  }

  # safe column names for formula parsing
  xp_names <- colnames(Xp)
  if (is.null(xp_names)) {
    xp_names <- paste0("x", seq_len(ncol(Xp)))
  }
  xp_names <- make.names(xp_names, unique = TRUE)

  H_df <- as.data.frame(Xp)
  colnames(H_df) <- paste0("h_", xp_names)

  # attach estimating equation columns to the existing design object
  sp_des_h <- sp_des
  sp_des_h$variables <- cbind(sp_des_h$variables, H_df)

  # no intercept
  fml <- stats::reformulate(colnames(H_df), intercept = FALSE)

  tot <- survey::svytotal(fml, sp_des_h)
  D <- stats::vcov(tot)

  return(D)
}

#' Validate inputs and build variable sets for multi-reference calibration
#'
#' @param sc Data frame. The nonprobability sample.
#' @param sp_list List of reference sample data frames.
#' @param vars_list List of character vectors of predictor variable names
#'   shared between `sc` and each reference sample.
#' @param wts_cols Character vector of survey-weight column names,
#'   one per element of `sp_list`.
#' @param verbose Logical. If `TRUE`, variable-set messages are printed.
#'
#' @return A list with components:
#'   \item{sc}{The nonprobability sample data frame (unchanged).}
#'   \item{sp_list}{The reference sample list (unchanged).}
#'   \item{vars_XC}{Character vector: union of all per-sample variable sets,
#'     used to build the `sc` design matrix `Xc`.}
#'   \item{xcol}{List of character vectors: the variables contributed by each
#'     reference sample after removing variables already covered by earlier
#'     samples (no intercept).}
#'   \item{wts_cols}{Character vector of weight column names (unchanged).}
#'   \item{log}{Character vector of log messages describing the variable sets.}
#'
#' @keywords internal
check_input_multi <- function(sc,
                              sp_list,
                              vars_list,
                              wts_cols,
                              verbose = FALSE) {

  log_messages <- character()

  spn <- if (!is.null(names(sp_list)) && all(nzchar(names(sp_list)))) {
    names(sp_list)
  } else {
    paste0("sp[[", seq_along(sp_list), "]]")
  }


  # ---- ensure all three inputs have the same length ----
  if (length(sp_list) != length(vars_list) ||
      length(sp_list) != length(wts_cols)) {
    stop("sp_list, vars_list, and wts_cols must all be the same length.",
         call. = FALSE)
  }

  # ============================================================
  #              PRINT SHARED VARIABLES PER SAMPLE
  # ============================================================
  for (i in seq_along(sp_list)) {
    msg <- sprintf(
      "Shared variables in %s:\n  %s\n",
      spn[i],
      paste(vars_list[[i]], collapse = ", ")
    )
    log_messages <- c(log_messages, msg)
  }


  # ============================================================
  #             BUILD xcol (NEW VARIABLES PER SAMPLE)
  # ============================================================
  n    <- length(vars_list)
  xcol <- vector("list", n)
  xcol[[1]] <- vars_list[[1]]

  if (n > 1) {
    for (i in seq.int(2, n)) {
      prev_vars <- Reduce(union, xcol[1:(i - 1)])
      new_vars  <- setdiff(vars_list[[i]], prev_vars)

      if (length(new_vars) == 0) {
        stop(sprintf("%s has no new variables; please remove this sample.", spn[i]),
             call. = FALSE)
      }
      xcol[[i]] <- new_vars
    }
  }

  for (i in seq_along(xcol)) {
    msg <- sprintf(
      "Variables used for calculation in %s:\n  %s\n",
      spn[i],
      paste(xcol[[i]], collapse = ", ")
    )
    log_messages <- c(log_messages, msg)
  }


  # ---- union of all xcol sets ----
  vars_XC <- Reduce(union, xcol)

  # ---- return validated objects ----
  list(
    sc       = sc,
    sp_list  = sp_list,
    vars_XC  = vars_XC,
    xcol     = xcol,
    wts_cols = wts_cols,
    log      = log_messages
  )
}

#' Estimate step for the ALP estimator
#'
#' Computes the domain-specific pseudo-weighted Hájek mean and its
#' Taylor-linearized variance for the adjusted logistic propensity estimator.
#'
#' @param Y Outcome vector for the nonprobability sample.
#' @param Z Domain indicator vector. Use `rep(1, length(Y))` for the
#'   overall mean.
#' @param w Estimated pseudo-weights from the ALP build step.
#' @param X Design matrix for the nonprobability sample.
#' @param D Design-based variance-covariance matrix of the estimated
#'   auxiliary totals from the reference survey.
#' @param S_beta Sensitivity matrix for the ALP estimating equations.
#'
#' @return A list with components `mean` and `variance`.
#'
#' @keywords internal
alp_estimate <- function(Y, Z, w, X, D, S_beta) {

  # --- pseudo-weighted Hajek mean ---
  T1 <- sum((Y * Z) * w)
  T2 <- sum(Z * w)
  mu <- T1 / T2

  # --- linearization coefficient for beta estimation ---
  U_beta <- t((Y * Z) - mu * Z) %*% (w * X)
  b_vec  <- U_beta %*% qr.solve(S_beta)   # 1 x p

  # --- linearized residual after accounting for beta estimation ---
  p_sc  <- 1 / (1 + w)
  resid <- (Y * Z) - mu * Z - p_sc * as.vector(X %*% t(b_vec))

  # v1: nonprobability sample component under the Poisson approximation
  v1 <- sum(w * (w - 1) * (resid^2))

  # v2: reference survey component from estimated auxiliary totals
  v2 <- b_vec %*% D %*% t(b_vec)

  variance <- as.vector((v1 + v2) / T2^2)

  list(
    mean = mu,
    variance = variance
  )
}

#' Extract analysis data from a survey design object
#'
#' Remove design variables (cluster IDs, strata, FPC, probabilities, weights)
#' from a survey design object and attaches the sampling weights as a single
#' column, returning a plain data frame ready for modeling.
#'
#' Cluster and strata variables are identified from the design object's own
#' fields (`des$cluster`, `des$strata`) rather than the original call, so the
#' result is correct even when `des` has been subset after construction.
#' Weights, FPC, and probability variables are still read from `des$call`.
#' For replicate designs (`svyrep.design`), replicate weight columns are also
#' dropped.
#'
#' @param des A survey design object of class `survey.design2`
#'   or `svyrep.design`.
#' @param weight_name A single string giving the name of the weight column
#'   added to the output. Defaults to `"sp_wts"`. Must not already exist
#'   among the analysis variables after design variables are removed.
#'
#' @return A data frame containing the analysis variables from `des$variables`
#'   (design variables removed) plus one column named `weight_name` storing
#'   the sampling weights.
#'
#' @keywords internal
extract_analysis_data <- function(des, weight_name = "sp_wts") {

  if (!inherits(des, c("survey.design2", "svyrep.design"))) {
    stop("`des` must be a survey design object.", call. = FALSE)
  }

  vars <- des$variables

  if (is.null(vars) || !is.data.frame(vars)) {
    stop("`des$variables` is missing or not a data.frame.", call. = FALSE)
  }

  get_call_vars <- function(x) {
    if (is.null(x)) return(character(0))
    out <- tryCatch(all.vars(x), error = function(e) character(0))
    setdiff(out, "1")
  }

  # Use design fields for cluster/strata -- robust after subsetting
  cluster_vars <- if (!is.null(des$cluster)) {
    setdiff(names(des$cluster), "1")
  } else {
    get_call_vars(des$call$ids)
  }

  strata_vars <- if (!is.null(des$strata)) {
    setdiff(names(des$strata), "1")
  } else {
    get_call_vars(des$call$strata)
  }

  # Weights, FPC, probs have no direct name field -- parse from call
  call_vars <- unique(unlist(lapply(
    list(des$call$fpc, des$call$probs, des$call$weights),
    get_call_vars
  )))

  drop_vars <- unique(c(cluster_vars, strata_vars, call_vars))

  if (inherits(des, "svyrep.design")) {
    rep_cols <- tryCatch(colnames(des$repweights), error = function(e) NULL)
    if (!is.null(rep_cols)) {
      drop_vars <- unique(c(drop_vars, rep_cols))
    }
  }

  drop_vars <- intersect(drop_vars, names(vars))
  out <- vars[, !(names(vars) %in% drop_vars), drop = FALSE]

  if (weight_name %in% names(out)) {
    stop(sprintf(
      "Column '%s' already exists in the analysis data. Use a different `weight_name`.",
      weight_name
    ), call. = FALSE)
  }

  if (inherits(des, "svyrep.design")) {
    out[[weight_name]] <- as.numeric(weights(des, type = "sampling"))
  } else {
    out[[weight_name]] <- as.numeric(weights(des))
  }

  out
}

#' Compute the naive mean for one domain
#'
#' Computes the unweighted sample mean and its variance for observations
#' belonging to a single domain, identified by a 0/1 indicator vector.
#'
#' @param yvec Numeric or integer outcome vector of length \eqn{n}.
#' @param zvec Integer 0/1 domain indicator of length \eqn{n}, or NULL
#'   (treated as all-ones, i.e., the overall mean).
#'
#' @return A list with components:
#'
#' - `mean`: unweighted sample mean of `yvec` within the domain,
#'   or NA if the domain is empty.
#' - `variance`: estimated variance of the mean (\eqn{s^2/n}),
#'   or NA if fewer than two observations are available.
#'
#' @keywords internal
naive_mean_one_domain <- function(yvec, zvec = NULL) {

  if (!is.numeric(yvec)) {
    stop("`yvec` must be numeric.", call. = FALSE)
  }

  if (is.null(zvec)) {
    zvec <- rep.int(1L, length(yvec))
  }

  if (length(zvec) != length(yvec)) {
    stop("Length mismatch between `yvec` and `zvec`.", call. = FALSE)
  }

  keep <- (zvec == 1L)
  y_sub <- yvec[keep]
  n <- length(y_sub)

  out <- list(
    mean     = if (n > 0L) mean(y_sub) else NA_real_,
    variance = if (n > 1L) stats::var(y_sub) / n else NA_real_
  )

  out
}


#' Compute naive (unweighted) means from the convenience sample
#'
#' Filters the convenience sample to complete cases, then computes the
#' unweighted sample mean (and its variance) for each domain using
#' `naive_mean_one_domain()`. The domain structure is standardized via
#' `standardize_zcol()`.
#'
#' @param df A data frame containing the convenience sample
#'   (typically `build$internal$raw_sc`).
#' @param domain_var Single character string naming the domain variable in
#'   `df`, or NULL for the overall mean.
#' @param y Single character string naming the outcome variable in `df`.
#'
#' @return A list with components:
#'
#' - `type`: `"single"` for overall or binary domains;
#'   `"multi"` for factor/character domains.
#' - `labels`: character vector of domain labels.
#' - `estimates`: for `type = "single"`, a list with `mean` and `variance`.
#'   For `type = "multi"`, a list of such lists, one per domain level.
#'
#' @keywords internal
naive_mean <- function(df, domain_var = NULL, y) {

  #----------------------------------#
  # Step 1: basic checks
  #----------------------------------#
  if (!is.data.frame(df)) {
    stop("`df` must be a data.frame.", call. = FALSE)
  }

  if (!is.character(y) || length(y) != 1L || is.na(y) || !nzchar(y)) {
    stop("`y` must be a single non-empty character string.", call. = FALSE)
  }

  if (!y %in% names(df)) {
    stop(sprintf("Outcome variable '%s' not found in `df`.", y), call. = FALSE)
  }

  #----------------------------------#
  # Step 2: complete-case filtering for y and domain_var
  #----------------------------------#
  if (is.null(domain_var)) {
    ok <- !is.na(df[[y]])
  } else {
    if (!is.character(domain_var) || length(domain_var) != 1L ||
        is.na(domain_var) || !nzchar(domain_var)) {
      stop("`domain_var` must be NULL or a single non-empty character string.",
           call. = FALSE)
    }

    if (!domain_var %in% names(df)) {
      stop(sprintf("Domain variable '%s' not found in `df`.", domain_var),
           call. = FALSE)
    }

    ok <- !is.na(df[[y]]) & !is.na(df[[domain_var]])
  }

  df_cc <- df[ok, , drop = FALSE]
  yvec <- df_cc[[y]]

  #----------------------------------#
  # Step 3: standardize domain variable
  #----------------------------------#
  domain_info <- standardize_zcol(data = df_cc, zcol = domain_var)

  #----------------------------------#
  # Step 4: dispatch by domain mode
  #----------------------------------#
  if (identical(domain_info$mode, "overall")) {
    est <- naive_mean_one_domain(yvec = yvec, zvec = NULL)

    return(list(
      type = "single",
      labels = "Overall",
      estimates = est
    ))
  }

  if (identical(domain_info$mode, "binary")) {
    zvec <- domain_info$indicators[[1]]
    est <- naive_mean_one_domain(yvec = yvec, zvec = zvec)

    return(list(
      type = "single",
      labels = domain_info$labels,
      estimates = est
    ))
  }

  if (identical(domain_info$mode, "factor")) {
    est_list <- lapply(seq_along(domain_info$labels), function(j) {
      naive_mean_one_domain(
        yvec = yvec,
        zvec = domain_info$indicators[[j]]
      )
    })

    return(list(
      type = "multi",
      labels = domain_info$labels,
      estimates = est_list
    ))
  }

  stop("Unsupported domain mode returned by `standardize_zcol()`.",
       call. = FALSE)
}

# helpers ----
#
# Synthetic data frame with known analytical results.
#
# y = c(1, 2, 4, 7)
#
# overall:
#   mean     = (1+2+4+7)/4 = 3.5
#   variance = var(c(1,2,4,7)) / 4 = 7 / 4 = 1.75
#
# grp_bin (indicator z = c(1,1,0,0)):
#   y_sub = c(1, 2),  mean = 1.5
#   variance = var(c(1,2)) / 2 = 0.5 / 2 = 0.25
#
# grp_fac / grp_chr level A (obs 1, 3): y_sub = c(1, 4)
#   mean = 2.5,  variance = var(c(1,4)) / 2 = 4.5 / 2 = 2.25
#
# grp_fac / grp_chr level B (obs 2, 4): y_sub = c(2, 7)
#   mean = 4.5,  variance = var(c(2,7)) / 2 = 12.5 / 2 = 6.25

make_naive_df <- function() {
  data.frame(
    y       = c(1.0, 2.0, 4.0, 7.0),
    grp_bin = c(1L, 1L, 0L, 0L),
    grp_fac = factor(c("A", "B", "A", "B")),
    grp_chr = c("A", "B", "A", "B"),
    stringsAsFactors = FALSE
  )
}

df <- make_naive_df()


# naive_mean_one_domain: output structure ----

test_that("naive_mean_one_domain: returns a named list with mean and variance", {
  res <- naive_mean_one_domain(yvec = c(1, 2, 4, 7))
  expect_type(res, "list")
  expect_named(res, c("mean", "variance"))
})

test_that("naive_mean_one_domain: mean and variance are length-1 scalars", {
  res <- naive_mean_one_domain(yvec = c(1, 2, 4, 7))
  expect_length(res$mean,     1L)
  expect_length(res$variance, 1L)
})


# naive_mean_one_domain: zvec = NULL (overall) ----

test_that("naive_mean_one_domain: zvec = NULL uses all observations for the mean", {
  res <- naive_mean_one_domain(yvec = c(1, 2, 4, 7))
  expect_equal(res$mean, 3.5)
})

test_that("naive_mean_one_domain: zvec = NULL gives variance = var(y) / n", {
  res <- naive_mean_one_domain(yvec = c(1, 2, 4, 7))
  expect_equal(res$variance, 1.75)
})


# naive_mean_one_domain: domain subset ----

test_that("naive_mean_one_domain: zvec = c(1,1,0,0) selects the first two observations", {
  res <- naive_mean_one_domain(yvec = c(1, 2, 4, 7), zvec = c(1L, 1L, 0L, 0L))
  expect_equal(res$mean, 1.5)
})

test_that("naive_mean_one_domain: zvec = c(1,1,0,0) gives variance = var(c(1,2)) / 2", {
  res <- naive_mean_one_domain(yvec = c(1, 2, 4, 7), zvec = c(1L, 1L, 0L, 0L))
  expect_equal(res$variance, 0.25)
})


# naive_mean_one_domain: edge cases ----

test_that("naive_mean_one_domain: empty domain (all zeros) returns NA mean and NA variance", {
  res <- naive_mean_one_domain(yvec = c(1, 2, 4, 7), zvec = c(0L, 0L, 0L, 0L))
  expect_true(is.na(res$mean))
  expect_true(is.na(res$variance))
})

test_that("naive_mean_one_domain: single observation returns valid mean and NA variance", {
  res <- naive_mean_one_domain(yvec = c(5.0), zvec = c(1L))
  expect_equal(res$mean, 5.0)
  expect_true(is.na(res$variance))
})


# naive_mean_one_domain: input validation ----

test_that("naive_mean_one_domain: rejects non-numeric yvec", {
  expect_error(
    naive_mean_one_domain(yvec = c("a", "b")),
    "`yvec` must be numeric.",
    fixed = TRUE
  )
})

test_that("naive_mean_one_domain: rejects zvec with different length than yvec", {
  expect_error(
    naive_mean_one_domain(yvec = c(1, 2, 4, 7), zvec = c(1L, 0L)),
    "Length mismatch between `yvec` and `zvec`.",
    fixed = TRUE
  )
})


# naive_mean: output structure (overall) ----

test_that("naive_mean: overall returns list with type, labels, estimates", {
  res <- naive_mean(df, domain_var = NULL, y = "y")
  expect_named(res, c("type", "labels", "estimates"))
})

test_that("naive_mean: overall returns type = 'single'", {
  res <- naive_mean(df, domain_var = NULL, y = "y")
  expect_equal(res$type, "single")
})

test_that("naive_mean: overall returns label 'Overall'", {
  res <- naive_mean(df, domain_var = NULL, y = "y")
  expect_equal(res$labels, "Overall")
})

test_that("naive_mean: overall estimates has mean and variance", {
  res <- naive_mean(df, domain_var = NULL, y = "y")
  expect_named(res$estimates, c("mean", "variance"))
})


# naive_mean: mean and variance correctness (overall) ----

test_that("naive_mean: overall mean equals 3.5", {
  res <- naive_mean(df, domain_var = NULL, y = "y")
  expect_equal(res$estimates$mean, 3.5)
})

test_that("naive_mean: overall variance equals 1.75", {
  res <- naive_mean(df, domain_var = NULL, y = "y")
  expect_equal(res$estimates$variance, 1.75)
})


# naive_mean: binary domain ----

test_that("naive_mean: binary domain returns type = 'single'", {
  res <- naive_mean(df, domain_var = "grp_bin", y = "y")
  expect_equal(res$type, "single")
})

test_that("naive_mean: binary domain returns label 'grp_bin = 1'", {
  res <- naive_mean(df, domain_var = "grp_bin", y = "y")
  expect_equal(res$labels, "grp_bin = 1")
})

test_that("naive_mean: binary domain estimates has mean and variance", {
  res <- naive_mean(df, domain_var = "grp_bin", y = "y")
  expect_named(res$estimates, c("mean", "variance"))
})

test_that("naive_mean: binary domain mean equals 1.5", {
  res <- naive_mean(df, domain_var = "grp_bin", y = "y")
  expect_equal(res$estimates$mean, 1.5)
})

test_that("naive_mean: binary domain variance equals 0.25", {
  res <- naive_mean(df, domain_var = "grp_bin", y = "y")
  expect_equal(res$estimates$variance, 0.25)
})


# naive_mean: factor domain ----

test_that("naive_mean: factor domain returns type = 'multi'", {
  res <- naive_mean(df, domain_var = "grp_fac", y = "y")
  expect_equal(res$type, "multi")
})

test_that("naive_mean: factor domain has two labels and two estimates", {
  res <- naive_mean(df, domain_var = "grp_fac", y = "y")
  expect_length(res$labels,    2L)
  expect_length(res$estimates, 2L)
})

test_that("naive_mean: factor domain labels include the domain variable name", {
  res <- naive_mean(df, domain_var = "grp_fac", y = "y")
  expect_equal(res$labels, c("grp_fac = A", "grp_fac = B"))
})

test_that("naive_mean: factor domain each estimate has mean and variance", {
  res <- naive_mean(df, domain_var = "grp_fac", y = "y")
  for (est in res$estimates) {
    expect_named(est, c("mean", "variance"))
  }
})

test_that("naive_mean: factor domain level A mean equals 2.5", {
  res   <- naive_mean(df, domain_var = "grp_fac", y = "y")
  idx_A <- which(res$labels == "grp_fac = A")
  expect_equal(res$estimates[[idx_A]]$mean, 2.5)
})

test_that("naive_mean: factor domain level A variance equals 2.25", {
  res   <- naive_mean(df, domain_var = "grp_fac", y = "y")
  idx_A <- which(res$labels == "grp_fac = A")
  expect_equal(res$estimates[[idx_A]]$variance, 2.25)
})

test_that("naive_mean: factor domain level B mean equals 4.5", {
  res   <- naive_mean(df, domain_var = "grp_fac", y = "y")
  idx_B <- which(res$labels == "grp_fac = B")
  expect_equal(res$estimates[[idx_B]]$mean, 4.5)
})

test_that("naive_mean: factor domain level B variance equals 6.25", {
  res   <- naive_mean(df, domain_var = "grp_fac", y = "y")
  idx_B <- which(res$labels == "grp_fac = B")
  expect_equal(res$estimates[[idx_B]]$variance, 6.25)
})


# naive_mean: character domain ----

test_that("naive_mean: character domain returns type = 'multi'", {
  res <- naive_mean(df, domain_var = "grp_chr", y = "y")
  expect_equal(res$type, "multi")
})

test_that("naive_mean: character domain has two labels and two estimates", {
  res <- naive_mean(df, domain_var = "grp_chr", y = "y")
  expect_length(res$labels,    2L)
  expect_length(res$estimates, 2L)
})

test_that("naive_mean: character domain labels include the domain variable name", {
  res <- naive_mean(df, domain_var = "grp_chr", y = "y")
  expect_equal(res$labels, c("grp_chr = A", "grp_chr = B"))
})

test_that("naive_mean: character domain each estimate has mean and variance", {
  res <- naive_mean(df, domain_var = "grp_chr", y = "y")
  for (est in res$estimates) {
    expect_named(est, c("mean", "variance"))
  }
})


# naive_mean: NA filtering ----

test_that("naive_mean: NA in y is silently dropped before computing mean", {
  df_na       <- make_naive_df()
  df_na$y[1L] <- NA_real_
  # After dropping obs 1, y = c(2, 4, 7), overall mean = 13/3
  res <- naive_mean(df_na, domain_var = NULL, y = "y")
  expect_equal(res$estimates$mean, 13 / 3)
})

test_that("naive_mean: NA in domain_var is silently dropped before computing", {
  df_na             <- make_naive_df()
  df_na$grp_fac[1L] <- NA
  # Obs 1 dropped; level A keeps only obs 3 (y=4): single obs -> NA variance
  res   <- naive_mean(df_na, domain_var = "grp_fac", y = "y")
  idx_A <- which(res$labels == "grp_fac = A")
  expect_equal(res$estimates[[idx_A]]$mean, 4.0)
  expect_true(is.na(res$estimates[[idx_A]]$variance))
})


# naive_mean: input validation ----

test_that("naive_mean: rejects non-data.frame df", {
  expect_error(
    naive_mean(list(y = c(1, 2)), domain_var = NULL, y = "y"),
    "`df` must be a data.frame.",
    fixed = TRUE
  )
})

test_that("naive_mean: rejects non-character y", {
  expect_error(
    naive_mean(df, domain_var = NULL, y = 1L),
    "`y` must be a single non-empty character string.",
    fixed = TRUE
  )
})

test_that("naive_mean: rejects NA y", {
  expect_error(
    naive_mean(df, domain_var = NULL, y = NA_character_),
    "`y` must be a single non-empty character string.",
    fixed = TRUE
  )
})

test_that("naive_mean: rejects empty string y", {
  expect_error(
    naive_mean(df, domain_var = NULL, y = ""),
    "`y` must be a single non-empty character string.",
    fixed = TRUE
  )
})

test_that("naive_mean: rejects y not found in df", {
  expect_error(
    naive_mean(df, domain_var = NULL, y = "not_a_col"),
    "Outcome variable 'not_a_col' not found in `df`.",
    fixed = TRUE
  )
})

test_that("naive_mean: rejects non-character domain_var", {
  expect_error(
    naive_mean(df, domain_var = 1L, y = "y"),
    "`domain_var` must be NULL or a single non-empty character string.",
    fixed = TRUE
  )
})

test_that("naive_mean: rejects empty string domain_var", {
  expect_error(
    naive_mean(df, domain_var = "", y = "y"),
    "`domain_var` must be NULL or a single non-empty character string.",
    fixed = TRUE
  )
})

test_that("naive_mean: rejects domain_var not found in df", {
  expect_error(
    naive_mean(df, domain_var = "not_a_col", y = "y"),
    "Domain variable 'not_a_col' not found in `df`.",
    fixed = TRUE
  )
})

#' Estimate Pseudo-Weights for Nonprobability Samples
#'
#' @description
#' \code{est_pw()} estimates pseudo-weights for a nonprobability sample using
#' one reference survey or multiple reference surveys. The function specifies the
#' participation model, handles missing values in the participation model
#' variables, solves the estimating equations, and stores the quantities needed
#' for downstream point and variance estimation.
#'
#' Users should harmonize variable names and coding before calling
#' \code{est_pw()}. Variables used in the participation model must have
#' consistent names and compatible definitions across the nonprobability sample
#' and the reference survey data used for estimation.
#'
#' With one reference survey, the available methods include the
#' raking ratio calibration method described in Landsman et al. (2026),
#' the adjusted logistic propensity weighting (ALP) method proposed by Wang, Valliant,
#' and Li (2021), and the CLW method proposed by Chen, Li, and Wu (2020).
#' With multiple reference surveys, pseudo-weights are estimated using the
#' multi-reference calibration method proposed by Landsman et al. (2026).
#'
#' The returned object is designed to be passed to \code{\link{pwmean}}.
#'
#' @details
#' \code{est_pw()} performs pseudo-weight estimation for the nonprobability
#' sample and stores the method-specific internal objects needed later by
#' \code{\link{pwmean}}. It does not require an outcome variable.
#'
#' The input \code{data} must be provided as a list, where the first element is
#' the nonprobability sample and the remaining elements are reference survey
#' design objects. Reference survey designs can be created with
#' \code{\link[survey]{svydesign}} for standard complex survey designs or
#' \code{\link[survey]{svrepdesign}} for surveys with replicate weights. These
#' objects preserve the sampling structure needed for design-consistent
#' variance estimation.
#'
#' \strong{Variable harmonization.}
#' Variables are matched by name, not by meaning. Before applying
#' \code{est_pw()}, shared variables must be harmonized across the
#' nonprobability sample and reference survey data. For example, if a
#' categorical variable is named \code{agecat} in the nonprobability sample and
#' \code{age_group} in the reference survey, the user should rename one of the
#' variables before estimation.
#'
#' Categorical variables should be encoded as factors with compatible category
#' definitions and identical levels in the same order. Even when categories are
#' substantively equivalent, mismatched factor levels may cause
#' \code{est_pw()} to return an error. Continuous variables included in the
#' participation model should also be measured on comparable scales across
#' datasets.
#'
#' Internally, \code{est_pw()} performs the following steps:
#' \enumerate{
#'   \item \strong{Input validation}\cr
#'     Validates the structure and required components of the input data.
#'   \item \strong{Reference survey detection}\cr
#'     Determines whether the input contains a single reference survey or
#'     multiple reference surveys.
#'   \item \strong{Method selection}\cr
#'     Selects the pseudo-weighting method based on the specified argument(s).
#'   \item \strong{Participation model specification}\cr
#'     Constructs a default participation model formula when
#'     \code{p_formula = NULL}.
#'   \item \strong{Missing data handling}\cr
#'     Applies missing-data handling procedures to variables used in the
#'     participation model.
#'   \item \strong{Model matrix construction}\cr
#'     Generates model matrices from the participation model variables.
#'   \item \strong{Pseudo-weight estimation}\cr
#'     Estimates pseudo-weights using the selected method.
#'   \item \strong{Output augmentation}\cr
#'     Appends the estimated pseudo-weights as a new column to the
#'     nonprobability sample.
#'   \item \strong{Metadata storage}\cr
#'     Stores information related to missing-data handling and other internal
#'     objects for later use or diagnostics.
#' }
#'
#' @section One-reference method and multi-reference method:
#' If \code{data} contains one reference survey design object, \code{est_pw()}
#' fits a one-reference method. If \code{data} contains more
#' than one reference survey design objects, \code{est_pw()} fits the
#' multi-reference calibration method.
#' In both settings, the auxiliary variables used for pseudo-weight estimation
#' should be harmonized across all data sources before calling
#' \code{est_pw()}.
#'
#' @section Multi-reference precalibration:
#' When \code{precali = TRUE}, cumulative precalibration is performed before the
#' main multi-reference calibration step. For overlapping
#' auxiliary variables, this procedure calibrates the survey weights of a
#' reference survey so that its weighted totals of the overlapping variables
#' and its sum of weights match the corresponding totals from the preceding
#' reference survey in the cumulative order. If there are no overlapping
#' auxiliary variables, cumulative precalibration is applied only to the sum of
#' weights.
#'
#' The order of the reference surveys is controlled by \code{sp_order}. If
#' \code{sp_order = "size"}, reference surveys are ordered by sample size, from
#' largest to smallest. If \code{sp_order = "given"}, the user-specified order
#' of the reference surveys is used.
#'
#' Cumulative precalibration is based only on overlapping variables that are
#' specified in \code{p_formula}, rather than on all overlapping variables in
#' the reference surveys. This choice avoids excluding observations because
#' of missing values in variables that are not used for pseudo-weight
#' estimation.
#'
#' @section Missing data handling:
#' Missing values are handled only for variables used in the participation
#' model. The selected \code{na.action} is recorded in the returned object,
#' together with the row indices of the nonprobability sample observations
#' retained for pseudo-weight estimation.
#'
#' With \code{stats::na.omit}, rows with missing participation model variables
#' are removed from \code{sc_updated}. With \code{stats::na.exclude}, the
#' original rows are retained in \code{sc_updated}, but excluded rows receive
#' \code{NA} in the pseudo-weight column. This can be useful when users want to
#' preserve row alignment with the original nonprobability sample for later imputation or merging.
#'
#' @section Numerical control:
#' Numerical settings are supplied through the \code{control} argument, which
#' should be created by \code{\link{pw_solver_control}}. This object controls
#' solver choice, convergence tolerance, maximum iterations, tracing, and
#' optional solver-specific arguments.
#'
#' The top-level \code{ftol}, \code{xtol}, and \code{maxit} values in
#' \code{\link{pw_solver_control}} are the package-level convergence controls
#' used by pseudo-weight estimation stages. When the selected solver is
#' \code{"nleqslv"}, additional arguments can be passed through
#' \code{nleqslv_control}. These are forwarded to \code{nleqslv::nleqslv()}.
#'
#' @usage
#' est_pw(
#'   data,
#'   sp_order = c("size", "given"),
#'   precali = TRUE,
#'   p_formula = NULL,
#'   method = NULL,
#'   na.action = stats::na.omit,
#'   sc_wname = "pseudo_wts",
#'   control = pw_solver_control(),
#'   verbose = FALSE
#' )
#'
#' @param data A list of input data objects of the form
#'   \code{list(sc, sp1_design, sp2_design, ...)}. The first element must be a
#'   data frame corresponding to the nonprobability sample. Each remaining
#'   element must be a valid survey design object corresponding to a reference
#'   probability survey, such as an object created by
#'   \code{\link[survey]{svydesign}} or \code{\link[survey]{svrepdesign}}.
#'
#' @param sp_order Character string controlling the order of reference surveys
#'   when multiple reference surveys are used. Supported values are
#'   \code{"size"} and \code{"given"}. \code{"size"} orders reference surveys
#'   by sample size, from largest to smallest. \code{"given"} uses the
#'   user-specified order of the reference surveys in \code{data}. Default is
#'   \code{"size"}. With one reference survey, this argument is ignored; a
#'   warning is issued if a non-default value is supplied.
#'
#' @param precali Logical. Used only with multiple reference surveys. If
#'   \code{TRUE}, cumulative precalibration is applied before the main
#'   multi-reference estimation step; see the
#'   \strong{Multi-reference precalibration} section for details. Default is
#'   \code{TRUE}. With one reference survey, this argument is ignored; a
#'   warning is issued if \code{FALSE} is supplied.
#'
#' @param p_formula Optional participation model formula. Must always be
#'   one-sided (no response variable on the left-hand side). A two-sided formula
#'   such as \code{y ~ x} will raise an error.
#'
#'   With one reference survey, supply a single one-sided formula, for example
#'   \code{~ age + sex + income}. With multiple reference surveys, supply a
#'   list of one-sided formulas with one formula per reference survey, for
#'   example \code{list(~ age + sex, ~ age + income)}. If \code{NULL}, a
#'   default formula is constructed automatically from variables shared across
#'   the data sources used for estimation. Since shared variables are identified
#'   by name, their names should be harmonized across data sources before
#'   estimation.
#'
#' @param method Character string specifying the pseudo-weighting method, or
#'   \code{NULL} (default). If \code{NULL}, \code{"calibration"} is used when
#'   \code{data} contains one reference survey, and \code{"multi"} is used when
#'   \code{data} contains more than one reference survey.
#'
#'   To override the default, supply one of the following values. For a
#'   one-reference method: \code{"alp"}, \code{"clw"}, or
#'   \code{"calibration"} (or \code{"cali"}). For the
#'   multi-reference method: \code{"multi"}.
#'
#'   The argument is case-insensitive, so inputs such as
#'   \code{"ALP"}, \code{"Clw"}, or \code{"CALI"} are also accepted.
#'
#' @param na.action Function specifying how missing values should be handled for
#'   variables used in the participation model. Common choices include
#'   \code{stats::na.omit}, \code{stats::na.exclude}, and
#'   \code{stats::na.fail}. Default is \code{stats::na.omit}.
#'
#' @param sc_wname Character string giving the name of the pseudo-weight column
#'   added to the returned nonprobability sample. Default is
#'   \code{"pseudo_wts"}. An error is raised at input validation if this name
#'   already exists as a column in \code{sc}.
#'
#' @param control A solver control object created by
#'   \code{\link{pw_solver_control}}. This object stores numerical settings
#'   for solving estimating equations, including the solver, convergence
#'   tolerance, maximum number of iterations, tracing behavior, and other
#'   options.
#'
#' @param verbose Logical. If \code{TRUE}, progress messages and diagnostics are
#'   printed during pseudo-weight estimation. Default is \code{FALSE}. Must be a
#'   single \code{TRUE} or \code{FALSE}; an error is raised otherwise.
#'
#' @return
#' An object of class \code{"pw_fit"}. This is a list containing user-facing
#' outputs and internal objects required by \code{\link{pwmean}}.
#'
#' Important components include:
#'
#' \describe{
#'   \item{\code{sc_updated}}{
#'     A data frame containing the nonprobability sample with an added
#'     pseudo-weight column named by \code{sc_wname}.
#'   }
#'   \item{\code{pseudo_weights}}{
#'     The estimated pseudo-weight vector. With \code{stats::na.omit}, the
#'     vector contains only observations retained for pseudo-weight estimation.
#'     With \code{stats::na.exclude}, excluded observations receive \code{NA}
#'     and the vector has length \code{nrow(sc)}.
#'   }
#'   \item{\code{coefficients}}{
#'     Estimated coefficients for the participation model variables.
#'   }
#'   \item{\code{solver_diagnostics}}{
#'     A list of solver diagnostics: \code{solver} (solver name),
#'     \code{termcd} (termination code), \code{message} (solver message),
#'     \code{iter} (number of iterations), and \code{fmax} (maximum absolute
#'     value of the final estimating equations at convergence).
#'   }
#'   \item{\code{method}}{
#'     The pseudo-weighting method used by the function.
#'   }
#'   \item{\code{internal}}{
#'     A list of internal objects needed for downstream estimation.
#'   }
#'   \item{\code{na_summary}}{
#'     An object of class \code{"pw_na_summary"} summarizing the number of rows
#'     excluded from the nonprobability sample and each reference survey due to
#'     missing participation model variables. \code{NULL} if no rows were
#'     excluded.
#'   }
#'   \item{\code{call}}{
#'     The matched function call.
#'   }
#' }
#'
#' @seealso
#' \code{\link{pw_solver_control}},
#' \code{\link{pwmean}}
#'
#' @references
#' Chen, Y., Li, P., and Wu, C. (2020).
#' Doubly robust inference with nonprobability survey samples.
#' \emph{Journal of the American Statistical Association},
#' 115(532), 2011--2021.
#' doi:10.1080/01621459.2019.1677241
#'
#' Wang, L., Valliant, R., and Li, Y. (2021).
#' Adjusted logistic propensity weighting methods for population inference
#' using nonprobability volunteer-based epidemiologic cohorts.
#' \emph{Statistics in Medicine}, 40(24), 5237--5250.
#' doi:10.1002/sim.9122
#'
#' Landsman, V., Wang, L., Carrillo-Garcia, I., Mitani, A. A.,
#' Smith, P. M., Graubard, B. I., Bui, T., and Carnide, N. (2026).
#' Correction for Participation Bias in Nonprobability Samples
#' Using Multiple Reference Surveys.
#' \emph{Statistics in Medicine}, 45(3--5).
#' doi:10.1002/sim.70403
#'
#' @examples
#' \donttest{
#' data(sc)
#' data(sp1)
#' data(sp2)
#'
#' ## One-reference example
#'
#' ref1_design <- survey::svydesign(
#'   ids     = ~psu_sp1,
#'   strata  = ~strata_sp1,
#'   weights = ~wts_sp1,
#'   data    = sp1,
#'   nest    = TRUE
#' )
#'
#' fit1 <- est_pw(
#'   data      = list(sc, ref1_design),
#'   p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
#'   method    = "calibration",
#'   control   = pw_solver_control(ftol = 1e-6)
#' )
#'
#' print(fit1)
#'
#' summary(fit1)
#'
#' ## Multi-reference example
#'
#' ref2_design <- survey::svydesign(
#'   ids     = ~psu_sp2,
#'   strata  = ~strata_sp2,
#'   weights = ~wts_sp2,
#'   data    = sp2,
#'   nest    = TRUE
#' )
#'
#' fit2 <- est_pw(
#'   data = list(sc, ref1_design, ref2_design),
#'   p_formula = list(
#'     ~ agecat + race + education + psa_level + pros_enlarged + comorbidity,
#'     ~ agecat + race + BMI + diabetes + comorbidity
#'   ),
#'   sp_order = "size",
#'   precali = TRUE,
#'   control = pw_solver_control(ftol = 1e-6)
#' )
#'
#' print(fit2)
#'
#' summary(fit2)
#' }
#' @export
#' @name est_pw
NULL

---
title: "Introduction to the nonprobsampling Package"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to the nonprobsampling Package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

The `nonprobsampling` package provides methods for correcting participation bias in nonprobability samples using one or multiple external probability reference surveys. The package is based on the generalized estimating equation framework proposed by Landsman et al. (2026), in which pseudo-weights are constructed by aligning auxiliary information from the nonprobability sample with corresponding information from probability reference surveys.

The package implements three one-reference methods within the framework: the raking ratio calibration method of Landsman et al. (2026), the adjusted logistic propensity (ALP) method of Wang et al. (2021), and the Chen–Li–Wu (CLW) method of Chen et al. (2020). It also implements the multi-reference calibration method of Landsman et al. (2026), which integrates auxiliary information from multiple reference surveys when no single reference survey contains all variables relevant to participation.


## Datasets

The package includes three datasets derived from real studies and used throughout this vignette to illustrate estimation of the mean serum PSA (prostate-specific antigen) level among U.S. men aged 55 years and older.

| Dataset | Type | Source | n |
|:-------:|------|--------|---|
| `sc` | Nonprobability sample | A synthetic sample drawn from a finite population, which was constructed based on NHANES survey data from the 1999–2010 cycles | 2,404 |
| `sp1` | First probability reference survey | National Health and Nutrition Examination Survey (NHANES) 1999–2010 | 3,494 |
| `sp2` | Second probability reference survey | National Health Interview Survey (NHIS) 1997–2008 | 35,525 |

The dataset `sc` represents the nonprobability sample and contains the outcome variable `psa_level`, defined as serum PSA level (ng/mL). The datasets `sp1` and `sp2` are probability reference surveys used to estimate pseudo-weights.

For illustration purposes, all shared variables have already been harmonized across the example datasets. Variable names, factor levels, and category definitions are consistent. The variables shared between `sc` and `sp1` are `agecat`, `race`, `education`, `BMI`, `comorbidity`, `diabetes` and `pros_enlarged`. The variables shared between `sc` and `sp2` are `agecat`, `race`, `BMI`, `comorbidity`, and `diabetes`. 

For user-supplied data, shared variables must be harmonized before estimation. Variable names must match exactly, and continuous variables included in the participation model should be measured on comparable scales across datasets.



```{r load-data}
library(nonprobsampling)

data("sc")
data("sp1")
data("sp1_bootstrap")
data("sp2")

str(sc)
str(sp1)
str(sp2)
```

```{r peek-data-sc}
head(sc)
```

---

## Step 1 — Create Survey Design Objects

Before calling `est_pw()`, each probability reference survey must be converted to a survey design object using either `survey::svydesign()` or `survey::svrepdesign()`. These objects retain the sampling information needed for design-based variance estimation.
```{r PSU-survey-designs}
ref1_design <- survey::svydesign(
  ids     = ~psu_sp1,
  strata  = ~strata_sp1,
  weights = ~wts_sp1,
  data    = sp1,
  nest    = TRUE
)

ref2_design <- survey::svydesign(
  ids     = ~psu_sp2,
  strata  = ~strata_sp2,
  weights = ~wts_sp2,
  data    = sp2,
  nest    = TRUE
)
```

The replicate weights in the example reference survey `sp1_bootstrap`
were generated using the Rao-Wu rescaling bootstrap method (Rust and Rao, 1996). They are provided to illustrate variance estimation with a replicate-weight survey design
object.
In real applications, users should use the replicate weights supplied by the
survey data provider when available.
```{r replicate-wts-survey-design}
ref1_design_rep_wts <- survey::svrepdesign(
   data    = sp1_bootstrap,
   weights = ~wts_sp1,
   repweights = "bw[0-9]+",
   type    = "bootstrap"
)

```

---

## Step 2.1 — Estimate Pseudo-Weights (One Reference Survey)

`est_pw()` is the main function of the package. It takes a list in which the first element is the nonprobability sample data frame and the remaining elements are probability reference surveys represented as survey design objects.

The `method` argument specifies the pseudo-weighting method to use, and
`p_formula` defines the participation model by selecting the shared auxiliary variables used to estimate the pseudo-weights.

### Calibration Estimator
The raking ratio calibration estimator is a one-reference pseudo-weighting method within the generalized estimating equation framework.
```{r one-ref-calibration-classic}

fit_cali <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "calibration",
  control   = pw_solver_control(ftol = 1e-6)
)

print(fit_cali)

summary(fit_cali)
```

When no `p_formula` is provided, it is constructed automatically from all variables shared between the nonprobability sample and the reference survey(s). The generated formula is shown in the model summary.

```{r one-ref-calibration-autopfml}
fit_cali_auto <- est_pw(
  data   = list(sc, ref1_design),
  method = "calibration",
  control   = pw_solver_control(ftol = 1e-6)
)
summary(fit_cali_auto)
```

Interaction and higher-order terms can be specified through `p_formula`.
```{r one-ref-calibration-int}

fit_cali_interaction <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ BMI * psa_level + diabetes + pros_enlarged + I(psa_level^2),
  method    = "calibration",
  control   = pw_solver_control(ftol = 1e-6)
)

summary(fit_cali_interaction)
```




Setting `verbose = TRUE` prints a brief processing log, including the number of observations used after missing-data filtering, the estimation method, participation model dimension, and solver convergence diagnostics such as iteration count and the final maximum absolute estimating equation value.

```{r one-ref-calibration-verbose}
fit_cali <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "calibration",
  verbose   = TRUE,
  control   = pw_solver_control(ftol = 1e-6)
)
```

The updated `sc` data frame is returned with a `pseudo_wts` column added.
```{r one-ref-calibration-sc_updated}
head(fit_cali$sc_updated)
```

The estimated pseudo-weights are also returned as a numeric vector.
```{r one-ref-calibration-pseudo_weights}
head(fit_cali$pseudo_weights)
```


### ALP Estimator

The adjusted logistic propensity (ALP) estimator is another one-reference pseudo-weighting method within the generalized estimating equation framework.

```{r one-ref-alp}
fit_alp <- est_pw(
  data      = list(sc, ref1_design_rep_wts),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "alp"
)

print(fit_alp)
```

### CLW Estimator

The Chen--Li--Wu (CLW) estimator is another one-reference pseudo-weighting method within the generalized estimating equation framework.
```{r one-ref-clw}
fit_clw <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "clw"
)

print(fit_clw)
```

## Step 2.2 — Estimate Pseudo-Weights (Multiple Reference Surveys)

When two or more probability reference surveys are available, use `method = "multi"` (or leave `method = NULL` for automatic selection). The argument `p_formula` should be supplied as a list containing one participation model formula per reference survey.

By default, `precali = TRUE` applies cumulative precalibration before pseudo-weight estimation. Under this procedure, the reference surveys are sequentially calibrated to previously processed surveys using their overlapping variables in the `p_formula`. This helps preserve information on the relationships between overlapping and survey-specific auxiliary variables within each reference survey.

The argument `sp_order` controls the order in which reference surveys are processed during cumulative precalibration and pseudo-weight estimation. With `sp_order = "size"` (default), surveys are ordered from largest to smallest sample size, so larger surveys are used as reference and overlapping variables are preferentially taken from them. With `sp_order = "given"`, surveys are processed in the user-supplied order, which may be preferable when a particular reference survey is considered more reliable or important.

### Order Surveys by Sample Size (Default)
```{r multi-ref-by-size}
fit_multi_order_by_size <- est_pw(
  data      = list(sc, ref1_design, ref2_design),
  p_formula = list(
    ~ agecat + race + education + psa_level  + pros_enlarged + comorbidity + BMI + diabetes,
    ~ agecat + race + diabetes  + BMI + comorbidity
  ),
  sp_order  = "size",
  precali   = TRUE,
  control   = pw_solver_control(ftol=1e-6)
)

summary(fit_multi_order_by_size)
```


### Use the User-Specified Survey Order
```{r multi-ref-as-given}
fit_multi_as_given <- est_pw(
  data      = list(sc, ref1_design, ref2_design),
  p_formula = list(
    ~ agecat + race + education + psa_level + pros_enlarged + comorbidity + diabetes,
    ~ agecat + race + BMI + diabetes + comorbidity
  ),
  sp_order  = "given",
  precali   = TRUE
)

summary(fit_multi_as_given)
```


---

## Step 3 — Estimate Population Means and Prevalences

Once pseudo-weights have been estimated, `pwmean()` uses them to estimate population means for numeric outcomes and population prevalences of each observed category for categorical outcomes. In both cases, the returned object includes unweighted and pseudo-weighted estimates, standard errors, and confidence intervals.

For a numeric outcome, `pwmean()` estimates the overall population mean. Since `psa_level` is numeric, the estimate is reported as a mean. Binary 0/1 outcomes are also treated as numeric and reported as means.
```{r pwmean-numeric}
est_cali_numeric <- pwmean(fit_cali, y = "psa_level")
print(est_cali_numeric)
summary(est_cali_numeric)
```

For a categorical outcome, `pwmean()` estimates the population prevalence of each observed category. Here, `BMI` is a categorical outcome, so the output is printed with category and prevalence labels.
```{r pwmean-factor}
est_cali_factor <- pwmean(fit_cali, y = "BMI")
print(est_cali_factor)
summary(est_cali_factor)
```

### Domain (Subgroup) Estimates

When a categorical domain variable is passed to `zcol`, the function returns separate estimates for each domain.

With a numeric outcome and a domain variable, `pwmean()` estimates a separate population mean within each domain. In this example, the mean of `psa_level` is estimated within each `BMI` domain.
```{r pwmean-domain-numeric}
est_by_bmi_numeric <- pwmean(fit_cali, y = "psa_level", zcol = "BMI")
print(est_by_bmi_numeric)
summary(est_by_bmi_numeric)
```

With a categorical outcome and a domain variable, `pwmean()` estimates a prevalence of each outcome category within each domain. In this example, the prevalence of each `agecat` category is estimated within each `BMI` domain.
```{r pwmean-domain-factor}
est_by_bmi_factor <- pwmean(fit_cali, y = "agecat", zcol = "BMI")
print(est_by_bmi_factor)
summary(est_by_bmi_factor)
```

## Handling Missing Data

`nonprobsampling` handles missing data in two separate stages.

First, `est_pw()` handles missing values in the participation model variables specified by `p_formula`. By default, `est_pw()` uses `na.action = stats::na.omit`, which removes observations with missing values before estimating pseudo-weights.

Alternatively, users may specify `na.action = stats::na.exclude`. In this case, excluded observations from the nonprobability sample are retained in the returned objects, but their pseudo-weights are set to `NA`. This is useful when users want to keep the original row alignment of the nonprobability sample for later merging or imputation.

Second, `pwmean()` handles missing values in the outcome variable and, if specified, the categorical domain variable. This second layer of missing-data handling is applied after pseudo-weights have already been estimated.

For illustration, we create temporary versions of the example datasets. In the nonprobability sample, we insert three missing values into the participation variable `agecat` and four missing values into the outcome variable `psa_level`. In the reference survey `sp1`, we insert two missing values into the participation variable `education`.

```{r na-setup}
sc_with_na <- sc
sc_with_na$agecat[c(1, 2, 4)] <- NA
sc_with_na$psa_level[c(6, 7, 8, 9)] <- NA

sp1_with_na <- sp1
sp1_with_na$education[c(3, 4)] <- NA

ref1_design_with_na <- survey::svydesign(
  ids     = ~psu_sp1,
  strata  = ~strata_sp1,
  weights = ~wts_sp1,
  data    = sp1_with_na,
  nest    = TRUE
)
```



### Missing-Data Handling in est_pw() with na.exclude
The following example uses the calibration method with `na.action = stats::na.exclude`. A brief summary of missing-data handling is displayed at the bottom of the model summary.

```{r fit-with-na}
fit_na_exclude <- est_pw(
  data      = list(sc_with_na, ref1_design_with_na),
  method    = "calibration",
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  na.action = stats::na.exclude,
  control   = pw_solver_control(ftol = 1e-6)
)
summary(fit_na_exclude)

```

A summary is available in the `na_summary` component.
```{r na_summary}
fit_na_exclude$na_summary
```

Information about observations excluded because of missing values in the participation model variables can be obtained using the S3 `na.action()` method.
```{r na.action}
na.action(fit_na_exclude)
```

With `na.exclude`, the returned pseudo-weight vector and `sc_updated` data frame preserve the original row structure. Observations excluded from pseudo-weight estimation remain in the output, with their pseudo-weights recorded as `NA`.
```{r na-head}
head(fit_na_exclude$pseudo_weights)
head(fit_na_exclude$sc_updated)
```


### Missing-Data Handling in pwmean()

Missing values in the outcome variable are handled when `pwmean()` is called. This is separate from the missing-value handling performed by `est_pw()`.
```{r na-2nd-layer}
result <- pwmean(fit_na_exclude, "psa_level")
summary(result)
```


A summary of outcome-level missing-value handling is stored in the returned object. The corresponding `na.action` information can also be obtained using the S3 `na.action()` method.
```{r na-info-2nd}
result$na.action
na.action(result)
```

### Missing-Data Handling in est_pw() with na.omit (Default)

Using `na.action = stats::na.omit` removes observations with missing participation model variables from the returned objects. This is the default behavior.

```{r fit-na-omit}
fit_na_omit <- est_pw(
  data      = list(sc_with_na, ref1_design_with_na),
  method    = "calibration",
  na.action = stats::na.omit,
  control   = pw_solver_control(ftol = 1e-6)
)
```

Information about the excluded observations can be obtained using the S3 `na.action()` method.
```{r na-omit-na.action}
na.action(fit_na_omit)
```

With `na.omit`, the returned pseudo-weight vector and `sc_updated` contain only the observations used for pseudo-weight estimation.
```{r na-omit-head}
head(fit_na_omit$pseudo_weights)
head(fit_na_omit$sc_updated)
```

In summary, `est_pw()` handles missing values in participation model variables, while `pwmean()` handles missing values in the outcome and domain variables. With `na.omit`, excluded observations are removed from the returned objects; with `na.exclude`, the original row structure is preserved and excluded observations receive `NA` pseudo-weights.





## Controlling the Numerical Solver

The `control` argument accepts an object created by `pw_solver_control()`, which is passed to the underlying `nleqslv` nonlinear equation solver. It controls the numerical algorithm used to solve the estimating equations, including convergence tolerances, iteration limits, and diagnostic output. The default settings are suitable for most applications, but advanced users may adjust them when troubleshooting convergence issues.

```{r solver-control}
ctrl <- pw_solver_control(
  method   = "Broyden",
  xtol     = 1e-10,
  ftol     = 1e-4,
  maxit    = 30,
  trace    = TRUE
)

fit_ctrl <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI,
  method    = "calibration",
  control   = ctrl
)
```

---

## Quick Reference

| Function | Purpose |
|----------|---------|
| `est_pw()` | Estimate pseudo-weights from a nonprobability sample and one or multiple reference surveys |
| `pwmean()` | Estimate pseudo-weighted population means for numeric outcomes and population prevalences for categorical outcomes, overall or by domain |
| `pw_solver_control()` | Configure numerical solver settings |

| Argument | Values | Notes |
|----------|--------|-------|
| `method` | `"alp"`, `"clw"`, `"calibration"`, `"multi"`, `NULL` | `NULL` auto-selects based on number of reference surveys |
| `sp_order` | `"size"`, `"given"` | Only relevant for multi-reference calibration method; `"size"` sorts largest first |
| `precali` | `TRUE`, `FALSE` | Precalibration before multi-reference calibration step |
| `na.action` | `na.omit`, `na.exclude`, `na.fail`| Controls missing-data handling |


# References

Chen, Y., Li, P., and Wu, C. (2020). Doubly robust inference with nonprobability survey samples. Journal of the American Statistical Association, 115(532), 2011–2021. [doi:10.1080/01621459.2019.1677241](https://doi.org/10.1080/01621459.2019.1677241)

Landsman, V., Wang, L., Carrillo-Garcia, I., Mitani, A. A., Smith, P. M., Graubard, B. I., Bui, T., and Carnide, N. (2026). Correction for Participation Bias in Nonprobability Samples Using Multiple Reference Surveys. Statistics in Medicine, 45(3–5). [doi:10.1002/sim.70403](https://doi.org/10.1002/sim.70403)

Rust, K. F., and Rao, J. N. K. (1996). Variance estimation for complex surveys using replication techniques. Statistical Methods in Medical Research, 5(3), 283–310. [doi:10.1177/096228029600500305](https://doi.org/10.1177/096228029600500305)

Wang, L., Valliant, R., and Li, Y. (2021). Adjusted logistic propensity weighting methods for population inference using nonprobability volunteer-based epidemiologic cohorts. Statistics in Medicine, 40(24), 5237–5250. [doi:10.1002/sim.9122](https://doi.org/10.1002/sim.9122)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s5_naive.R
\name{naive_mean_one_domain}
\alias{naive_mean_one_domain}
\title{Compute the naive mean for one domain}
\usage{
naive_mean_one_domain(yvec, zvec = NULL)
}
\arguments{
\item{yvec}{Numeric or integer outcome vector of length \eqn{n}.}

\item{zvec}{Integer 0/1 domain indicator of length \eqn{n}, or NULL
(treated as all-ones, i.e., the overall mean).}
}
\value{
A list with components:

- `mean`: unweighted sample mean of `yvec` within the domain,
  or NA if the domain is empty.
- `variance`: estimated variance of the mean (\eqn{s^2/n}),
  or NA if fewer than two observations are available.
}
\description{
Computes the unweighted sample mean and its variance for observations
belonging to a single domain, identified by a 0/1 indicator vector.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s5_naive.R
\name{naive_mean}
\alias{naive_mean}
\title{Compute naive (unweighted) means from the convenience sample}
\usage{
naive_mean(df, domain_var = NULL, y)
}
\arguments{
\item{df}{A data frame containing the convenience sample
(typically `build$internal$raw_sc`).}

\item{domain_var}{Single character string naming the domain variable in
`df`, or NULL for the overall mean.}

\item{y}{Single character string naming the outcome variable in `df`.}
}
\value{
A list with components:

- `type`: `"single"` for overall or binary domains;
  `"multi"` for factor/character domains.
- `labels`: character vector of domain labels.
- `estimates`: for `type = "single"`, a list with `mean` and `variance`.
  For `type = "multi"`, a list of such lists, one per domain level.
}
\description{
Filters the convenience sample to complete cases, then computes the
unweighted sample mean (and its variance) for each domain using
`naive_mean_one_domain()`. The domain structure is standardized via
`standardize_zcol()`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s3_process_na.R
\name{process_na_build}
\alias{process_na_build}
\title{Handle missing values at the build stage (first layer)}
\usage{
process_na_build(sc, sp, sp_des, p_formula, na.action, n_ref, verbose = FALSE)
}
\arguments{
\item{sc}{A data frame. The nonprobability sample (before NA removal).}

\item{sp}{A data frame (one-reference case) or a named list of data frames
(multi-reference case). Analysis data extracted from the reference
survey design(s).}

\item{sp_des}{A single survey design object (one-reference case) or a
named list of survey design objects (multi-reference case). Subsetted
to align with the filtered `sp`.}

\item{p_formula}{A one-sided formula (one-reference case) or a list of
one-sided formulas (multi-reference case) specifying the participation
model variables used to identify rows with missing values.}

\item{na.action}{A function (`stats::na.omit`, `stats::na.exclude`,
`stats::na.fail`, or `stats::na.pass`), an equivalent character string,
or `NULL` (which inherits from `getOption("na.action")`).}

\item{n_ref}{Integer. Number of reference surveys. Controls whether
one-reference or multi-reference logic is applied when subsetting
`sp_des` and computing `n_sp_orig`.}

\item{verbose}{Logical. If `TRUE`, prints per-dataset row counts and
exclusion totals via `message()`.}
}
\value{
A list with the following elements:
\describe{
  \item{`na_mode`}{Character string: one of `"omit"`, `"exclude"`,
    `"fail"`, or `"pass"`.}
  \item{`sc`}{The cleaned nonprobability sample data frame.}
  \item{`sp`}{The cleaned reference survey data frame(s).}
  \item{`sp_des`}{The subsetted survey design object(s).}
  \item{`keep_sc`}{Logical vector indicating which rows of the original
    `sc` are retained.}
  \item{`keep_sp`}{Logical vector (one-reference) or list of logical
    vectors (multi-reference) indicating which rows of each `sp` are
    retained.}
  \item{`n_sp_orig`}{Original row count(s) of `sp_des` before
    subsetting.}
  \item{`na_action_obj`}{An lm-style `na.action` attribute for `sc`,
    or `NULL` if no rows were removed.}
  \item{`log_messages`}{Character vector of per-variable NA detail
    messages from `handle_na_for_ipwm()`, suitable for appending to
    the running `log_messages` in `est_pw()`.}
  \item{`na_summary`}{A `pw_na_summary` object with row counts before
    and after NA removal, or `NULL` if no rows were excluded.}
}
}
\description{
Process NA for `est_pw()` before estimation step.
Resolves the `na.action` argument to a standard mode string, applies NA
filtering to `sc` and `sp`, subsets the survey design object(s) in
`sp_des` to match the filtered `sp`, and produces a summary of
exclusions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc.R
\docType{data}
\name{sc}
\alias{sc}
\title{Nonprobability Sample (sc)}
\format{
A data frame with 2404 observations and 8 variables:
\describe{
  \item{psa_level}{Outcome variable: serum prostate-specific antigen level (numeric)}
  \item{BMI}{Body mass index category (factor with 4 levels: "Normal", "Overweight", "Obese", "Morbidly Obese")}
  \item{race}{Race category (factor with 4 levels: 1 = White, 2 = Black, 3 = Hispanic, 4 = Other)}
  \item{agecat}{Age category (factor with 4 levels: 1 = 55--59, 2 = 60--64, 3 = 65--69, 4 = 70+)}
  \item{education}{Education level (factor with 5 levels: 1 = Less Than 8 Years, 2 = 8--11 Years, 3 = 12 Years Or Completed High School, 4 = College Graduate, 5 = Postgraduate)}
  \item{pros_enlarged}{Prostate enlargement indicator (factor with 2 levels: 0 = No, 1 = Yes)}
  \item{comorbidity}{General comorbidity indicator (factor with 2 levels: 0 = No, 1 = Yes)}
  \item{diabetes}{Diabetes diagnosis indicator (factor with 2 levels: 0 = No, 1 = Yes)}
}
}
\source{
Synthetic data generated by the package authors. The underlying finite
population was constructed from the National Health and Nutrition
Examination Survey (NHANES), 1999--2010 cycles, conducted by the U.S.
National Center for Health Statistics (NCHS).
}
\usage{
data(sc)
}
\description{
This dataset represents a synthetic nonprobability sample generated via
Poisson sampling from a finite population constructed from the National
Health and Nutrition Examination Survey (NHANES) cycles 1999--2010. It is
intended to illustrate the pseudo-weighting methods implemented in the
\code{nonprobsampling} package.
}
\details{
The dataset has 2,404 complete-case observations, with
\code{psa_level} serving as the outcome variable. Auxiliary variables
shared with the probability reference surveys \code{sp1} and \code{sp2}
are used to construct pseudo-weights aimed at correcting for participation
bias.
}
\examples{
data(sc)
str(sc)
summary(sc)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s5_one.R
\name{ipwm_one_build}
\alias{ipwm_one_build}
\title{Dispatch one-reference pseudo-weight estimation}
\usage{
ipwm_one_build(
  sc,
  sp,
  sp_des,
  vars = NULL,
  weight,
  method,
  control,
  verbose = FALSE,
  log_messages = NULL
)
}
\arguments{
\item{sc}{Data frame. The nonprobability sample.}

\item{sp}{Data frame. The single probability reference sample (already
NA-processed and p_formula-processed).}

\item{sp_des}{A `survey.design2` or `svyrep.design` object for `sp`.}

\item{vars}{Character vector of predictor variable names to use in the
participation model design matrices.}

\item{weight}{Character scalar. Name of the survey-weight column in `sp`.}

\item{method}{Character scalar. One of `"alp"`, `"clw"`, `"calibration"`,
or `"cali"`.}

\item{control}{A list created by `pw_solver_control()`.}

\item{verbose}{Logical. If `TRUE`, progress messages are printed.}

\item{log_messages}{Character vector of messages accumulated in earlier
steps, forwarded into the returned object.}
}
\value{
A list with components:
  \item{pseudo_weights}{Numeric vector of estimated pseudo-weights for `sc`.}
  \item{coefficients}{Named numeric vector of participation model coefficients.}
  \item{method}{Character scalar identifying the method used.}
  \item{solver_diagnostics}{List of solver convergence diagnostics.}
  \item{internal}{List of intermediate objects (design matrices, fitted
    probabilities, sandwich components) needed by the estimation stage.}
}
\description{
Dispatch one-reference pseudo-weight estimation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s1_extract_data.R
\name{extract_analysis_data}
\alias{extract_analysis_data}
\title{Extract analysis data from a survey design object}
\usage{
extract_analysis_data(des, weight_name = "sp_wts")
}
\arguments{
\item{des}{A survey design object of class `survey.design2`
or `svyrep.design`.}

\item{weight_name}{A single string giving the name of the weight column
added to the output. Defaults to `"sp_wts"`. Must not already exist
among the analysis variables after design variables are removed.}
}
\value{
A data frame containing the analysis variables from `des$variables`
  (design variables removed) plus one column named `weight_name` storing
  the sampling weights.
}
\description{
Remove design variables (cluster IDs, strata, FPC, probabilities, weights)
from a survey design object and attaches the sampling weights as a single
column, returning a plain data frame ready for modeling.
}
\details{
Cluster and strata variables are identified from the design object's own
fields (`des$cluster`, `des$strata`) rather than the original call, so the
result is correct even when `des` has been subset after construction.
Weights, FPC, and probability variables are still read from `des$call`.
For replicate designs (`svyrep.design`), replicate weight columns are also
dropped.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_d_matrix_multi.R
\name{make_block_D_multi}
\alias{make_block_D_multi}
\title{Construct block-diagonal D matrix for multi-reference raking}
\usage{
make_block_D_multi(sp_des_list, Xp_list)
}
\arguments{
\item{sp_des_list}{A non-empty list of survey design objects, each of class
`"survey.design2"` or `"svyrep.design"`.}

\item{Xp_list}{A list of reference-sample design matrices, one for each
reference survey. Each matrix must have the same number of rows as the
corresponding survey design object in `sp_des_list`.}
}
\value{
A block-diagonal covariance matrix \eqn{D}.
}
\description{
Internal helper to assemble the block-diagonal covariance matrix
\eqn{D} for the multi-reference raking estimator. Each reference survey
contributes one design-based covariance block computed from
`compute_D_raking()`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_doc.R
\name{est_pw}
\alias{est_pw}
\title{Estimate Pseudo-Weights for Nonprobability Samples}
\usage{
est_pw(
  data,
  sp_order = c("size", "given"),
  precali = TRUE,
  p_formula = NULL,
  method = NULL,
  na.action = stats::na.omit,
  sc_wname = "pseudo_wts",
  control = pw_solver_control(),
  verbose = FALSE
)
}
\arguments{
\item{data}{A list of input data objects of the form
\code{list(sc, sp1_design, sp2_design, ...)}. The first element must be a
data frame corresponding to the nonprobability sample. Each remaining
element must be a valid survey design object corresponding to a reference
probability survey, such as an object created by
\code{\link[survey]{svydesign}} or \code{\link[survey]{svrepdesign}}.}

\item{sp_order}{Character string controlling the order of reference surveys
when multiple reference surveys are used. Supported values are
\code{"size"} and \code{"given"}. \code{"size"} orders reference surveys
by sample size, from largest to smallest. \code{"given"} uses the
user-specified order of the reference surveys in \code{data}. Default is
\code{"size"}. With one reference survey, this argument is ignored; a
warning is issued if a non-default value is supplied.}

\item{precali}{Logical. Used only with multiple reference surveys. If
\code{TRUE}, cumulative precalibration is applied before the main
multi-reference estimation step; see the
\strong{Multi-reference precalibration} section for details. Default is
\code{TRUE}. With one reference survey, this argument is ignored; a
warning is issued if \code{FALSE} is supplied.}

\item{p_formula}{Optional participation model formula. Must always be
  one-sided (no response variable on the left-hand side). A two-sided formula
  such as \code{y ~ x} will raise an error.

  With one reference survey, supply a single one-sided formula, for example
  \code{~ age + sex + income}. With multiple reference surveys, supply a
  list of one-sided formulas with one formula per reference survey, for
  example \code{list(~ age + sex, ~ age + income)}. If \code{NULL}, a
  default formula is constructed automatically from variables shared across
  the data sources used for estimation. Since shared variables are identified
  by name, their names should be harmonized across data sources before
  estimation.}

\item{method}{Character string specifying the pseudo-weighting method, or
  \code{NULL} (default). If \code{NULL}, \code{"calibration"} is used when
  \code{data} contains one reference survey, and \code{"multi"} is used when
  \code{data} contains more than one reference survey.

  To override the default, supply one of the following values. For a
  one-reference method: \code{"alp"}, \code{"clw"}, or
  \code{"calibration"} (or \code{"cali"}). For the
  multi-reference method: \code{"multi"}.

  The argument is case-insensitive, so inputs such as
  \code{"ALP"}, \code{"Clw"}, or \code{"CALI"} are also accepted.}

\item{na.action}{Function specifying how missing values should be handled for
variables used in the participation model. Common choices include
\code{stats::na.omit}, \code{stats::na.exclude}, and
\code{stats::na.fail}. Default is \code{stats::na.omit}.}

\item{sc_wname}{Character string giving the name of the pseudo-weight column
added to the returned nonprobability sample. Default is
\code{"pseudo_wts"}. An error is raised at input validation if this name
already exists as a column in \code{sc}.}

\item{control}{A solver control object created by
\code{\link{pw_solver_control}}. This object stores numerical settings
for solving estimating equations, including the solver, convergence
tolerance, maximum number of iterations, tracing behavior, and other
options.}

\item{verbose}{Logical. If \code{TRUE}, progress messages and diagnostics are
printed during pseudo-weight estimation. Default is \code{FALSE}. Must be a
single \code{TRUE} or \code{FALSE}; an error is raised otherwise.}
}
\value{
An object of class \code{"pw_fit"}. This is a list containing user-facing
outputs and internal objects required by \code{\link{pwmean}}.

Important components include:

\describe{
  \item{\code{sc_updated}}{
    A data frame containing the nonprobability sample with an added
    pseudo-weight column named by \code{sc_wname}.
  }
  \item{\code{pseudo_weights}}{
    The estimated pseudo-weight vector. With \code{stats::na.omit}, the
    vector contains only observations retained for pseudo-weight estimation.
    With \code{stats::na.exclude}, excluded observations receive \code{NA}
    and the vector has length \code{nrow(sc)}.
  }
  \item{\code{coefficients}}{
    Estimated coefficients for the participation model variables.
  }
  \item{\code{solver_diagnostics}}{
    A list of solver diagnostics: \code{solver} (solver name),
    \code{termcd} (termination code), \code{message} (solver message),
    \code{iter} (number of iterations), and \code{fmax} (maximum absolute
    value of the final estimating equations at convergence).
  }
  \item{\code{method}}{
    The pseudo-weighting method used by the function.
  }
  \item{\code{internal}}{
    A list of internal objects needed for downstream estimation.
  }
  \item{\code{na_summary}}{
    An object of class \code{"pw_na_summary"} summarizing the number of rows
    excluded from the nonprobability sample and each reference survey due to
    missing participation model variables. \code{NULL} if no rows were
    excluded.
  }
  \item{\code{call}}{
    The matched function call.
  }
}
}
\description{
\code{est_pw()} estimates pseudo-weights for a nonprobability sample using
one reference survey or multiple reference surveys. The function specifies the
participation model, handles missing values in the participation model
variables, solves the estimating equations, and stores the quantities needed
for downstream point and variance estimation.

Users should harmonize variable names and coding before calling
\code{est_pw()}. Variables used in the participation model must have
consistent names and compatible definitions across the nonprobability sample
and the reference survey data used for estimation.

With one reference survey, the available methods include the
raking ratio calibration method described in Landsman et al. (2026),
the adjusted logistic propensity weighting (ALP) method proposed by Wang, Valliant,
and Li (2021), and the CLW method proposed by Chen, Li, and Wu (2020).
With multiple reference surveys, pseudo-weights are estimated using the
multi-reference calibration method proposed by Landsman et al. (2026).

The returned object is designed to be passed to \code{\link{pwmean}}.
}
\details{
\code{est_pw()} performs pseudo-weight estimation for the nonprobability
sample and stores the method-specific internal objects needed later by
\code{\link{pwmean}}. It does not require an outcome variable.

The input \code{data} must be provided as a list, where the first element is
the nonprobability sample and the remaining elements are reference survey
design objects. Reference survey designs can be created with
\code{\link[survey]{svydesign}} for standard complex survey designs or
\code{\link[survey]{svrepdesign}} for surveys with replicate weights. These
objects preserve the sampling structure needed for design-consistent
variance estimation.

\strong{Variable harmonization.}
Variables are matched by name, not by meaning. Before applying
\code{est_pw()}, shared variables must be harmonized across the
nonprobability sample and reference survey data. For example, if a
categorical variable is named \code{agecat} in the nonprobability sample and
\code{age_group} in the reference survey, the user should rename one of the
variables before estimation.

Categorical variables should be encoded as factors with compatible category
definitions and identical levels in the same order. Even when categories are
substantively equivalent, mismatched factor levels may cause
\code{est_pw()} to return an error. Continuous variables included in the
participation model should also be measured on comparable scales across
datasets.

Internally, \code{est_pw()} performs the following steps:
\enumerate{
  \item \strong{Input validation}\cr
    Validates the structure and required components of the input data.
  \item \strong{Reference survey detection}\cr
    Determines whether the input contains a single reference survey or
    multiple reference surveys.
  \item \strong{Method selection}\cr
    Selects the pseudo-weighting method based on the specified argument(s).
  \item \strong{Participation model specification}\cr
    Constructs a default participation model formula when
    \code{p_formula = NULL}.
  \item \strong{Missing data handling}\cr
    Applies missing-data handling procedures to variables used in the
    participation model.
  \item \strong{Model matrix construction}\cr
    Generates model matrices from the participation model variables.
  \item \strong{Pseudo-weight estimation}\cr
    Estimates pseudo-weights using the selected method.
  \item \strong{Output augmentation}\cr
    Appends the estimated pseudo-weights as a new column to the
    nonprobability sample.
  \item \strong{Metadata storage}\cr
    Stores information related to missing-data handling and other internal
    objects for later use or diagnostics.
}
}
\section{One-reference method and multi-reference method}{

If \code{data} contains one reference survey design object, \code{est_pw()}
fits a one-reference method. If \code{data} contains more
than one reference survey design objects, \code{est_pw()} fits the
multi-reference calibration method.
In both settings, the auxiliary variables used for pseudo-weight estimation
should be harmonized across all data sources before calling
\code{est_pw()}.
}

\section{Multi-reference precalibration}{

When \code{precali = TRUE}, cumulative precalibration is performed before the
main multi-reference calibration step. For overlapping
auxiliary variables, this procedure calibrates the survey weights of a
reference survey so that its weighted totals of the overlapping variables
and its sum of weights match the corresponding totals from the preceding
reference survey in the cumulative order. If there are no overlapping
auxiliary variables, cumulative precalibration is applied only to the sum of
weights.

The order of the reference surveys is controlled by \code{sp_order}. If
\code{sp_order = "size"}, reference surveys are ordered by sample size, from
largest to smallest. If \code{sp_order = "given"}, the user-specified order
of the reference surveys is used.

Cumulative precalibration is based only on overlapping variables that are
specified in \code{p_formula}, rather than on all overlapping variables in
the reference surveys. This choice avoids excluding observations because
of missing values in variables that are not used for pseudo-weight
estimation.
}

\section{Missing data handling}{

Missing values are handled only for variables used in the participation
model. The selected \code{na.action} is recorded in the returned object,
together with the row indices of the nonprobability sample observations
retained for pseudo-weight estimation.

With \code{stats::na.omit}, rows with missing participation model variables
are removed from \code{sc_updated}. With \code{stats::na.exclude}, the
original rows are retained in \code{sc_updated}, but excluded rows receive
\code{NA} in the pseudo-weight column. This can be useful when users want to
preserve row alignment with the original nonprobability sample for later imputation or merging.
}

\section{Numerical control}{

Numerical settings are supplied through the \code{control} argument, which
should be created by \code{\link{pw_solver_control}}. This object controls
solver choice, convergence tolerance, maximum iterations, tracing, and
optional solver-specific arguments.

The top-level \code{ftol}, \code{xtol}, and \code{maxit} values in
\code{\link{pw_solver_control}} are the package-level convergence controls
used by pseudo-weight estimation stages. When the selected solver is
\code{"nleqslv"}, additional arguments can be passed through
\code{nleqslv_control}. These are forwarded to \code{nleqslv::nleqslv()}.
}

\examples{
\donttest{
data(sc)
data(sp1)
data(sp2)

## One-reference example

ref1_design <- survey::svydesign(
  ids     = ~psu_sp1,
  strata  = ~strata_sp1,
  weights = ~wts_sp1,
  data    = sp1,
  nest    = TRUE
)

fit1 <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "calibration",
  control   = pw_solver_control(ftol = 1e-6)
)

print(fit1)

summary(fit1)

## Multi-reference example

ref2_design <- survey::svydesign(
  ids     = ~psu_sp2,
  strata  = ~strata_sp2,
  weights = ~wts_sp2,
  data    = sp2,
  nest    = TRUE
)

fit2 <- est_pw(
  data = list(sc, ref1_design, ref2_design),
  p_formula = list(
    ~ agecat + race + education + psa_level + pros_enlarged + comorbidity,
    ~ agecat + race + BMI + diabetes + comorbidity
  ),
  sp_order = "size",
  precali = TRUE,
  control = pw_solver_control(ftol = 1e-6)
)

print(fit2)

summary(fit2)
}
}
\references{
Chen, Y., Li, P., and Wu, C. (2020).
Doubly robust inference with nonprobability survey samples.
\emph{Journal of the American Statistical Association},
115(532), 2011--2021.
doi:10.1080/01621459.2019.1677241

Wang, L., Valliant, R., and Li, Y. (2021).
Adjusted logistic propensity weighting methods for population inference
using nonprobability volunteer-based epidemiologic cohorts.
\emph{Statistics in Medicine}, 40(24), 5237--5250.
doi:10.1002/sim.9122

Landsman, V., Wang, L., Carrillo-Garcia, I., Mitani, A. A.,
Smith, P. M., Graubard, B. I., Bui, T., and Carnide, N. (2026).
Correction for Participation Bias in Nonprobability Samples
Using Multiple Reference Surveys.
\emph{Statistics in Medicine}, 45(3--5).
doi:10.1002/sim.70403
}
\seealso{
\code{\link{pw_solver_control}},
\code{\link{pwmean}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pw_solver_control.R
\name{pw_solver_control}
\alias{pw_solver_control}
\title{Control Solver Settings for Pseudo-Weight Estimation}
\usage{
pw_solver_control(
  solver = "nleqslv",
  maxit = NULL,
  trace = FALSE,
  method = c("Newton", "Broyden"),
  global = c("dbldog", "cline", "pwldog", "qline", "gline", "hook", "none"),
  xscalm = c("fixed", "auto"),
  ftol = 1e-08,
  xtol = 1e-08,
  nleqslv_control = list()
)
}
\arguments{
\item{solver}{Character string specifying the numerical solver used for
solving the estimating equations. Currently, only \code{"nleqslv"} is
supported. Default is \code{"nleqslv"}.}

\item{maxit}{Positive finite numeric value passed to
\code{nleqslv::nleqslv()} as the maximum number of solver iterations. The
value is converted to an integer before being stored in the returned
control object. Default is \code{150} when a global strategy is specified
(i.e., \code{global != "none"}), and \code{20} when no global strategy is
used (\code{global = "none"}), matching \code{nleqslv}'s own defaults.
Since the default global strategy is \code{"dbldog"}, the effective default
is \code{150} unless \code{global = "none"} is explicitly specified.}

\item{trace}{Logical. If \code{TRUE}, tracing or solver progress information
may be requested from the underlying numerical routine when supported.
Default is \code{FALSE}.}

\item{method}{Character string specifying the numerical method passed to
\code{nleqslv::nleqslv()}. Supported values are \code{"Newton"} and
\code{"Broyden"}. Default is \code{"Newton"}.}

\item{global}{Character string specifying the global strategy passed to
\code{nleqslv::nleqslv()}. Supported values are \code{"dbldog"},
\code{"cline"}, \code{"pwldog"}, \code{"qline"}, \code{"gline"},
\code{"hook"}, and \code{"none"}. Default is \code{"dbldog"}.}

\item{xscalm}{Character string specifying the scaling method passed to
\code{nleqslv::nleqslv()}. Supported values are \code{"fixed"} and
\code{"auto"}. Default is \code{"fixed"}.}

\item{ftol}{Positive finite numeric value passed to
\code{nleqslv::nleqslv()} as the function-value convergence tolerance. This
controls convergence based on the size of the estimating function. Default
is \code{1e-8}.}

\item{xtol}{Positive finite numeric value passed to
\code{nleqslv::nleqslv()} as the parameter-step convergence tolerance. This
controls convergence based on changes in the parameter vector. Default is
\code{1e-8}.}

\item{nleqslv_control}{A list of additional control options
passed to \code{nleqslv::nleqslv()}. This can include less commonly used
control options, such as \code{btol}, \code{cndtol},
\code{sigma}, and \code{scalex}. See
\code{\link[nleqslv]{nleqslv}} for details.}
}
\value{
A flat list containing all solver control settings for pseudo-weight
estimation:

\describe{
  \item{\code{solver}}{The selected numerical solver.}
  \item{\code{method}}{The nleqslv numerical method.}
  \item{\code{global}}{The nleqslv global strategy.}
  \item{\code{xscalm}}{The nleqslv scaling method.}
  \item{\code{ftol}}{The function-value convergence tolerance.}
  \item{\code{xtol}}{The parameter-step convergence tolerance.}
  \item{\code{maxit}}{The maximum number of solver iterations, stored as an integer.
    \code{150} if a global strategy is used; \code{20} if \code{global = "none"}.
    Since the default global strategy is \code{"dbldog"}, the effective default
    is \code{150} unless \code{global = "none"} is explicitly specified.}
  \item{\code{trace}}{Logical value indicating whether tracing information is requested.}
  \item{\code{nleqslv_control}}{A list of additional options passed to
    \code{nleqslv::nleqslv()}.}
}
}
\description{
\code{pw_solver_control()} creates a solver control object used by
\code{\link{est_pw}} to manage numerical settings for pseudo-weight
estimation.
}
\details{
The control object stores solver settings used by pseudo-weight estimation
step. It is passed to \code{\link{est_pw}} through the \code{control}
argument.

Currently, only \code{solver = "nleqslv"} is supported. The arguments
\code{method}, \code{global}, \code{xscalm}, \code{ftol}, \code{xtol}, and
\code{maxit} correspond to options used by \code{nleqslv::nleqslv()}.
They are collected internally and passed to \code{nleqslv::nleqslv()} at the
pseudo-weight estimation step.

The argument \code{ftol} is the function-value convergence tolerance. It
controls convergence based on the size of the estimating function. The
argument \code{xtol} is the parameter-step convergence tolerance. It controls
convergence based on changes in the parameter vector. The argument
\code{maxit} controls the maximum number of solver iterations.

Additional, less commonly used \code{nleqslv} control options can be supplied
through \code{nleqslv_control}. To avoid ambiguity, do not supply
\code{ftol}, \code{xtol}, \code{maxit}, or \code{trace} inside
\code{nleqslv_control}; use the main arguments instead.
}
\examples{
## Default solver control settings
ctrl <- pw_solver_control()

## Custom nleqslv solver settings
ctrl <- pw_solver_control(
  maxit  = 20,
  trace  = FALSE,
  method = "Newton",
  global = "cline",
  xscalm = "auto",
  ftol   = 1e-8,
  xtol   = 1e-10
)

## Additional nleqslv control options
ctrl <- pw_solver_control(
  method = "Newton",
  global = "dbldog",
  nleqslv_control = list(
    btol = 1e-3
  )
)

}
\seealso{
\code{\link{est_pw}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s4_cali.R
\name{raking_estimate}
\alias{raking_estimate}
\title{Estimate step for raking-ratio calibration}
\usage{
raking_estimate(Y, Z, w, X, D, S_beta)
}
\arguments{
\item{Y}{Outcome vector for the nonprobability sample.}

\item{Z}{Domain indicator vector. Use `rep(1, length(Y))` for the
overall mean.}

\item{w}{Estimated pseudo-weights from the calibration build step.}

\item{X}{Calibration design matrix for the nonprobability sample.}

\item{D}{Design-based variance-covariance matrix of the estimated
auxiliary totals from the reference survey or surveys. For multiple
reference surveys, this is block diagonal.}

\item{S_beta}{Sensitivity matrix for the calibration estimating equations,
typically \code{t(w * X) \%*\% X} for raking-ratio calibration.}
}
\value{
A list with components `mean` and `variance`.
}
\description{
Computes the domain-specific pseudo-weighted Hájek mean and its
Taylor-linearized variance for the raking-ratio calibration estimator.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s3_helper.R
\name{print.pw_na_summary}
\alias{print.pw_na_summary}
\title{Print method for pw_na_summary}
\usage{
\method{print}{pw_na_summary}(x, ...)
}
\arguments{
\item{x}{A `pw_na_summary` object returned by `.report_na_exclusions()`.}

\item{...}{Further arguments passed to or from other methods (unused).}
}
\value{
Invisibly returns `x`.
}
\description{
Prints a formatted table showing the original row count, rows used,
and rows excluded due to missing participation model variables, for
each dataset (`sc` and each reference survey).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s4_clw.R
\name{clw_estimate}
\alias{clw_estimate}
\title{Estimate step for the CLW estimator}
\usage{
clw_estimate(Y, Z, w, X, D, S_beta)
}
\arguments{
\item{Y}{Outcome vector for the nonprobability sample.}

\item{Z}{Domain indicator vector. Use `rep(1, length(Y))` for the
overall mean.}

\item{w}{Estimated pseudo-weights from the CLW build step.}

\item{X}{Design matrix for the nonprobability sample.}

\item{D}{Design-based variance-covariance matrix of the estimated
auxiliary totals from the reference survey.}

\item{S_beta}{Sensitivity matrix for the CLW estimating equations.}
}
\value{
A list with components `mean` and `variance`.
}
\description{
Computes the domain-specific pseudo-weighted Hájek mean and its
Taylor-linearized variance for the CLW estimator.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s6_output.R
\name{assemble_output}
\alias{assemble_output}
\title{Assemble the final output object}
\usage{
assemble_output(build, est, naive, na_info)
}
\arguments{
\item{build}{A `pw_fit` object returned by the build step.}

\item{est}{A list returned by `dispatch_estimator()`, with components
`type`, `labels`, and `estimates`.}

\item{naive}{A list returned by `naive_mean()`, with the same structure
as `est`.}

\item{na_info}{NA-handling information returned by `process_na_yz()`.}
}
\value{
A list with components `method`, `estimates`, and `na.action`.
  The S3 class is assigned by the calling function `pwmean()`.
}
\description{
Combines the pseudo-weighted estimates and naive (unweighted) estimates
into a plain list to be returned by `pwmean()`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s5_multi_s2_validation.R
\name{check_input_multi}
\alias{check_input_multi}
\title{Validate inputs and build variable sets for multi-reference calibration}
\usage{
check_input_multi(sc, sp_list, vars_list, wts_cols, verbose = FALSE)
}
\arguments{
\item{sc}{Data frame. The nonprobability sample.}

\item{sp_list}{List of reference sample data frames.}

\item{vars_list}{List of character vectors of predictor variable names
shared between `sc` and each reference sample.}

\item{wts_cols}{Character vector of survey-weight column names,
one per element of `sp_list`.}

\item{verbose}{Logical. If `TRUE`, variable-set messages are printed.}
}
\value{
A list with components:
  \item{sc}{The nonprobability sample data frame (unchanged).}
  \item{sp_list}{The reference sample list (unchanged).}
  \item{vars_XC}{Character vector: union of all per-sample variable sets,
    used to build the `sc` design matrix `Xc`.}
  \item{xcol}{List of character vectors: the variables contributed by each
    reference sample after removing variables already covered by earlier
    samples (no intercept).}
  \item{wts_cols}{Character vector of weight column names (unchanged).}
  \item{log}{Character vector of log messages describing the variable sets.}
}
\description{
Validate inputs and build variable sets for multi-reference calibration
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.pwmean.R
\name{summary.pwmean}
\alias{summary.pwmean}
\title{Summary method for pwmean objects}
\usage{
\method{summary}{pwmean}(object, ...)
}
\arguments{
\item{object}{An object of class \code{"pwmean"}, returned by
\code{\link{pwmean}}.}

\item{...}{Additional arguments, currently unused.}
}
\value{
Invisibly returns \code{object}.
}
\description{
Provides console output for objects of class \code{"pwmean"}, including
unweighted and pseudo-weighted mean estimates, standard errors, confidence
intervals, and optional domain-level summaries.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.pwmean.R
\name{print.pwmean_factor}
\alias{print.pwmean_factor}
\title{Print method for pwmean objects with categorical outcomes}
\usage{
\method{print}{pwmean_factor}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"pwmean_factor"}, returned by
\code{\link{pwmean}} when \code{y} is a factor.}

\item{...}{Additional arguments, currently unused.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Displays pseudo-weighted prevalence estimates and their uncertainty.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s5_multi_solver_helpers.R
\name{multi_raking_fp}
\alias{multi_raking_fp}
\title{Compute reference weighted totals for multi-reference calibration}
\usage{
multi_raking_fp(Xp_list, wts_list, label = "Multi_Calibration")
}
\arguments{
\item{Xp_list}{List of reference-sample design matrices.}

\item{wts_list}{List of numeric weight vectors, one per element of
`Xp_list`.}

\item{label}{Character scalar used as a prefix in error messages.}
}
\value{
A named numeric vector of length equal to the total number of
  columns across all elements of `Xp_list`, containing the survey-weighted
  column totals \eqn{\sum_j w_{ij} X_{ij}} stacked in block order.
}
\description{
Compute reference weighted totals for multi-reference calibration
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s6_output.R
\name{build_domains_df}
\alias{build_domains_df}
\title{Build the estimates summary data frame}
\usage{
build_domains_df(labels, mean_unw, se_unw, mean_adj, se_adj)
}
\arguments{
\item{labels}{Character vector of domain labels, one entry per domain.}

\item{mean_unw}{Numeric vector of unweighted (naive) domain means.}

\item{se_unw}{Numeric vector of standard errors for the unweighted means.}

\item{mean_adj}{Numeric vector of pseudo-weighted (adjusted) domain means.}

\item{se_adj}{Numeric vector of standard errors for the adjusted means.}
}
\value{
A data frame with one row per domain and columns `domain`,
  `unweighted_mean`, `unweighted_se`, `unweighted_lower`,
  `unweighted_upper`, `adjusted_mean`, `adjusted_se`, `adjusted_lower`,
  and `adjusted_upper`.
}
\description{
Constructs the `estimates` data frame that is stored in a `pwmean` object.
Each row corresponds to one estimate. The 95% confidence intervals are
computed using the 0.975 normal quantile (approximately 1.96).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s5_alp.R
\name{alp_build}
\alias{alp_build}
\title{Build pseudo-weights using the ALP method}
\usage{
alp_build(
  vars,
  sc,
  sp,
  sp_des,
  wts.col,
  control,
  verbose = FALSE,
  log_messages = NULL
)
}
\arguments{
\item{vars}{Character vector of predictor variable names.}

\item{sc}{Data frame. The nonprobability sample.}

\item{sp}{Data frame. The probability reference sample.}

\item{sp_des}{A `survey.design2` or `svyrep.design` object for `sp`.}

\item{wts.col}{Character scalar. Name of the survey-weight column in `sp`.}

\item{control}{A list created by `pw_solver_control()`.}

\item{verbose}{Logical. If `TRUE`, convergence messages are printed.}

\item{log_messages}{Character vector of messages accumulated in earlier
steps, appended to the returned object unchanged.}
}
\value{
A list with components:
  \item{weights}{Numeric vector of ALP pseudo-weights for `sc`.}
  \item{coefficients}{Named numeric vector of participation model coefficients.}
  \item{solver_diagnostics}{List of solver convergence diagnostics.}
  \item{log_messages}{Updated character vector of log messages.}
  \item{internal}{List containing design matrices (`Xc`, `Xp`), fitted
    participation probabilities (`p_sc`, `p_sp`), and sandwich components
    (`S_beta`, `D`) needed by the estimation stage.}
}
\description{
Build pseudo-weights using the ALP method
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nleqslv.R
\name{check_nleqslv_result}
\alias{check_nleqslv_result}
\title{Check the result returned by nleqslv}
\usage{
check_nleqslv_result(root, label, ftol)
}
\arguments{
\item{root}{The list returned by `nleqslv::nleqslv()`.}

\item{label}{Character string used as a prefix in messages to identify
the calling method.}

\item{ftol}{The function-value convergence tolerance, used to assess
whether partial convergence is acceptable.}
}
\value{
`invisible(TRUE)` on success or acceptable convergence;
  otherwise stops or warns.
}
\description{
Inspects the termination code (`termcd`) from `nleqslv::nleqslv()` and
either returns silently on success, issues a warning for partial
convergence, or stops with an informative error message on failure.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.pwmean.R
\name{print.pwmean}
\alias{print.pwmean}
\title{Print method for pwmean objects}
\usage{
\method{print}{pwmean}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"pwmean"}, returned by \code{\link{pwmean}}.}

\item{...}{Additional arguments, currently unused.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Displays the pseudo-weighted mean estimate and its uncertainty.
For factor-like domain variables, prints one row per domain level.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.na.action.R
\name{na.action.pw_fit}
\alias{na.action.pw_fit}
\title{Extract NA action from a pw_fit object}
\usage{
\method{na.action}{pw_fit}(object, ...)
}
\arguments{
\item{object}{An object of class \code{"pw_fit"} returned by \code{\link{est_pw}}.}

\item{...}{Additional arguments (not used).}
}
\value{
The \code{na.action} object recorded by \code{\link{est_pw}} during
  the build step: an integer vector of the nonprobability-sample rows omitted
  because of missing participation model variables (of class \code{"omit"} or
  \code{"exclude"}), or \code{NULL} if no rows were omitted.
}
\description{
Returns the \code{na.action} component recorded during the build step.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sp1_bootstrap.R
\docType{data}
\name{sp1_bootstrap}
\alias{sp1_bootstrap}
\title{Probability Reference Sample 1 with Bootstrap Replicate Weights (sp1_bootstrap)}
\format{
A data frame with 3494 rows and 512 columns. The first 12 columns
  are the substantive survey variables from \code{sp1} (\code{agecat},
  \code{marital}, \code{race}, \code{education}, \code{employment},
  \code{smoking}, \code{comorbidity}, \code{psa_level}, \code{BMI},
  \code{diabetes}, \code{pros_enlarged}, \code{wts_sp1}). The remaining
  500 columns are bootstrap replicate weights named \code{bw1} through
  \code{bw500} (numeric).
}
\source{
Derived from \code{sp1} (National Health and Nutrition Examination Survey,
NHANES, 1999--2010 cycles), with bootstrap replicate weights added by the
package authors using the Rao-Wu rescaling bootstrap.
}
\usage{
data(sp1_bootstrap)
}
\description{
A replicate-weight version of \code{sp1}, including the main survey weight
and 500 bootstrap replicate weights. It is provided to illustrate
design-based variance estimation with \code{\link[survey]{svrepdesign}}.
The original primary sampling unit and stratum identifiers are not included.
}
\details{
The bootstrap replicate weights were constructed from the original
stratified cluster design of \code{sp1}, using the Rao-Wu rescaling
bootstrap method. A total of \code{R = 500} replicates were
produced with \code{seed = 2026}.

The variables \code{psu_sp1} and \code{strata_sp1} are not included in this
dataset because they are not needed when using replicate weights for
variance estimation. The main survey weight \code{wts_sp1} and the replicate
weight columns \code{bw1}--\code{bw500} are sufficient for constructing a
replicate-weight survey design object via \code{survey::svrepdesign()}.
}
\examples{
data(sp1_bootstrap)

# Example: create replicate-weight survey design object
des_boot <- survey::svrepdesign(
  data    = sp1_bootstrap,
  weights = ~wts_sp1,
  repweights = "bw[0-9]+",
  type    = "bootstrap",
  combined.weights = FALSE
)

summary(des_boot)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s5_multi_s3_matrix.R
\name{multi_matrix_construction}
\alias{multi_matrix_construction}
\title{Construct block design matrices for multi-reference calibration}
\usage{
multi_matrix_construction(vars_XC, sc, sp_list, xcol, wts_cols)
}
\arguments{
\item{vars_XC}{Character vector of all predictor variables (union across
reference samples), used to build the `sc` design matrix `Xc`.}

\item{sc}{Data frame. The nonprobability sample.}

\item{sp_list}{List of reference sample data frames.}

\item{xcol}{List of character vectors: the variables contributed by each
reference sample (as returned by `check_input_multi`).}

\item{wts_cols}{Character vector of survey-weight column names,
one per element of `sp_list`.}
}
\value{
A list with components:
  \item{Xc}{Numeric design matrix for `sc`, with intercept prepended,
    of dimension \eqn{n_c \times p}.}
  \item{Xp_list}{List of numeric design matrices for each reference sample.
    The first block includes an intercept column; subsequent blocks do not.}
  \item{wts_list}{List of numeric weight vectors, one per reference sample.}
}
\description{
Construct block design matrices for multi-reference calibration
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nleqslv.R
\name{solve_participation_model}
\alias{solve_participation_model}
\title{Solve the participation model estimating equations}
\usage{
solve_participation_model(beta_start, fn, jac, label, control = NULL, ...)
}
\arguments{
\item{beta_start}{Numeric vector of starting values for the coefficient
vector \eqn{\beta}.}

\item{fn}{Function returning the estimating equation vector \eqn{g(\beta)}.}

\item{jac}{Function returning the Jacobian matrix \eqn{J(\beta)}.}

\item{label}{Character string used as a prefix in error messages to
identify the calling method.}

\item{control}{A list created by `pw_solver_control()`.}

\item{...}{Additional arguments passed to `fn` and `jac`
(e.g., design matrices and weights).}
}
\value{
A list with components `coefficients`, `iterations`, `solver`,
  `solver_result`, `solver_method`, `fvec`, `termcd`, and `message`.
}
\description{
A unified wrapper that calls `nleqslv::nleqslv()` to solve the
system of estimating equations \eqn{g(\beta) = 0}, then validates the
result via `check_nleqslv_result()`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nleqslv.R
\name{prepare_nleqslv_args}
\alias{prepare_nleqslv_args}
\title{Assemble arguments for nleqslv}
\usage{
prepare_nleqslv_args(control)
}
\arguments{
\item{control}{A list created by `pw_solver_control()`.}
}
\value{
A list with components `method`, `global`, `xscalm`, and `control`
  (containing `ftol`, `xtol`, `maxit`, `trace`, and any user-supplied
  extras from `nleqslv_control`).
}
\description{
Extracts and assembles the solver settings from a control object created
by `pw_solver_control()` into the format expected by `nleqslv::nleqslv()`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s6_sc_update.R
\name{reconstruct_sc_output}
\alias{reconstruct_sc_output}
\title{Reconstruct the sample data frame with fitted pseudo-weights}
\usage{
reconstruct_sc_output(sc0, w_fit, keep_sc, na_mode, na_action_obj, sc_wname)
}
\arguments{
\item{sc0}{The original sample data frame (before any NA removal).}

\item{w_fit}{Numeric vector of fitted pseudo-weights, one per kept row.}

\item{keep_sc}{Logical vector identifying which rows of `sc0` were retained
for fitting.}

\item{na_mode}{Character string; one of `"omit"`, `"exclude"`,
`"fail"`, or `"pass"`.}

\item{na_action_obj}{The `na.action` attribute produced during NA
removal, or NULL.}

\item{sc_wname}{Name of the column in which pseudo-weights will be stored.}
}
\value{
A data frame derived from `sc0` with a pseudo-weight column
  named `sc_wname`.
}
\description{
Inserts the fitted pseudo-weights `w_fit` back into the original sample
data frame `sc0`, using the NA handling strategy used during
estimation. For `"omit"`, only the rows that were kept are returned.
For `"exclude"`, all original rows are returned with NA weights
for rows that were dropped. For `"fail"` or `"pass"`, all rows
are assumed to be present and weights are attached directly.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s2_prepare_sc.R
\name{prepare_sc_data}
\alias{prepare_sc_data}
\title{Extract and align sc data from a build object}
\usage{
prepare_sc_data(build)
}
\arguments{
\item{build}{A `pw_fit` object returned by the build step.}
}
\value{
A list with components:

- `sc`: data frame of build-stage complete cases.
- `X`: covariate matrix `build$internal$Xc`.
- `w`: pseudo-weight vector with NA placeholders removed.
- `idx_keep`: integer index of which rows of `raw_sc` were retained.
}
\description{
Reads the raw sc sample, the build-stage complete-case index
(`keep_sc`), the covariate matrix `Xc`, and the pseudo-weights from
`build`. Remove NA placeholders inserted by `na.exclude` so that all
returned vectors and matrices have the same length (the number of
build-stage complete cases).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s1_check_inputs.R
\name{check_ipwm_inputs_estimate}
\alias{check_ipwm_inputs_estimate}
\title{Validate inputs at the estimation stage}
\usage{
check_ipwm_inputs_estimate(build, y, zcol = NULL)
}
\arguments{
\item{build}{A `pw_fit` object returned by the build step.}

\item{y}{Single character string naming the numeric or categorical outcome
variable in `build$internal$raw_sc`.}

\item{zcol}{Single character string naming the domain variable in
`build$internal$raw_sc`, or NULL for the overall mean.}
}
\value{
`invisible(TRUE)` on success; stops with an informative message
  on failure.
}
\description{
Checks that `build` is a valid `pw_fit` object with all required internal
fields, that `y` names a numeric or factor column in the raw convenience
sample, and
that `zcol` (if supplied) is a supported domain variable with at least two
non-missing levels in both the full sample and the build-stage complete cases.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.na.action.R
\name{na.action.pwmean}
\alias{na.action.pwmean}
\title{Extract NA action from a pwmean object}
\usage{
\method{na.action}{pwmean}(object, ...)
}
\arguments{
\item{object}{An object of class \code{"pwmean"} returned by \code{\link{pwmean}}.}

\item{...}{Additional arguments (not used).}
}
\value{
The \code{na.action} object recorded by \code{\link{pwmean}}: an
  integer vector of rows omitted because of missing outcome or domain values
  (of class \code{"omit"} or \code{"exclude"}), or \code{NULL} if no rows were
  omitted.
}
\description{
Returns the \code{na.action} component recorded during estimation,
mimicking \code{\link[stats]{na.action}} behavior for fitted model objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s4_multi.R
\name{multi_estimate}
\alias{multi_estimate}
\title{Estimate step for multi-reference raking}
\usage{
multi_estimate(Y, Z, w, X, D, S_beta)
}
\arguments{
\item{Y}{Outcome vector for the nonprobability sample.}

\item{Z}{Domain indicator vector. Use `rep(1, length(Y))` for the
overall mean.}

\item{w}{Estimated pseudo-weights from the multi-reference raking
build step.}

\item{X}{Design matrix for the nonprobability sample.}

\item{D}{Block-diagonal design-based variance-covariance matrix of the
estimated auxiliary totals from the reference surveys.}

\item{S_beta}{Sensitivity matrix for the multi-reference raking
estimating equations, typically \code{t(w * X) \%*\% X}.}
}
\value{
A list with components `mean` and `variance`.
}
\description{
Computes the domain-specific pseudo-weighted Hájek mean and its
Taylor-linearized variance for the multi-reference raking estimator.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s5_multi_solver_helpers.R
\name{multi_raking_start}
\alias{multi_raking_start}
\title{Compute starting values for multi-reference calibration}
\usage{
multi_raking_start(Xc, Xp_list, block_cols, f_p, label = "Multi_Calibration")
}
\arguments{
\item{Xc}{Numeric design matrix for the nonprobability sample.}

\item{Xp_list}{List of reference-sample design matrices.}

\item{block_cols}{List of integer vectors mapping each reference-sample
block to its column indices in `Xc`, as returned by
`multi_raking_block_cols`.}

\item{f_p}{Named numeric vector of reference weighted totals, as returned
by `multi_raking_fp`.}

\item{label}{Character scalar used as a prefix in error messages.}
}
\value{
A numeric vector of length `ncol(Xc)` containing the starting
  values for the coefficient vector \eqn{\beta}, with the intercept
  initialized to match the ratio of reference to sample weighted totals
  and all other coefficients set to zero.
}
\description{
Compute starting values for multi-reference calibration
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s4_dispatch.R
\name{dispatch_estimator_one_domain}
\alias{dispatch_estimator_one_domain}
\title{Run the estimator for a single domain}
\usage{
dispatch_estimator_one_domain(build, Y, zvec, w, X)
}
\arguments{
\item{build}{A `pw_fit` object returned by the build step.}

\item{Y}{Outcome vector of length `n`.}

\item{zvec}{Domain indicator vector of length `n`.}

\item{w}{Pseudo-weight vector of length `n`.}

\item{X}{Design matrix for the convenience sample, with dimension
`n x p`.}
}
\value{
A list with components `mean` and `variance`.
}
\description{
Selects and calls the specific estimator function based on
`build$method`, passing the relevant matrices and vectors
for one domain at a time.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.est_pw.R
\name{summary.pw_fit}
\alias{summary.pw_fit}
\title{Summarize a Pseudo-Weight Fit}
\usage{
\method{summary}{pw_fit}(object, ...)
}
\arguments{
\item{object}{An object of class \code{"pw_fit"}, returned by \code{\link{est_pw}}.}

\item{...}{Additional arguments, currently unused.}
}
\value{
Invisibly returns \code{object}.
}
\description{
Summarize a Pseudo-Weight Fit
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s5_multi_s1_reorder.R
\name{sort_by_sp_size}
\alias{sort_by_sp_size}
\title{Sort reference samples by size}
\usage{
sort_by_sp_size(sp, vars, weight, design, sp_order, verbose = FALSE)
}
\arguments{
\item{sp}{List of reference sample data frames.}

\item{vars}{List of character vectors of predictor variable names,
one per element of `sp`.}

\item{weight}{Character vector of survey-weight column names,
one per element of `sp`.}

\item{design}{List of `survey.design2` or `svyrep.design` objects,
one per element of `sp`.}

\item{sp_order}{Character scalar. `"size"` reorders reference samples
largest-first; `"given"` keeps the user-supplied order.}

\item{verbose}{Logical. If `TRUE`, a summary of the ordering is printed.}
}
\value{
A list with components:
  \item{sp}{Reordered list of reference sample data frames.}
  \item{vars}{Reordered list of predictor variable name vectors.}
  \item{weight}{Reordered character vector of weight column names.}
  \item{design}{Reordered list of survey design objects.}
  \item{order_used}{Integer vector giving the reordering index.}
  \item{log}{Character vector of log messages describing the ordering.}
}
\description{
Sort reference samples by size
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s1_parse_design.R
\name{parse_ipwm_data}
\alias{parse_ipwm_data}
\title{Parse the data list passed to est_pw}
\usage{
parse_ipwm_data(data)
}
\arguments{
\item{data}{A named or unnamed list. The first element is the
nonprobability sample as a data frame. Every remaining element is a
survey design object (`survey.design2`, `survey.design`, or
`svyrep.design`).}
}
\value{
A list with four elements:

- `sc`: the nonprobability sample (first element of `data`).
- `sp_des`: named list of the survey design objects.
- `sp_vars`: named list of plain data frames, one per reference survey,
  with design variables removed and a `sp_wts` column added.
- `n_ref`: integer giving the number of reference surveys.
}
\description{
Splits the input list into its two parts: the nonprobability sample
(`sc`) and one or more reference survey designs (`sp_des`).
For each design it calls `extract_analysis_data()` to remove design
variables and attach a `sp_wts` column, giving back a plain data
frame ready for modeling. If the list elements have no names, or if
any element name is an empty string (partially named list), default
names `"sp[[1]]"`, `"sp[[2]]"`, ... are assigned to all elements.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s5_multi_solver_helpers.R
\name{multi_raking_block_cols}
\alias{multi_raking_block_cols}
\title{Map each reference-sample block to its column indices in Xc}
\usage{
multi_raking_block_cols(Xc, Xp_list, label = "Multi_Calibration")
}
\arguments{
\item{Xc}{Numeric design matrix for the nonprobability sample, with named
columns.}

\item{Xp_list}{List of reference-sample design matrices, each with named
columns that are a subset of `colnames(Xc)`.}

\item{label}{Character scalar used as a prefix in error messages.}
}
\value{
A list of integer vectors, one per element of `Xp_list`, giving the
  column positions in `Xc` that correspond to the columns of each
  `Xp_list[[j]]`.
}
\description{
Map each reference-sample block to its column indices in Xc
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nleqslv.R
\name{check_estimating_system}
\alias{check_estimating_system}
\title{Validate the Jacobian and estimating equation vector}
\usage{
check_estimating_system(J, g, label = "method", tol_singular = 1e-07)
}
\arguments{
\item{J}{Square numeric Jacobian matrix of dimension \eqn{p \times p}.}

\item{g}{Numeric estimating equation vector of length \eqn{p}.}

\item{label}{Character string used as a prefix in error messages to
identify the calling method.}

\item{tol_singular}{Tolerance passed to `qr()` for detecting rank
deficiency. Default is `1e-7`, matching R's `qr()` default.}
}
\value{
`invisible(TRUE)` if all checks pass; otherwise stops with
  an informative error message.
}
\description{
Checks that the Jacobian `J` and estimating equation vector `g`
are numerically valid and that `J` is square and full rank before
a Newton-Raphson step is taken.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_doc.R
\name{pwmean}
\alias{pwmean}
\title{Estimate Pseudo-Weighted Means, Prevalences, and Standard Errors}
\usage{
pwmean(object, y, zcol = NULL, na.action = stats::na.omit)
}
\arguments{
\item{object}{An object of class \code{"pw_fit"} returned by
\code{\link{est_pw}}.}

\item{y}{A character string specifying the name of the outcome variable in the
nonprobability sample stored in \code{object}. The outcome must be numeric
for mean estimation, including binary 0/1 outcomes for prevalence
estimation, or a factor for category prevalence estimation.}

\item{zcol}{Optional character string giving the name of a categorical domain
variable in the nonprobability sample stored in the \code{object}. If
\code{NULL}, the overall mean is estimated. If supplied, estimates are
computed within domains defined by this variable. The following column
types are supported: \code{logical} (must contain both \code{TRUE} and
\code{FALSE}); \code{numeric} or \code{integer} containing only \code{0}
and \code{1} after removing missing values; \code{character} (empty
strings are treated as missing values); and \code{factor} (unused levels
are dropped).}

\item{na.action}{Function specifying how missing values in \code{y} and
\code{zcol} should be handled. Default is \code{stats::na.omit}.}
}
\value{
An object of class \code{"pwmean"} containing unweighted and pseudo-weighted
estimates, standard errors, and confidence intervals. For categorical outcomes,
the estimate columns contain category prevalences.

\describe{
  \item{\code{method}}{
    Character. The pseudo-weighting method used.
  }

  \item{\code{estimates}}{
    A data frame containing the unweighted and pseudo-weighted estimates.

    For numeric outcomes, the first column is \code{domain}. If
    \code{zcol = NULL}, \code{domain} is \code{"Overall"}. If \code{zcol}
    is a \code{logical} variable or a \code{numeric}/\code{integer}
    variable containing only \code{0} and \code{1}, there is one row with
    \code{domain} labeled \code{"<zcol> = 1"}. If \code{zcol} is a
    \code{factor} or \code{character} variable, there is one row per
    \code{zcol} level, with \code{domain} labeled
    \code{"<zcol> = <level>"}.

    For categorical outcomes, the first two columns are
    \code{category} and \code{domain}. \code{category} identifies the
    outcome level as \code{"<y> = <level>"}. If \code{zcol = NULL},
    \code{domain} is \code{"Overall"} for each outcome level. If
    \code{zcol} is supplied, the rows are formed by each outcome category
    within each domain, and \code{domain} follows the same labels described
    above for \code{zcol}.

    The columns are:
    \describe{
      \item{\code{category}}{Category label for categorical outcomes only.}
      \item{\code{domain}}{Domain label.}
      \item{\code{unweighted_mean}, \code{unweighted_se}}{Unweighted
        mean of \code{y} and its standard error.}
      \item{\code{unweighted_lower}, \code{unweighted_upper}}{Bounds of the
        95\% confidence interval for the unweighted mean, based on the
        normal approximation.}
      \item{\code{adjusted_mean}, \code{adjusted_se}}{Pseudo-weighted mean
        of \code{y} and its standard error.}
      \item{\code{adjusted_lower}, \code{adjusted_upper}}{Bounds of the
        95\% confidence interval for the pseudo-weighted mean, based on the
        normal approximation.}
    }
  }

  \item{\code{na.action}}{
    Integer vector of row indices omitted at the outcome-estimation step,
    with class \code{"omit"} or \code{"exclude"} matching the
    \code{na.action} argument, or \code{NULL} if no observations were
    omitted. The indices refer to the nonprobability sample available to
    \code{pwmean()} after missing-data handling in
    \code{est_pw()}.
  }

  \item{\code{call}}{
    The matched function call.
  }
}
}
\description{
Computes pseudo-weighted means and standard errors using a fitted
pseudo-weight object of class \code{"pw_fit"} returned by
\code{\link{est_pw}}. The function applies second-layer missing-data
handling for the outcome and optional domain variable, and then estimates
overall or domain-specific means or prevalences using the pseudo-weighting
method stored in \code{object}.
}
\details{
\strong{Missing data handling (layer 2).}
After pseudo-weights are constructed by \code{est_pw()}, estimation of the
mean requires complete cases for the outcome \code{y} and, if supplied, the
domain variable \code{zcol}. The argument \code{na.action} controls how
these missing values are handled at the outcome-estimation step.

\strong{Input \code{object}.}
The \code{object} argument should be an object of class \code{"pw_fit"}
returned by \code{\link{est_pw}}. It stores the estimated pseudo-weights,
participation model information, and design-based quantities required for point
and variance estimation.

\strong{Categorical outcomes.}
When \code{y} is a categorical variable (defined as a factor in R),
\code{pwmean()} estimates the prevalence (proportion) of each category.
To do so, each category is internally converted into a 0/1 indicator
variable, and the pseudo-weighted mean estimator is then computed for each
indicator.
}
\examples{
\donttest{
data(sc)
data(sp1)

ref1_design <- survey::svydesign(
  ids     = ~psu_sp1,
  strata  = ~strata_sp1,
  weights = ~wts_sp1,
  data    = sp1,
  nest    = TRUE
)

fit <- est_pw(
  data      = list(sc, ref1_design),
  p_formula = ~ agecat + race + education + comorbidity + BMI + diabetes,
  method    = "calibration",
  control   = pw_solver_control(ftol=1e-6)
)

out <- pwmean(fit, y = "psa_level", zcol = "BMI")

print(out)

summary(out)
}

}
\seealso{
\code{\link{est_pw}},
\code{\link{summary.pwmean}},
\code{\link{print.pwmean}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s5_multi.R
\name{ipwm_multi_build}
\alias{ipwm_multi_build}
\title{Build pseudo-weights for the multi-reference calibration method}
\usage{
ipwm_multi_build(
  sc,
  sp,
  vars,
  weight,
  sp_des,
  sp_order,
  control,
  verbose = FALSE,
  log_messages = NULL
)
}
\arguments{
\item{sc}{Data frame. The nonprobability sample.}

\item{sp}{List of data frames. Each element is one probability reference
sample (already NA-processed and p_formula-processed).}

\item{vars}{List of character vectors, one per reference sample, giving the
predictor variable names shared with `sc`.}

\item{weight}{Character vector, one per reference sample, giving the
survey-weight column name in each element of `sp`.}

\item{sp_des}{List of `survey.design2` or `svyrep.design` objects,
one per reference sample.}

\item{sp_order}{Character scalar, either `"size"` (sort reference samples
largest-first) or `"given"` (keep user order).}

\item{control}{A list created by `pw_solver_control()`.}

\item{verbose}{Logical. If `TRUE`, progress messages are printed.}

\item{log_messages}{Character vector of messages accumulated in earlier
steps, forwarded into the returned object.}
}
\value{
A list with components:
  \item{pseudo_weights}{Numeric vector of estimated pseudo-weights for `sc`.}
  \item{coefficients}{Named numeric vector of participation model coefficients.}
  \item{method}{Character scalar, always `"multi"`.}
  \item{solver_diagnostics}{List of solver convergence diagnostics.}
  \item{internal}{List of intermediate objects (block design matrices,
    reference weighted totals, block column indices, sandwich components,
    and sorted metadata) needed by the estimation stage.}
}
\description{
Build pseudo-weights for the multi-reference calibration method
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_d_matrix_clw.R
\name{compute_D_CLW}
\alias{compute_D_CLW}
\title{Compute design-based covariance matrix D for CLW}
\usage{
compute_D_CLW(sp_des, pi_sp, Xp)
}
\arguments{
\item{sp_des}{A survey design object of class `"survey.design2"` or
`"svyrep.design"`.}

\item{pi_sp}{Numeric vector of estimated values multiplying `X_p`
in the CLW estimating equation. Length must equal the number of
rows in `sp_des$variables`.}

\item{Xp}{Numeric matrix of dimension `n_p x p`. Each row is the
covariate vector `x_j` used in the CLW estimating equation.}
}
\value{
A `p x p` covariance matrix `D`.
}
\description{
Internal helper to compute
\deqn{D = Var_p\left( \sum_{j \in s_p} d_j \pi_j x_j \right)}
from an existing survey design object created by
survey::svydesign() or survey::svrepdesign().
}
\details{
In the CLW method, the probability sample contribution in the
estimating equation is \eqn{\pi_j d_j x_j}, where `pi_j` corresponds
to the inverse of the estimated pseudo-weight. Thus the design-based
covariance matrix is obtained by treating \eqn{h_j = \pi_j x_j}
as the survey total variable.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.pw_fit.R
\name{print.pw_fit}
\alias{print.pw_fit}
\title{Print method for pw_fit objects}
\usage{
\method{print}{pw_fit}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"pw_fit"}, returned by \code{\link{est_pw}}.}

\item{...}{Additional arguments, currently unused.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Compact one-screen overview of a fitted pseudo-weight object: the call, the
pseudo-weighting method, the participation model size, solver convergence,
and a summary of the estimated pseudo-weights. For the full coefficient
table and detailed solver diagnostics, use \code{\link{summary.pw_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_d_matrix_cali.R
\name{compute_D_raking}
\alias{compute_D_raking}
\title{Compute design-based covariance matrix D for calibration}
\usage{
compute_D_raking(sp_des, Xp)
}
\arguments{
\item{sp_des}{A survey design object of class `"survey.design2"` or
`"svyrep.design"`.}

\item{Xp}{Numeric matrix of dimension `n_p x p`. Each row is the
covariate vector `x_j` used in the calibration estimating equation.}
}
\value{
A `p x p` covariance matrix `D`.
}
\description{
Internal helper to compute
\deqn{D = Var_p\left( \sum_{j \in s_p} d_j x_j \right)}
from an existing survey design object created by
survey::svydesign() or survey::svrepdesign().
}
\details{
In the raking ratio calibration method, the probability sample
contribution in the estimating equation is \eqn{d_j x_j}, so the
design-based covariance matrix is obtained by treating
\eqn{h_j = x_j} as the survey total variable.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s2_auto_pfml.R
\name{p_formula_construction}
\alias{p_formula_construction}
\title{Auto-build a default participation model formula}
\usage{
p_formula_construction(sc, sp, weight)
}
\arguments{
\item{sc}{A data frame. The nonprobability sample.}

\item{sp}{A data frame (one-reference case) or a named list of data frames
(multi-reference case). Each data frame contains the analysis variables
of one reference survey, with design variables removed and a sampling
weight column appended (as produced by `extract_analysis_data()`).}

\item{weight}{A single character string (one-reference case) or a
character vector of the same length as `sp` (multi-reference case)
giving the name of the sampling weight column in each reference survey
data frame. This column is excluded from the candidate covariate set.}
}
\value{
A list with two elements:
\describe{
  \item{`p_formula`}{A one-sided formula (one-reference case) or a named
    list of one-sided formulas (multi-reference case).}
  \item{`log_messages`}{A character vector of messages describing the
    auto-generated formula(s), used for downstream printing when
    `verbose = TRUE`.}
}
}
\description{
Constructs a one-sided participation model formula from the variables
shared between the nonprobability sample (`sc`) and each reference
survey, excluding the sampling weight column. Called by `est_pw()` when
`p_formula = NULL`.
}
\details{
For each reference survey, the candidate covariates are the column names
present in both `sc` and the reference survey data frame after dropping
the sampling weight column named by `weight`. A one-sided formula of the
form `~ var1 + var2 + ...` is built from the remaining shared variables.
An error is raised if no shared covariates remain after excluding the
weight column.

The function distinguishes the one-reference case from the multi-reference
case by the type of `sp`: a plain data frame triggers the one-reference
path; a list of data frames triggers the multi-reference path. Note that
in R a `data.frame` is also a `list`, so the `is.data.frame()` check is
performed first.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.pwmean.R
\name{summary.pwmean_factor}
\alias{summary.pwmean_factor}
\title{Summary method for pwmean objects with categorical outcomes}
\usage{
\method{summary}{pwmean_factor}(object, ...)
}
\arguments{
\item{object}{An object of class \code{"pwmean_factor"}, returned by
\code{\link{pwmean}} when \code{y} is a factor.}

\item{...}{Additional arguments, currently unused.}
}
\value{
Invisibly returns \code{object}.
}
\description{
Provides console output for objects of class \code{"pwmean_factor"},
including unweighted and pseudo-weighted prevalence estimates, standard
errors, and confidence intervals.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nonprobsampling-package.R
\docType{package}
\name{nonprobsampling-package}
\alias{nonprobsampling}
\alias{nonprobsampling-package}
\title{nonprobsampling: Inference for Nonprobability Samples Using Multiple Reference Surveys}
\description{
\pkg{nonprobsampling} implements pseudo-weighting methods for
finite population inference from nonprobability samples, such as
convenience samples, volunteer cohorts, and opt-in panels. Because the
participation mechanism in a nonprobability sample is unknown, unadjusted
estimates of population means and prevalences may be biased. The package
addresses this issue by leveraging auxiliary information from one or more
probability reference surveys to estimate participation probabilities and
using their inverses as pseudo-weights to obtain bias-corrected estimates
of finite population means and prevalences.
}
\details{
The package implements the generalized estimating equation framework of
Landsman et al. (2026). Pseudo-weights are obtained by solving estimating
equations that equate auxiliary variable totals estimated from the
nonprobability sample with the corresponding totals estimated from one or
more probability reference surveys. Totals from the nonprobability sample are
computed using estimated pseudo-weights, whereas totals from the reference
surveys are computed using the known survey sampling weights.

Several one reference methods arise as special cases of this framework under
different choices of the weight and estimating functions:
\itemize{
  \item the raking ratio calibration method of
    Landsman et al. (2026);
  \item the adjusted logistic propensity (ALP) method of
    Wang, Valliant, and Li (2021);
  \item the Chen--Li--Wu (CLW) method of Chen, Li, and Wu (2020).
}

A key feature of the package is the multi-reference extension of the
calibration method. This extension enables the integration of auxiliary
information across multiple surveys when no single reference survey
contains all variables relevant to participation (Landsman et al., 2026),
with an optional cumulative precalibration step to preserve information on
the relationships between overlapping and unique auxiliary variables across
reference surveys.

Variance estimation is based on Taylor linearization. The resulting analytic
variance estimator accounts for uncertainty from pseudo-weight estimation and
for the sampling designs of the reference surveys through integration with
the \pkg{survey} package. The package also supports bootstrap-based variance
estimation when bootstrap weights are provided with the reference surveys.
}
\section{Typical workflow}{

Estimation is carried out in two steps:
\enumerate{
  \item \code{\link{est_pw}()} estimates pseudo-weights using the
    nonprobability sample and one or more reference survey design objects.
    This step fits the participation model and stores the internal quantities
    needed for subsequent estimation. It does not require an outcome variable.
  \item \code{\link{pwmean}()} uses the object returned by \code{est_pw()}
    to estimate a pseudo-weighted mean or prevalence for an outcome, either
    overall or by category. It also returns the standard error and confidence
    interval.
}

Numerical settings for solving the estimating equations can be specified with
\code{\link{pw_solver_control}()}.
}

\section{Datasets}{

The package includes example datasets for demonstrating one reference and
multiple reference analyses: \code{\link{sc}} is a nonprobability sample,
\code{\link{sp1}} and \code{\link{sp2}} are probability reference surveys,
and \code{\link{sp1_bootstrap}} contains replicate weights for
\code{sp1}. See the package vignette for worked examples.
}

\references{
Chen, Y., Li, P., and Wu, C. (2020).
Doubly robust inference with nonprobability survey samples.
\emph{Journal of the American Statistical Association},
115(532), 2011--2021.
doi:10.1080/01621459.2019.1677241

Landsman, V., Wang, L., Carrillo-Garcia, I., Mitani, A. A.,
Smith, P. M., Graubard, B. I., Bui, T., and Carnide, N. (2026).
Correction for participation bias in nonprobability samples using multiple
reference surveys.
\emph{Statistics in Medicine}, 45(3--5).
doi:10.1002/sim.70403

Wang, L., Valliant, R., and Li, Y. (2021).
Adjusted logistic propensity weighting methods for population inference
using nonprobability volunteer-based epidemiologic cohorts.
\emph{Statistics in Medicine}, 40(24), 5237--5250.
doi:10.1002/sim.9122
}
\seealso{
\code{\link{est_pw}}, \code{\link{pwmean}}, \code{\link{pw_solver_control}}
}
\author{
\strong{Maintainer}: Jiakun Lin \email{jiak.lin@alumni.utoronto.ca}

Authors:
\itemize{
  \item Victoria Landsman
  \item Aya A. Mitani
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_d_matrix_alp.R
\name{compute_D_ALP}
\alias{compute_D_ALP}
\title{Compute design-based covariance matrix D for ALP}
\usage{
compute_D_ALP(sp_des, p_sp, Xp)
}
\arguments{
\item{sp_des}{A survey design object of class `"survey.design2"` or
`"svyrep.design"`.}

\item{p_sp}{Numeric vector of estimated participation probabilities
for the probability sample. Length must equal the number of rows in
`sp_des$variables`.}

\item{Xp}{Numeric matrix of dimension `n_p x p`. Each row is the
covariate vector `x_j` used in the ALP estimating equation.}
}
\value{
A `p x p` covariance matrix `D`.
}
\description{
Internal helper to compute
\deqn{D = Var_p\left( \sum_{j \in s_p} d_j p_j x_j \right)}
from an existing survey design object created by
survey::svydesign() or survey::svrepdesign().
}
\details{
In the ALP method, the probability sample contribution in the
estimating equation is \eqn{p_j d_j x_j}, so the design-based
covariance matrix is obtained by treating \eqn{h_j = p_j x_j}
as the survey total variable.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s4_dispatch.R
\name{dispatch_estimator}
\alias{dispatch_estimator}
\title{Dispatch the estimator across domains}
\usage{
dispatch_estimator(build, yz_data)
}
\arguments{
\item{build}{A `pw_fit` object returned by the build step.}

\item{yz_data}{A list returned by `process_na_yz()`, containing
`Y`, `w`, `X`, and `domain`.}
}
\value{
A list with components:

- `type`: either `"single"` or `"multi"`.
- `labels`: character vector of domain labels.
- `estimates`: a list of per-domain results, each with `mean` and `variance`.
}
\description{
Determines the domain mode (overall, binary, or factor) from
`yz_data` and calls `dispatch_estimator_one_domain()` for each domain,
returning a unified result list.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s3_helper.R
\name{standardize_zcol}
\alias{standardize_zcol}
\title{Standardize a domain variable into a common internal format}
\usage{
standardize_zcol(data, zcol = NULL)
}
\arguments{
\item{data}{A data frame containing the column named by `zcol`.}

\item{zcol}{Single character string naming the domain variable in `data`,
or NULL for the overall (no-domain) case.}
}
\value{
A list with components:

- `mode`: one of `"overall"`, `"binary"`, or `"factor"`.
- `z_name`: the value of `zcol`, or NULL when `zcol` is NULL.
- `labels`: character vector of domain labels shown to the user.
- `indicators`: data frame of integer 0/1 indicator columns (one column
  per domain level), or NULL in the overall case.
}
\description{
Converts `zcol` in `data` to a standard form used throughout the
estimation pipeline. Logical and binary numeric (0/1) variables become a
single integer indicator column; character variables are trimmed and
coerced to a factor; factor variables are dropped of unused levels and
expanded to one integer indicator column per level. The returned list
always has the same structure so downstream code can branch on `mode`
alone.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sp1.R
\docType{data}
\name{sp1}
\alias{sp1}
\title{Probability Reference Sample 1 (sp1)}
\format{
A data frame with 3494 observations and 14 variables:
\describe{
  \item{agecat}{Age category (factor with 4 levels: 1 = 55--59, 2 = 60--64, 3 = 65--69, 4 = 70+)}
  \item{marital}{Marital status (factor with 4 levels: 1 = Married Or Living As Married, 2 = Widowed, 3 = Divorced or Separated, 4 = Never Married)}
  \item{race}{Race category (factor with 4 levels: 1 = White, 2 = Black, 3 = Hispanic, 4 = Other)}
  \item{education}{Education level (factor with 5 levels: 1 = Less Than 8 Years, 2 = 8--11 Years, 3 = 12 Years Or Completed High School, 4 = College Graduate, 5 = Postgraduate)}
  \item{employment}{Employment status (factor with 2 levels: 0 = Not Working, 1 = Working)}
  \item{smoking}{Smoking status (factor with 3 levels: 1 = Never Smoker, 2 = Former Smoker, 3 = Current Smoker)}
  \item{comorbidity}{General comorbidity indicator (factor with 2 levels: 0 = No, 1 = Yes)}
  \item{psa_level}{Serum prostate-specific antigen level (numeric)}
  \item{BMI}{Body mass index category (factor with 4 levels: "Normal", "Overweight", "Obese", "Morbidly Obese")}
  \item{diabetes}{Diabetes diagnosis indicator (factor with 2 levels: 0 = No, 1 = Yes)}
  \item{pros_enlarged}{Prostate enlargement indicator (factor with 2 levels: 0 = No, 1 = Yes)}
  \item{strata_sp1}{Stratum identifier for complex survey design (numeric)}
  \item{psu_sp1}{Primary sampling unit identifier for complex survey design (numeric)}
  \item{wts_sp1}{10-year interview sampling weights (numeric)}
}
}
\source{
Derived from the National Health and Nutrition Examination Survey (NHANES),
1999--2010 cycles, conducted by the U.S. National Center for Health
Statistics (NCHS).
}
\usage{
data(sp1)
}
\description{
This dataset represents a probability sample derived from the
1999--2010 cycles of the National Health and Nutrition
Examination Survey (NHANES). It is used as a probability
reference survey to support the pseudo-weighting methods
implemented in the \code{nonprobsampling} package.
}
\details{
The dataset includes auxiliary variables shared with the nonprobability
sample \code{sc}, enabling the construction of pseudo-weights to adjust
for participation bias. Survey design variables and sampling weights are
provided to support design-consistent estimation.

The \code{sp1} dataset
contains the outcome variable \code{psa_level}, which is also observed in
\code{sc}, allowing for the evaluation of pseudo-weighted estimators against
estimates based on true sampling weights. It may also be incorporated into
the participation model, potentially enhancing bias reduction when
participation depends on the outcome.
}
\examples{
data(sp1)
str(sp1)
summary(sp1)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s7_finalize_output.R
\name{finalize_pw_fit}
\alias{finalize_pw_fit}
\title{Finalize and assemble the `pw_fit` return object}
\usage{
finalize_pw_fit(result, sc_out, sc0, sc_wname, na_mode, keep_sc, na_action_obj)
}
\arguments{
\item{result}{A list accumulating outputs from the estimation pipeline.}

\item{sc_out}{The reconstructed sample data frame produced by
`reconstruct_sc_output()`.}

\item{sc0}{The original (pre-NA-removal) sample data frame, stored for
diagnostic access via `result$internal$raw_sc`.}

\item{sc_wname}{Name of the pseudo-weight column in `sc_out`.}

\item{na_mode}{Character string describing how NAs were handled; stored in
`result$internal$na`.}

\item{keep_sc}{Logical vector of retained rows; stored in
`result$internal$na`.}

\item{na_action_obj}{The `na.action` attribute from NA removal; stored
in `result$internal$na`.}
}
\value{
`result` with `sc_updated`, `pseudo_weights`, and
  `internal` fields populated, and class set to `"pw_fit"`.
}
\description{
Attaches the reconstructed sample data frame, the pseudo-weight vector, and
NA-handling metadata to `result`, then assigns the `"pw_fit"`
class so that downstream S3 methods can dispatch on it.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwmean_s3_process_na.R
\name{process_na_yz}
\alias{process_na_yz}
\title{Handle missing values in the outcome and domain variables}
\usage{
process_na_yz(sc_data, y, zcol = NULL, na.action = stats::na.omit)
}
\arguments{
\item{sc_data}{A list returned by `prepare_sc_data()`, with components
`sc`, `X`, `w`, and `idx_keep`.}

\item{y}{Single character string naming the outcome variable in `sc_data$sc`.}

\item{zcol}{Single character string naming the domain variable in
`sc_data$sc`, or NULL for the overall mean.}

\item{na.action}{NA-handling function; one of `stats::na.omit` (default),
`stats::na.exclude`, or `stats::na.fail`. `na.pass` is not supported.}
}
\value{
A list with components:

- `Y`: numeric outcome vector (complete cases only).
- `X`: covariate matrix (complete cases only).
- `w`: pseudo-weight vector (complete cases only).
- `sc`: data frame (complete cases only).
- `y_name`: the value of `y`.
- `zcol`: the value of `zcol`.
- `domain`: list returned by `standardize_zcol()`.
- `na_info`: list with `na_action`, `n_omitted`, `n_used`,
  `omitted_raw`, and `kept_raw`.
}
\description{
Identifies rows with missing values in `y` (and `zcol` if supplied),
applies the chosen `na.action` strategy, and returns the complete-case
subsets of the outcome vector, covariate matrix, and pseudo-weight vector.
Also standardizes the domain variable via `standardize_zcol()`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estpw_s4_precali.R
\name{precal_cumulative_order}
\alias{precal_cumulative_order}
\title{Cumulative precalibration across multiple probability samples}
\usage{
precal_cumulative_order(sp_raw, sp_new, weight, sp_order)
}
\arguments{
\item{sp_raw}{A list of data frames (raw reference samples).}

\item{sp_new}{A list of data frames (working reference samples) with the same
structure as \code{sp_raw}. Calibrated weights are written back in place.}

\item{weight}{A list of weight column name strings, one per sample.}

\item{sp_order}{\code{"size"} to process largest sample first; otherwise
the original list order is used.}
}
\value{
A list with components \code{sp_new}, \code{total_vector},
  \code{log_messages}, and \code{order_used}.
}
\description{
Performs a cumulative precalibration across multiple probability samples by
sequentially calibrating each sample's weights to align with the marginal
totals of the previous samples. Called internally by \code{est_pw()}.
}
\keyword{internal}
